open OUnit

(* tests *)
let test_register_field _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_field
      fc (Xnc_serializable.new_field "Int" 1 Xnc_serializable.wrap_int); 
    OUnit.assert_bool
      "Failed registering field in fc"
      (Xnc_field_collection.is_registered fc "Int");
;;

let test_register_serializable _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_serializable
      fc "Int_option"
      (Xnc_serializable.wrap_option
         (Some (Xnc_serializable.wrap_int 2)));
    OUnit.assert_bool
      "Failed registering field in fc"
      (Xnc_field_collection.is_registered fc "Int_option");
;;

let test_register_value _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_value
      fc "Int_option" (Some 2)
      (fun v -> Xnc_serializable.wrap_option
                  (Option.map Xnc_serializable.wrap_int v));
    OUnit.assert_bool
      "Failed registering field in fc"
      (Xnc_field_collection.is_registered fc "Int_option");
;;

let test_register _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_int fc "Int_value" 42;
    Xnc_field_collection.register_float fc "Float_value" 3.14;
    Xnc_field_collection.register_char fc "Char_value" 'c';
    Xnc_field_collection.register_string fc "String_value" "test";
    Xnc_field_collection.register_bool fc "Bool_value" false;
    Xnc_field_collection.register_date fc "Date_value"
      { Netdate.year = 2007;
        Netdate.month = 2;
        Netdate.day = 12;
        Netdate.hour = 23;
        Netdate.minute = 11;
        Netdate.second = 30;
        Netdate.zone = 0;
        Netdate.week_day = -1
      };
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      6
      (Xnc_field_collection.length fc);
    OUnit.assert_bool
      "Failed registering int field in fc"
      (Xnc_field_collection.is_registered fc "Int_value");
    OUnit.assert_bool
      "Failed registering float field in fc"
      (Xnc_field_collection.is_registered fc "Float_value");
    OUnit.assert_bool
      "Failed registering char field in fc"
      (Xnc_field_collection.is_registered fc "Char_value");
    OUnit.assert_bool
      "Failed registering string field in fc"
      (Xnc_field_collection.is_registered fc "String_value");
    OUnit.assert_bool
      "Failed registering bool field in fc"
      (Xnc_field_collection.is_registered fc "Bool_value");
    OUnit.assert_bool
      "Failed registering date field in fc"
      (Xnc_field_collection.is_registered fc "Date_value");
;;

let test_register_option _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_option
      fc "None_value" None Xnc_serializable.wrap_int;
    Xnc_field_collection.register_option fc
      "Float_option" (Some 1.) Xnc_serializable.wrap_float;
    Xnc_field_collection.register_option fc
      "Char_option" (Some 'a') Xnc_serializable.wrap_char;
    Xnc_field_collection.register_option fc
      "String_option" (Some "Some") Xnc_serializable.wrap_string;
    Xnc_field_collection.register_option fc
      "Bool_option" (Some true) Xnc_serializable.wrap_bool;
    Xnc_field_collection.register_option fc "Date_value"
      (Some { Netdate.year = 2007;
              Netdate.month = 2;
              Netdate.day = 12;
              Netdate.hour = 23;
              Netdate.minute = 11;
              Netdate.second = 30;
              Netdate.zone = 0;
              Netdate.week_day = -1
            }) Xnc_serializable.wrap_date;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      6
      (Xnc_field_collection.length fc);
    OUnit.assert_bool
      "Failed registering int option (None) field in fc"
      (Xnc_field_collection.is_registered fc "None_value");
    OUnit.assert_bool
      "Failed registering float option (1.) field in fc"
      (Xnc_field_collection.is_registered fc "Float_option");
    OUnit.assert_bool
      "Failed registering char option ('a') field in fc"
      (Xnc_field_collection.is_registered fc "Char_option");
    OUnit.assert_bool
      "Failed registering string option (\"Some\") field in fc"
      (Xnc_field_collection.is_registered fc "String_option");
    OUnit.assert_bool
      "Failed registering bool option (true) field in fc"
      (Xnc_field_collection.is_registered fc "Bool_option");
;;

let test_register_object _ =
  let fc = Xnc_field_collection.create "Test" in
  let o = new Serialization_test_helper.test_serializable 3 4 in
    Xnc_field_collection.register_object fc "Object" o;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      1
      (Xnc_field_collection.length fc);
    OUnit.assert_bool
      "Failed registering object field in fc"
      (Xnc_field_collection.is_registered fc "Object");
;;

let test_register_custom _ =
  let write_value ?properties writer value =
    match value with
        Xnc_serializable.String s ->
          Xnc_writer.write_string writer (String.uppercase s)
      | _ -> failwith "unsupported" in
  let read_value ?properties reader value =
    match value with
        Xnc_serializable.String _ ->
          Xnc_serializable.String
            (String.lowercase (Xnc_reader.read_string reader))
      | _ -> failwith "unsupported" in
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_custom fc "Custom"
      "test" Xnc_serializable.wrap_string write_value read_value;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      1
      (Xnc_field_collection.length fc);
    OUnit.assert_bool
      "Failed registering custom field in fc"
      (Xnc_field_collection.is_registered fc "Custom");
;;

let test_register_dynarray _ =
  let fc = Xnc_field_collection.create "Test" in
  let dynarray = DynArray.of_array [|1; 2; 3; 4|] in
    Xnc_field_collection.register_dynarray
      fc "Int_dynarray" dynarray Xnc_serializable.wrap_int;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      4
      (Xnc_field_collection.length fc);
    OUnit.assert_bool
      "Failed registering int DynArray.t field in fc"
      (Xnc_field_collection.is_registered fc "Int_dynarray");
;;

let test_register_array _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_array
      fc "Int_array" [|1; 2; 3; 4|] Xnc_serializable.wrap_int;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      4
      (Xnc_field_collection.length fc);
    OUnit.assert_bool
      "Failed registering int array field in fc"
      (Xnc_field_collection.is_registered fc "Int_array");
;;

let test_register_list _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_list
      fc "Int_list" [1; 2; 3; 4] Xnc_serializable.wrap_int;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      4
      (Xnc_field_collection.length fc);
    OUnit.assert_bool
      "Failed registering int list field in fc"
      (Xnc_field_collection.is_registered fc "Int_list");
;;

let test_get_field _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_int fc "Int" 1;
    Xnc_field_collection.register_string
      ~properties:(Xnc_property.from_list
                     [Xnc_property.attribute;
                      Xnc_property.Length 10])
      ~prefix:"xnc"
      fc "String" "test";
    OUnit.assert_equal
      ~printer:Xnc_serializable.string_of_field
      ~msg:"Unexpected registered int field"
      (Xnc_serializable.new_field "Int" 1 Xnc_serializable.wrap_int)
      (Xnc_field_collection.get_field
         fc "Int");
    OUnit.assert_equal
      ~printer:Xnc_serializable.string_of_field
      ~msg:"Unexpected registered string field"
      (Xnc_serializable.new_field
         ~properties:(Xnc_property.from_list
                        [Xnc_property.attribute;
                         Xnc_property.Length 10])
         ~prefix:"xnc"
         "String" "test" Xnc_serializable.wrap_string)
      (Xnc_field_collection.get_field
         ~properties:(Xnc_property.singleton Xnc_property.attribute)
         ~prefix:"xnc"
         fc "String");
;;

let test_get_serializable _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_int fc "Int" 1;
    OUnit.assert_equal
      ~printer:Xnc_serializable.string_of_value
      ~msg:"Unexpected registered int value"
      (Xnc_serializable.wrap_int 1)
      (Xnc_field_collection.get_serializable fc "Int");
;;

let test_get_value _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_int fc "Int_value" 42;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected registered int value"
      42
      (Xnc_field_collection.get_value
         fc "Int_value" Xnc_serializable.unwrap_int);
;;

let test_get _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_int fc "Int_value" 42;
    Xnc_field_collection.register_float fc "Float_value" 3.14;
    Xnc_field_collection.register_char fc "Char_value" 'c';
    Xnc_field_collection.register_string fc "String_value" "test";
    Xnc_field_collection.register_bool fc "Bool_value" false;
    Xnc_field_collection.register_date fc "Date_value"
      { Netdate.year = 2007;
        Netdate.month = 2;
        Netdate.day = 12;
        Netdate.hour = 23;
        Netdate.minute = 11;
        Netdate.second = 30;
        Netdate.zone = 0;
        Netdate.week_day = -1
      };
    Xnc_field_collection.register_option
      fc "None_value" None Xnc_serializable.wrap_int;
    Xnc_field_collection.register_option
      fc "Int_option" (Some 2) Xnc_serializable.wrap_int;
    Xnc_field_collection.register_option
      fc "Float_option" (Some 0.25) Xnc_serializable.wrap_float;
    Xnc_field_collection.register_option
      fc "Char_option" (Some 'q') Xnc_serializable.wrap_char;
    Xnc_field_collection.register_option
      fc "String_option" (Some "String") Xnc_serializable.wrap_string;
    Xnc_field_collection.register_option
      fc "Bool_option" (Some false) Xnc_serializable.wrap_bool;
    Xnc_field_collection.register_option fc "Date_option"
      (Some { Netdate.year = 2006;
              Netdate.month = 2;
              Netdate.day = 12;
              Netdate.hour = 23;
              Netdate.minute = 11;
              Netdate.second = 30;
              Netdate.zone = 0;
              Netdate.week_day = -1
            }) Xnc_serializable.wrap_date;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected registered int value"
      42
      (Xnc_field_collection.get_int fc "Int_value");
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected registered float value"
      3.14
      (Xnc_field_collection.get_float fc "Float_value");
    OUnit.assert_equal
      ~printer:Std.string_of_char
      ~msg:"Unexpected registered char value"
      'c'
      (Xnc_field_collection.get_char fc "Char_value");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected registered string value"
      "test"
      (Xnc_field_collection.get_string fc "String_value");
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected registered bool value"
      false
      (Xnc_field_collection.get_bool fc "Bool_value");
    OUnit.assert_equal
      ~printer:(Netdate.format ~fmt:"%D %T")
      ~msg:"Unexpected registered date value"
      { Netdate.year = 2007;
        Netdate.month = 2;
        Netdate.day = 12;
        Netdate.hour = 23;
        Netdate.minute = 11;
        Netdate.second = 30;
        Netdate.zone = 0;
        Netdate.week_day = -1
      }
      (Xnc_field_collection.get_date fc "Date_value");
    OUnit.assert_equal
      ~msg:"Unexpected string option (None) value"
      None
      (Xnc_field_collection.get_option
         fc "None_value" Xnc_serializable.unwrap_string);
    OUnit.assert_equal
      ~msg:"Unexpected int option (2) value"
      (Some 2)
      (Xnc_field_collection.get_option
         fc "Int_option" Xnc_serializable.unwrap_int);
    OUnit.assert_equal
      ~msg:"Unexpected float option (0.25) value"
      (Some 0.25)
      (Xnc_field_collection.get_option
         fc "Float_option" Xnc_serializable.unwrap_float);
    OUnit.assert_equal
      ~msg:"Unexpected char option ('q') value"
      (Some 'q')
      (Xnc_field_collection.get_option
         fc "Char_option" Xnc_serializable.unwrap_char);
    OUnit.assert_equal
      ~msg:"Unexpected string option (\"String\") value"
      (Some "String")
      (Xnc_field_collection.get_option
         fc "String_option" Xnc_serializable.unwrap_string);
    OUnit.assert_equal
      ~msg:"Unexpected bool option (false) value"
      (Some false)
      (Xnc_field_collection.get_option
         fc "Bool_option" Xnc_serializable.unwrap_bool);
    OUnit.assert_equal
      ~msg:"Unexpected date option (2/12/2006 23:11:30) value"
      (Some { Netdate.year = 2006;
              Netdate.month = 2;
              Netdate.day = 12;
              Netdate.hour = 23;
              Netdate.minute = 11;
              Netdate.second = 30;
              Netdate.zone = 0;
              Netdate.week_day = -1
            })
      (Xnc_field_collection.get_option
         fc "Date_option" Xnc_serializable.unwrap_date);
;;

let test_get_option _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_option
      fc "None_value" None Xnc_serializable.wrap_string;
    Xnc_field_collection.register_option
      fc "Float_option" (Some 0.25) Xnc_serializable.wrap_float;
    OUnit.assert_equal
      ~msg:"Unexpected string option (None) value"
      None
      (Xnc_field_collection.get_option
         fc "None_value" Xnc_serializable.unwrap_string);
    OUnit.assert_equal
      ~msg:"Unexpected float option (0.25) value"
      (Some 0.25)
      (Xnc_field_collection.get_option
         fc "Float_option" Xnc_serializable.unwrap_float);
;;

let test_get_object _ =
  let fc = Xnc_field_collection.create "Test" in
  let o = new Serialization_test_helper.test_serializable 3 4 in
    Xnc_field_collection.register_object fc "Object" o;
    OUnit.assert_equal
      ~cmp:(fun o1 o2 ->
              Serialization_test_helper.print_object o1 =
                Serialization_test_helper.print_object o2)
      ~printer:Serialization_test_helper.print_object
      ~msg:"Unexpected object value"
      (new Serialization_test_helper.test_serializable 3 4
         :> Xnc_serialization.serializable)
      (Xnc_field_collection.get_object fc "Object")
;;

let test_get_custom _ =
  let write_value ?properties writer value =
    match value with
        Xnc_serializable.String s ->
          Xnc_writer.write_string writer (String.uppercase s)
      | _ -> failwith "unsupported" in
  let read_value ?properties reader value =
    match value with
        Xnc_serializable.String _ ->
          Xnc_serializable.String
            (String.lowercase (Xnc_reader.read_string reader))
      | _ -> failwith "unsupported" in
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_custom fc "Custom"
      "test" Xnc_serializable.wrap_string write_value read_value;
    OUnit.assert_equal
      ~msg:"Unexpected custom value"
      "test"
      (Xnc_field_collection.get_custom
         fc "Custom" Xnc_serializable.unwrap_string);
;;

let test_get_dynarray _ =
  let fc = Xnc_field_collection.create "Test" in
  let dynarray = DynArray.of_array [|1; 2; 3; 4|] in
    Xnc_field_collection.register_dynarray
      fc "Int_dynarray" dynarray Xnc_serializable.wrap_int;
    OUnit.assert_equal
      ~msg:"Unexpected int array value"
      [|1; 2; 3; 4|]
      (DynArray.to_array
         (Xnc_field_collection.get_dynarray
            fc "Int_dynarray" Xnc_serializable.unwrap_int));
;;

let test_get_array _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_array
      fc "Int_array" [|1; 2; 3; 4|] Xnc_serializable.wrap_int;
    OUnit.assert_equal
      ~msg:"Unexpected int array value"
      [|1; 2; 3; 4|]
      (Xnc_field_collection.get_array
         fc "Int_array" Xnc_serializable.unwrap_int);
;;

let test_get_list _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_list
      fc "Int_list" [1; 2; 3; 4] Xnc_serializable.wrap_int;
    OUnit.assert_equal
      ~msg:"Unexpected int list value"
      [1; 2; 3; 4]
      (Xnc_field_collection.get_list
         fc "Int_list" Xnc_serializable.unwrap_int);
;;

let test_update_field _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_field fc
      (Xnc_serializable.new_field "Int_value" 42 Xnc_serializable.wrap_int);
    Xnc_field_collection.update_field fc
      (Xnc_serializable.new_field "Int_value" 15 Xnc_serializable.wrap_int);
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected registered int value"
      15
      (Xnc_field_collection.get_int fc "Int_value");
;;

let test_update_serializable _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_serializable
      ~prefix:"xnc" fc "Int_option"
      (Xnc_serializable.wrap_option
         (Some (Xnc_serializable.wrap_int 2)));
    Xnc_field_collection.register_serializable
      ~prefix:"xnc" fc "Int"
      (Xnc_serializable.wrap_int 2);
    Xnc_field_collection.update_serializable
      ~prefix:"xnc" fc "Int_option"
      (Xnc_serializable.wrap_option
         (Some (Xnc_serializable.wrap_int 15)));
    Xnc_field_collection.update_serializable
      ~prefix:"tst" fc "Int"
      (Xnc_serializable.wrap_int 15);
    OUnit.assert_equal
      ~msg:"Unexpected registered option int value"
      (Some 15)
      (Xnc_field_collection.get_option
         ~prefix:"xnc" fc "Int_option" Xnc_serializable.unwrap_int);
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected registered int value"
      2
      (Xnc_field_collection.get_int
         ~prefix:"xnc" fc "Int");
;;

let test_update_value _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_int fc "Int_value" 42;
    Xnc_field_collection.update_value
      fc "Int_value" 15 Xnc_serializable.wrap_int;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected registered int value"
      15
      (Xnc_field_collection.get_int fc "Int_value");
;;

let test_update _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_int fc "Int_value" 42;
    Xnc_field_collection.register_float fc "Float_value" 3.14;
    Xnc_field_collection.register_char fc "Char_value" 'c';
    Xnc_field_collection.register_string fc "String_value" "test";
    Xnc_field_collection.register_bool fc "Bool_value" false;
    Xnc_field_collection.register_date fc "Date_value"
      { Netdate.year = 1970;
        Netdate.month = 1;
        Netdate.day = 1;
        Netdate.hour = 0;
        Netdate.minute = 0;
        Netdate.second = 0;
        Netdate.zone = 0;
        Netdate.week_day = -1
      };
    Xnc_field_collection.register_option
      fc "None_value" None Xnc_serializable.wrap_int;
    Xnc_field_collection.register_option
      fc "Int_option" (Some 2) Xnc_serializable.wrap_int;
    Xnc_field_collection.register_option
      fc "Float_option" (Some 0.25) Xnc_serializable.wrap_float;
    Xnc_field_collection.register_option
      fc "Char_option" (Some 'q') Xnc_serializable.wrap_char;
    Xnc_field_collection.register_option
      fc "String_option" (Some "String") Xnc_serializable.wrap_string;
    Xnc_field_collection.register_option
      fc "Bool_option" (Some false) Xnc_serializable.wrap_bool;
    Xnc_field_collection.register_option fc "Date_option"
      (Some { Netdate.year = 2006;
              Netdate.month = 2;
              Netdate.day = 12;
              Netdate.hour = 23;
              Netdate.minute = 11;
              Netdate.second = 30;
              Netdate.zone = 0;
              Netdate.week_day = -1
            }) Xnc_serializable.wrap_date;
    Xnc_field_collection.update_int fc "Int_value" 15;
    Xnc_field_collection.update_float fc "Float_value" 2.72;
    Xnc_field_collection.update_char fc "Char_value" 'a';
    Xnc_field_collection.update_string fc "String_value" "string";
    Xnc_field_collection.update_bool fc "Bool_value" true;
    Xnc_field_collection.update_date fc "Date_value"
      { Netdate.year = 2007;
        Netdate.month = 2;
        Netdate.day = 12;
        Netdate.hour = 23;
        Netdate.minute = 11;
        Netdate.second = 30;
        Netdate.zone = 0;
        Netdate.week_day = -1
      };
    Xnc_field_collection.update_option
      fc "None_value" (Some "None") Xnc_serializable.wrap_string;
    Xnc_field_collection.update_option
      fc "Int_option" (Some 7) Xnc_serializable.wrap_int;
    Xnc_field_collection.update_option
      fc "Float_option" None Xnc_serializable.wrap_float;
    Xnc_field_collection.update_option
      fc "Char_option" (Some 'o') Xnc_serializable.wrap_char;
    Xnc_field_collection.update_option
      fc "String_option" (Some "test") Xnc_serializable.wrap_string;
    Xnc_field_collection.update_option
      fc "Bool_option" (Some true) Xnc_serializable.wrap_bool;
    Xnc_field_collection.update_option fc "Date_option"
      (Some { Netdate.year = 2006;
              Netdate.month = 2;
              Netdate.day = 12;
              Netdate.hour = 23;
              Netdate.minute = 11;
              Netdate.second = 30;
              Netdate.zone = 0;
              Netdate.week_day = -1
            }) Xnc_serializable.wrap_date;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected registered int value"
      15
      (Xnc_field_collection.get_int fc "Int_value");
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected registered float value"
      2.72
      (Xnc_field_collection.get_float fc "Float_value");
    OUnit.assert_equal
      ~printer:Std.string_of_char
      ~msg:"Unexpected registered char value"
      'a'
      (Xnc_field_collection.get_char fc "Char_value");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected registered string value"
      "string"
      (Xnc_field_collection.get_string fc "String_value");
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected registered bool value"
      true
      (Xnc_field_collection.get_bool fc "Bool_value");
    OUnit.assert_equal
      ~printer:(Netdate.format ~fmt:"%D %T")
      ~msg:"Unexpected registered date value"
      { Netdate.year = 2007;
        Netdate.month = 2;
        Netdate.day = 12;
        Netdate.hour = 23;
        Netdate.minute = 11;
        Netdate.second = 30;
        Netdate.zone = 0;
        Netdate.week_day = -1
      }
      (Xnc_field_collection.get_date fc "Date_value");
    OUnit.assert_equal
      ~msg:"Unexpected string option (\"None\") value"
      (Some "None")
      (Xnc_field_collection.get_option
         fc "None_value" Xnc_serializable.unwrap_string);
    OUnit.assert_equal
      ~msg:"Unexpected int option (7) value"
      (Some 7)
      (Xnc_field_collection.get_option
         fc "Int_option" Xnc_serializable.unwrap_int);
    OUnit.assert_equal
      ~msg:"Unexpected float option (None) value"
      None
      (Xnc_field_collection.get_option
         fc "Float_option" Xnc_serializable.unwrap_float);
    OUnit.assert_equal
      ~msg:"Unexpected char option ('o') value"
      (Some 'o')
      (Xnc_field_collection.get_option
         fc "Char_option" Xnc_serializable.unwrap_char);
    OUnit.assert_equal
      ~msg:"Unexpected string option (\"test\") value"
      (Some "test")
      (Xnc_field_collection.get_option
         fc "String_option" Xnc_serializable.unwrap_string);
    OUnit.assert_equal
      ~msg:"Unexpected bool option (true) value"
      (Some true)
      (Xnc_field_collection.get_option
         fc "Bool_option" Xnc_serializable.unwrap_bool);
    OUnit.assert_equal
      ~msg:"Unexpected date option (2/12/2006 23:11:30) value"
      (Some { Netdate.year = 2006;
              Netdate.month = 2;
              Netdate.day = 12;
              Netdate.hour = 23;
              Netdate.minute = 11;
              Netdate.second = 30;
              Netdate.zone = 0;
              Netdate.week_day = -1
            })
      (Xnc_field_collection.get_option
         fc "Date_option" Xnc_serializable.unwrap_date);
;;

let test_update_option _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_option
      fc "Int_option" (Some 2) Xnc_serializable.wrap_int;
    Xnc_field_collection.update_option
      fc "Int_option" (Some 7) Xnc_serializable.wrap_int;
    OUnit.assert_equal
      ~msg:"Unexpected int option (7) value"
      (Some 7)
      (Xnc_field_collection.get_option
         fc "Int_option" Xnc_serializable.unwrap_int);
;;

let test_update_object _ =
  let fc = Xnc_field_collection.create "Test" in
  let o = new Serialization_test_helper.test_serializable 3 4 in
  let o' = new Serialization_test_helper.test_serializable 1 2 in
    Xnc_field_collection.register_object fc "Object" o;
    Xnc_field_collection.update_object fc "Object" o';
    OUnit.assert_equal
      ~cmp:(fun o1 o2 ->
              Serialization_test_helper.print_object o1 =
                Serialization_test_helper.print_object o2)
      ~printer:Serialization_test_helper.print_object
      ~msg:"Unexpected object value"
      (new Serialization_test_helper.test_serializable 1 2
         :> Xnc_serialization.serializable)
      (Xnc_field_collection.get_object fc "Object")
;;

let test_update_custom _ =
  let write_value ?properties writer value =
    match value with
        Xnc_serializable.String s ->
          Xnc_writer.write_string writer (String.uppercase s)
      | Xnc_serializable.Int i ->
          Xnc_writer.write_int writer (succ i)
      | _ -> failwith "unsupported" in
  let read_value ?properties reader value =
    match value with
        Xnc_serializable.String _ ->
          Xnc_serializable.String
            (String.lowercase (Xnc_reader.read_string reader))
      | Xnc_serializable.Int _ ->
          Xnc_serializable.Int
            (pred (Xnc_reader.read_int reader))
      | _ -> failwith "unsupported" in
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_custom fc "Custom"
      "test" Xnc_serializable.wrap_string write_value read_value;
    Xnc_field_collection.update_custom fc "Custom"
      1 Xnc_serializable.wrap_int write_value read_value;
    OUnit.assert_equal
      ~msg:"Unexpected custom value"
      1
      (Xnc_field_collection.get_custom
         fc "Custom" Xnc_serializable.unwrap_int);
;;

let test_update_dynarray _ =
  let fc = Xnc_field_collection.create "Test" in
  let dynarray = DynArray.of_array [|1; 2; 3; 4|] in
  let dynarray' = DynArray.of_array [|'a'; 'b'; 'c'|] in
    Xnc_field_collection.register_dynarray
      fc "DynArray" dynarray Xnc_serializable.wrap_int;
    Xnc_field_collection.update_dynarray
      fc "DynArray" dynarray' Xnc_serializable.wrap_char;
    OUnit.assert_equal
      ~msg:"Unexpected dynarray value"
      [|'a'; 'b'; 'c'|]
      (DynArray.to_array
         (Xnc_field_collection.get_dynarray
            fc "DynArray" Xnc_serializable.unwrap_char));
;;

let test_update_array _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_array
      fc "Array" [|1; 2; 3; 4|] Xnc_serializable.wrap_int;
    Xnc_field_collection.update_array
      fc "Array" [|'a'; 'b'; 'c'|] Xnc_serializable.wrap_char;
    OUnit.assert_equal
      ~msg:"Unexpected array value"
      [|'a'; 'b'; 'c'|]
      (Xnc_field_collection.get_array
         fc "Array" Xnc_serializable.unwrap_char);
;;

let test_update_list _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_list
      fc "List" [1; 2; 3; 4] Xnc_serializable.wrap_int;
    Xnc_field_collection.update_list
      fc "List" ['a'; 'b'; 'c'] Xnc_serializable.wrap_char;
    OUnit.assert_equal
      ~msg:"Unexpected list value"
      ['a'; 'b'; 'c']
      (Xnc_field_collection.get_list
         fc "List" Xnc_serializable.unwrap_char);
;;

let test_to_string _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_int fc "Int_value" 42;
    Xnc_field_collection.register_float fc "Float_value" 3.14;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected serialization"
      "Test=[Int_value=42;Float_value=3.14]"
      (Xnc_field_collection.to_string fc)
;;

let test_to_netchannel _ =
  let buf = Buffer.create 32 in
  let out_ch = new Netchannels.output_buffer buf in
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_int fc "Int_value" 42;
    Xnc_field_collection.register_float fc "Float_value" 3.14;
    Xnc_field_collection.to_netchannel fc out_ch;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected serialization"
      "Test=[Int_value=42;Float_value=3.14]"
      (Buffer.contents buf)
;;

let test_to_string_custom_serializer _ =
  let serializer =
    new Xnc_serializing.simple_serializer
      ~config:{ Xnc_serialization.default_config
                  with Xnc_serialization.start_of_list = "(";
                       Xnc_serialization.end_of_list = ")";
                       Xnc_serialization.list_separator = "," } () in
  let fc = Xnc_field_collection.create
             ~serializer
             "Test" in
    Xnc_field_collection.register_int fc "Int_value" 42;
    Xnc_field_collection.register_float fc "Float_value" 3.14;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected serialization"
      "Test=(Int_value=42,Float_value=3.14)"
      (Xnc_field_collection.to_string fc)
;;

let test_from_string _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_serializable
      fc "Int_value" Xnc_serializable.default_int;
    Xnc_field_collection.register_serializable
      fc "Float_value" Xnc_serializable.default_float;
    Xnc_field_collection.from_string fc
      "Test=[Int_value=42;Float_value=3.14]";
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Failed deserializing int field"
      42
      (Xnc_field_collection.get_int fc "Int_value");
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Failed deserializing float field"
      3.14
      (Xnc_field_collection.get_float fc "Float_value")
;;

let test_from_netchannel _ =
  let in_ch = new Netchannels.input_string
                "Test=[Int_value=42;Float_value=3.14]" in
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_serializable
      fc "Int_value" Xnc_serializable.default_int;
    Xnc_field_collection.register_serializable
      fc "Float_value" Xnc_serializable.default_float;
    Xnc_field_collection.from_netchannel fc in_ch;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Failed deserializing int field"
      42
      (Xnc_field_collection.get_int fc "Int_value");
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Failed deserializing float field"
      3.14
      (Xnc_field_collection.get_float fc "Float_value")
;;

let test_from_string_custom_serializer _ =
  let serializer =
    new Xnc_serializing.simple_serializer
      ~config:{ Xnc_serialization.default_config
                  with Xnc_serialization.start_of_list = "(";
                       Xnc_serialization.end_of_list = ")";
                       Xnc_serialization.list_separator = "," } () in
  let fc = Xnc_field_collection.create
             "Test" in
    Xnc_field_collection.register_serializable
      fc "Int_value" Xnc_serializable.default_int;
    Xnc_field_collection.register_serializable
      fc "Float_value" Xnc_serializable.default_float;
    Xnc_field_collection.from_string ~serializer fc
      "Test=(Int_value=42,Float_value=3.14)";
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Failed deserializing int field"
      42
      (Xnc_field_collection.get_int fc "Int_value");
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Failed deserializing float field"
      3.14
      (Xnc_field_collection.get_float fc "Float_value")
;;

let test_to_name_value_string_list _ =
  let fc = Xnc_field_collection.create "Test" in
    Xnc_field_collection.register_int fc "Int_value" 42;
    Xnc_field_collection.register_float fc "Float_value" 3.14;
    OUnit.assert_equal
      ~msg:"Unexpected name/value list serialization"
      [("Int_value", "42"); ("Float_value", "3.14")]
      (Xnc_field_collection.to_name_value_string_list fc)
;;

(* test suite *)
let suite = "Field collection test" >:::
  ["test_register_field" >:: test_register_field;
   "test_register_serializable" >:: test_register_serializable;
   "test_register_value" >:: test_register_value;
   "test_register" >:: test_register;
   "test_register_option" >:: test_register_option;
   "test_register_object" >:: test_register_object;
   "test_register_custom" >:: test_register_custom;
   "test_register_dynarray" >:: test_register_dynarray;
   "test_register_array" >:: test_register_array;
   "test_register_list" >:: test_register_list;
   "test_get_serializable" >:: test_get_serializable;
   "test_get_field" >:: test_get_field;
   "test_get_value" >:: test_get_value;
   "test_get" >:: test_get;
   "test_get_option" >:: test_get_option;
   "test_get_object" >:: test_get_object;
   "test_get_custom" >:: test_get_custom;
   "test_get_dynarray" >:: test_get_dynarray;
   "test_get_array" >:: test_get_array;
   "test_get_list" >:: test_get_list;
   "test_update_field" >:: test_update_field;
   "test_update_serializable" >:: test_update_serializable;
   "test_update_value" >:: test_update_value;
   "test_update" >:: test_update;
   "test_update_option" >:: test_update_option;
   "test_update_object" >:: test_update_object;
   "test_update_custom" >:: test_update_custom;
   "test_update_dynarray" >:: test_update_dynarray;
   "test_update_array" >:: test_update_array;
   "test_update_list" >:: test_update_list;
   "test_to_string" >:: test_to_string;
   "test_to_netchannel" >:: test_to_netchannel;
   "test_to_string_custom_serializer" >:: test_to_string_custom_serializer;
   "test_from_string" >:: test_from_string;
   "test_from_netchannel" >:: test_from_netchannel;
   "test_from_string_custom_serializer" >:: test_from_string_custom_serializer;
   "test_to_name_value_string_list" >:: test_to_name_value_string_list]
;;

