open OUnit

module S = Xnc_serializable

class fixed_length_formatter () : [string] Xnc_formatting.formatter =
object(self)
  inherit Xnc_formatting.basic_formatter ()

  method format_null ?(properties = Xnc_property.empty) () =
    let width = Xnc_property.get_length properties 10 in
      String.make width ' '

  method format_int ?(properties = Xnc_property.empty) i =
    let width = Xnc_property.get_length properties 10 in
      Printf.sprintf "%0*d" width i

  method format_float ?(properties = Xnc_property.empty) f =
    let width = Xnc_property.get_length properties 10 in
    let precision = Xnc_property.get_precision properties 10 in
      Printf.sprintf "%0*.*f" width precision f

  method format_char ?properties c =
    self # format_raw_string ?properties (Std.string_of_char c)

  method format_bool ?properties b =
    self # format_raw_string ?properties (string_of_bool b)

  method format_raw_string ?(properties = Xnc_property.empty) s =
    let width = Xnc_property.get_length properties 10 in
      Printf.sprintf "%-*s" width s

  method format_string ?properties s =
    self # format_raw_string ?properties s

  method format_date ?properties d =
    Netdate.format ~fmt:"%Y%m%d" d

end;;

module Make_text_printer
  (Output : Xnc_printing.OUTPUT)(Pretty : Xnc_printing.PRETTY) =
struct
  include Xnc_printing.Make_text_printer(Output)(Pretty)

  class fixed_length_printer out =
  object
    inherit text_printer out

    method print_field :
        'a. string -> string -> 'a -> Xnc_property.properties ->
        ('a -> unit) -> unit =
      fun prefix name value properties print_value ->
        print_value value

  end

end;;

module Fixed_length_printer =
  Make_text_printer(Xnc_printing.Netch)(Xnc_printing.Non_pp)
;;

class fixed_length_printer
  out_ch : [string] Xnc_printing.printer =
  Fixed_length_printer.fixed_length_printer out_ch
;;

class fixed_length_scanner
  sb : [string] Xnc_scanning.scanner =
object(self)
  inherit Xnc_scanning.base_scanner ()

  val scanbuf = sb

  method skip n =
    Xnc_scanning.Scanning_buffer.skip scanbuf n

  method private scan_value ?(properties = Xnc_property.empty) action =
    try
      let width = Xnc_property.get_length properties 10 in
        Xnc_scanning.Scanning_buffer.next_string scanbuf width
    with Xnc_scanning.Scan_failure d ->
      raise (Xnc_scanning.Scan_failure (action ^ ": " ^ d))

end;;

class fixed_length_evaluator () : [string] Xnc_token_eval.token_evaluator =
object(self)
  method get_int ?properties token =
    self # parse_value token int_of_string "to int"

  method get_float ?properties token =
    self # parse_value token float_of_string "to float"

  method get_char ?properties token =
    self # parse_value token (fun t -> t.[0]) "to char"

  method get_bool ?properties token =
    self # parse_value token bool_of_string "to bool"

  method get_string ?properties token =
    self # parse_value token Std.identity "to string"

  method get_date ?properties token =
    try
      let year = String.sub token 0 4 in
      let month = String.sub token 4 2 in
      let day = String.sub token 6 2 in
        { Netdate.year = int_of_string year;
          Netdate.month = int_of_string month;
          Netdate.day = int_of_string day;
          Netdate.hour = 0;
          Netdate.minute = 0;
          Netdate.second = 0;
          Netdate.zone = 0;
          Netdate.week_day = -1;
        }
    with _ ->
      raise (Xnc_token_eval.Token_evaluation_failed
               ("error trying to convert \"" ^ token ^ "\" to date"))

  method private parse_value :
      'a. string -> (string -> 'a) -> string -> 'a =
    fun token parse_token type_name ->
      let ltrimmed = Str.global_replace (Str.regexp "^ +") "" token in
      let trimmed = Str.global_replace (Str.regexp " +$") "" ltrimmed in
        try
          parse_token trimmed
        with _ ->
          raise (Xnc_token_eval.Token_evaluation_failed
                   (Printf.sprintf 
                      "error trying to convert \"%s\" %s"
                      token
                      type_name))

end;;

class fixed_length_lexer
  sb : [string] Xnc_lexing.lexical_analyzer =
object
  inherit Xnc_lexing.text_lexer ()

  val scanbuf = sb

  method try_parse_null ?(properties = Xnc_property.empty) () =
    let width = Xnc_property.get_length properties 10 in
    let none = String.make width ' ' in
      if Xnc_scanning.Scanning_buffer.peek_string scanbuf width = none then
        begin
          ignore(Xnc_scanning.Scanning_buffer.next_string scanbuf width);
          true
        end
      else
        false

  method parse_homogeneous_enum : 'a. (unit -> 'a) -> 'a Enum.t =
    fun parse_element ->
      failwith "Not implemented method"

  method parse_heterogeneous_enum : 'a. 'a Enum.t -> ('a -> 'a) -> 'a Enum.t =
    fun enum_schema parse_element ->
      Enum.map 
        (fun schema_element ->
           parse_element schema_element)
        enum_schema

  method parse_value_by_name :
      'a. string -> string -> Xnc_property.properties ->
      (unit -> 'a) -> 'a =
    fun prefix name properties parse_value ->
      parse_value ()

  method close () =
    Xnc_scanning.Scanning_buffer.close scanbuf

end;;

class fixed_length_serializer () : [string] Xnc_serializing.serializer =
object(self)
  inherit Xnc_serializing.netchannel_serializer ()

  method to_netchannel serializable_field out_ch =
    let formatter = new fixed_length_formatter () in
    let printer = new fixed_length_printer out_ch in
    let writer = Xnc_writer.create formatter printer in
      S.write writer serializable_field;
      Xnc_writer.flush writer

  method from_netchannel prototype in_ch =
    let scanbuf = Xnc_scanning.Scanning_buffer.create_from_netchannel in_ch in
    let scanner = new fixed_length_scanner scanbuf in
    let lexer = new fixed_length_lexer scanbuf in
    let token_evaluator = new fixed_length_evaluator () in
    let reader = Xnc_reader.create scanner token_evaluator lexer in
      S.read reader prototype

end;;

(* tests *)
let test_serialize _ =
  let serializer = new fixed_length_serializer () in
  let int_field =
    S.new_field
      ~properties:(Xnc_property.singleton (Xnc_property.Length 5))
      "int" 42 S.wrap_int in
  let float_field =
    S.new_field
      ~properties:(Xnc_property.from_list [Xnc_property.Length 5;
                                           Xnc_property.Precision 2])
      "float" 3.141592 S.wrap_float in
  let char_field =
    S.new_field
      ~properties:(Xnc_property.singleton (Xnc_property.Length 2))
      "char" 'a' S.wrap_char in
  let none_field =
    S.new_field
      ~properties:(Xnc_property.singleton (Xnc_property.Length 5))
      "null" None S.wrap_option in
  let bool_option_field =
    S.new_field
      ~properties:(Xnc_property.singleton (Xnc_property.Length 5))
      "option" (Some (S.Bool true)) S.wrap_option in
  let string_field =
    S.new_field
      ~properties:(Xnc_property.singleton (Xnc_property.Length 10))
      "string" "test" S.wrap_string in
  let object_field =
    S.new_field
      ~properties:(Xnc_property.singleton (Xnc_property.Length 10))
      "object"
      (new Serialization_test_helper.test_serializable 8 9) S.wrap_object in
  let date_field =
    S.new_field
      ~properties:(Xnc_property.singleton (Xnc_property.Length 8))
      "date" { Netdate.year = 2006;
               Netdate.month = 12;
               Netdate.day = 8;
               Netdate.hour = 19;
               Netdate.minute = 30;
               Netdate.second = 50;
               Netdate.zone = 0;
               Netdate.week_day = -1
             } S.wrap_date in
  let field = S.new_field "list"
                [int_field;
                float_field;
                char_field;
                none_field;
                bool_option_field;
                string_field;
                object_field;
                date_field] S.wrap_list in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected fixed length record"
      ("00042\
        03.14\
        a " ^
       "     " ^
       "true \
        test      \
        8/9       \
        20061208")
      (serializer # serialize field);
;;

let test_deserialize _ =
  let serializer = new fixed_length_serializer () in
  let int_field =
    S.create
      ~properties:(Xnc_property.singleton (Xnc_property.Length 5))
      "int" S.default_int in
  let float_field =
    S.create
      ~properties:(Xnc_property.from_list [Xnc_property.Length 5;
                                           Xnc_property.Precision 2])
      "float" S.default_float in
  let char_field =
    S.create
      ~properties:(Xnc_property.singleton (Xnc_property.Length 2))
      "char" S.default_char in
  let none_field =
    S.create
      ~properties:(Xnc_property.singleton (Xnc_property.Length 5))
      "null" (S.default_option S.default_int) in
  let bool_option_field =
    S.create
      ~properties:(Xnc_property.singleton (Xnc_property.Length 5))
      "option" (S.default_option
                  S.default_bool) in
  let string_field =
    S.create
      ~properties:(Xnc_property.singleton (Xnc_property.Length 10))
      "string" S.default_string in
  let object_field =
    S.new_field "object"
      ~properties:(Xnc_property.singleton (Xnc_property.Length 10))
      (new Serialization_test_helper.test_serializable 0 0) S.wrap_object in
  let date_field =
    S.create
      ~properties:(Xnc_property.singleton (Xnc_property.Length 8))
      "date" S.default_date in
  let field = S.new_field "list"
                [int_field;
                 float_field;
                 char_field;
                 none_field;
                 bool_option_field;
                 string_field;
                 object_field;
                 date_field] S.wrap_list in
    serializer # deserialize field
      ("00042\
        03.14\
        a " ^
       "     " ^
       "true \
        test      \
        8/9       \
        20061208");
    OUnit.assert_equal
      ~cmp:(S.custom_equal
              ~obj_eq:(fun o1 o2 ->
                         (o1 # to_string ()) = (o2 # to_string ())))
      ~printer:S.string_of_field
      ~msg:"Unexpected deserialized field"
      (S.new_field "list"
         [ S.new_field
             ~properties:(Xnc_property.singleton (Xnc_property.Length 5))
             "int" 42 S.wrap_int;
           S.new_field
             ~properties:(Xnc_property.from_list [Xnc_property.Length 5;
                                                  Xnc_property.Precision 2])
             "float" 3.14 S.wrap_float;
           S.new_field
             ~properties:(Xnc_property.singleton (Xnc_property.Length 2))
             "char" 'a' S.wrap_char;
           S.new_field
             ~properties:(Xnc_property.singleton (Xnc_property.Length 5))
             "null" None S.wrap_option;
           S.new_field "option"
             ~properties:(Xnc_property.singleton (Xnc_property.Length 5))
             (Some (S.wrap_bool true))
             S.wrap_option;
           S.new_field
             ~properties:(Xnc_property.singleton (Xnc_property.Length 10))
             "string" "test" S.wrap_string;
           S.new_field "object"
             ~properties:(Xnc_property.singleton (Xnc_property.Length 10))
             (new Serialization_test_helper.test_serializable 8 9)
             S.wrap_object;
           S.new_field
             ~properties:(Xnc_property.singleton (Xnc_property.Length 8))
             "date" { Netdate.year = 2006;
                      Netdate.month = 12;
                      Netdate.day = 8;
                      Netdate.hour = 0;
                      Netdate.minute = 0;
                      Netdate.second = 0;
                      Netdate.zone = 0;
                      Netdate.week_day = -1
                    } S.wrap_date
         ] S.wrap_list)
      field;
;;

(* test suite *)
let suite = "Fixed-length record serializer test" >:::
  ["test_serialize" >:: test_serialize;
   "test_deserialize" >:: test_deserialize]
;;

