open OUnit

(* tests *)
let test_parse_value _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "1\
                   \"test_string\"\
                   5.0\
                   false\
                   t" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
  let lexer = new Xnc_lexing.xnc_scanbuf_lexer scanbuf in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected int value"
      1
      (lexer # parse_value
         (scanner # scan_int)
         int_of_string);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string value"
      "test_string"
      (lexer # parse_value
         (scanner # scan_string)
         Std.identity);
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected float value"
      5.
      (lexer # parse_value
         (scanner # scan_float)
         float_of_string);
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected bool value"
      false
      (lexer # parse_value
         (scanner # scan_bool)
         bool_of_string);
    OUnit.assert_equal
      ~printer:Std.string_of_char
      ~msg:"Unexpected char value"
      't'
      (lexer # parse_value
         (scanner # scan_char)
         (fun s -> String.get s 0));
    lexer # close ();
;;

let test_parse_option _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "None\
                   false" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
  let lexer = new Xnc_lexing.xnc_scanbuf_lexer scanbuf in
  let print_bool_option = Serialization_test_helper.print_option
                            string_of_bool in
  let try_parse_null () =
    lexer # try_parse_null () in
  let parse_bool () =
    lexer # parse_value (scanner # scan_bool) bool_of_string
  in
    OUnit.assert_equal
      ~printer:print_bool_option
      ~msg:"Unexpected bool option value"
      None
      (lexer # parse_option try_parse_null parse_bool);
    OUnit.assert_equal
      ~printer:print_bool_option
      ~msg:"Unexpected bool option value"
      (Some false)
      (lexer # parse_option try_parse_null parse_bool);
    lexer # close ();
;;

let test_parse_object _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string "\"1/2\"" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
  let lexer = new Xnc_lexing.xnc_scanbuf_lexer scanbuf in
  let obj = new Serialization_test_helper.test_serializable 0 0 in
    lexer # parse_object obj (scanner # scan_string);
    OUnit.assert_equal
      ~cmp:Serialization_test_helper.cmp_object
      ~printer:Serialization_test_helper.print_object
      ~msg:"Unexpected object"
      (new Serialization_test_helper.test_serializable 1 2)
      (obj);
    lexer # close ();
;;

let test_parse_homogeneous_enum _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "[1.;2.;3.;4.]" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
  let lexer = new Xnc_lexing.xnc_scanbuf_lexer scanbuf in
  let parse_float () =
    lexer # parse_value (scanner # scan_float) float_of_string in
  let enum = lexer # parse_homogeneous_enum parse_float in
  let l = ExtList.List.of_enum enum in
    OUnit.assert_equal
      ~printer:(Serialization_test_helper.print_list string_of_float)
      ~msg:"Unexpected enum value"
      [1.; 2.; 3.; 4.]
      l;
    lexer # close ();
;;

type test_schema_type =
    Int of int
  | Char of char
  | String of string

let test_parse_heterogeneous_enum _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "[a;\"test\";1]" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
  let lexer = new Xnc_lexing.xnc_scanbuf_lexer scanbuf in
  let enum_schema = ExtList.List.enum [Char '?'; String ""; Int 0] in
  let parse_int () =
    lexer # parse_value (scanner # scan_int) int_of_string in
  let parse_char () =
    lexer # parse_value (scanner # scan_char) (fun s -> String.get s 0) in
  let parse_string () =
    lexer # parse_value (scanner # scan_string) Std.identity in
  let parse_element e =
    match e with
        Int _ -> Int (parse_int ())
      | Char _ -> Char (parse_char ())
      | String _ -> String (parse_string ()) in
  let print_list =
    Serialization_test_helper.print_list
      (fun e -> match e with
           Int i -> string_of_int i
         | Char c -> Std.string_of_char c
         | String s -> s) in
  let enum = lexer # parse_heterogeneous_enum enum_schema parse_element in
    OUnit.assert_equal
      ~printer:print_list
      ~msg:"Unexpected enum value"
      [Char 'a'; String "test"; Int 1]
      (ExtList.List.of_enum enum);
    lexer # close ();
;;

let test_parse_value_by_name _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "name=1\
                   @n=\"xyz\"\
                   xnc:name=1\
                   @xnc:n=\"xyz\"\
                   error=false" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
  let lexer = new Xnc_lexing.xnc_scanbuf_lexer scanbuf in
  let parse_int () =
    lexer # parse_value (scanner # scan_int) int_of_string in
  let parse_bool () =
    lexer # parse_value (scanner # scan_bool) bool_of_string in
  let parse_string () =
    lexer # parse_value (scanner # scan_string) Std.identity
  in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected value"
      1
      (lexer # parse_value_by_name
         "" "name" Xnc_property.empty parse_int);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value"
      "xyz"
      (lexer # parse_value_by_name
         "" "n"
         (Xnc_property.singleton Xnc_property.attribute)
         parse_string);
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected value"
      1
      (lexer # parse_value_by_name
         "xnc" "name" Xnc_property.empty parse_int);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value"
      "xyz"
      (lexer # parse_value_by_name
         "xnc" "n"
         (Xnc_property.singleton Xnc_property.attribute)
         parse_string);
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected value"
      false
      (lexer # parse_value_by_name
         "" "error" Xnc_property.empty parse_bool);
    lexer # close ();
;;

(* test suite *)
let suite = "Lexing test" >:::
  ["test_parse_value" >:: test_parse_value;
   "test_parse_option" >:: test_parse_option;
   "test_parse_object" >:: test_parse_object;
   "test_parse_homogeneous_enum" >:: test_parse_homogeneous_enum;
   "test_parse_heterogeneous_enum" >:: test_parse_heterogeneous_enum;
   "test_parse_value_by_name" >:: test_parse_value_by_name]
;;

