open OUnit

(* tests *)
let test_convert _ =
  let id_transform = Std.identity in
  let conv = Xnc_pipe_conv.convert id_transform in
  let pipe = new Netchannels.pipe ~conv () in
  let buf = Buffer.create 256 in
  let endpoint_ch = new Netchannels.output_buffer buf in
  let out_ch = new Netchannels.output_filter pipe endpoint_ch in
    out_ch # output_string "Pipe ";
    out_ch # flush ();
    out_ch # output_string "conversion ";
    out_ch # flush ();
    out_ch # output_string "test";
    out_ch # close_out ();
    endpoint_ch # close_out ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected endpoint contents"
      "Pipe conversion test"
      (Buffer.contents buf)
;;

let test_xor_conv _ =
  let pipe = new Netchannels.pipe
               ~conv:(Xnc_pipe_conv.xor_conv "ELQF") () in
  let buf = Buffer.create 256 in
  let endpoint_ch = new Netchannels.output_buffer buf in
  let out_ch = new Netchannels.output_filter pipe endpoint_ch in
    out_ch # output_string "pipe conversion test";
    out_ch # close_out ();
    endpoint_ch # close_out ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected endpoint contents"
      "5%!#e/>(3)#5,#?f1)\"2"
      (Buffer.contents buf)
;;

let test_escape_xml_conv _ =
  let pipe = new Netchannels.pipe
               ~conv:Xnc_pipe_conv.escape_xml_conv () in
  let buf = Buffer.create 256 in
  let endpoint_ch = new Netchannels.output_buffer buf in
  let out_ch = new Netchannels.output_filter pipe endpoint_ch in
    out_ch # output_string "'<\" test \">'";
    out_ch # close_out ();
    endpoint_ch # close_out ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected endpoint contents"
      "&apos;&lt;&quot; test &quot;&gt;&apos;"
      (Buffer.contents buf)
;;

let test_unescape_xml_conv _ =
  let pipe = new Netchannels.pipe
               ~conv:Xnc_pipe_conv.unescape_xml_conv () in
  let buf = Buffer.create 256 in
  let endpoint_ch = new Netchannels.output_buffer buf in
  let out_ch = new Netchannels.output_filter pipe endpoint_ch in
    out_ch # output_string "&apos;&lt;&quot; test &quot;&gt;&apos;";
    out_ch # close_out ();
    endpoint_ch # close_out ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected endpoint contents"
      "'<\" test \">'"
      (Buffer.contents buf)
;;

let test_cryptokit_zlib_hexa_conv _ =
  let transform = Xnc_string_transform.compose_cryptokit_transform
                    [Cryptokit.Zlib.compress ();
                     Cryptokit.Hexa.encode ()] in
  let pipe = new Netchannels.pipe
               ~conv:(Xnc_pipe_conv.cryptokit_conv transform) () in
  let buf = Buffer.create 256 in
  let endpoint_ch = new Netchannels.output_buffer buf in
  let out_ch = new Netchannels.output_filter pipe endpoint_ch in
    out_ch # output_string "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa";
    out_ch # close_out ();
    endpoint_ch # close_out ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected endpoint contents"
      "4b4c240000"
      (Buffer.contents buf)
;;

let test_cryptokit_hexa_zlib_conv _ =
  let transform = Xnc_string_transform.compose_cryptokit_transform
                    [Cryptokit.Hexa.decode ();
                     Cryptokit.Zlib.uncompress ()] in
  let pipe = new Netchannels.pipe
               ~conv:(Xnc_pipe_conv.cryptokit_conv transform) () in
  let buf = Buffer.create 256 in
  let endpoint_ch = new Netchannels.output_buffer buf in
  let out_ch = new Netchannels.output_filter pipe endpoint_ch in
    out_ch # output_string "4b4c240000";
    out_ch # close_out ();
    endpoint_ch # close_out ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected endpoint contents"
      "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
      (Buffer.contents buf)
;;

let test_digest_conv _ =
  let hash = Cryptokit.MAC.hmac_md5
               "1234567890abcdefghijklmnopqrstuvwxyz" in
  let pipe = new Netchannels.pipe
               ~conv:(Xnc_pipe_conv.digest_conv hash) () in
  let buf = Buffer.create 256 in
  let endpoint_ch = new Netchannels.output_buffer buf in
  let out_ch = new Netchannels.output_filter pipe endpoint_ch in
    out_ch # output_string "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa";
    out_ch # close_out ();
    endpoint_ch # close_out ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected endpoint contents"
      "IqJ/8GDpwRSC6YyWvQH46GFh\
       YWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYQ=="
      (Xnc_string_transform.base64_encode (Buffer.contents buf))
;;

let test_verify_digest_conv _ =
  let hash = Cryptokit.MAC.hmac_md5
               "1234567890abcdefghijklmnopqrstuvwxyz" in
  let pipe = new Netchannels.pipe
               ~conv:(Xnc_pipe_conv.verify_digest_conv hash) () in
  let buf = Buffer.create 256 in
  let endpoint_ch = new Netchannels.output_buffer buf in
  let out_ch = new Netchannels.output_filter pipe endpoint_ch in
    out_ch # output_string
      (Xnc_string_transform.base64_decode
         "IqJ/8GDpwRSC6YyWvQH46GFh\
          YWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYQ==");
    out_ch # close_out ();
    endpoint_ch # close_out ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected endpoint contents"
      "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"
      (Buffer.contents buf)
;;

let test_invalid_digest _ =
  let hash = Cryptokit.MAC.hmac_md5
               "1234567890abcdefghijklmnopqrstuvwxyz" in
  let pipe = new Netchannels.pipe
               ~conv:(Xnc_pipe_conv.verify_digest_conv hash) () in
  let buf = Buffer.create 256 in
  let endpoint_ch = new Netchannels.output_buffer buf in
  let out_ch = new Netchannels.output_filter pipe endpoint_ch in
    out_ch # output_string
      (Xnc_string_transform.base64_decode
         "JqJ/8GDpwRSC6YyWvQH46GFh\
          YWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYWFhYQ==");
    OUnit.assert_raises
      Xnc_pipe_conv.Invalid_digest
      (out_ch # close_out);
    endpoint_ch # close_out ()
;;

(* test suite *)
let suite = "Pipe conversion test" >:::
  ["test_convert" >:: test_convert;
   "test_xor_conv" >:: test_xor_conv;
   "test_escape_xml_conv" >:: test_escape_xml_conv;
   "test_unescape_xml_conv" >:: test_unescape_xml_conv;
   "test_cryptokit_zlib_hexa_conv" >:: test_cryptokit_zlib_hexa_conv;
   "test_cryptokit_hexa_zlib_conv" >:: test_cryptokit_hexa_zlib_conv;
   "test_digest_conv" >:: test_digest_conv;
   "test_verify_digest_conv" >:: test_verify_digest_conv;
   "test_invalid_digest" >:: test_invalid_digest]
;;

