open OUnit

(* fixture setup and teardown *)
let set_up _ =
  let buf = Buffer.create 16 in
  let printer = new Xnc_printing.Buffer_printer.default_printer buf in
    (buf, printer)
;;

let tear_down _ =
  ()
;;

(* tests *)
let test_print_value
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
    printer # print_value 1 string_of_int;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer contents after print_value"
      "1"
      (Buffer.contents buf);
;;
 
let test_print_option
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
  let print_none v = printer # print_value () (fun () -> "None") in
  let print_value v = printer # print_value v string_of_int in
    printer # print_option None print_none print_value;
    printer # print_option (Some 1) print_none print_value;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer contents after 2 print_option"
      "None1"
      (Buffer.contents buf);
;;
 
let test_print_object
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
  let print_value v = printer # print_value v Std.identity in
  let obj = new Serialization_test_helper.test_serializable 1 2 in
    printer # print_object obj print_value;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer contents after print_object"
      "1/2"
      (Buffer.contents buf);
;;
 
let test_print_enum
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
  let print_value v = printer # print_value v string_of_int in
  let enum = ExtList.List.enum [1; 2; 3; 4] in
    printer # print_enum enum print_value;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_enum"
      "[1;2;3;4]"
      (Buffer.contents buf);
;;
 
let test_print_field
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
  let print_value v = printer # print_value v string_of_int in
    printer # print_field
      "" "key" 10 Xnc_property.empty print_value;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field (no prefix, no attr)"
      "key=10"
      (Buffer.contents buf);
    Buffer.clear buf;
    printer # print_field
      "" "key" 10
      (Xnc_property.singleton Xnc_property.attribute)
      print_value;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field (no prefix, attr)"
      "@key=10"
      (Buffer.contents buf);
    Buffer.clear buf;
    printer # print_field
      "tst" "key" 10 Xnc_property.empty print_value;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field (prefix, no attr)"
      "tst:key=10"
      (Buffer.contents buf);
    Buffer.clear buf;
    printer # print_field
      "tst" "key" 10
      (Xnc_property.singleton Xnc_property.attribute)
      print_value;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field (prefix, attr)"
      "@tst:key=10"
      (Buffer.contents buf);
;;

let test_pretty_print_list _ =
  let buf = Buffer.create 16 in
  let printer = new Xnc_printing.Buffer_pretty_printer.default_printer buf in
  let print_value v = printer # print_value v Std.identity in
  let enum = ExtList.List.enum
               ["short text";
                "loooooooooooooooooooooooooooooooooooooooooooooong text";
                "another text";
                "test"] in
    printer # print_enum enum print_value;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after pretty print_enum"
      "\
[
 short text;
 loooooooooooooooooooooooooooooooooooooooooooooong text;
 another text;
 test
]
"
      (Buffer.contents buf);
;;

(* test suite *)
let suite = "Printing test" >:::
  ["test_print_value" >:: (bracket set_up test_print_value tear_down);
   "test_print_option" >:: (bracket set_up test_print_option tear_down);
   "test_print_object" >:: (bracket set_up test_print_object tear_down);
   "test_print_enum" >:: (bracket set_up test_print_enum tear_down);
   "test_print_field" >:: (bracket set_up test_print_field tear_down);
   "test_pretty_print_list" >:: test_pretty_print_list]
;;

