open OUnit

(* tests *)
let test_read_int _ =
  let reader = Xnc_reader.create_from_string "01abc" in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected int value"
      1
      (Xnc_reader.read_int reader);
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning int: \
          bad input at line 1, char 3: \
          'a' (unscanned buffer: \"abc\", EOF: false), \
          pattern not matched")
      (fun () -> Xnc_reader.read_int reader);
    Xnc_reader.close reader
;;

let test_read_float _ =
  let reader = Xnc_reader.create_from_string "3.1416\n3.\n3.5e-2" in
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected float value"
      3.1416
      (Xnc_reader.read_float reader);
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected float value"
      3.
      (Xnc_reader.read_float reader);
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected float value"
      0.035
      (Xnc_reader.read_float reader);
    Xnc_reader.close reader
;;

let test_read_char _ =
  let reader = Xnc_reader.create_from_string "s\n\n" in
    OUnit.assert_equal
      ~printer:Std.string_of_char
      ~msg:"Unexpected char value"
      's'
      (Xnc_reader.read_char reader);
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning char: \
          bad input at line 3, char 1: \
          '' (unscanned buffer: \"\", EOF: true), \
          pattern not matched")
      (fun () -> Xnc_reader.read_char reader);
    Xnc_reader.close reader
;;

let test_read_bool _ =
  let reader = Xnc_reader.create_from_string "true123\nfalsetrue\n\n" in
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected bool value"
      true
      (Xnc_reader.read_bool reader);
    Xnc_reader.skip reader 3;
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected bool value"
      false
      (Xnc_reader.read_bool reader);
    Xnc_reader.skip reader 4;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning bool: \
          bad input at line 4, char 1: \
          '' (unscanned buffer: \"\", EOF: true), \
          pattern not matched")
      (fun () -> Xnc_reader.read_bool reader);
    Xnc_reader.close reader
;;

let test_read_string _ =
  let config = { Xnc_serialization.default_config
                   with Xnc_serialization.delimiter_char = '#' } in
  let reader = Xnc_reader.create_from_string
                 ~config
                 "#hash#delimited#c\n#quoted\\#delimiter#ok#ko" in
  let reader' = Xnc_reader.create_from_string
                  ~config:{ config
                              with Xnc_serialization.escape_char = '%' }
                  "#%##\n##" in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string value"
      "hash"
      (Xnc_reader.read_string reader);
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning '#'-delimited string: \
          bad input at line 1, char 7: \
          'd' (unscanned buffer: \"delimited#c\n\
                                   #quoted\\#delimiter#ok#ko\", EOF: false), \
          pattern not matched")
      (fun () -> Xnc_reader.read_string reader);
    Xnc_reader.skip reader 11;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string value"
      "quoted#delimiter"
      (Xnc_reader.read_string reader);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string value"
      "#"
      (Xnc_reader.read_string reader');
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string value"
      ""
      (Xnc_reader.read_string reader');
    Xnc_reader.close reader;
    Xnc_reader.close reader'
;;

let test_read_date _ =
  let reader = Xnc_reader.create_from_string
                 "01/10/2005 09:00:10\n10/01/2005 00:09:10\n\n" in
    OUnit.assert_equal
      ~printer:(Netdate.format ~fmt:"%D %T")
      ~msg:"Unexpected date value"
      { Netdate.year = 2005;
        Netdate.month = 1;
        Netdate.day = 10;
        Netdate.hour = 9;
        Netdate.minute = 0;
        Netdate.second = 10;
        Netdate.zone = 0;
        Netdate.week_day = -1
      }
      (Xnc_reader.read_date reader);
    OUnit.assert_equal
      ~printer:(Netdate.format ~fmt:"%D %T")
      ~msg:"Unexpected date value"
      { Netdate.year = 2005;
        Netdate.month = 10;
        Netdate.day = 1;
        Netdate.hour = 0;
        Netdate.minute = 9;
        Netdate.second = 10;
        Netdate.zone = 0;
        Netdate.week_day = -1
      }
      (Xnc_reader.read_date reader);
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning date: \
          bad input at line 3, char 1: \
          '' (unscanned buffer: \"\", EOF: false), \
          expecting 19-character long string")
      (fun () -> Xnc_reader.read_date reader);
    Xnc_reader.close reader
;;

let test_read_value _ =
  let in_ch = new Netchannels.input_string "3\n5.10\n-\nfalse\n/test" in
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_netchannel in_ch in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
  let token_evaluator = new Xnc_token_eval.standard_evaluator () in
  let lexical_analyzer = new Xnc_lexing.xnc_scanbuf_lexer scanbuf in
  let reader =
    Xnc_reader.create scanner token_evaluator lexical_analyzer
  in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected int value"
      3
      (Xnc_reader.read_value reader (scanner # scan_int) int_of_string);
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected float value"
      5.1
      (Xnc_reader.read_value reader (scanner # scan_float) float_of_string);
    OUnit.assert_equal
      ~printer:Std.string_of_char
      ~msg:"Unexpected char value"
      '-'
      (Xnc_reader.read_value reader
         (fun () -> Xnc_scanning.Scanning_buffer.next_string scanbuf 1)
         (fun s -> String.get s 0));
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected bool value"
      false
      (Xnc_reader.read_value reader (scanner # scan_bool) bool_of_string);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string value"
      "/test"
      (Xnc_reader.read_value reader
         (fun () -> Xnc_scanning.Scanning_buffer.next_string scanbuf 5)
         Std.identity);
    Xnc_reader.close reader
;;

let test_read_option _ =
  let reader = Xnc_reader.create_from_string "None\n15" in
  let print_int_option = Serialization_test_helper.print_option
                           string_of_int in
    OUnit.assert_equal
      ~printer:print_int_option
      ~msg:"Unexpected int option value"
      None
      (Xnc_reader.read_option reader
         (fun () -> Xnc_reader.try_parse_null reader)
         (fun () -> Xnc_reader.read_int reader));
    OUnit.assert_equal
      ~printer:print_int_option
      ~msg:"Unexpected int option value"
      (Some 15)
      (Xnc_reader.read_option reader
         (fun () -> Xnc_reader.try_parse_null reader)
         (fun () -> Xnc_reader.read_int reader));
    Xnc_reader.close reader
;;

let test_read_object _ =
  let reader = Xnc_reader.create_from_string "\"3/4\"\n\"1/10\"" in
  let obj = new Serialization_test_helper.test_serializable 0 0 in
    Xnc_reader.read_object reader obj;
    OUnit.assert_equal
      ~cmp:Serialization_test_helper.cmp_object
      ~printer:Serialization_test_helper.print_object
      ~msg:"Unexpected object value"
      (new Serialization_test_helper.test_serializable 3 4)
      obj;
    Xnc_reader.read_object reader obj;
    OUnit.assert_equal
      ~cmp:Serialization_test_helper.cmp_object
      ~printer:Serialization_test_helper.print_object
      ~msg:"Unexpected object value"
      (new Serialization_test_helper.test_serializable 1 10)
      obj;
    Xnc_reader.close reader
;;

let test_read_homogeneous_enum _ =
  let reader = Xnc_reader.create_from_string "[1;2;3;4]" in
  let enum = Xnc_reader.read_homogeneous_enum reader
               (fun () -> Xnc_reader.read_int reader) in
    OUnit.assert_equal
      ~printer:(Serialization_test_helper.print_list string_of_int)
      ~msg:"Unexpected enum value"
      [1; 2; 3; 4]
      (ExtList.List.of_enum enum);
    Xnc_reader.close reader
;;

let test_read_homogeneous_list _ =
  let reader = Xnc_reader.create_from_string "[1;2;3;4]\n/1\n[1:" in
    OUnit.assert_equal
      ~printer:(Serialization_test_helper.print_list string_of_int)
      ~msg:"Unexpected list value"
      [1; 2; 3; 4]
      (Xnc_reader.read_homogeneous_collection reader
         (fun () -> Xnc_reader.read_int reader)
         ExtList.List.of_enum);
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 2, char 1: \
          '/' (unscanned buffer: \"/1\n[1:\", EOF: false), \
          expecting \";\"")
      (fun () -> Xnc_reader.read_homogeneous_collection reader
                   (fun () -> Xnc_reader.read_int reader)
                   ExtList.List.of_enum);
    Xnc_reader.skip reader 2;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 3, char 3: \
          ':' (unscanned buffer: \":\", EOF: false), \
          expecting \";\"")
      (fun () -> Xnc_reader.read_homogeneous_collection reader
                   (fun () -> Xnc_reader.read_int reader)
                   ExtList.List.of_enum);
    Xnc_reader.close reader
;;

let test_read_homogeneous_dllist _ =
  let reader = Xnc_reader.create_from_string "[1;2;3;4]" in
  let dllist = Xnc_reader.read_homogeneous_collection reader
                 (fun () -> Xnc_reader.read_int reader)
                 Dllist.of_enum in
    OUnit.assert_equal
      ~printer:(Serialization_test_helper.print_list string_of_int)
      ~msg:"Unexpected dllist value"
      [1; 2; 3; 4]
      (Dllist.to_list dllist);
    Xnc_reader.close reader
;;

let test_read_homogeneous_array _ =
  let reader = Xnc_reader.create_from_string "[1;2;3;4]\n|]1\n[1:" in
    OUnit.assert_equal
      ~printer:(Serialization_test_helper.print_array string_of_int)
      ~msg:"Unexpected array value"
      [|1; 2; 3; 4|]
      (Xnc_reader.read_homogeneous_collection reader
         (fun () -> Xnc_reader.read_int reader)
         ExtArray.Array.of_enum);
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 2, char 1: \
          '|' (unscanned buffer: \"|]1\n[1:\", EOF: false), \
          expecting \";\"")
      (fun () -> Xnc_reader.read_homogeneous_collection reader
                   (fun () -> Xnc_reader.read_int reader)
                   ExtArray.Array.of_enum);
    Xnc_reader.skip reader 3;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 3, char 3: \
          ':' (unscanned buffer: \":\", EOF: false), \
          expecting \";\"")
      (fun () -> Xnc_reader.read_homogeneous_collection reader
                   (fun () -> Xnc_reader.read_int reader)
                   ExtArray.Array.of_enum);
    Xnc_reader.close reader
;;

let test_read_homogeneous_dynarray _ =
  let reader = Xnc_reader.create_from_string "[1;2;3;4]" in
  let dynarray = Xnc_reader.read_homogeneous_collection reader
                   (fun () -> Xnc_reader.read_int reader)
                   DynArray.of_enum in
    OUnit.assert_equal
      ~printer:(Serialization_test_helper.print_array string_of_int)
      ~msg:"Unexpected array value"
      [|1; 2; 3; 4|]
      (DynArray.to_array dynarray);
    Xnc_reader.close reader
;;

type test_schema_type =
    Int of int
  | Bool of bool
  | Float of float

let test_read_heterogeneous_enum _ =
  let reader = Xnc_reader.create_from_string "[1;true;3.]" in
  let enum_schema = ExtList.List.enum [Int 0; Bool false; Float 0.] in
  let read_element e =
    match e with
        Int _ -> Int (Xnc_reader.read_int reader)
      | Bool _ -> Bool (Xnc_reader.read_bool reader)
      | Float _ -> Float (Xnc_reader.read_float reader) in
  let print_list =
    Serialization_test_helper.print_list
      (fun e -> match e with
           Int i -> string_of_int i
         | Bool b -> string_of_bool b
         | Float f -> string_of_float f) in
  let enum = Xnc_reader.read_heterogeneous_enum
               reader enum_schema read_element
  in
    OUnit.assert_equal
      ~printer:print_list
      ~msg:"Unexpected list value"
      [Int 1; Bool true; Float 3.]
      (ExtList.List.of_enum enum);
    Xnc_reader.close reader
;;

let test_read_heterogeneous_list _ =
  let reader = Xnc_reader.create_from_string "[1;true;3.]\n]1\n[1[" in
  let list_schema = [Int 0; Bool false; Float 0.] in
  let read_element e =
    match e with
        Int _ -> Int (Xnc_reader.read_int reader)
      | Bool _ -> Bool (Xnc_reader.read_bool reader)
      | Float _ -> Float (Xnc_reader.read_float reader) in
  let print_list =
    Serialization_test_helper.print_list
      (fun e -> match e with
           Int i -> string_of_int i
         | Bool b -> string_of_bool b
         | Float f -> string_of_float f)
  in
    OUnit.assert_equal
      ~printer:print_list
      ~msg:"Unexpected list value"
      [Int 1; Bool true; Float 3.]
      (Xnc_reader.read_heterogeneous_collection reader
         list_schema ExtList.List.enum
         read_element ExtList.List.of_enum);
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 2, char 1: \
          ']' (unscanned buffer: \"]1\n[1[\", EOF: false), \
          expecting \"[\"")
      (fun () -> Xnc_reader.read_heterogeneous_collection reader
                   list_schema ExtList.List.enum
                   read_element ExtList.List.of_enum);
    Xnc_reader.skip reader 2;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 3, char 3: \
          '[' (unscanned buffer: \"[\", EOF: false), \
          expecting \"]\"")
      (fun () -> Xnc_reader.read_heterogeneous_collection reader
                   list_schema ExtList.List.enum
                   read_element ExtList.List.of_enum);
    Xnc_reader.close reader
;;

let test_read_heterogeneous_dllist _ =
  let reader = Xnc_reader.create_from_string "[1;true;3.]\n]1\n[1[" in
  let dllist_schema = Dllist.of_list [Int 0; Bool false; Float 0.] in
  let read_element e =
    match e with
        Int _ -> Int (Xnc_reader.read_int reader)
      | Bool _ -> Bool (Xnc_reader.read_bool reader)
      | Float _ -> Float (Xnc_reader.read_float reader) in
  let print_list =
    Serialization_test_helper.print_list
      (fun e -> match e with
           Int i -> string_of_int i
         | Bool b -> string_of_bool b
         | Float f -> string_of_float f) in
  let dllist =
    Xnc_reader.read_heterogeneous_collection reader
      dllist_schema Dllist.enum
      read_element Dllist.of_enum
  in
    OUnit.assert_equal
      ~printer:print_list
      ~msg:"Unexpected list value"
      [Int 1; Bool true; Float 3.]
      (Dllist.to_list dllist);
    Xnc_reader.close reader
;;

let test_read_heterogeneous_array _ =
  let reader = Xnc_reader.create_from_string "[1;true;3.]\n|1\n[1[" in
  let array_schema = [|Int 0; Bool false; Float 0.|] in
  let read_element e =
    match e with
        Int _ -> Int (Xnc_reader.read_int reader)
      | Bool _ -> Bool (Xnc_reader.read_bool reader)
      | Float _ -> Float (Xnc_reader.read_float reader) in
  let print_array =
    Serialization_test_helper.print_array
      (fun e -> match e with
           Int i -> string_of_int i
         | Bool b -> string_of_bool b
         | Float f -> string_of_float f)
  in
    OUnit.assert_equal
      ~printer:print_array
      ~msg:"Unexpected array value"
      [|Int 1; Bool true; Float 3.|]
      (Xnc_reader.read_heterogeneous_collection reader
         array_schema ExtArray.Array.enum
         read_element ExtArray.Array.of_enum);
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 2, char 1: \
          '|' (unscanned buffer: \"|1\n[1[\", EOF: false), \
          expecting \"[\"")
      (fun () -> Xnc_reader.read_heterogeneous_collection reader
                   array_schema ExtArray.Array.enum
                   read_element ExtArray.Array.of_enum);
    Xnc_reader.skip reader 2;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 3, char 3: \
          '[' (unscanned buffer: \"[\", EOF: false), \
          expecting \"]\"")
      (fun () -> Xnc_reader.read_heterogeneous_collection reader
                   array_schema ExtArray.Array.enum
                   read_element ExtArray.Array.of_enum);
    Xnc_reader.close reader
;;

let test_read_heterogeneous_dynarray _ =
  let reader = Xnc_reader.create_from_string "[1;true;3.]" in
  let dynarray_schema = DynArray.of_array [|Int 0; Bool false; Float 0.|] in
  let read_element e =
    match e with
        Int _ -> Int (Xnc_reader.read_int reader)
      | Bool _ -> Bool (Xnc_reader.read_bool reader)
      | Float _ -> Float (Xnc_reader.read_float reader) in
  let print_array =
    Serialization_test_helper.print_array
      (fun e -> match e with
           Int i -> string_of_int i
         | Bool b -> string_of_bool b
         | Float f -> string_of_float f) in
  let dynarray =
    Xnc_reader.read_heterogeneous_collection reader
      dynarray_schema DynArray.enum
      read_element DynArray.of_enum
  in
    OUnit.assert_equal
      ~printer:print_array
      ~msg:"Unexpected array value"
      [|Int 1; Bool true; Float 3.|]
      (DynArray.to_array dynarray);
    Xnc_reader.close reader
;;

let test_read_value_by_name _ =
  let reader = Xnc_reader.create_from_string
                 "name=10\n\
                  @attribute=1.1\n\
                  xnc:prefixed=true\n\
                  @xnc:attr=x\n\
                  =xyz\nwrong=value" in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected value"
      10
      (Xnc_reader.read_value_by_name reader
         "" "name" Xnc_property.empty
         (fun () -> Xnc_reader.read_int reader));
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected value"
      1.1
      (Xnc_reader.read_value_by_name reader
         "" "attribute"
         (Xnc_property.singleton Xnc_property.attribute)
         (fun () -> Xnc_reader.read_float reader));
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected value"
      true
      (Xnc_reader.read_value_by_name reader
         "xnc" "prefixed" Xnc_property.empty
         (fun () -> Xnc_reader.read_bool reader));
    OUnit.assert_equal
      ~printer:Std.string_of_char
      ~msg:"Unexpected value"
      'x'
      (Xnc_reader.read_value_by_name reader
         "xnc" "attr"
         (Xnc_property.singleton Xnc_property.attribute)
         (fun () -> Xnc_reader.read_char reader));
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 5, char 1: \
          '=' (unscanned buffer: \"=xyz\nwrong=value\", EOF: false), \
          expecting \"name\"")
      (fun () -> Xnc_reader.read_value_by_name reader
                   "" "name" Xnc_property.empty
                   (fun () -> Xnc_reader.read_int reader));
    Xnc_reader.skip reader 4;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 6, char 1: \
          'w' (unscanned buffer: \"wrong=value\", EOF: false), \
          expecting \"name\"")
      (fun () -> Xnc_reader.read_value_by_name reader
                   "" "name" Xnc_property.empty
                   (fun () -> Xnc_reader.read_int reader));
;;

(* test suite *)
let suite = "Reader test" >:::
  ["test_read_int" >:: test_read_int;
   "test_read_float" >:: test_read_float;
   "test_read_char" >:: test_read_char;
   "test_read_bool" >:: test_read_bool;
   "test_read_string" >:: test_read_string;
   "test_read_value" >:: test_read_value;
   "test_read_option" >:: test_read_option;
   "test_read_object" >:: test_read_object;
   "test_read_homogeneous_enum" >:: test_read_homogeneous_enum;
   "test_read_homogeneous_list" >:: test_read_homogeneous_list;
   "test_read_homogeneous_dllist" >:: test_read_homogeneous_dllist;
   "test_read_homogeneous_array" >:: test_read_homogeneous_array;
   "test_read_homogeneous_dynarray" >:: test_read_homogeneous_dynarray;
   "test_read_heterogeneous_enum" >:: test_read_heterogeneous_enum;
   "test_read_heterogeneous_list" >:: test_read_heterogeneous_list;
   "test_read_heterogeneous_dllist" >:: test_read_heterogeneous_dllist;
   "test_read_heterogeneous_array" >:: test_read_heterogeneous_array;
   "test_read_heterogeneous_dynarray" >:: test_read_heterogeneous_dynarray;
   "test_read_value_by_name" >:: test_read_value_by_name]
;;

