open OUnit

(* tests *)
let test_scan_int _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "01\
                   -5\
                   +007\
                   abc" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected int token"
      "01"
      (scanner # scan_int ());
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected int token"
      "-5"
      (scanner # scan_int ());
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected int token"
      "+007"
      (scanner # scan_int ());
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning int: \
          bad input at line 1, char 9: \
          'a' (unscanned buffer: \"abc\", EOF: false), \
          pattern not matched")
      (scanner # scan_int);
;;
 
let test_scan_float _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "3.1416garbage\
                   3.abc\
                   3.5e-2\
                   -42.0\
                   +2.72\
                   pqrs" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected float token"
      "3.1416"
      (scanner # scan_float ());
    scanner # skip 7;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected float token"
      "3."
      (scanner # scan_float ());
    scanner # skip 3;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected float token"
      "3.5e-2"
      (scanner # scan_float ());
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected float token"
      "-42.0"
      (scanner # scan_float ());
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected float token"
      "+2.72"
      (scanner # scan_float ());
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning float: \
          bad input at line 1, char 35: \
          'p' (unscanned buffer: \"pqrs\", EOF: false), \
          pattern not matched")
      (scanner # scan_float);
;;

let test_scan_char _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "stringtest" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected char token"
      "s"
      (scanner # scan_char ());
    scanner # skip 9;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning char: \
          bad input at line 1, char 11: \
          '' (unscanned buffer: \"\", EOF: true), \
          pattern not matched")
      (scanner # scan_char)
;;

let test_scan_bool _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "true123\
                   falsetrue" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected bool token"
      "true"
      (scanner # scan_bool ());
    scanner # skip 3;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected bool token"
      "false"
      (scanner # scan_bool ());
    scanner # skip 4;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning bool: \
          bad input at line 1, char 17: \
          '' (unscanned buffer: \"\", EOF: true), \
          pattern not matched")
      (scanner # scan_bool)
;;

let test_scan_string _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "#hash#delimited\
                   ab#c\
                   #quoted\\#delimiter#\
                   #%##\
                   ##" in
  let config = { Xnc_serialization.default_config
                   with Xnc_serialization.delimiter_char = '#' } in
  let scanner = new Xnc_scanning.standard_scanner ~config scanbuf in
  let scanner' = new Xnc_scanning.standard_scanner
                   ~config:{ config
                               with Xnc_serialization.escape_char = '%' }
                   scanbuf in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string token"
      "hash"
      (scanner # scan_string ());
    scanner # skip 9;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning '#'-delimited string: \
          bad input at line 1, char 16: \
          'a' (unscanned buffer: \
                 \"ab#c#quoted\\#delimiter##%####\", EOF: false), \
          pattern not matched")
      (scanner # scan_string);
    scanner # skip 4;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string token"
      "quoted#delimiter"
      (scanner # scan_string ());
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string token"
      "#"
      (scanner' # scan_string ());
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string token"
      ""
      (scanner # scan_string ())
;;

let test_scan_date _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  "01/02/2006 10:11:12\
                   01/a" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected date token"
      "01/02/2006 10:11:12"
      (scanner # scan_date ());
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning date: \
          bad input at line 1, char 20: \
          '0' (unscanned buffer: \"01/a\", EOF: false), \
          expecting 19-character long string")
      (scanner # scan_date)
;;

let test_skip_whitespace _ =
  let scanbuf = Xnc_scanning.Scanning_buffer.create_from_string
                  ~skip_whitespace:true
                  "1  \
                   \"test_string\"\t\
                   5.0\n\
                   false\
                   t" in
  let scanner = new Xnc_scanning.standard_scanner scanbuf in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected int value"
      "1"
      (scanner # scan_int ());
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string value"
      "test_string"
      (scanner # scan_string ());
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected float value"
      "5.0"
      (scanner # scan_float ());
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected bool value"
      "false"
      (scanner # scan_bool ());
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected char value"
      "t"
      (scanner # scan_char ());
;;

(* test suite *)
let suite = "Scanning test" >:::
  ["test_scan_int" >:: test_scan_int;
   "test_scan_float" >:: test_scan_float;
   "test_scan_char" >:: test_scan_char;
   "test_scan_bool" >:: test_scan_bool;
   "test_scan_string" >:: test_scan_string;
   "test_scan_date" >:: test_scan_date;
   "test_skip_whitespace" >:: test_skip_whitespace]
;;

