open OUnit

module S = Xnc_serializable

(* tests *)
let test_create _ =
  let field = S.create
                ~properties:(Xnc_property.singleton Xnc_property.attribute)
                ~prefix:"xnc"
                "Int_attribute"
                (S.Int 3) in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected field name"
      "Int_attribute"
      (S.get_name field);
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected field value"
      3
      (S.unwrap_int (S.get_value field));
    OUnit.assert_equal
      ~msg:"Unexpected field properties"
      (Xnc_property.singleton Xnc_property.attribute)
      (S.get_properties field);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected field prefix"
      "xnc"
      (S.get_prefix field);
;;

let test_new_field _ =
  let field = S.new_field "Int_field" 1 S.wrap_int in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected field value"
      1
      (S.get_int field);
  let field = S.new_field "Float_field" 3.14 S.wrap_float in
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected field value"
      3.14
      (S.get_float field);
  let field = S.new_field "Char_field" 'a' S.wrap_char in
    OUnit.assert_equal
      ~printer:Std.string_of_char
      ~msg:"Unexpected character"
      'a'
      (S.get_char field);
  let field = S.new_field "Bool_field" true S.wrap_bool in
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected field value"
      true
      (S.get_bool field);
  let field = S.new_field "String_field" "test" S.wrap_string in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected field value"
      "test"
      (S.get_string field);
  let field = S.new_field "Date_field"
                { Netdate.year = 2006;
                  Netdate.month = 10;
                  Netdate.day = 22;
                  Netdate.hour = 16;
                  Netdate.minute = 46;
                  Netdate.second = 30;
                  Netdate.zone = 0;
                  Netdate.week_day = -1
                } S.wrap_date
  in
    OUnit.assert_equal
      ~printer:(Netdate.format ~fmt:"%D %T")
      ~msg:"Unexpected field value"
      { Netdate.year = 2006;
        Netdate.month = 10;
        Netdate.day = 22;
        Netdate.hour = 16;
        Netdate.minute = 46;
        Netdate.second = 30;
        Netdate.zone = 0;
        Netdate.week_day = -1
      }
      (S.get_date field);
  let field = S.new_field "Option_field" (Some (S.Int 1)) S.wrap_option in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected field value"
      1
      (S.unwrap_int (Option.get (S.get_option field)));
  let o = new Serialization_test_helper.test_serializable 0 0 in
  let field = S.new_field "Object_field" o S.wrap_object in
    OUnit.assert_equal
      ~cmp:(fun o1 o2 ->
              Serialization_test_helper.print_object o1 =
                Serialization_test_helper.print_object o2)
      ~printer:Serialization_test_helper.print_object
      ~msg:"Unexpected field value"
      (new Serialization_test_helper.test_serializable 0 0
         :> Xnc_serialization.serializable)
      (S.get_object field)
;;

let test_new_list_field _ =
  let int_field = S.new_field "Int_field" 1 S.wrap_int in
  let float_field = S.new_field "Float_field" 3.14 S.wrap_float in
  let char_field = S.new_field "Char_field" 'a' S.wrap_char in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let bool_field = S.new_field "Bool_field" true S.wrap_bool in
  let field = S.new_field "List_field"
                [int_field;
                 float_field;
                 char_field;
                 string_field;
                 bool_field] S.wrap_list in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected list length"
      5
      (S.count_leaves field)
;;

let test_new_array_field _ =
  let int_field = S.new_field "Int_field" 2 S.wrap_int in
  let float_field = S.new_field "Float_field" 6.28 S.wrap_float in
  let char_field = S.new_field "Char_field" 'b' S.wrap_char in
  let bool_field = S.new_field "Bool_field" false S.wrap_bool in
  let field = S.new_field "Array_field"
                [|int_field;
                  float_field;
                  char_field;
                  bool_field|] S.wrap_array in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected array length"
      4
      (S.count_leaves field)
;;

let test_equal _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let float_field = S.new_field "Float_field" 5. S.wrap_float in
  let bool_field = S.new_field "Bool_field" true S.wrap_bool in
  let array_field = S.new_field "Array_field"
                      [|int_field; bool_field|] S.wrap_array in
  let field = S.new_field "List_field"
                [int_field; string_field; float_field; array_field] S.wrap_list in
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"A field should be equal to itself"
      true
      (S.equal field field);
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Should be field <> array_field"
      false
      (S.equal field array_field);
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"A field should be equal to a field with the same name and value"
      true
      (S.equal field
         (S.new_field "List_field"
            [ S.new_field "Int_field" 10 S.wrap_int;
              S.new_field "String_field" "test" S.wrap_string;
              S.new_field "Float_field" 5. S.wrap_float;
              S.new_field "Array_field"
                 [| S.new_field "Int_field" 10 S.wrap_int;
                    S.new_field "Bool_field" true S.wrap_bool
                 |] S.wrap_array
            ] S.wrap_list
         )
      );
;;

let test_clone _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let float_field = S.new_field "Float_field" 5. S.wrap_float in
  let bool_field = S.new_field "Bool_field" true S.wrap_bool in
  let array_field = S.new_field "Array_field"
                      [|int_field; bool_field|] S.wrap_array in
  let object_field =
    S.new_field "Object_field"
      (new Serialization_test_helper.test_serializable 8 9) S.wrap_object in
  let field = S.new_field "List_field"
                [int_field;
                 string_field;
                 float_field;
                 array_field;
                 object_field] S.wrap_list in
  let field' = S.clone_field field in
  let array_field' = S.find_field field' "Array_field" in
  let object_field' = S.find_field field' "Object_field" in
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Cloned field should be equal to its original"
      true
      (S.custom_equal
         ~obj_eq:(fun o1 o2 ->
                    (o1 # to_string ()) = (o2 # to_string ()))
         field field');
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Cloned field should be physically different from its original"
      false
      (field == field');
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Cloned array should be physically different from its original"
      false
      (array_field == array_field');
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Cloned object should be physically different from its original"
      false
      (object_field == object_field');
;;

let test_append_field _ =
  let field = S.create_empty_list "List_field" in
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let float_field = S.new_field "Float_field" 5. S.wrap_float in
  let bool_field = S.new_field "Bool_field" true S.wrap_bool in
  let array_field = S.create_empty_array "Array_field" in
    S.append_field array_field int_field;
    S.append_field array_field float_field;
    S.append_field field int_field;
    S.append_field field string_field;
    S.append_field field array_field;
    S.append_field field float_field;
    S.append_field field bool_field;
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected field content after appending 4 fields"
      (S.new_field "List_field"
         [ S.new_field "Int_field" 10 S.wrap_int;
           S.new_field "String_field" "test" S.wrap_string;
           S.new_field "Array_field"
             [| S.new_field "Int_field" 10 S.wrap_int;
                S.new_field "Float_field" 5. S.wrap_float
             |] S.wrap_array;
           S.new_field "Float_field" 5. S.wrap_float;
           S.new_field "Bool_field" true S.wrap_bool
         ] S.wrap_list
      )
      field;
;;

let test_replace_field _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let new_int_field = S.new_field "Int_field" 15 S.wrap_int in
  let array_field = S.new_field "Array_field"
                      [|int_field; string_field|] S.wrap_array in
  let new_string_field = S.new_field "String_field" "new" S.wrap_string in
  let field = S.new_field "List_field"
                [array_field; int_field; string_field] S.wrap_list in
    S.replace_field field new_int_field;
    S.replace_field array_field new_string_field;
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected field content after 2 replace_field"
      (S.new_field "List_field"
         [ S.new_field "Array_field"
             [| S.new_field "Int_field" 10 S.wrap_int;
                S.new_field "String_field" "new" S.wrap_string
             |] S.wrap_array;
           S.new_field "Int_field" 15 S.wrap_int;
           S.new_field "String_field" "test" S.wrap_string
         ] S.wrap_list
      )
      field;
;;

let test_iter_leaves _ =
  let int_field = S.new_field "Field" 10 S.wrap_int in
  let char_field = S.new_field "Field" 'a' S.wrap_char in
  let string_field = S.new_field "Field" "test" S.wrap_string in
  let bool_field = S.new_field "Field" false S.wrap_bool in
  let inner_field = S.new_field "Inner_array_field"
                      [|int_field; char_field|] S.wrap_array in
  let inner_field' = S.new_field "Second_inner_list_field"
                       [string_field; bool_field] S.wrap_list in
  let field = S.new_field "List_field"
                [inner_field; inner_field'] S.wrap_list in
  let default_int_field = S.create "Field"
                            S.default_int in
    S.iter ~traversal:S.Leaves_only
      (fun e ->
         S.set_value e S.default_int)
      field;
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected result after leaf iteration"
      (S.new_field "List_field"
         [ S.new_field "Inner_array_field"
             [| default_int_field; default_int_field |] S.wrap_array;
           S.new_field "Second_inner_list_field"
             [ default_int_field; default_int_field ] S.wrap_list
         ] S.wrap_list
      )
      field;
;;

let test_iter_pre _ =
  let int_field = S.new_field "Int" 10 S.wrap_int in
  let char_field = S.new_field "Char" 'a' S.wrap_char in
  let string_field = S.new_field "String" "test" S.wrap_string in
  let bool_field = S.new_field "Bool" false S.wrap_bool in
  let inner_field = S.new_field "Array"
                      [|int_field; char_field|] S.wrap_array in
  let inner_field' = S.new_field "List2"
                       [string_field; bool_field] S.wrap_list in
  let field = S.new_field "List"
                [inner_field; inner_field'] S.wrap_list in
  let buf = Buffer.create 128 in
    S.iter ~traversal:S.Pre_order
      (fun e ->
         let name = S.get_name e in
           Buffer.add_string buf name;
           Buffer.add_char buf ' ')
      field;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected result after pre-order iteration"
      "List Array Int Char List2 String Bool "
      (Buffer.contents buf);
;;

let test_iter_post _ =
  let int_field = S.new_field "Int" 10 S.wrap_int in
  let char_field = S.new_field "Char" 'a' S.wrap_char in
  let string_field = S.new_field "String" "test" S.wrap_string in
  let bool_field = S.new_field "Bool" false S.wrap_bool in
  let inner_field = S.new_field "Array"
                      [|int_field; char_field|] S.wrap_array in
  let inner_field' = S.new_field "List2"
                       [string_field; bool_field] S.wrap_list in
  let field = S.new_field "List"
                [inner_field; inner_field'] S.wrap_list in
  let buf = Buffer.create 128 in
    S.iter ~traversal:S.Post_order
      (fun e ->
         let name = S.get_name e in
           Buffer.add_string buf name;
           Buffer.add_char buf ' ')
      field;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected result after post-order iteration"
      "Int Char Array String Bool List2 List "
      (Buffer.contents buf);
;;

let test_fold_leaves _ =
  let int_field = S.new_field "Int" 10 S.wrap_int in
  let char_field = S.new_field "Char" 'a' S.wrap_char in
  let string_field = S.new_field "String" "test" S.wrap_string in
  let bool_field = S.new_field "Bool" false S.wrap_bool in
  let inner_field = S.new_field "Array"
                      [|int_field; char_field|] S.wrap_array in
  let inner_field' = S.new_field "List2"
                       [string_field; bool_field] S.wrap_list in
  let field = S.new_field "List"
                [inner_field; inner_field'] S.wrap_list in
  let s =
    S.fold ~traversal:S.Leaves_only
      (fun accu e ->
         let name = S.get_name e in
           accu ^ name ^ " ")
      ""
      field
  in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected result after leaf folding"
      "Int Char String Bool "
      s;
;;

let test_fold_pre _ =
  let int_field = S.new_field "Int" 10 S.wrap_int in
  let char_field = S.new_field "Char" 'a' S.wrap_char in
  let string_field = S.new_field "String" "test" S.wrap_string in
  let bool_field = S.new_field "Bool" false S.wrap_bool in
  let inner_field = S.new_field "Array"
                      [|int_field; char_field|] S.wrap_array in
  let inner_field' = S.new_field "List2"
                       [string_field; bool_field] S.wrap_list in
  let field = S.new_field "List"
                [inner_field; inner_field'] S.wrap_list in
  let s =
    S.fold ~traversal:S.Pre_order
      (fun accu e ->
         let name = S.get_name e in
           accu ^ name ^ " ")
      ""
      field
  in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected result after pre-order folding"
      "List Array Int Char List2 String Bool "
      s
;;

let test_fold_post _ =
  let int_field = S.new_field "Int" 10 S.wrap_int in
  let char_field = S.new_field "Char" 'a' S.wrap_char in
  let string_field = S.new_field "String" "test" S.wrap_string in
  let bool_field = S.new_field "Bool" false S.wrap_bool in
  let inner_field = S.new_field "Array"
                      [|int_field; char_field|] S.wrap_array in
  let inner_field' = S.new_field "List2"
                       [string_field; bool_field] S.wrap_list in
  let field = S.new_field "List"
                [inner_field; inner_field'] S.wrap_list in
  let s =
    S.fold ~traversal:S.Post_order
      (fun accu e ->
         let name = S.get_name e in
           accu ^ name ^ " ")
      ""
      field
  in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected result after pre-order folding"
      "Int Char Array String Bool List2 List "
      s
;;

let test_map_leaves _ =
  let int_field = S.new_field "Int" 10 S.wrap_int in
  let char_field = S.new_field "Char" 'a' S.wrap_char in
  let string_field = S.new_field "String" "test" S.wrap_string in
  let bool_field = S.new_field "Bool" false S.wrap_bool in
  let inner_field = S.new_field "Array"
                      [|int_field; char_field|] S.wrap_array in
  let inner_field' = S.new_field "List2"
                       [string_field; bool_field] S.wrap_list in
  let field = S.new_field "List"
                [inner_field; inner_field'] S.wrap_list in
  let new_leaf = S.new_field "Bool" true S.wrap_bool in
  let field' =
    S.map ~traversal:S.Leaves_only
      (fun e -> new_leaf)
      field
  in
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected result after leaf mapping"
      (S.new_field "List"
         [ S.new_field "Array"
             [| new_leaf; new_leaf |] S.wrap_array;
           S.new_field "List2"
             [ new_leaf; new_leaf ] S.wrap_list
         ] S.wrap_list
      )
      field';
;;

let test_map_pre _ =
  let int_field = S.new_field "Int" 10 S.wrap_int in
  let char_field = S.new_field "Char" 'a' S.wrap_char in
  let string_field = S.new_field "String" "test" S.wrap_string in
  let bool_field = S.new_field "Bool" false S.wrap_bool in
  let inner_field = S.new_field "Array"
                      [|int_field; char_field|] S.wrap_array in
  let inner_field' = S.new_field "List2"
                       [string_field; bool_field] S.wrap_list in
  let field = S.new_field "List"
                [inner_field; inner_field'] S.wrap_list in
  let field' =
    S.map ~traversal:S.Pre_order
      (fun e ->
         let v = S.get_value e in
         let n =
           match v with
               S.List l ->
                 string_of_bool
                   (List.for_all
                      (fun e ->
                         (S.get_name e) = "Mapped") l)
             | _ -> "Mapped"
         in
           S.create n v
      )
      field
  in
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected result after pre-order mapping"
      (S.new_field "false"
         [ S.new_field "Mapped"
             [| S.new_field "Mapped" 10 S.wrap_int;
                S.new_field "Mapped" 'a' S.wrap_char |] S.wrap_array;
           S.new_field "false"
             [ S.new_field "Mapped" "test" S.wrap_string;
               S.new_field "Mapped" false S.wrap_bool ] S.wrap_list
         ] S.wrap_list
      )
      field';
;;

let test_map_post _ =
  let int_field = S.new_field "Int" 10 S.wrap_int in
  let char_field = S.new_field "Char" 'a' S.wrap_char in
  let string_field = S.new_field "String" "test" S.wrap_string in
  let bool_field = S.new_field "Bool" false S.wrap_bool in
  let inner_field = S.new_field "Array"
                      [|int_field; char_field|] S.wrap_array in
  let inner_field' = S.new_field "List2"
                       [string_field; bool_field] S.wrap_list in
  let field = S.new_field "List"
                [inner_field; inner_field'] S.wrap_list in
  let field' =
    S.map ~traversal:S.Post_order
      (fun e ->
         let v = S.get_value e in
         let n =
           match v with
               S.List l ->
                 string_of_bool
                   (List.for_all
                      (fun e ->
                         (S.get_name e) = "Mapped") l)
             | _ -> "Mapped"
         in
           S.create n v
      )
      field
  in
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected result after post-order mapping"
      (S.new_field "false"
         [ S.new_field "Mapped"
             [| S.new_field "Mapped" 10 S.wrap_int;
                S.new_field "Mapped" 'a' S.wrap_char |] S.wrap_array;
           S.new_field "true"
             [ S.new_field "Mapped" "test" S.wrap_string;
               S.new_field "Mapped" false S.wrap_bool ] S.wrap_list
         ] S.wrap_list
      )
      field';
;;

let test_to_stream _ =
  let field = 
    (S.new_field "List_field"
       [ S.new_field "Inner_list_field"
           [ S.new_field "Int_field" 10 S.wrap_int;
             S.new_field "Float_field" 2.72 S.wrap_float
           ] S.wrap_list;
         S.new_field "Second_inner_list_field"
           [ S.new_field "String_field" "test" S.wrap_string;
             S.new_field "Bool_field" true S.wrap_bool
           ] S.wrap_list
       ] S.wrap_list
    ) in
  let buf = Buffer.create 128 in
  let stream = S.to_stream field in
    Stream.iter
      (fun field ->
         Buffer.add_string buf (S.get_name field))
      stream;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected stream content"
      "List_field\
       Inner_list_field\
       Int_field\
       Float_field\
       Second_inner_list_field\
       String_field\
       Bool_field"
      (Buffer.contents buf);
;;

let test_to_name_value_stream _ =
  let field = 
    (S.new_field "List_field"
       [ S.new_field "Inner_list_field"
           [ S.new_field
               ~properties:(Xnc_property.singleton Xnc_property.attribute)
               "Int_field" 10 S.wrap_int;
             S.new_field
               ~prefix:"xnc"
               "Float_field" 2.72 S.wrap_float
           ] S.wrap_list;
         S.new_field "Second_inner_list_field"
           [ S.new_field
               ~prefix:"xnc"
               ~properties:(Xnc_property.singleton Xnc_property.attribute)
               "String_field" "test" S.wrap_string;
             S.new_field "Bool_field" true S.wrap_bool
           ] S.wrap_list
       ] S.wrap_list
    ) in
  let buf = Buffer.create 128 in
  let stream = S.to_name_value_stream field in
    Stream.iter
      (fun (name, _) ->
         Buffer.add_string buf name)
      stream;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected stream content"
      "/List_field\
       /List_field/Inner_list_field\
       /List_field/Inner_list_field/@Int_field\
       /List_field/Inner_list_field/xnc:Float_field\
       /List_field/Second_inner_list_field\
       /List_field/Second_inner_list_field/@xnc:String_field\
       /List_field/Second_inner_list_field/Bool_field"
      (Buffer.contents buf);
;;

let test_value_present _ =
  let field = S.new_field "Int_field" 1 S.wrap_int in
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Expected field not found"
      true
      (S.value_present field "Int_field");
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected field found"
      false
      (S.value_present field "Not present")
;;

let test_value_present_in_list _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let inner_field = S.new_field "Inner_list_field"
                      [int_field; string_field] S.wrap_list in
  let field = S.new_field "List_field"
                [inner_field] S.wrap_list in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected list length"
      2
      (S.count_leaves field);
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Expected field not found"
      true
      (S.value_present field "String_field");
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected field found"
      false
      (S.value_present field "Not_present")
;;

let test_value_present_in_array _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let inner_field = S.new_field "Inner_array_field"
                      [|int_field; string_field|] S.wrap_array in
  let field = S.new_field "Array_field"
                [|inner_field|] S.wrap_array in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected array length"
      2
      (S.count_leaves field);
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Expected field not found"
      true
      (S.value_present field "String_field");
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected field found"
      false
      (S.value_present field "Not_present")
;;

let test_find_value _ =
  let field = S.new_field "Int_field" 1 S.wrap_int in
  let field' = S.new_field ~prefix:"xnc" "Bool_field" false S.wrap_bool in
  let field'' = S.new_field
                  ~properties:(Xnc_property.singleton Xnc_property.attribute)
                  "String_field" "attr" S.wrap_string in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Int_field not found"
      1
      (S.unwrap_int
         (S.find_value field "Int_field"));
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"xnc:Bool_field not found"
      false
      (S.unwrap_bool
         (S.find_value ~prefix:"xnc" field' "Bool_field"));
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"String_field not found"
      "attr"
      (S.unwrap_string
         (S.find_value
            ~properties:(Xnc_property.singleton Xnc_property.attribute)
            field'' "String_field"));
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"String_field attribute not found"
      "attr"
      (S.unwrap_string
         (S.find_value field'' "String_field"));
    OUnit.assert_raises
      ~msg:"Not_present found"
      Not_found
      (fun () -> S.find_value field "Not_present");
    OUnit.assert_raises
      ~msg:"xnc:Int_field found"
      Not_found
      (fun () -> S.find_value ~prefix:"xnc" field "Int_value");
    OUnit.assert_raises
      ~msg:"String_field element found"
      Not_found
      (fun () -> S.find_value 
                   ~properties:(Xnc_property.singleton Xnc_property.element)
                   field'' "String_field");
;;

let test_find_value_in_list _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let string_field = S.new_field
                       ~prefix:"xnc" "String_field" "test" S.wrap_string in
  let inner_field = S.new_field "Inner_list_field"
                      [int_field] S.wrap_list in
  let inner_field' = S.new_field "Second_inner list_field"
                       [string_field] S.wrap_list in
  let field = S.new_field "List_field"
                [inner_field; inner_field'] S.wrap_list in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Int_field not found"
      10
      (S.unwrap_int
         (S.find_value field "Int_field"));
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"xnc:String_field not found"
      "test"
      (S.unwrap_string
         (S.find_value ~prefix:"xnc" field "String_field"));
    OUnit.assert_raises
      ~msg:"Not_present found"
      Not_found
      (fun () -> S.find_value field "Not_present");
    OUnit.assert_raises
      ~msg:"xnc:Int_field found"
      Not_found
      (fun () -> S.find_value ~prefix:"xnc" field "Int_field");
    OUnit.assert_raises
      ~msg:"Int_field attribute found"
      Not_found
      (fun () -> S.find_value
                   ~properties:(Xnc_property.singleton Xnc_property.attribute)
                   ~prefix:"xnc"
                   field "Int_field");
;;

let test_find_value_in_array _ =
  let int_field = S.new_field ~prefix:"xnc" "Int_field" 10 S.wrap_int in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let inner_field = S.new_field "Inner_array_field"
                      [|int_field|] S.wrap_array in
  let inner_field' = S.new_field "Second_inner array_field"
                       [|string_field|] S.wrap_array in
  let field = S.new_field "Array_field"
                [|inner_field; inner_field'|] S.wrap_array in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"xnc:Int_field not found"
      10
      (S.unwrap_int
         (S.find_value ~prefix:"xnc" field "Int_field"));
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"String_field not found"
      "test"
      (S.unwrap_string
         (S.find_value field "String_field"));
    OUnit.assert_raises
      ~msg:"Not_present found"
      Not_found
      (fun () -> S.find_value field "Not_present")
;;

let test_find_all_values _ =
  let int_field = S.new_field ~prefix:"xnc" "Field" 10 S.wrap_int in
  let char_field = S.new_field ~prefix:"xnc" "Field" 'a' S.wrap_char in
  let string_field = S.new_field "Field" "test" S.wrap_string in
  let bool_field = S.new_field ~prefix:"xnc" "Field" false S.wrap_bool in
  let inner_field = S.new_field "Inner_array_field"
                      [|int_field; char_field|] S.wrap_array in
  let inner_field' = S.new_field "Second_inner_list_field"
                       [string_field; bool_field] S.wrap_list in
  let field = S.new_field "List_field"
                [inner_field; inner_field'] S.wrap_list in
  let found_list = S.find_all_values
                     ~prefix:"xnc" field "Field" in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"First xnc:Field (int) not found"
      10
      (S.unwrap_int
         (List.nth found_list 0));
    OUnit.assert_equal
      ~printer:Std.string_of_char
      ~msg:"Second xnc:Field (char) not found"
      'a'
      (S.unwrap_char
         (List.nth found_list 1));
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Third xnc:Field (bool) not found"
      false
      (S.unwrap_bool
         (List.nth found_list 2));
;;

let test_write_value _ =
  let buf = Buffer.create 128 in
  let writer = Xnc_writer.create_from_buffer buf in
  let int_value = S.wrap_int 1 in
  let float_value = S.wrap_float 3.141592 in
  let date_value = S.wrap_date
                     { Netdate.year = 2006;
                       Netdate.month = 10;
                       Netdate.day = 22;
                       Netdate.hour = 16;
                       Netdate.minute = 46;
                       Netdate.second = 30;
                       Netdate.zone = 0;
                       Netdate.week_day = -1
                     } in
  let option_value = S.wrap_option None in
  let option_value' = S.wrap_option
                        (Some (S.Bool false)) in
  let object_value =
    S.wrap_object
      (new Serialization_test_helper.test_serializable 1 2)
  in
    S.write_value writer int_value;
    S.write_value writer float_value;
    S.write_value writer date_value;
    S.write_value writer option_value;
    S.write_value writer option_value';
    S.write_value writer object_value;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after 4 write_value"
      "13.14159210/22/2006 16:46:30Nonefalse\"1/2\""
      (Buffer.contents buf)
;;

let test_write_simple_field _ =
  let buf = Buffer.create 128 in
  let writer = Xnc_writer.create_from_buffer buf in
  let int_field = S.new_field "Int_field" 1 S.wrap_int in
  let float_field = S.new_field "Float_field" 3.141592 S.wrap_float in
  let none_field = S.new_field "Null" None S.wrap_option in
  let bool_option_field = S.new_field "Option"
                            (Some (S.Bool false))
                            S.wrap_option in
  let object_field =
    S.new_field "Object_field"
      (new Serialization_test_helper.test_serializable 8 9) S.wrap_object
  in
    S.write writer int_field;
    S.write writer float_field;
    S.write writer none_field;
    S.write writer bool_option_field;
    S.write writer object_field;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer contents after writing "
      "Int_field=1Float_field=3.141592Null=NoneOption=falseObject_field=\"8/9\""
      (Buffer.contents buf);
    Xnc_writer.close writer
;;

let test_write_list_field _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let float_field = S.new_field "Float_field" 2.72 S.wrap_float in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let bool_field = S.new_field "Bool_field" true S.wrap_bool in
  let inner_field = S.new_field "Inner_list_field"
                      [int_field; float_field] S.wrap_list in
  let inner_field' = S.new_field "Second_inner_list_field"
                       [string_field; bool_field] S.wrap_list in
  let list_field = S.new_field "List_field"
                     [inner_field; inner_field'] S.wrap_list in
  let buf = Buffer.create 128 in
  let writer = Xnc_writer.create_from_buffer buf in
    S.write writer list_field;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string serialization"
      "List_field=[\
         Inner_list_field=[\
            Int_field=10;\
            Float_field=2.72\
         ];\
         Second_inner_list_field=[\
            String_field=\"test\";\
            Bool_field=true\
         ]\
       ]"
      (Buffer.contents buf);
    Xnc_writer.close writer
;;

let test_write_array_field _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let float_field = S.new_field "Float_field" 2.72 S.wrap_float in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let bool_field = S.new_field "Bool_field" true S.wrap_bool in
  let inner_field = S.new_field "Inner_array_field"
                      [|int_field; float_field|] S.wrap_array in
  let inner_field' = S.new_field "Second_inner_array_field"
                       [|string_field; bool_field|] S.wrap_array in
  let array_field = S.new_field "Array_field"
                     [|inner_field; inner_field'|] S.wrap_array in
  let buf = Buffer.create 128 in
  let writer = Xnc_writer.create_from_buffer buf in
    S.write writer array_field;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string serialization"
      "Array_field=[\
         Inner_array_field=[\
            Int_field=10;\
            Float_field=2.72\
         ];\
         Second_inner_array_field=[\
            String_field=\"test\";\
            Bool_field=true\
         ]\
       ]"
      (Buffer.contents buf);
    Xnc_writer.close writer
;;

let test_write_using_pretty_printer _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let float_field = S.new_field "Float_field" 2.72 S.wrap_float in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let bool_field = S.new_field "Bool_field" true S.wrap_bool in
  let inner_field = S.new_field "Inner_list_field"
                      [int_field; float_field] S.wrap_list in
  let inner_field' = S.new_field "Second_inner_list_field"
                       [string_field; bool_field] S.wrap_list in
  let list_field = S.new_field "List_field"
                     [inner_field; inner_field'] S.wrap_list in
  let buf = Buffer.create 128 in
  let formatter = new Xnc_formatting.standard_formatter () in
  let printer = new Xnc_printing.Buffer_pretty_printer.default_printer buf in
  let writer = Xnc_writer.create formatter printer in
    S.write writer list_field;
    Xnc_writer.close writer;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected pretty-printed serialization"
      "\
List_field=[
            Inner_list_field=[Int_field=10;Float_field=2.72];
            Second_inner_list_field=[String_field=\"test\";Bool_field=true]
           ]
"
      (Buffer.contents buf);
;;

let test_write_using_xml_printer _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let float_field = S.new_field "Float_field" 2.72 S.wrap_float in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let bool_field = S.new_field
                     ~properties:(Xnc_property.singleton Xnc_property.attribute)
                     "Bool_field" true S.wrap_bool in
  let inner_field = S.new_field "Inner_list_field"
                      [int_field; float_field] S.wrap_list in
  let inner_field' = S.new_field "Second_inner_list_field"
                       [string_field; bool_field] S.wrap_list in
  let list_field = S.new_field ~prefix:"xnc" "List_field"
                     [inner_field; inner_field'] S.wrap_list in
  let buf = Buffer.create 128 in
  let formatter = new Xnc_formatting.xml_formatter () in
  let printer = new Xnc_printing.Buffer_printer.xml_printer buf in
  let writer = Xnc_writer.create formatter printer in
    S.write writer list_field;
    Xnc_writer.close writer;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected xml serialization"
      "<xnc:List_field>\
         <Inner_list_field>\
           <Int_field>10</Int_field>\
           <Float_field>2.72</Float_field>\
         </Inner_list_field>\
         <Second_inner_list_field Bool_field=\"true\">\
           <String_field>test</String_field>\
         </Second_inner_list_field>\
       </xnc:List_field>"
      (Buffer.contents buf);
;;

let test_write_using_xml_pretty_printer _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let float_field = S.new_field "Float_field" 3.141592 S.wrap_float in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let bool_field = S.new_field
                     ~properties:(Xnc_property.singleton Xnc_property.attribute)
                     "Bool_field" true S.wrap_bool in
  let inner_field = S.new_field "Inner_list_field"
                      [int_field; float_field] S.wrap_list in
  let inner_field' = S.new_field "Second_inner_list_field"
                       [string_field; bool_field] S.wrap_list in
  let list_field = S.new_field ~prefix:"xnc" "List_field"
                     [inner_field; inner_field'] S.wrap_list in
  let buf = Buffer.create 128 in
  let formatter = new Xnc_formatting.xml_formatter () in
  let printer = new Xnc_printing.Buffer_pretty_printer.xml_printer buf in
  let writer = Xnc_writer.create formatter printer in
    S.write writer list_field;
    Xnc_writer.close writer;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected xml serialization"
      "\
<xnc:List_field>
  <Inner_list_field>
    <Int_field>10</Int_field>
    <Float_field>3.141592</Float_field>
  </Inner_list_field>
  <Second_inner_list_field Bool_field=\"true\">
    <String_field>test</String_field>
  </Second_inner_list_field>
</xnc:List_field>
"
      (Buffer.contents buf);
;;

let test_read_value _ =
  let ch = new Netchannels.input_string
             "01\n3.1416\n10/22/2006 16:46:30\nNone\nfalse\"5/5\"" in
  let reader = Xnc_reader.create_from_netchannel ch in
  let int_value = S.default_int in
  let float_value = S.default_float in
  let date_value = S.default_date in
  let int_option_value = S.default_option
                           S.default_int in
  let bool_option_value = S.default_option
                            S.default_bool in
  let object_value =
    S.wrap_object
      (new Serialization_test_helper.test_serializable 0 0) in
  let print_serializable_option = Serialization_test_helper.print_option
                                    S.string_of_value in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected int deserialization"
      1
      (S.unwrap_int
         (S.read_value
            reader
            int_value));
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected float deserialization"
      3.1416
      (S.unwrap_float
         (S.read_value
            reader
            float_value));
    OUnit.assert_equal
      ~printer:(Netdate.format ~fmt:"%D %T")
      ~msg:"Unexpected date deserialization"
      { Netdate.year = 2006;
        Netdate.month = 10;
        Netdate.day = 22;
        Netdate.hour = 16;
        Netdate.minute = 46;
        Netdate.second = 30;
        Netdate.zone = 0;
        Netdate.week_day = -1
      }
      (S.unwrap_date
         (S.read_value
            reader
            date_value));
    OUnit.assert_equal
      ~printer:print_serializable_option
      ~msg:"Unexpected int option deserialization"
      None
      (S.unwrap_option
         (S.read_value
            reader
            int_option_value));
    OUnit.assert_equal
      ~printer:print_serializable_option
      ~msg:"Unexpected int option deserialization"
      (Some (S.Bool false))
      (S.unwrap_option
         (S.read_value
            reader
            bool_option_value));
    OUnit.assert_equal
      ~cmp:(fun o1 o2 ->
              Serialization_test_helper.print_object o1 =
                Serialization_test_helper.print_object o2)
      ~printer:Serialization_test_helper.print_object
      ~msg:"Unexpected object deserialization"
      (new Serialization_test_helper.test_serializable 5 5
         :> Xnc_serialization.serializable)
      (S.unwrap_object
         (S.read_value
            reader
            object_value));
    Xnc_reader.close reader
;;

let test_read_simple_field _ =
  let ch = new Netchannels.input_string "Int_field=05\n\
                                         Float_field=3.14\n\
                                         Wrong_name=3.14\n\
                                         Null=None\n\
                                         Option=false\n\
                                         Object=\"7/2\"" in
  let reader = Xnc_reader.create_from_netchannel ch in
  let int_field = S.create "Int_field"
                    S.default_int in
  let float_field = S.create "Float_field"
                      S.default_float in
  let float_field' = S.create "Wrong_name"
                       S.default_float in
  let bool_option_field = S.create "Null"
                            (S.default_option
                               S.default_bool) in
  let bool_option_field' = S.create "Option"
                             (S.default_option
                                S.default_bool) in
  let object_field =
    S.new_field "Object"
      (new Serialization_test_helper.test_serializable 0 0) S.wrap_object
  in
    S.read reader int_field;
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected int field deserialization"
      (S.create
          "Int_field" (S.Int 5))
      int_field;
    S.read reader float_field;
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected float field deserialization"
      ( S.create
          "Float_field" (S.Float 3.14))
      float_field;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "bad input at line 3, char 1: \
          'W' (unscanned buffer: \"Wrong_name=3.14\", EOF: false), \
          expecting \"Float_field\"")
      (fun () -> S.read reader float_field);
    S.read reader float_field';
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected float field deserialization"
      (S.create
         "Wrong_name" (S.Float 3.14))
      float_field';
    S.read reader bool_option_field;
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected bool option deserialization"
      (S.create
         "Null" (S.Option None))
      bool_option_field;
    S.read reader bool_option_field';
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected bool option deserialization"
      (S.create
         "Option" (S.Option
                     (Some (S.Bool false))))
      bool_option_field';
    S.read reader object_field;
    OUnit.assert_equal
      ~cmp:(S.custom_equal
              ~obj_eq:(fun o1 o2 ->
                         (o1 # to_string ()) = (o2 # to_string ())))
      ~printer:S.string_of_field
      ~msg:"Unexpected object deserialization"
      (S.new_field
         "Object"
         (new Serialization_test_helper.test_serializable 7 2)
         S.wrap_object
      )
      object_field;
    Xnc_reader.close reader;
;;

let test_read_list_field _ =
  let ch = new Netchannels.input_string
             "List_field=[\
                Inner_list_field=[\
                   Int_field=10;\
                   Float_field=2.72\
                ];\
                Second_inner_list_field=[\
                   String_field=\"test\";\
                   Bool_field=true\
                ]\
              ]" in
  let reader = Xnc_reader.create_from_netchannel ch in
  let int_field = S.create "Int_field"
                    S.default_int in
  let float_field = S.create "Float_field"
                      S.default_float in
  let string_field = S.create "String_field"
                       S.default_string in
  let bool_field = S.create "Bool_field"
                     S.default_bool in
  let inner_field = S.new_field "Inner_list_field"
                      [int_field; float_field] S.wrap_list in
  let inner2_field = S.new_field "Second_inner_list_field"
                       [string_field; bool_field] S.wrap_list in
  let list_field = S.new_field "List_field"
                     [inner_field; inner2_field] S.wrap_list in
    S.read reader list_field;
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected list deserialization"
      (S.new_field "List_field"
         [ S.new_field "Inner_list_field"
             [ S.new_field "Int_field" 10 S.wrap_int;
               S.new_field "Float_field" 2.72 S.wrap_float
             ] S.wrap_list;
           S.new_field "Second_inner_list_field"
             [ S.new_field "String_field" "test" S.wrap_string;
               S.new_field "Bool_field" true S.wrap_bool
             ] S.wrap_list
         ] S.wrap_list
      )
      list_field;
    Xnc_reader.close reader
;;

let test_read_array_field _ =
  let ch = new Netchannels.input_string
             "Array_field=[\
                Inner_array_field=[\
                   Int_field=10;\
                   Float_field=2.72\
                ];\
                Second_inner_array_field=[\
                   String_field=\"test\";\
                   Bool_field=true\
                ]\
              ]" in
  let reader = Xnc_reader.create_from_netchannel ch in
  let int_field = S.create "Int_field"
                    S.default_int in
  let float_field = S.create "Float_field"
                      S.default_float in
  let string_field = S.create "String_field"
                       S.default_string in
  let bool_field = S.create "Bool_field"
                     S.default_bool in
  let inner_field = S.new_field "Inner_array_field"
                      [|int_field; float_field|] S.wrap_array in
  let inner2_field = S.new_field "Second_inner_array_field"
                       [|string_field; bool_field|] S.wrap_array in
  let array_field = S.new_field "Array_field"
                     [|inner_field; inner2_field|] S.wrap_array in
    S.read reader array_field;
    Xnc_reader.close reader;
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected array deserialization"
      (S.new_field "Array_field"
         [| S.new_field "Inner_array_field"
              [| S.new_field "Int_field" 10 S.wrap_int;
                 S.new_field "Float_field" 2.72 S.wrap_float
              |] S.wrap_array;
            S.new_field "Second_inner_array_field"
              [| S.new_field "String_field" "test" S.wrap_string;
                 S.new_field "Bool_field" true S.wrap_bool
              |] S.wrap_array;
         |] S.wrap_array
      )
      array_field;
;;

let test_read_using_xml_lexer _ =
  let doc = Pxp_tree_parser.parse_wfdocument_entity
              Pxp_types.default_config
              (Pxp_types.from_string
                 "<?xml version=\"1.0\" encoding=\"iso-8859-1\" ?>
                  <xnc:List_field>
                    <Inner_list_field>
                      <Int_field>10</Int_field>
                      <Float_field>2.72</Float_field>
                    </Inner_list_field>
                    <Second_inner_list_field Bool_field=\"true\">
                      <String_field>test</String_field>
                      <attributes name1=\"value1\" name2=\"value2\" />
                    </Second_inner_list_field>
                  </xnc:List_field>")
              Pxp_tree_parser.default_spec in
  let reader = Xnc_reader.create_from_pxp_node (doc # root) in
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let float_field = S.new_field "Float_field" 2.72 S.wrap_float in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let bool_field = S.new_field
                     ~properties:(Xnc_property.singleton Xnc_property.attribute)
                     "Bool_field" true S.wrap_bool in
  let attribute1_field = S.new_field
                           ~properties:(Xnc_property.singleton
                                          Xnc_property.attribute)
                           "name1" "value1" S.wrap_string in
  let attribute2_field = S.new_field
                           ~properties:(Xnc_property.singleton
                                          Xnc_property.attribute)
                           "name2" "value2" S.wrap_string in
  let attributes_field = S.new_field "attributes"
                           [attribute1_field;
                            attribute2_field] S.wrap_list in
  let inner_field = S.new_field "Inner_list_field"
                      [int_field; float_field] S.wrap_list in
  let inner_field' = S.new_field "Second_inner_list_field"
                       [bool_field;
                        string_field;
                        attributes_field] S.wrap_list in
  let list_field = S.new_field ~prefix:"xnc" "List_field"
                     [inner_field; inner_field'] S.wrap_list in
    S.read reader list_field;
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected list deserialization"
      (S.new_field ~prefix:"xnc" "List_field"
         [ S.new_field "Inner_list_field"
             [ S.new_field "Int_field" 10 S.wrap_int;
               S.new_field "Float_field" 2.72 S.wrap_float
             ] S.wrap_list;
           S.new_field "Second_inner_list_field"
             [ S.new_field
                 ~properties:(Xnc_property.singleton Xnc_property.attribute)
                 "Bool_field" true S.wrap_bool;
               S.new_field "String_field" "test" S.wrap_string;
               S.new_field "attributes"
                 [ S.new_field
                     ~properties:(Xnc_property.singleton Xnc_property.attribute)
                     "name1" "value1" S.wrap_string;
                   S.new_field
                     ~properties:(Xnc_property.singleton Xnc_property.attribute)
                     "name2" "value2" S.wrap_string;
                 ] S.wrap_list
             ] S.wrap_list
         ] S.wrap_list
      )
      list_field;
    Xnc_reader.close reader;
;;

let test_read_using_xml_stream_lexer _ =
  let config = Pxp_types.default_config in
  let mgr = Pxp_ev_parser.create_entity_manager
              config
              (Pxp_types.from_string
                 "<?xml version=\"1.0\" encoding=\"iso-8859-1\" ?>
                  <xnc:List_field>
                    <Inner_list_field>
                      <Int_field>10</Int_field>
                      <Float_field>2.72</Float_field>
                    </Inner_list_field>
                    <Second_inner_list_field Bool_field=\"true\">
                      <String_field>test</String_field>
                      <attributes name1=\"value1\" name2=\"value2\" />
                    </Second_inner_list_field>
                  </xnc:List_field>") in
  let stream = Xnc_scanning.pull_parser_to_string_stream config mgr in
  let reader = Xnc_reader.create_from_stream stream in
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let float_field = S.new_field "Float_field" 2.72 S.wrap_float in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let bool_field = S.new_field
                     ~properties:(Xnc_property.singleton Xnc_property.attribute)
                     "Bool_field" true S.wrap_bool in
  let attribute1_field = S.new_field
                           ~properties:(Xnc_property.singleton
                                          Xnc_property.attribute)
                           "name1" "value1" S.wrap_string in
  let attribute2_field = S.new_field
                           ~properties:(Xnc_property.singleton
                                          Xnc_property.attribute)
                           "name2" "value2" S.wrap_string in
  let attributes_field = S.new_field "attributes"
                           [attribute1_field;
                            attribute2_field] S.wrap_list in
  let inner_field = S.new_field "Inner_list_field"
                      [int_field; float_field] S.wrap_list in
  let inner_field' = S.new_field "Second_inner_list_field"
                       [bool_field;
                        string_field;
                        attributes_field] S.wrap_list in
  let list_field = S.new_field ~prefix:"xnc" "List_field"
                     [inner_field; inner_field'] S.wrap_list in
    S.read reader list_field;
    OUnit.assert_equal
      ~cmp:S.equal
      ~printer:S.string_of_field
      ~msg:"Unexpected list deserialization"
      (S.new_field ~prefix:"xnc" "List_field"
         [ S.new_field "Inner_list_field"
             [ S.new_field "Int_field" 10 S.wrap_int;
               S.new_field "Float_field" 2.72 S.wrap_float
             ] S.wrap_list;
           S.new_field "Second_inner_list_field"
             [ S.new_field
                 ~properties:(Xnc_property.singleton Xnc_property.attribute)
                 "Bool_field" true S.wrap_bool;
               S.new_field "String_field" "test" S.wrap_string;
               S.new_field "attributes"
                 [ S.new_field
                     ~properties:(Xnc_property.singleton Xnc_property.attribute)
                     "name1" "value1" S.wrap_string;
                   S.new_field
                     ~properties:(Xnc_property.singleton Xnc_property.attribute)
                     "name2" "value2" S.wrap_string;
                 ] S.wrap_list
             ] S.wrap_list
         ] S.wrap_list
      )
      list_field;
    Xnc_reader.close reader;
;;

let test_custom_value _ =
  let write_value ?properties writer value =
    match value with
        S.String s ->
          Xnc_writer.write_string writer (String.uppercase s)
      | _ -> failwith "unsupported" in
  let read_value ?properties reader value =
    match value with
        S.String _ ->
          S.String
            (String.lowercase (Xnc_reader.read_string reader))
      | _ -> failwith "unsupported" in
  let buf = Buffer.create 128 in
  let writer = Xnc_writer.create_from_buffer buf in
  let reader = Xnc_reader.create_from_string "Custom_field=\"ANOTHER TEST\"" in
  let field = S.create_custom
                "Custom_field" write_value read_value
                (S.String "test") in
    S.write writer field;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected serialization"
      "Custom_field=\"TEST\""
      (Buffer.contents buf);
    S.read reader field;
    OUnit.assert_equal
      ~cmp:S.value_equal
      ~printer:S.string_of_value
      ~msg:"Unexpected deserialization"
      (S.String "another test")
      (S.get_custom
         field);
;;

let test_generic_serializer _ =
  let module M = struct
    type t = A | B | C

    let to_string t =
      match t with
          A -> "A"
        | B -> "B"
        | C -> "C"

    let from_string s =
      match s with
          "A" -> A
        | "B" -> B
        | "C" -> C
        | _ -> failwith "unexpected serialization"

  end in
  let t = M.A in
  let generic = new Xnc_serialization.generic_serializer
                  t
                  M.to_string
                  M.from_string in
  let field = S.new_field "Generic" generic S.wrap_object in
  let buf = Buffer.create 128 in
  let writer = Xnc_writer.create_from_buffer buf in
  let reader = Xnc_reader.create_from_string "Generic=\"B\"" in
    S.write writer field;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected serialization"
      "Generic=\"A\""
      (Buffer.contents buf);
    S.read reader field;
    OUnit.assert_equal
      ~printer:M.to_string
      ~msg:"Unexpected deserialization"
      (M.B)
      (generic # get_value);
;;

let test_to_name_value_string_list _ =
  let int_field = S.new_field "Int_field" 10 S.wrap_int in
  let string_field = S.new_field "String_field" "test" S.wrap_string in
  let field = S.new_field "List_field"
                [int_field; string_field] S.wrap_list in
  let print_list =
    Serialization_test_helper.print_list
      (fun (n, v) -> n ^ "=" ^ v)
  in
    OUnit.assert_equal
      ~printer:print_list
      ~msg:"Unexpected name/value string list"
      [("Int_field", "10"); ("String_field", "\"test\"")]
      (S.to_name_value_string_list
         field)
;;

(* test suite *)
let suite = "Serializable test" >:::
  ["test_create" >:: test_create;
   "test_new_field" >:: test_new_field;
   "test_new_list_field" >:: test_new_list_field;
   "test_new_array_field" >:: test_new_array_field;
   "test_equal" >:: test_equal;
   "test_clone" >:: test_clone;
   "test_append_field" >:: test_append_field;
   "test_replace_field" >:: test_replace_field;
   "test_iter_leaves" >:: test_iter_leaves;
   "test_iter_pre" >:: test_iter_pre;
   "test_iter_post" >:: test_iter_post;
   "test_fold_leaves" >:: test_iter_leaves;
   "test_fold_pre" >:: test_fold_pre;
   "test_fold_post" >:: test_fold_post;
   "test_map_leaves" >:: test_map_leaves;
   "test_map_pre" >:: test_map_pre;
   "test_map_post" >:: test_map_post;
   "test_to_stream" >:: test_to_stream;
   "test_to_name_value_stream" >:: test_to_name_value_stream;
   "test_value_present" >:: test_value_present;
   "test_value_present_in_list" >:: test_value_present_in_list;
   "test_value_present_in_array" >:: test_value_present_in_array;
   "test_find_value" >:: test_find_value;
   "test_find_value_in_list" >:: test_find_value_in_list;
   "test_find_value_in_array" >:: test_find_value_in_array;
   "test_find_all_values" >:: test_find_all_values;
   "test_write_value" >:: test_write_value;
   "test_write_simple_field" >:: test_write_simple_field;
   "test_write_list_field" >:: test_write_list_field;
   "test_write_array_field" >:: test_write_array_field;
   "test_write_using_pretty_printer" >:: test_write_using_pretty_printer;
   "test_write_using_xml_printer" >:: test_write_using_xml_printer;
   "test_write_using_xml_pretty_printer"
     >:: test_write_using_xml_pretty_printer;
   "test_read_value" >:: test_read_value;
   "test_read_simple_field" >:: test_read_simple_field;
   "test_read_list_field" >:: test_read_list_field;
   "test_read_array_field" >:: test_read_array_field;
   "test_read_using_xml_lexer" >:: test_read_using_xml_lexer;
   "test_read_using_xml_stream_lexer" >:: test_read_using_xml_stream_lexer;
   "test_custom_value" >:: test_custom_value;
   "test_generic_serializer" >:: test_generic_serializer;
   "test_to_name_value_string_list" >:: test_to_name_value_string_list]
;;

