open OUnit

(* tests *)
let test_simple_serializer_serialize _ =
  let serializer = new Xnc_serializing.simple_serializer () in
  let serializer' =
    new Xnc_serializing.simple_serializer
      ~config:{ Xnc_serialization.default_config
                  with Xnc_serialization.delimiter_char = '\'' } () in
  let field = Xnc_serializable.new_field
                "String_field"
                "Test_string:&>'<"
                Xnc_serializable.wrap_string in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected test-serialized string"
      "String_field=&quot;Test_string:&amp;&gt;&apos;&lt;&quot;"
      (serializer # serialize field);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected test-serialized string"
      "String_field=&apos;Test_string:&amp;&gt;\\&apos;&lt;&apos;"
      (serializer' # serialize field)
;;

let test_simple_serializer_to_netchannel _ =
  let serializer = new Xnc_serializing.simple_serializer () in
  let buf = Buffer.create 128 in
  let out_ch = new Netchannels.output_buffer buf in
  let field = Xnc_serializable.new_field
                "String_field"
                "Test_string:&>'<"
                Xnc_serializable.wrap_string in
    serializer # to_netchannel field out_ch;
    out_ch # close_out ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected test-serialized string"
      "String_field=&quot;Test_string:&amp;&gt;&apos;&lt;&quot;"
      (Buffer.contents buf)
;;

let test_simple_serializer_from_netchannel _ =
  let serializer = new Xnc_serializing.simple_serializer () in
  let field = Xnc_serializable.create
                "String_field" Xnc_serializable.default_string in
  let in_ch = new Netchannels.input_string
                "String_field=&quot;Test_string:&amp;&gt;&apos;&lt;&quot;" in
    serializer # from_netchannel field in_ch;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected deserialized field name"
      "String_field"
      (Xnc_serializable.get_name field);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected deserialized field value"
      "Test_string:&>'<"
      (Xnc_serializable.get_string field)
;;

let test_simple_serializer_deserialize _ =
  let serializer = new Xnc_serializing.simple_serializer () in
  let field = Xnc_serializable.create
                "String_field" Xnc_serializable.default_string in
    serializer # deserialize field
      "String_field=&quot;Test_string:&amp;&gt;&apos;&lt;&quot;";
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected deserialized field name"
      "String_field"
      (Xnc_serializable.get_name field);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected deserialized field value"
      "Test_string:&>'<"
      (Xnc_serializable.get_string field)
;;

let test_xml_serializer_serialize _ =
  let serializer = new Xnc_serializing.xml_serializer () in
  let field = Xnc_serializable.new_field
                "string"
                "Test string:&>'<"
                Xnc_serializable.wrap_string in
  let root = Xnc_serializable.new_field "root"
               [field] Xnc_serializable.wrap_list in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected XML-serialized string"
      "<root><string>Test string:&amp;&gt;&apos;&lt;</string></root>"
      (serializer # serialize root);
;;

let test_xml_serializer_deserialize _ =
  let serializer = new Xnc_serializing.xml_serializer () in
  let field = Xnc_serializable.create
                "string" Xnc_serializable.default_string in
  let root = Xnc_serializable.new_field "root"
               [field] Xnc_serializable.wrap_list in
    serializer # deserialize root
      "<root><string>Test string:&amp;&gt;&apos;&lt;</string></root>";
    OUnit.assert_equal
      ~cmp:Xnc_serializable.equal
      ~printer:Xnc_serializable.string_of_field
      ~msg:"Unexpected deserialized field"
      (Xnc_serializable.new_field "root"
         [ Xnc_serializable.new_field
             "string" "Test string:&>'<" Xnc_serializable.wrap_string ]
         Xnc_serializable.wrap_list)
      root;
;;

let test_validating_xml_serializer_deserialize _ =
  let serializer = new Xnc_serializing.validating_xml_serializer () in
  let field = Xnc_serializable.create
                "string" Xnc_serializable.default_string in
  let root = Xnc_serializable.new_field "root"
               [field] Xnc_serializable.wrap_list in
    serializer # deserialize root
      "<?xml version=\"1.0\" ?>
       <!DOCTYPE root [
         <!ELEMENT root (string)>
         <!ELEMENT string (#PCDATA)>
       ]>
       <root>
         <string>Test string:&amp;&gt;&apos;&lt;</string>
       </root>";
    OUnit.assert_equal
      ~cmp:Xnc_serializable.equal
      ~printer:Xnc_serializable.string_of_field
      ~msg:"Unexpected deserialized field"
      (Xnc_serializable.new_field "root"
         [ Xnc_serializable.new_field
             "string" "Test string:&>'<" Xnc_serializable.wrap_string ]
         Xnc_serializable.wrap_list)
      root;
;;

(* test suite *)
let suite = "Serializer test" >:::
  ["test_simple_serializer_serialize"
     >:: test_simple_serializer_serialize;
   "test_simple_serializer_to_netchannel"
     >:: test_simple_serializer_to_netchannel;
   "test_simple_serializer_deserialize"
     >:: test_simple_serializer_deserialize;
   "test_simple_serializer_from_netchannel"
     >:: test_simple_serializer_from_netchannel;
   "test_xml_serializer_serialize"
     >:: test_xml_serializer_serialize;
   "test_xml_serializer_deserialize"
     >:: test_xml_serializer_deserialize;
   "test_validating_xml_serializer_deserialize"
     >:: test_validating_xml_serializer_deserialize]
;;

