open OUnit

(* tests *)
let test_xor_transform _ =
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected transformed string"
    "1)\"2"
    (Xnc_string_transform.xor_transform "Test" "eLQF")
;;

let test_base64_transform _ =
  let encoded = Xnc_string_transform.base64_encode "" in
  let decoded = Xnc_string_transform.base64_decode encoded in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      ""
      decoded
;;

let test_base16_transform _ =
  let encoded = Xnc_string_transform.base16_encode "" in
  let decoded = Xnc_string_transform.base16_decode encoded in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected encoded string"
      "e0e8ecf2f9"
      encoded;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected decoded string"
      ""
      decoded
;;

let test_escape_xml _ =
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected transformed string"
    "&amp;&gt;&apos;&lt;"
    (Xnc_string_transform.escape_xml "&>'<")
;;

let test_unescape_xml _ =
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected transformed string"
    "&>'<"
    (Xnc_string_transform.unescape_xml "&amp;&gt;&apos;&lt;")
;;

let test_aes_transform _ =
  let key = "16-byte-test-key" in
  let encrypted = Xnc_string_transform.Cipher.encrypt
                    Xnc_string_transform.Cipher.Aes
                    key
                    "This is the string that will be \
                    encryped by Cryptokit using AES" in
  let decrypted = Xnc_string_transform.Cipher.decrypt
                    Xnc_string_transform.Cipher.Aes
                    key
                    encrypted in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "This is the string that will be encryped by Cryptokit using AES"
      decrypted
;;

let test_des_transform _ =
  let key = "8-byte-k" in
  let encrypted = Xnc_string_transform.Cipher.encrypt
                    Xnc_string_transform.Cipher.Des
                    key
                    "This is the string that will be \
                    encryped by Cryptokit using DES" in
  let decrypted = Xnc_string_transform.Cipher.decrypt
                    Xnc_string_transform.Cipher.Des
                    key
                    encrypted in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "This is the string that will be encryped by Cryptokit using DES"
      decrypted
;;

let test_triple_des_transform _ =
  let key = "16-byte-test-key" in
  let encrypted = Xnc_string_transform.Cipher.encrypt
                    Xnc_string_transform.Cipher.Triple_des
                    key
                    "This is the string that will be \
                    encrypted by Cryptokit using Triple DES" in
  let decrypted = Xnc_string_transform.Cipher.decrypt
                    Xnc_string_transform.Cipher.Triple_des
                    key
                    encrypted in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "This is the string that will be encrypted by Cryptokit using Triple DES"
      decrypted
;;

let test_hmac_md5_transform _ =
  let key = "1234567890abcdefghijklmnopqrstuvwxyz" in
  let mac = Xnc_string_transform.Mac.digest
              Xnc_string_transform.Mac.Md5
              key
              "This transform returns a MAC based on the \
               HMAC construction (RFC2104) applied to MD5" in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "bIipxUJJEriU7ZXinKfufA=="
      (Xnc_string_transform.base64_encode mac)
;;

let test_hmac_sha1_transform _ =
  let key = "1234567890abcdefghijklmnopqrstuvwxyz" in
  let mac = Xnc_string_transform.Mac.digest
              Xnc_string_transform.Mac.Sha1
              key
              "This transform returns a MAC based on the \
               HMAC construction (RFC2104) applied to SHA1" in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "LmmAUXES1Lp0kTSXKCFJmsx5MR4="
      (Xnc_string_transform.base64_encode mac)
;;

let test_mac_aes_transform _ =
  let key = "1234567890abcdef" in
  let mac = Xnc_string_transform.Mac.digest
              Xnc_string_transform.Mac.Aes
              key
              "This transform returns a MAC returns \
               a MAC based on AES encryption in CBC mode" in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "cXMShWkx+oYA0F5XBVrICA=="
      (Xnc_string_transform.base64_encode mac)
;;

let test_mac_des_transform _ =
  let key = "12345678" in
  let mac = Xnc_string_transform.Mac.digest
              Xnc_string_transform.Mac.Des
              key
              "This transform returns a MAC returns \
               a MAC based on DES encryption in CBC mode" in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "Jeogeu+QtUc="
      (Xnc_string_transform.base64_encode mac)
;;

let test_mac_triple_des_transform _ =
  let key = "1234567890abcdef" in
  let mac = Xnc_string_transform.Mac.digest
              Xnc_string_transform.Mac.Triple_des
              key
              "This transform returns a MAC returns \
               a MAC based on triple DES encryption in CBC mode" in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "C9RYEUVOPjo="
      (Xnc_string_transform.base64_encode mac)
;;

let test_mac_dftd_transform _ =
  let key = "1234567890abcdefghijklmn" in
  let mac = Xnc_string_transform.Mac.digest
              Xnc_string_transform.Mac.Des_final_triple_des
              key
              "This transform returns a MAC that uses \
               DES CBC with the first 8 bytes of the key \
               as key. The final initialization vector \
               is then DES-decrypted with bytes 8 to 15 of key, \
               and DES-encrypted again with the last 8 bytes of key" in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "cSv84MQhBx4="
      (Xnc_string_transform.base64_encode mac)
;;

let test_zlib_transform _ =
  let compressed = Xnc_string_transform.zlib_compress
                     "This is the string that will be \
                     compressed using using the zlib library." in
  let uncompressed = Xnc_string_transform.zlib_uncompress
                    compressed in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "This is the string that will be compressed using using the zlib library."
      uncompressed
;;

let test_compose_cryptokit_transform _ =
  let transform1 = Cryptokit.Zlib.compress ~level:9 () in
  let transform2 = Cryptokit.Cipher.aes
                     ~pad:Cryptokit.Padding.length
                     "16-byte-test-key"
                     Cryptokit.Cipher.Encrypt in
  let transform3 = Cryptokit.Hexa.encode () in
  let transform = Xnc_string_transform.compose_cryptokit_transform
                    [transform1;
                     transform2;
                     transform3] in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected transformed string"
      "661f6025bead09ea2612bc4bc03cf1a0"
      (Cryptokit.transform_string
         transform
         "AAAAAAAAAAAAAAAAAAAA")
;;

(* test suite *)
let suite = "String transform test" >:::
  ["test_xor_transform" >:: test_xor_transform;
   "test_base64_transform" >:: test_base64_transform;
   "test_base16_transform" >:: test_base16_transform;
   "test_escape_xml" >:: test_escape_xml;
   "test_unescape_xml" >:: test_unescape_xml;
   "test_aes_transform" >:: test_aes_transform;
   "test_des_transform" >:: test_des_transform;
   "test_triple_des_transform" >:: test_triple_des_transform;
   "test_hmac_md5_transform" >:: test_hmac_md5_transform;
   "test_hmac_sha1_transform" >:: test_hmac_sha1_transform;
   "test_mac_aes_transform" >:: test_mac_aes_transform;
   "test_mac_des_transform" >:: test_mac_des_transform;
   "test_mac_triple_des_transform" >:: test_mac_triple_des_transform;
   "test_mac_dftd_transform" >:: test_mac_dftd_transform;
   "test_zlib_transform" >:: test_zlib_transform;
   "test_compose_cryptokit_transform" >:: test_compose_cryptokit_transform]
;;

