open OUnit

(* fixture setup and teardown *)
let set_up _ =
  let buf = Buffer.create 128 in
  let writer = Xnc_writer.create_from_buffer buf in
    (buf, writer)
;;

let tear_down (buf, writer) =
  Xnc_writer.close writer
;;

(* tests *)
let test_write_null (buf, writer) =
  Xnc_writer.write_null writer ();
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected buffer content after write_null"
    "None"
    (Buffer.contents buf)
;;
 
let test_write_int (buf, writer) =
  Xnc_writer.write_int writer 1;
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected buffer content after write_int"
    "1"
    (Buffer.contents buf)
;;
 
let test_write_float (buf, writer) =
  Xnc_writer.write_float writer 3.141592;
  Xnc_writer.write_float writer 3.0;
  Xnc_writer.write_float writer 0.035;
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected buffer content after 3 write_float"
    "3.1415923.0.035"
    (Buffer.contents buf)
;;

let test_write_char (buf, writer) =
  Xnc_writer.write_char writer 'c';
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected buffer content after write_char"
    "c"
    (Buffer.contents buf)
;;

let test_write_bool (buf, writer) =
  Xnc_writer.write_bool writer true;
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected buffer content after write_bool"
    "true"
    (Buffer.contents buf)
;;

let test_write_raw_string (buf, writer) =
  Xnc_writer.write_raw_string writer "test";
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected buffer content after write_raw_string"
    "test"
    (Buffer.contents buf)
;;

let test_write_string (buf, writer) =
  let config = { Xnc_serialization.default_config
                   with Xnc_serialization.delimiter_char = '#' } in
  let writer' = Xnc_writer.create_from_buffer ~config:config buf in
    Xnc_writer.write_string writer "test";
    Xnc_writer.write_string writer' "test";
    Xnc_writer.write_string writer' "test#escape#";
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after 3 write_string"
      "\"test\"#test##test\\#escape\\##"
      (Buffer.contents buf);
    Xnc_writer.close writer'
;;

let test_write_date (buf, writer) =
  Xnc_writer.write_date writer
    { Netdate.year = 2006;
      Netdate.month = 10;
      Netdate.day = 22;
      Netdate.hour = 16;
      Netdate.minute = 46;
      Netdate.second = 30;
      Netdate.zone = 0;
      Netdate.week_day = -1
    };
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected buffer content after write_date"
    "10/22/2006 16:46:30"
    (Buffer.contents buf)
;;

let test_write_value (buf, writer) =
  Xnc_writer.write_value writer 3 string_of_int;
  Xnc_writer.write_value writer 5.10 string_of_float;
  Xnc_writer.write_value writer '-' Std.string_of_char;
  Xnc_writer.write_value writer false string_of_bool;
  Xnc_writer.write_value writer "/test" Std.identity;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after 5 write_value"
      "35.1-false/test"
      (Buffer.contents buf)
;;

let test_write_option (buf, writer) =
  let write_null = Xnc_writer.write_null writer in
  let write_value = Xnc_writer.write_int writer in
    Xnc_writer.write_option writer None write_null write_value;
    Xnc_writer.write_option writer (Some 3) write_null write_value;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after 2 write_option"
      "None3"
      (Buffer.contents buf)
;;

let test_write_object (buf, writer) =
  let o1 = new Serialization_test_helper.test_serializable 10 20 in
  let o2 = new Serialization_test_helper.test_serializable 20 30 in
    Xnc_writer.write_object writer o1;
    Xnc_writer.write_object writer o2;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after 2 write_object"
      "\"10/20\"\"20/30\""
      (Buffer.contents buf)
;;

let test_write_enum (buf, writer) =
  let config = { Xnc_serialization.default_config
                   with Xnc_serialization.start_of_list = "(";
                        Xnc_serialization.end_of_list = ")";
                        Xnc_serialization.list_separator = ","} in
  let writer' = Xnc_writer.create_from_buffer ~config:config buf in
  let write_element e = Xnc_writer.write_raw_string writer e in
  let empty_enum = Enum.empty () in
  let enum = ExtList.List.enum ["test1"; "test2"] in
  let enum' = ExtList.List.enum ['a'; 'b'; 'c'] in
    Xnc_writer.write_enum writer empty_enum write_element;
    Xnc_writer.write_enum writer enum write_element;
    Xnc_writer.write_enum writer' enum'
      (fun e -> Xnc_writer.write_char writer' e);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after 3 write_enum"
      "[][test1;test2](a,b,c)"
      (Buffer.contents buf);
    Xnc_writer.close writer'
;;

let test_write_collection (buf, writer) =
  let config = { Xnc_serialization.default_config
                   with Xnc_serialization.start_of_list = "(";
                        Xnc_serialization.end_of_list = ")";
                        Xnc_serialization.list_separator = ","} in
  let writer' = Xnc_writer.create_from_buffer ~config:config buf in
  let dynarray = DynArray.of_array [|3; 4|] in
  let dllist = Dllist.of_list [true; false; false] in
    Xnc_writer.write_collection writer
      [] ExtList.List.enum
      (fun e -> Xnc_writer.write_raw_string writer' e);
    Xnc_writer.write_collection writer
      ["test1"; "test2"] ExtList.List.enum
      (fun e -> Xnc_writer.write_raw_string writer' e);
    Xnc_writer.write_collection writer'
      [|'a'; 'b'; 'c'|] ExtArray.Array.enum
      (fun e -> Xnc_writer.write_char writer' e);
    Xnc_writer.write_collection writer
      dynarray DynArray.enum
      (fun e -> Xnc_writer.write_int writer e);
    Xnc_writer.write_collection writer'
      dllist Dllist.enum
      (fun e -> Xnc_writer.write_bool writer' e);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after 5 write_collection"
      "[][test1;test2](a,b,c)[3;4](true,false,false)"
      (Buffer.contents buf);
    Xnc_writer.close writer'
;;

let test_write_field (buf, writer) =
  let config = { Xnc_serialization.default_config
                   with Xnc_serialization.name_value_separator = ":";
                        Xnc_serialization.attribute_marker = "$" } in
  let writer' = Xnc_writer.create_from_buffer ~config:config buf in
  let write_value = Xnc_writer.write_raw_string writer in
    Xnc_writer.write_field writer
      "" "key" "value" Xnc_property.empty write_value;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after write_field (no prefix, no attr)"
      "key=value"
      (Buffer.contents buf);
    Buffer.clear buf;
    Xnc_writer.write_field writer'
      "" "attribute" "value"
      (Xnc_property.singleton Xnc_property.attribute)
      write_value;
    Xnc_writer.close writer';
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after write_field (no prefix, attr)"
      "$attribute:value"
      (Buffer.contents buf);
    Buffer.clear buf;
    Xnc_writer.write_field writer
      "tst" "key" "value" Xnc_property.empty write_value;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after write_field (prefix, no attr)"
      "tst:key=value"
      (Buffer.contents buf);
    Buffer.clear buf;
    Xnc_writer.write_field writer
      "xnc" "attribute" "value"
      (Xnc_property.singleton Xnc_property.attribute)
      write_value;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after write_field (prefix, attr)"
      "@xnc:attribute=value"
      (Buffer.contents buf);
;;

let test_write_on_netchannel _ =
  let buf = Buffer.create 128 in
  let out_ch = new Netchannels.output_buffer buf in
  let writer = Xnc_writer.create_from_netchannel out_ch in
    Xnc_writer.write_int writer 1;
    Xnc_writer.write_char writer '-';
    Xnc_writer.write_float writer 1.11;
    Xnc_writer.write_string writer "test";
    Xnc_writer.write_bool writer false;
    Xnc_writer.close writer;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content"
      "1-1.11\"test\"false"
      (Buffer.contents buf)
;;
 
(* test suite *)
let suite = "Writer test" >:::
  ["test_write_null" >:: (bracket set_up test_write_null tear_down);
   "test_write_int" >:: (bracket set_up test_write_int tear_down);
   "test_write_float" >:: (bracket set_up test_write_float tear_down);
   "test_write_char" >:: (bracket set_up test_write_char tear_down);
   "test_write_bool" >:: (bracket set_up test_write_bool tear_down);
   "test_write_raw_string" >:: (bracket set_up test_write_raw_string tear_down);
   "test_write_string" >:: (bracket set_up test_write_string tear_down);
   "test_write_date" >:: (bracket set_up test_write_date tear_down);
   "test_write_value" >:: (bracket set_up test_write_value tear_down);
   "test_write_option" >:: (bracket set_up test_write_option tear_down);
   "test_write_object" >:: (bracket set_up test_write_object tear_down);
   "test_write_enum" >:: (bracket set_up test_write_enum tear_down);
   "test_write_collection" >:: (bracket set_up test_write_collection tear_down);
   "test_write_field" >:: (bracket set_up test_write_field tear_down);
   "test_write_on_netchannel" >:: test_write_on_netchannel]
;;

