open OUnit

let get_name node = 
  match node # node_type with
      Pxp_document.T_element n -> n
    | _ -> "T_data"
;;

(* fixture setup and teardown *)
let set_up _ =
  let doc = Pxp_tree_parser.parse_wfdocument_entity
              Pxp_types.default_config
              (Pxp_types.from_string
                 "<root>\
                    <list>\
                      <int>10</int>\
                      <char>x</char>\
                      <bool>false</bool>\
                    </list>\
                    <list2>\
                      <innerlist>\
                        <float>0.5</float>\
                      </innerlist>\
                      <string>test</string>\
                    </list2>\
                  </root>")
              Pxp_tree_parser.default_spec
  in
    Xnc_scanning.pxp_node_to_stream (doc # root)
;;

let tear_down _ =
  ()
;;

(* tests *)
let test_has_next stream =
  let not_empty = Option.is_some (Stream.peek stream) in
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Iterator should have next element"
      true
      (not_empty);
;;

let test_peek stream =
  let node = Option.get (Stream.peek stream) in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected root node"
      "root"
      (get_name node);
;;

let test_next stream =
  let rec build_name_list accu =
    if Option.is_some (Stream.peek stream) then
      begin
        let node = Stream.next stream in
        let l = (get_name node) :: accu in
          build_name_list l
      end
    else
      accu in
  let name_list = build_name_list [] in
    OUnit.assert_equal
      ~printer:Std.dump
      ~msg:"Unexpected node name list"
      ["T_data"; "string";
       "T_data"; "float";
       "innerlist";
       "list2";
       "T_data"; "bool";
       "T_data"; "char";
       "T_data"; "int";
       "list";
       "root"]
      name_list;
;;
 
let test_scan_int _ =
  let doc = Pxp_tree_parser.parse_wfdocument_entity
              Pxp_types.default_config
              (Pxp_types.from_string "<int>10</int>")
              Pxp_tree_parser.default_spec in
  let stream = Xnc_scanning.pxp_node_to_stream (doc # root) in
  let scanner = new Xnc_scanning.pxp_node_stream_scanner stream in
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected int token"
      "10"
      (scanner # scan_int ());
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning int: \
          Stream.Failure")
      (scanner # scan_int);
;;

let test_parse_value _ =
  let doc = Pxp_tree_parser.parse_wfdocument_entity
              Pxp_types.default_config
              (Pxp_types.from_string
                 "<root>\
                    <int>10</int>\
                    <char>x</char>\
                    <bool>false</bool>\
                    <float>0.5</float>\
                    <string>test</string>\
                  </root>")
              Pxp_tree_parser.default_spec in
  let stream = Xnc_scanning.pxp_node_to_stream (doc # root) in
  let scanner = new Xnc_scanning.pxp_node_stream_scanner stream in
  let lexer = new Xnc_lexing.pxp_node_stream_lexer stream in
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected int value"
      10
      (lexer # parse_value
         (fun () ->
            scanner # skip 1;
            scanner # scan_int ())
         int_of_string);
    OUnit.assert_equal
      ~printer:Std.string_of_char
      ~msg:"Unexpected char value"
      'x'
      (lexer # parse_value
         (fun () ->
            scanner # skip 1;
            scanner # scan_char ())
         (fun s -> s.[0]));
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected bool value"
      false
      (lexer # parse_value
         (fun () ->
            scanner # skip 1;
            scanner # scan_bool ())
         bool_of_string);
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected float value"
      0.5
      (lexer # parse_value
         (fun () ->
            scanner # skip 1;
            scanner # scan_float ())
         float_of_string);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string value"
      "test"
      (lexer # parse_value
         (fun () ->
            scanner # skip 1;
            scanner # scan_string ())
         Std.identity);
    lexer # close ();
;;

let test_parse_option _ =
  let doc = Pxp_tree_parser.parse_wfdocument_entity
              Pxp_types.default_config
              (Pxp_types.from_string
                 "<root>\
                    <none></none>\
                    <some>false</some>\
                  </root>")
              Pxp_tree_parser.default_spec in
  let stream = Xnc_scanning.pxp_node_to_stream (doc # root) in
  let scanner = new Xnc_scanning.pxp_node_stream_scanner stream in
  let lexer = new Xnc_lexing.pxp_node_stream_lexer stream in
  let print_bool_option = Serialization_test_helper.print_option
                            string_of_bool in
  let try_parse_null () =
    lexer # try_parse_null () in
  let parse_bool () =
    lexer # parse_value scanner # scan_bool bool_of_string
  in
    scanner # skip 1;
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:print_bool_option
      ~msg:"Unexpected bool option value"
      None
      (lexer # parse_option try_parse_null parse_bool);
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:print_bool_option
      ~msg:"Unexpected bool option value"
      (Some false)
      (lexer # parse_option try_parse_null parse_bool);
    lexer # close ();
;;

let test_parse_object _ =
  let doc = Pxp_tree_parser.parse_wfdocument_entity
              Pxp_types.default_config
              (Pxp_types.from_string "<object>1/2</object>")
              Pxp_tree_parser.default_spec in
  let stream = Xnc_scanning.pxp_node_to_stream (doc # root) in
  let scanner = new Xnc_scanning.pxp_node_stream_scanner stream in
  let lexer = new Xnc_lexing.pxp_node_stream_lexer stream in
  let obj = new Serialization_test_helper.test_serializable 0 0 in
    scanner # skip 1;
    lexer # parse_object obj (scanner # scan_string);
    OUnit.assert_equal
      ~cmp:Serialization_test_helper.cmp_object
      ~printer:Serialization_test_helper.print_object
      ~msg:"Unexpected object"
      (new Serialization_test_helper.test_serializable 1 2)
      (obj);
    lexer # close ();
;;

let test_parse_homogeneous_enum _ =
  let doc = Pxp_tree_parser.parse_wfdocument_entity
              Pxp_types.default_config
              (Pxp_types.from_string
                 "<list>\
                    <elem>1.</elem>\
                    <elem>2.</elem>\
                    <elem>3.</elem>\
                    <elem>4.</elem>\
                  </list>")
              Pxp_tree_parser.default_spec in
  let stream = Xnc_scanning.pxp_node_to_stream (doc # root) in
  let scanner = new Xnc_scanning.pxp_node_stream_scanner stream in
  let lexer = new Xnc_lexing.pxp_node_stream_lexer stream in
  let parse_float () =
    lexer # parse_value
      (fun () ->
         scanner # skip 1;
         scanner # scan_float ())
      float_of_string in
  let _ = Stream.next stream in
  let enum = lexer # parse_homogeneous_enum parse_float in
  let l = ExtList.List.of_enum enum in
    OUnit.assert_equal
      ~printer:(Serialization_test_helper.print_list string_of_float)
      ~msg:"Unexpected enum value"
      [1.; 2.; 3.; 4.]
      l;
    lexer # close ();
;;

type test_schema_type =
    Int of int
  | Char of char
  | String of string

let test_parse_heterogeneous_enum _ =
  let doc = Pxp_tree_parser.parse_wfdocument_entity
              Pxp_types.default_config
              (Pxp_types.from_string
                 "<list>\
                    <char>a</char>\
                    <string>test</string>\
                    <int>1</int>\
                  </list>")
              Pxp_tree_parser.default_spec in
  let stream = Xnc_scanning.pxp_node_to_stream (doc # root) in
  let scanner = new Xnc_scanning.pxp_node_stream_scanner stream in
  let lexer = new Xnc_lexing.pxp_node_stream_lexer stream in
  let enum_schema = ExtList.List.enum [Char '?'; String ""; Int 0] in
  let parse_int () =
    lexer # parse_value
      (fun () ->
         scanner # skip 1;
         scanner # scan_int ())
      int_of_string in
  let parse_char () =
    lexer # parse_value
      (fun () ->
         scanner # skip 1;
         scanner # scan_char ())
      (fun s -> String.get s 0) in
  let parse_string () =
    lexer # parse_value
      (fun () ->
         scanner # skip 1;
         scanner # scan_string ())
      Std.identity in
  let parse_element e =
    match e with
        Int _ -> Int (parse_int ())
      | Char _ -> Char (parse_char ())
      | String _ -> String (parse_string ()) in
  let print_list =
    Serialization_test_helper.print_list
      (fun e -> match e with
           Int i -> string_of_int i
         | Char c -> Std.string_of_char c
         | String s -> s) in
  let _ = Stream.next stream in
  let enum = lexer # parse_heterogeneous_enum enum_schema parse_element in
    OUnit.assert_equal
      ~printer:print_list
      ~msg:"Unexpected enum value"
      [Char 'a'; String "test"; Int 1]
      (ExtList.List.of_enum enum);
    lexer # close ();
;;

let test_parse_value_by_name _ =
  let doc = Pxp_tree_parser.parse_wfdocument_entity
              Pxp_types.default_config
              (Pxp_types.from_string
                 "<list>\
                    <name>1</name>\
                    <x n=\"xyz\"></x>\
                    <xnc:name>test</xnc:name>\
                    <y xnc:n=\"xyz\"></y>\
                    <error>false</error>\
                  </list>")
              Pxp_tree_parser.default_spec in
  let stream = Xnc_scanning.pxp_node_to_stream (doc # root) in
  let scanner = new Xnc_scanning.pxp_node_stream_scanner stream in
  let lexer = new Xnc_lexing.pxp_node_stream_lexer stream in
  let parse_int () =
    lexer # parse_value (scanner # scan_int) int_of_string in
  let parse_bool () =
    lexer # parse_value (scanner # scan_bool) bool_of_string in
  let parse_string () =
    lexer # parse_value (scanner # scan_string) Std.identity
  in
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected value"
      1
      (lexer # parse_value_by_name
         "" "name" Xnc_property.empty parse_int);
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value"
      "xyz"
      (lexer # parse_value_by_name
         "" "n"
         (Xnc_property.singleton Xnc_property.attribute)
         parse_string);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value"
      "test"
      (lexer # parse_value_by_name
         "xnc" "name" Xnc_property.empty parse_string);
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value"
      "xyz"
      (lexer # parse_value_by_name
         "xnc" "n"
         (Xnc_property.singleton Xnc_property.attribute)
         parse_string);
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected value"
      false
      (lexer # parse_value_by_name
         "" "error" Xnc_property.empty parse_bool);
    lexer # close ();
;;

let test_stream_from_pxp_pull_parser _ =
  let config = Pxp_types.default_config in
  let mgr = Pxp_ev_parser.create_entity_manager
              config
              (Pxp_types.from_string
                 "<root>\
                    <list>\
                      <int att=\"val\">10</int>\
                      <char>x</char>\
                      <bool>false</bool>\
                    </list>\
                    <list2>\
                      <innerlist>\
                        <float>0.5</float>\
                      </innerlist>\
                      <string a1=\"v1\" a2=\"v2\">test</string>\
                    </list2>\
                  </root>") in
  let stream = Xnc_scanning.pull_parser_to_string_stream config mgr in
  let buf = Buffer.create 64 in
    Stream.iter
      (fun s -> Buffer.add_string buf s; Buffer.add_char buf ' ')
      stream;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected stream dump"
      "root \
         list \
           int att val 10 / \
           char x / \
           bool false / \
         / \
         list2 \
           innerlist \
             float 0.5 / \
           / \
           string a1 v1 a2 v2 test / \
         / \
       / "
      (Buffer.contents buf);
;;

let test_stream_scan_int _ =
  let config = Pxp_types.default_config in
  let mgr = Pxp_ev_parser.create_entity_manager
              config
              (Pxp_types.from_string "<int>10</int>") in
  let stream = Xnc_scanning.pull_parser_to_string_stream config mgr in
  let scanner = new Xnc_scanning.string_stream_scanner stream in
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected int token"
      "10"
      (scanner # scan_int ());
    scanner # skip 1;
    OUnit.assert_raises
      (Xnc_scanning.Scan_failure
         "Scanning int: \
          stream failure")
      (scanner # scan_int);
;;

let test_stream_parse_value _ =
  let config = Pxp_types.default_config in
  let mgr = Pxp_ev_parser.create_entity_manager
              config
              (Pxp_types.from_string
                 "<root>\
                    <int>10</int>\
                    <char>x</char>\
                    <bool>false</bool>\
                    <float>0.5</float>\
                    <string>test</string>\
                  </root>") in
  let stream = Xnc_scanning.pull_parser_to_string_stream config mgr in
  let scanner = new Xnc_scanning.string_stream_scanner stream in
  let lexer = new Xnc_lexing.string_stream_lexer stream in
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected int value"
      10
      (lexer # parse_value
         (fun () ->
            scanner # skip 1;
            scanner # scan_int ())
         int_of_string);
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:Std.string_of_char
      ~msg:"Unexpected char value"
      'x'
      (lexer # parse_value
         (fun () ->
            scanner # skip 1;
            scanner # scan_char ())
         (fun s -> s.[0]));
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected bool value"
      false
      (lexer # parse_value
         (fun () ->
            scanner # skip 1;
            scanner # scan_bool ())
         bool_of_string);
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected float value"
      0.5
      (lexer # parse_value
         (fun () ->
            scanner # skip 1;
            scanner # scan_float ())
         float_of_string);
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected string value"
      "test"
      (lexer # parse_value
         (fun () ->
            scanner # skip 1;
            scanner # scan_string ())
         Std.identity);
    scanner # skip 1;
    lexer # close ();
;;

let test_stream_parse_option _ =
  let config = Pxp_types.default_config in
  let mgr = Pxp_ev_parser.create_entity_manager
              config
              (Pxp_types.from_string
                 "<root>\
                    <none></none>\
                    <some>false</some>\
                  </root>") in
  let stream = Xnc_scanning.pull_parser_to_string_stream config mgr in
  let scanner = new Xnc_scanning.string_stream_scanner stream in
  let lexer = new Xnc_lexing.string_stream_lexer stream in
  let print_bool_option = Serialization_test_helper.print_option
                            string_of_bool in
  let try_parse_null () =
    lexer # try_parse_null () in
  let parse_bool () =
    lexer # parse_value scanner # scan_bool bool_of_string
  in
    scanner # skip 1;
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:print_bool_option
      ~msg:"Unexpected bool option value"
      None
      (lexer # parse_option try_parse_null parse_bool);
    scanner # skip 1;
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:print_bool_option
      ~msg:"Unexpected bool option value"
      (Some false)
      (lexer # parse_option try_parse_null parse_bool);
    scanner # skip 1;
    lexer # close ();
;;

let test_stream_parse_object _ =
  let config = Pxp_types.default_config in
  let mgr = Pxp_ev_parser.create_entity_manager
              config
              (Pxp_types.from_string "<object>1/2</object>") in
  let stream = Xnc_scanning.pull_parser_to_string_stream config mgr in
  let scanner = new Xnc_scanning.string_stream_scanner stream in
  let lexer = new Xnc_lexing.string_stream_lexer stream in
  let obj = new Serialization_test_helper.test_serializable 0 0 in
    scanner # skip 1;
    lexer # parse_object obj (scanner # scan_string);
    OUnit.assert_equal
      ~cmp:Serialization_test_helper.cmp_object
      ~printer:Serialization_test_helper.print_object
      ~msg:"Unexpected object"
      (new Serialization_test_helper.test_serializable 1 2)
      (obj);
    lexer # close ();
;;

let test_stream_parse_homogeneous_enum _ =
  let config = Pxp_types.default_config in
  let mgr = Pxp_ev_parser.create_entity_manager
              config
              (Pxp_types.from_string
                 "<list>\
                    <elem>1.</elem>\
                    <elem>2.</elem>\
                    <elem>3.</elem>\
                    <elem>4.</elem>\
                  </list>") in
  let stream = Xnc_scanning.pull_parser_to_string_stream config mgr in
  let scanner = new Xnc_scanning.string_stream_scanner stream in
  let lexer = new Xnc_lexing.string_stream_lexer stream in
  let parse_float () =
    lexer # parse_value
      (fun () ->
         let _ = scanner # skip 1 in
         let f = scanner # scan_float () in
         let _ = scanner # skip 1 in
           f)
      float_of_string in
  let _ = scanner # skip 1 in
  let enum = lexer # parse_homogeneous_enum parse_float in
  let l = ExtList.List.of_enum enum in
    OUnit.assert_equal
      ~printer:(Serialization_test_helper.print_list string_of_float)
      ~msg:"Unexpected enum value"
      [1.; 2.; 3.; 4.]
      l;
    lexer # close ();
;;

let test_stream_parse_heterogeneous_enum _ =
  let config = Pxp_types.default_config in
  let mgr = Pxp_ev_parser.create_entity_manager
              config
              (Pxp_types.from_string
                 "<list>\
                    <char>a</char>\
                    <string>test</string>\
                    <int>1</int>\
                  </list>") in
  let stream = Xnc_scanning.pull_parser_to_string_stream config mgr in
  let scanner = new Xnc_scanning.string_stream_scanner stream in
  let lexer = new Xnc_lexing.string_stream_lexer stream in
  let enum_schema = ExtList.List.enum [Char '?'; String ""; Int 0] in
  let parse_int () =
    lexer # parse_value
      (fun () ->
         let _ = scanner # skip 1 in
         let i = scanner # scan_int () in
         let _ = scanner # skip 1 in
           i)
      int_of_string in
  let parse_char () =
    lexer # parse_value
      (fun () ->
         let _ = scanner # skip 1 in
         let c = scanner # scan_char () in
         let _ = scanner # skip 1 in
           c)
      (fun s -> String.get s 0) in
  let parse_string () =
    lexer # parse_value
      (fun () ->
         let _ = scanner # skip 1 in
         let s = scanner # scan_string () in
         let _ = scanner # skip 1 in
           s)
      Std.identity in
  let parse_element e =
    match e with
        Int _ -> Int (parse_int ())
      | Char _ -> Char (parse_char ())
      | String _ -> String (parse_string ()) in
  let print_list =
    Serialization_test_helper.print_list
      (fun e -> match e with
           Int i -> string_of_int i
         | Char c -> Std.string_of_char c
         | String s -> s) in
  let _ = scanner # skip 1 in
  let enum = lexer # parse_heterogeneous_enum enum_schema parse_element in
    OUnit.assert_equal
      ~printer:print_list
      ~msg:"Unexpected enum value"
      [Char 'a'; String "test"; Int 1]
      (ExtList.List.of_enum enum);
    lexer # close ();
;;

let test_stream_parse_value_by_name _ =
  let config = Pxp_types.default_config in
  let mgr = Pxp_ev_parser.create_entity_manager
              config
              (Pxp_types.from_string
                 "<list>\
                    <name>1</name>\
                    <x n=\"xyz\"></x>\
                    <xnc:name>test</xnc:name>\
                    <y xnc:n=\"xyz\"></y>\
                    <error>false</error>\
                  </list>") in
  let stream = Xnc_scanning.pull_parser_to_string_stream config mgr in
  let scanner = new Xnc_scanning.string_stream_scanner stream in
  let lexer = new Xnc_lexing.string_stream_lexer stream in
  let parse_int () =
    lexer # parse_value (scanner # scan_int) int_of_string in
  let parse_bool () =
    lexer # parse_value (scanner # scan_bool) bool_of_string in
  let parse_string () =
    lexer # parse_value (scanner # scan_string) Std.identity
  in
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected value"
      1
      (lexer # parse_value_by_name
         "" "name" Xnc_property.empty parse_int);
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value"
      "xyz"
      (lexer # parse_value_by_name
         "" "n"
         (Xnc_property.singleton Xnc_property.attribute)
         parse_string);
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value"
      "test"
      (lexer # parse_value_by_name
         "xnc" "name" Xnc_property.empty parse_string);
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value"
      "xyz"
      (lexer # parse_value_by_name
         "xnc" "n"
         (Xnc_property.singleton Xnc_property.attribute)
         parse_string);
    scanner # skip 1;
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected value"
      false
      (lexer # parse_value_by_name
         "" "error" Xnc_property.empty parse_bool);
    lexer # close ();
;;

(* test suite *)
let suite = "Xml lexer test" >:::
  ["test_has_next" >:: (bracket set_up test_has_next tear_down);
   "test_peek" >:: (bracket set_up test_peek tear_down);
   "test_next" >:: (bracket set_up test_next tear_down);
   "test_scan_int" >:: test_scan_int;
   "test_parse_value" >:: test_parse_value;
   "test_parse_option" >:: test_parse_option;
   "test_parse_object" >:: test_parse_object;
   "test_parse_homogeneous_enum" >:: test_parse_homogeneous_enum;
   "test_parse_heterogeneous_enum" >:: test_parse_heterogeneous_enum;
   "test_parse_value_by_name" >:: test_parse_value_by_name;
   "test_stream_from_pxp_pull_parser" >:: test_stream_from_pxp_pull_parser;
   "test_stream_scan_int" >:: test_stream_scan_int;
   "test_stream_parse_value" >:: test_stream_parse_value;
   "test_stream_parse_option" >:: test_stream_parse_option;
   "test_stream_parse_object" >:: test_stream_parse_object;
   "test_stream_parse_homogeneous_enum" >:: test_stream_parse_homogeneous_enum;
   "test_stream_parse_heterogeneous_enum"
     >:: test_stream_parse_heterogeneous_enum;
   "test_stream_parse_value_by_name" >:: test_stream_parse_value_by_name]
;;

