open OUnit

(* fixture setup and teardown *)
let set_up _ =
  let buf = Buffer.create 16 in
  let printer = new Xnc_printing.Buffer_printer.xml_printer buf in
    (buf, printer)
;;

let tear_down _ =
  ()
;;

(* tests *)
let test_print_string
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
  let formatter = new Xnc_formatting.xml_formatter () in
  let print_value v = printer # print_value v (formatter # format_string) in
    printer # print_field
      "xnc" "key" "escape<>\"'&" Xnc_property.empty print_value;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field"
      "<xnc:key>escape&lt;&gt;&quot;&apos;&amp;</xnc:key>"
      (Buffer.contents buf)
;;

let test_print_element
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
  let print_value v = printer # print_value v string_of_int in
    printer # print_field
      "xnc" "key" 10 Xnc_property.empty print_value;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field"
      "<xnc:key>10</xnc:key>"
      (Buffer.contents buf)
;;

let test_print_attribute
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
  let print_value v = printer # print_value v string_of_int in
    printer # print_field
      "xnc" "key" 10 Xnc_property.empty
      (fun value ->
         printer # print_field
           "" "attr" 5
           (Xnc_property.singleton Xnc_property.attribute)
           print_value;
         print_value value);
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field"
      "<xnc:key attr=\"5\">10</xnc:key>"
      (Buffer.contents buf)
;;

let test_print_option_element
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
  let print_value v = printer # print_value v string_of_int in
  let print_option o = printer # print_option o (fun () -> ()) print_value in
    printer # print_field
      "xnc" "element" (Some 10) Xnc_property.empty print_option;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field (Some 10)"
      "<xnc:element>10</xnc:element>"
      (Buffer.contents buf);
    Buffer.clear buf;
    printer # print_field
      "xnc" "element" None Xnc_property.empty print_option;
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field (None)"
      "<xnc:element></xnc:element>"
      (Buffer.contents buf)
;;

let test_print_text
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
  let print_value v = printer # print_value v string_of_int in
    printer # print_field
      "xnc" "element"
      [ (fun () -> printer # print_field
                     "" "attr" 5
                     (Xnc_property.singleton Xnc_property.attribute)
                     print_value);
        (fun () -> printer # print_field
                     "" "text" 10
                     (Xnc_property.singleton Xnc_property.text)
                     print_value);
      ]
      Xnc_property.empty
      (fun l -> printer # print_enum (ExtList.List.enum l)
                  (fun e -> e ()));
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field"
      "<xnc:element attr=\"5\">10</xnc:element>"
      (Buffer.contents buf)
;;

let test_print_cdata
      ((buf, printer) : Buffer.t * _ Xnc_printing.printer) =
  let print_value v = printer # print_value v string_of_int in
    printer # print_field
      "xnc" "element"
      [ (fun () -> printer # print_field
                     "" "attr" 5
                     (Xnc_property.singleton Xnc_property.attribute)
                     print_value);
        (fun () -> printer # print_field
                     "" "cdata" 10
                     (Xnc_property.singleton Xnc_property.cdata)
                     print_value);
      ]
      Xnc_property.empty
      (fun l -> printer # print_enum (ExtList.List.enum l)
                  (fun e -> e ()));
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field"
      "<xnc:element attr=\"5\"><![CDATA[10]]</xnc:element>"
      (Buffer.contents buf)
;;

let test_pretty_printer _ =
  let buf = Buffer.create 16 in
  let printer = new Xnc_printing.Buffer_pretty_printer.xml_printer buf in
  let print_value v = printer # print_value v string_of_int in
    printer # print_field
      "xnc" "root"
      [ (fun () -> printer # print_field
                     "xmlns" "xnc" "urn:ex-nunc"
                     (Xnc_property.singleton Xnc_property.attribute)
                     (fun v -> printer # print_value v Std.identity));
        (fun () -> printer # print_field
                     "" "key" 10 Xnc_property.empty print_value);
        (fun () ->
           printer # print_field
             "xnc" "element"
             [ (fun () -> printer # print_field
                            "" "attr" 5
                            (Xnc_property.singleton Xnc_property.attribute)
                            print_value);
               (fun () -> printer # print_field
                            "" "cdata" 10
                            (Xnc_property.singleton Xnc_property.cdata)
                            print_value);
             ]
             Xnc_property.empty
             (fun l -> printer # print_enum (ExtList.List.enum l)
                              (fun e -> e ())));
      ]
      Xnc_property.empty
      (fun l -> printer # print_enum (ExtList.List.enum l)
                  (fun e -> e ()));
    printer # close ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected buffer content after print_field"
      "\
<xnc:root xmlns:xnc=\"urn:ex-nunc\">
  <key>10</key>
  <xnc:element attr=\"5\"><![CDATA[10]]</xnc:element>
</xnc:root>
"
      (Buffer.contents buf)
;;

(* test suite *)
let suite = "XML printer test" >:::
  ["test_print_string"
     >:: (bracket set_up test_print_string tear_down);
   "test_print_element"
     >:: (bracket set_up test_print_element tear_down);
   "test_print_attribute"
     >:: (bracket set_up test_print_attribute tear_down);
   "test_print_option_element"
     >:: (bracket set_up test_print_option_element tear_down);
   "test_print_text"
     >:: (bracket set_up test_print_text tear_down);
   "test_print_cdata"
     >:: (bracket set_up test_print_cdata tear_down);
   "test_pretty_printer" >:: test_pretty_printer]
;;

