type t =
    { fields : string Xnc_serializable.field;
      mutable serializer : string Xnc_serializing.serializer
    }
;;

let create
      ?(serializer = new Xnc_serializing.simple_serializer ())
      label =
  { fields = Xnc_serializable.create_empty_array label;
    serializer = serializer
  }
;;

let length fc =
  Xnc_serializable.count_leaves fc.fields
;;

let register_field fc field =
  Xnc_serializable.append_field fc.fields field
;;

let register_serializable ?properties ?prefix fc name value =
  let field = Xnc_serializable.create ?properties ?prefix name value in
    register_field fc field
;;

let register_value ?properties ?prefix fc name value wrap_value =
  let field = Xnc_serializable.new_field
                ?properties ?prefix
                name value wrap_value in
    register_field fc field
;;

let register_int ?properties ?prefix fc name value =
  register_value ?properties ?prefix fc name value Xnc_serializable.wrap_int
;;

let register_float ?properties ?prefix fc name value =
  register_value ?properties ?prefix fc name value Xnc_serializable.wrap_float
;;

let register_char ?properties ?prefix fc name value =
  register_value ?properties ?prefix fc name value Xnc_serializable.wrap_char
;;

let register_string ?properties ?prefix fc name value =
  register_value ?properties ?prefix fc name value Xnc_serializable.wrap_string
;;

let register_bool ?properties ?prefix fc name value =
  register_value ?properties ?prefix fc name value Xnc_serializable.wrap_bool
;;

let register_date ?properties ?prefix fc name value =
  register_value ?properties ?prefix fc name value Xnc_serializable.wrap_date
;;

let register_option ?properties ?prefix fc name o wrap_value =
  register_value ?properties ?prefix fc name o
    (fun v -> Xnc_serializable.wrap_option (Option.map wrap_value v));
;;

let register_custom ?properties ?prefix fc name value wrap_value writer reader =
  let field = Xnc_serializable.create_custom ?properties ?prefix name
                writer reader
                (wrap_value value) in
    register_field fc field
;;

let register_object ?properties ?prefix fc name o =
  register_value ?properties ?prefix fc name o Xnc_serializable.wrap_object;
;;

(* private *)
let handle_collection
      ?element_properties
      ?element_prefix
      ?(element_name = "Element")
      handle_field
      fc name c wrap_value
      wrap_collection map_collection =
  handle_field fc name c
    (fun v -> wrap_collection
                (map_collection
                   (fun e ->
                      Xnc_serializable.new_field
                        ?properties:element_properties
                        ?prefix:element_prefix
                        element_name e wrap_value)
                   v));
;;

let register_dynarray
      ?element_properties
      ?element_prefix
      ?element_name
      fc name d wrap_value =
  handle_collection
    ?element_properties
    ?element_prefix
    ?element_name
    register_value
    fc name d wrap_value
    Xnc_serializable.wrap_dynarray DynArray.map
;;

let register_array
      ?element_properties
      ?element_prefix
      ?element_name
      fc name a wrap_value =
  handle_collection
    ?element_properties
    ?element_prefix
    ?element_name
    register_value
    fc name a wrap_value
    Xnc_serializable.wrap_array Array.map
;;

let register_list
      ?element_properties
      ?element_prefix
      ?element_name
      fc name l wrap_value =
  handle_collection
    ?element_properties
    ?element_prefix
    ?element_name
    register_value
    fc name l wrap_value
    Xnc_serializable.wrap_list List.map
;;

let is_registered ?properties ?prefix fc name =
  Xnc_serializable.value_present ?properties ?prefix fc.fields name
;;

let get_field ?properties ?prefix fc name =
  Xnc_serializable.find_field ?properties ?prefix fc.fields name
;;

let get_serializable ?properties ?prefix fc name =
  Xnc_serializable.find_value ?properties ?prefix fc.fields name
;;

let get_value ?properties ?prefix fc name unwrap_value =
  let value = get_serializable ?properties ?prefix fc name in
    unwrap_value value 
;;

let get_int ?properties ?prefix fc name =
  get_value ?properties ?prefix fc name Xnc_serializable.unwrap_int
;;

let get_float ?properties ?prefix fc name =
  get_value ?properties ?prefix fc name Xnc_serializable.unwrap_float
;;

let get_char ?properties ?prefix fc name =
  get_value ?properties ?prefix fc name Xnc_serializable.unwrap_char
;;

let get_string ?properties ?prefix fc name =
  get_value ?properties ?prefix fc name Xnc_serializable.unwrap_string
;;

let get_bool ?properties ?prefix fc name =
  get_value ?properties ?prefix fc name Xnc_serializable.unwrap_bool
;;

let get_date ?properties ?prefix fc name =
  get_value ?properties ?prefix fc name Xnc_serializable.unwrap_date
;;

let get_option ?properties ?prefix fc name unwrap_value =
  get_value ?properties ?prefix fc name
    (fun v ->
       let option_value = Xnc_serializable.unwrap_option v in
         Option.map unwrap_value option_value)
;;

let get_object ?properties ?prefix fc name =
  get_value ?properties ?prefix fc name Xnc_serializable.unwrap_object
;;

let get_custom ?properties ?prefix fc name unwrap_value =
  get_value ?properties ?prefix fc name
    (fun v -> unwrap_value (Xnc_serializable.unwrap_custom v))
;;

(* private *)
let get_collection
      ?properties 
      ?prefix
      fc name unwrap_value
      unwrap_collection map_collection =
  get_value ?properties ?prefix fc name
    (fun v ->
       let c = unwrap_collection v in
         map_collection
           (fun e ->
              unwrap_value (Xnc_serializable.get_value e))
           c)
;;

let get_dynarray ?properties ?prefix fc name unwrap_value =
  get_collection
    ?properties 
    ?prefix 
    fc name unwrap_value
    Xnc_serializable.unwrap_dynarray DynArray.map
;;

let get_array ?properties ?prefix fc name unwrap_value =
  get_collection
    ?properties 
    ?prefix 
    fc name unwrap_value
    Xnc_serializable.unwrap_array Array.map
;;

let get_list ?properties ?prefix fc name unwrap_value =
  get_collection
    ?properties 
    ?prefix 
    fc name unwrap_value
    Xnc_serializable.unwrap_list List.map
;;

let update_field fc field =
  Xnc_serializable.replace_field fc.fields field
;;

let update_serializable ?properties ?prefix fc name value =
  let field = Xnc_serializable.create ?properties ?prefix name value in
    update_field fc field
;;

let update_value ?properties ?prefix fc name value wrap_value =
  let field = Xnc_serializable.new_field
                ?properties ?prefix
                name value wrap_value in
    update_field fc field
;;

let update_int ?properties ?prefix fc name value =
  update_value ?properties ?prefix fc name value Xnc_serializable.wrap_int
;;

let update_float ?properties ?prefix fc name value =
  update_value ?properties ?prefix fc name value Xnc_serializable.wrap_float
;;

let update_char ?properties ?prefix fc name value =
  update_value ?properties ?prefix fc name value Xnc_serializable.wrap_char
;;

let update_bool ?properties ?prefix fc name value =
  update_value ?properties ?prefix fc name value Xnc_serializable.wrap_bool
;;

let update_string ?properties ?prefix fc name value =
  update_value ?properties ?prefix fc name value Xnc_serializable.wrap_string
;;

let update_date ?properties ?prefix fc name value =
  update_value ?properties ?prefix fc name value Xnc_serializable.wrap_date
;;

let update_option ?properties ?prefix fc name value wrap_value =
  update_value ?properties ?prefix fc name value
    (fun v -> Xnc_serializable.wrap_option (Option.map wrap_value v))
;;

let update_object ?properties ?prefix fc name value =
  update_value ?properties ?prefix fc name value Xnc_serializable.wrap_object
;;

let update_custom ?properties ?prefix fc name value wrap_value writer reader =
  let field = Xnc_serializable.create_custom ?properties ?prefix name
                writer reader
                (wrap_value value) in
    update_field fc field
;;

let update_dynarray
      ?prefix
      ?element_properties
      ?element_prefix
      ?element_name
      fc name d wrap_value =
  handle_collection
    ?element_properties
    ?element_prefix
    ?element_name
    (update_value ?prefix)
    fc name d wrap_value
    Xnc_serializable.wrap_dynarray DynArray.map
;;

let update_array
      ?prefix
      ?element_properties
      ?element_prefix
      ?element_name
      fc name d wrap_value =
  handle_collection
    ?element_properties
    ?element_prefix
    ?element_name
    (update_value ?prefix)
    fc name d wrap_value
    Xnc_serializable.wrap_array Array.map
;;

let update_list
      ?prefix
      ?element_properties
      ?element_prefix
      ?element_name
      fc name d wrap_value =
  handle_collection
    ?element_properties
    ?element_prefix
    ?element_name
    (update_value ?prefix)
    fc name d wrap_value
    Xnc_serializable.wrap_list List.map
;;

let to_name_value_string_list fc =
  Xnc_serializable.to_name_value_string_list fc.fields
;;

(* Serialization/Deserialization *)

(* private *)
let set_optional_serializer fc serializer =
  match serializer with
      None ->
        ()
    | Some s ->
        fc.serializer <- s
;;

let to_string ?serializer fc =
  set_optional_serializer fc serializer;
  fc.serializer # serialize fc.fields
;;

let to_netchannel ?serializer fc out_ch =
  set_optional_serializer fc serializer;
  fc.serializer # to_netchannel fc.fields out_ch
;;

let from_string ?serializer fc serialization =
  set_optional_serializer fc serializer;
  fc.serializer # deserialize fc.fields serialization
;;

let from_netchannel ?serializer fc in_ch =
  set_optional_serializer fc serializer;
  fc.serializer # from_netchannel fc.fields in_ch
;;

