(** Serializable field collections
  *
  * This module implements a mutable collection of serializable fields. A
  * collection is built registering fields and may be serialized using a string
  * serializer. Functions are provided to extract values from the collection and
  * to update fields.
  *)

type t
(** The type of serializable field collections. *)

val create :
  ?serializer:Xnc_serializing.simple_serializer ->
  string -> t
(** [create serializer label] creates an empty field collection, using [label]
  * as the name of the container. The collection is serialized/deserialized
  * using [serializer] (the default serializer is
  * [Xnc_serializing.simple_serializer]).
  *)

val length : t -> int
(** [length fc] returns the number of elements stored in the collection [fc]. *)

val register_field : t -> string Xnc_serializable.field -> unit
(** [register_field fc f] adds the serializable field [f] to the collection
  * [fc].
  *)

val register_serializable :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> string Xnc_serializable.serializable_value -> unit
(** [register_serializable fc n v] adds a serializable field with value [v] and
  * name [n] to the collection [fc].
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_value :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> 'a ->
  ('a -> string Xnc_serializable.serializable_value) -> unit
(** [register_value fc n v w] adds a field with value [v] wrapped by [w], and
  * name [n] to the collection [fc].
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_int :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> int -> unit
(** [register_int fc n v] adds a field with value [v] and name [n] to the
  * collection [fc].
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_float :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> float -> unit
(** [register_float fc n v] adds a field with value [v] and name [n] to the
  * collection [fc].
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_char :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> char -> unit
(** [register_char fc n v] adds a field with value [v] and name [n] to the
  * collection [fc].
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_string :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> string -> unit
(** [register_string fc n v] adds a field with value [v] and name [n] to the
  * collection [fc].
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_bool :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> bool -> unit
(** [register_bool fc n v] adds a field with value [v] and name [n] to the
  * collection [fc].
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_date :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> Netdate.t -> unit
(** [register_date fc n v] adds a field with value [v] and name [n] to the
  * collection [fc].
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_option :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> 'a option ->
  ('a -> string Xnc_serializable.serializable_value) -> unit
(** [register_date fc n v w] adds a field with value [v] wrapped by [w], and
  * name [n] to the collection [fc].
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_object :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> #Xnc_serialization.serializable -> unit
(** [register_object fc n v] adds a field with value [v] and name [n] to the
  * collection [fc].
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_custom :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> 'a ->
  ('a -> string Xnc_serializable.serializable_value) ->
  string Xnc_serializable.value_writer ->
  string Xnc_serializable.value_reader -> unit
(** [register_custom fc n v w writer reader] adds a field with value [v] wrapped
  * by [w], and name [n] to the collection [fc]. A custom value is a value
  * serialized using a custom [writer] function and deserialized using a custom
  * [reader] function.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val register_dynarray :
  ?element_properties:Xnc_property.properties ->
  ?element_prefix:string ->
  ?element_name:string ->
  t -> string -> 'a DynArray.t ->
  ('a -> string Xnc_serializable.serializable_value) -> unit
(** [register_dynarray fc n v w] adds a field with value [v] and name [n] to the
  * collection [fc]. Each element of array [v] is wrapped by [w].
  *
  * @param element_properties Optional field properties added to each element.
  * @param element_prefix Optional name prefix (namespace) given to each
  * element.
  * @param element_name Optional name given to each element. Defaults to
  * ["Element"].
  *)

val register_array :
  ?element_properties:Xnc_property.properties ->
  ?element_prefix:string ->
  ?element_name:string ->
  t -> string -> 'a array ->
  ('a -> string Xnc_serializable.serializable_value) -> unit
(** [register_array fc n v w] adds a field with value [v] and name [n] to the
  * collection [fc]. Each element of array [v] is wrapped by [w].
  *
  * @param element_properties Optional field properties added to each element.
  * @param element_prefix Optional name prefix (namespace) given to each
  * element.
  * @param element_name Optional name given to each element. Defaults to
  * ["Element"].
  *)

val register_list :
  ?element_properties:Xnc_property.properties ->
  ?element_prefix:string ->
  ?element_name:string ->
  t -> string -> 'a list ->
  ('a -> string Xnc_serializable.serializable_value) -> unit
(** [register_list fc n v w] adds a field with value [v] and name [n] to the
  * collection [fc]. Each element of list [v] is wrapped by [w].
  *
  * @param element_properties Optional field properties added to each element.
  * @param element_prefix Optional name prefix (namespace) given to each
  * element.
  * @param element_name Optional name given to each element. Defaults to
  * ["Element"].
  *)

val is_registered :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> bool
(** [is_registered fc n] tests if a field named [n] is registered in the
  * collection [fc].
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_field :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> string Xnc_serializable.field
(** [get_field fc n] extracts the serializable field named [n] from the
  * collection [fc]. Raises [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_serializable :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> string Xnc_serializable.serializable_value
(** [get_field fc n] extracts the serializable value named [n] from the
  * collection [fc]. Raises [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_value :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> (string Xnc_serializable.serializable_value -> 'a) -> 'a
(** [get_field fc n unwrap] extracts the value named [n] from the collection
  * [fc], unwrapping it using function [unwrap]. Raises [Not_found] if no field
  * matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_int :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> int
(** [get_int fc n] extracts the [int] value named [n] from the collection [fc].
  * Raises [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_float :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> float
(** [get_float fc n] extracts the [float] value named [n] from the collection
  * [fc]. Raises [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_char :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> char
(** [get_char fc n] extracts the [char] value named [n] from the collection
  * [fc]. Raises [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_string :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> string
(** [get_string fc n] extracts the [string] value named [n] from the collection
  * [fc]. Raises [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_bool :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> bool
(** [get_bool fc n] extracts the [bool] value named [n] from the collection
  * [fc]. Raises [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_date :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> Netdate.t
(** [get_date fc n] extracts the [Netdate.t] value named [n] from the collection
  * [fc]. Raises [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_option :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string ->
  (string Xnc_serializable.serializable_value -> 'a) -> 'a option
(** [get_option fc n unwrap] extracts the [option] value named [n] from the
  * collection [fc], unwrapping it using [unwrap] function. Raises [Not_found]
  * if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_object :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> Xnc_serialization.serializable
(** [get_object fc n] extracts the serializable object named [n] from the
  * collection [fc]. Raises [Not_found] if no field matches.
  * 
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_custom :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string ->
  (string Xnc_serializable.serializable_value -> 'a) -> 'a
(** [get_custom fc n unwrap] extracts the custom value named [n] from the
  * collection [fc], unwrapping it using [unwrap] function. Raises [Not_found]
  * if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_dynarray :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string ->
  (string Xnc_serializable.serializable_value -> 'a) -> 'a DynArray.t
(** [get_dynarray fc n unwrap] extracts the [DynArray.t] value named [n] from
  * the collection [fc], unwrapping each element using [unwrap] function. Raises
  * [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_array :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string ->
  (string Xnc_serializable.serializable_value -> 'a) -> 'a array
(** [get_array fc n unwrap] extracts the [array] value named [n] from the
  * collection [fc], unwrapping each element using [unwrap] function. Raises
  * [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val get_list :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string ->
  (string Xnc_serializable.serializable_value -> 'a) -> 'a list
(** [get_list fc n unwrap] extracts the [list] value named [n] from the
  * collection [fc], unwrapping each element using [unwrap] function. Raises
  * [Not_found] if no field matches.
  *
  * @param properties Optional subset of properties to match.
  * @param prefix Optional name prefix to match.
  *)

val update_field : t -> string Xnc_serializable.field -> unit
(** [update_field fc f] replaces a serializable field that has the same
  * name and prefix of field [f] with field [f] in the collection [fc]. No
  * action is performed if no field matches.
  *)

val update_serializable :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> string Xnc_serializable.serializable_value -> unit
(** [update_serializable fc n v] updates a serializable field that has name [n]
  * with value [v] in the collection [fc]. No action is performed if no field
  * matches.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_value :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> 'a ->
  ('a -> string Xnc_serializable.serializable_value) -> unit
(** [update_value fc n v w] updates a serializable field that has name [n] with
  * the new value [v], wrapped by [w], in the collection [fc]. No action is
  * performed if no field matches.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_int :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> int -> unit
(** [update_int fc n v] updates a serializable field that has name [n] with
  * the new value [v]. No action is performed if no field matches.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_float :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> float -> unit
(** [update_float fc n v] updates a serializable field that has name [n] with
  * the new value [v]. No action is performed if no field matches.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_char :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> char -> unit
(** [update_char fc n v] updates a serializable field that has name [n] with
  * the new value [v]. No action is performed if no field matches.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_string :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> string -> unit
(** [update_string fc n v] updates a serializable field that has name [n] with
  * the new value [v]. No action is performed if no field matches.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_bool :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> bool -> unit
(** [update_bool fc n v] updates a serializable field that has name [n] with
  * the new value [v]. No action is performed if no field matches.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_date :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> Netdate.t -> unit
(** [update_date fc n v] updates a serializable field that has name [n] with
  * the new value [v]. No action is performed if no field matches.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_option :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> 'a option ->
  ('a -> string Xnc_serializable.serializable_value) -> unit
(** [update_option fc n v w] updates a serializable field that has name [n] with
  * the new value [v], wrapped by [w]. No action is performed if no field
  * matches.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_custom :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> 'a ->
  ('a -> string Xnc_serializable.serializable_value) ->
  string Xnc_serializable.value_writer ->
  string Xnc_serializable.value_reader -> unit
(** [update_custom fc n v w writer reader] updates a serializable field that has
  * name [n] with the new custom value [v], wrapped by [w]. No action is
  * performed if no field matches. A custom value is a value serialized using a
  * custom [writer] function and deserialized using a custom [reader] function.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_object :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  t -> string -> #Xnc_serialization.serializable -> unit
(** [update_object fc n v] updates a serializable field that has name [n] with
  * the new value [v]. No action is performed if no field matches.
  *
  * @param properties Optional field properties.
  * @param prefix Optional name prefix (namespace).
  *)

val update_dynarray :
  ?prefix:string ->
  ?element_properties:Xnc_property.properties ->
  ?element_prefix:string ->
  ?element_name:string ->
  t -> string -> 'a DynArray.t ->
  ('a -> string Xnc_serializable.serializable_value) -> unit
(** [update_dynarray fc n v w] updates a serializable field that has name [n]
  * with the new value [v], wrapping each element with [w] function. No action
  * is performed if no field matches.
  *
  * @param prefix Optional name prefix (namespace) used to match the right
  * field.
  * @param element_properties Optional field properties added to each element.
  * @param element_prefix Optional name prefix (namespace) given to each
  * element.
  * @param element_name Optional name given to each element. Defaults to
  * ["Element"].
  *)

val update_array :
  ?prefix:string ->
  ?element_properties:Xnc_property.properties ->
  ?element_prefix:string ->
  ?element_name:string ->
  t -> string -> 'a array ->
  ('a -> string Xnc_serializable.serializable_value) -> unit
(** [update_array fc n v w] updates a serializable field that has name [n] with
  * the new value [v], wrapping each element with [w]. No action is performed if
  * no field matches.
  *
  * @param prefix Optional name prefix (namespace) used to match the right
  * field.
  * @param element_properties Optional field properties added to each element.
  * @param element_prefix Optional name prefix (namespace) given to each
  * element.
  * @param element_name Optional name given to each element. Defaults to
  * ["Element"].
  *)

val update_list :
  ?prefix:string ->
  ?element_properties:Xnc_property.properties ->
  ?element_prefix:string ->
  ?element_name:string ->
  t -> string -> 'a list ->
  ('a -> string Xnc_serializable.serializable_value) -> unit
(** [update_list fc n v w] updates a serializable field that has name [n] with
  * the new value [v], wrapping each element with [w]. No action is performed if
  * no field matches.
  *
  * @param prefix Optional name prefix (namespace) used to match the right
  * field.
  * @param element_properties Optional field properties added to each element.
  * @param element_prefix Optional name prefix (namespace) given to each
  * element.
  * @param element_name Optional name given to each element. Defaults to
  * ["Element"].
  *)

val to_name_value_string_list : t -> (string * string) list
(** [to_name_value_string_list fc] builds a name/value pair [string list]
  * representing the content of [fc]. *)

val to_string :
  ?serializer:string Xnc_serializing.serializer ->
  t -> string
(** [to_string fc] serializes the collection [fc] to a [string].
  *
  * @param serializer Serializer to use. Defaults to current serializer. *)

val to_netchannel :
  ?serializer:string Xnc_serializing.serializer ->
  t -> Netchannels.out_obj_channel -> unit
(** [to_string fc ch] serializes the collection [fc] to a
  * the output channel [ch].
  *
  * @param serializer Serializer to use. Defaults to current serializer. *)

val from_string :
  ?serializer:string Xnc_serializing.serializer -> t -> string -> unit
(** [from_string fc s] deserializes the [string] value [s] that contains a valid
  * serialization of the collection [fc].
  *
  * @param serializer Serializer to use. Defaults to current serializer. *)

val from_netchannel :
  ?serializer:string Xnc_serializing.serializer ->
  t -> Netchannels.in_obj_channel -> unit
(** [from_netchannel fc ch] deserializes a valid serialization of the collection
  * [fc], reading from the input channel [ch].
  *
  * @param serializer Serializer to use. Defaults to current serializer. *)

