(** Basic formatting classes
  *
  * This module provides the low level interface and classes used to format
  * values.
  *)

class type ['a] formatter =
object
  method format_null :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  (** [format_null ()] formats the null value. *)

  method format_int :
    ?properties:Xnc_property.properties ->
    int -> 'a
  (** [format_int v] formats value [v]. *)

  method format_float :
    ?properties:Xnc_property.properties ->
    float -> 'a
  (** [format_float v] formats value [v]. *)

  method format_char :
    ?properties:Xnc_property.properties ->
    char -> 'a
  (** [format_char v] formats value [v]. *)

  method format_bool :
    ?properties:Xnc_property.properties ->
    bool -> 'a
  (** [format_bool v] formats value [v]. *)

  method format_raw_string :
    ?properties:Xnc_property.properties ->
    string -> 'a
  (** [format_raw_string s] formats string [s] as is. *)

  method format_string :
    ?properties:Xnc_property.properties ->
    string -> 'a
  (** [format_string s] formats string [s] replacing all the special characters
    * with proper escape sequences. *)

  method format_date :
    ?properties:Xnc_property.properties ->
    Netdate.t -> 'a
  (** [format_date v] formats value [v]. *)

end
(** This interface collects methods used to convert a value of a primitive type
  * to its serialized representation (a value of type ['a]). All the methods
  * share a [properties] parameter that is used to pass optional metadata.
  *)

class virtual basic_formatter :
  unit ->
  [string] formatter
(** This abstract class gives a default implemetation of the string formatter.
  * It implements the following behavior:
  * - converts null values to the empty string
  * - converts [int] values using [string_of_int]
  * - converts [float] values using [string_of_float]
  * - converts [char] values using [Std.string_of_char]
  * - converts [bool] values using [string_of_bool]
  * - [string] values are returned as they are
  * - converts [Netdate.t] values using this format: [%m/%d/%Y %T]
  * It can be used as a base class to write custom string formatters.
  *)

class standard_formatter :
  ?config:Xnc_serialization.common_config ->
  unit ->
  [string] formatter
(** This class implements the methods used to produce the simple serialization
  * format used by the framework. It inherits its behavior from
  * [basic_formatter] and reimplements [format_null] and [format_string]
  * methods:
  * - converts null values to the string specified by the [config] parameter
  * - converts [string] values enclosing them in double quotes and escaping
  * contained double quotes if present
  * @param config specifies the set of keywords of the language
  *)

class xml_formatter :
  unit ->
  [string] formatter
(** This class implements the methods used to produce XML documents. It inherits
  * its behavior from [basic_formatter] and reimplements [format_string] method:
  * converts [string] values, escaping characters that need protection in XML
  * text by converting them to entity references.
  *)

