class type ['a] lexical_analyzer =
object
  method try_parse_null : ?properties:Xnc_property.properties -> unit -> bool
  method parse_value : (unit -> 'a) -> ('a -> 'b) -> 'b
  method parse_option : (unit -> bool) -> (unit -> 'c) -> 'c option
  method parse_object :
    'd. (#Xnc_serialization.serializable as 'd) -> (unit -> string) -> unit
  method parse_homogeneous_enum : (unit -> 'e) -> 'e Enum.t
  method parse_heterogeneous_enum : 'f Enum.t -> ('f -> 'f) -> 'f Enum.t
  method parse_value_by_name :
    string -> string -> Xnc_property.properties -> (unit -> 'g) -> 'g
  method close : unit -> unit
end;;

class virtual text_lexer () =
object
  method parse_value : 'a. (unit -> string) -> (string -> 'a) -> 'a =
    fun scan_token parse_token ->
      let token = scan_token () in
        parse_token token

  method parse_option : 'a. (unit -> bool) -> (unit -> 'a) -> 'a option =
    fun try_parse_null parse_value ->
      if try_parse_null () then
        None
      else
        Some (parse_value ())

  method parse_object :
      'a. (#Xnc_serialization.serializable as 'a) ->
      (unit -> string) -> unit  =
    fun serializable scan_object ->
      let serialization = scan_object () in
        serializable # from_string serialization

end;;

class xnc_scanbuf_lexer
  ?(config = Xnc_serialization.default_config)
  scanbuf : [string] lexical_analyzer =
object
  inherit text_lexer ()

  val config = config
  val scanbuf = scanbuf

  method try_parse_null ?properties () =
    let none = config.Xnc_serialization.none_string in
      Xnc_scanning.Scanning_buffer.match_literal scanbuf none

  method parse_homogeneous_enum : 'a. (unit -> 'a) -> 'a Enum.t =
    fun parse_element ->
      Enum.from
        (fun () ->
           if Xnc_scanning.Scanning_buffer.match_literal scanbuf
                config.Xnc_serialization.end_of_list then
             raise Enum.No_more_elements
           else if not (Xnc_scanning.Scanning_buffer.match_literal scanbuf
                          config.Xnc_serialization.start_of_list) then
             Xnc_scanning.Scanning_buffer.parse_literal scanbuf
               config.Xnc_serialization.list_separator
           else
             ();
           parse_element ()
        )

  method parse_heterogeneous_enum : 'a. 'a Enum.t -> ('a -> 'a) -> 'a Enum.t =
    fun enum_schema parse_element ->
      Xnc_scanning.Scanning_buffer.parse_literal scanbuf
        config.Xnc_serialization.start_of_list;
      Enum.map 
        (fun schema_element ->
           let value = parse_element schema_element in
             if not (Xnc_scanning.Scanning_buffer.match_literal scanbuf
                       config.Xnc_serialization.list_separator) then
               Xnc_scanning.Scanning_buffer.parse_literal scanbuf
                 config.Xnc_serialization.end_of_list
             else
               ();
             value)
        enum_schema

  method parse_value_by_name :
      'a. string -> string -> Xnc_property.properties ->
      (unit -> 'a) -> 'a =
    fun prefix name properties parse_value ->
      if Xnc_property.is_attribute properties then
        Xnc_scanning.Scanning_buffer.parse_literal scanbuf
          config.Xnc_serialization.attribute_marker
      else
        ();
      if prefix <> "" then
        begin
          Xnc_scanning.Scanning_buffer.parse_literal scanbuf prefix;
          Xnc_scanning.Scanning_buffer.parse_literal scanbuf
            config.Xnc_serialization.prefix_separator
        end
      else
        ();
      Xnc_scanning.Scanning_buffer.parse_literal scanbuf name;
      Xnc_scanning.Scanning_buffer.parse_literal scanbuf
        config.Xnc_serialization.name_value_separator;
      parse_value ()

  method close () =
    Xnc_scanning.Scanning_buffer.close scanbuf

end;;

let whitespace_regexp = Str.regexp "^[ \t\r\n]+$";;

class virtual stream_lexer () =
object
  inherit text_lexer ()

  method parse_heterogeneous_enum : 'a. 'a Enum.t -> ('a -> 'a) -> 'a Enum.t =
    fun enum_schema parse_element ->
      Enum.map 
        (fun schema_element ->
           parse_element schema_element)
        enum_schema

  method close () =
    ()

end;;

(* TODO: refactor with string_stream_lexer *)
class ['ext] pxp_node_stream_lexer
  (stream : 'ext Pxp_document.node Stream.t) : [string] lexical_analyzer =
object
  inherit stream_lexer ()

  method try_parse_null ?properties () =
    let node = Option.get (Stream.peek stream) in
      match node # node_type with
          Pxp_document.T_data -> false
        | _ -> true

  method parse_homogeneous_enum : 'a. (unit -> 'a) -> 'a Enum.t =
    fun parse_element ->
      Enum.from
        (fun () ->
           let node = Stream.peek stream in
             match node with
                 None ->
                   raise Enum.No_more_elements
               | _ ->
                   parse_element ())

  method parse_value_by_name :
      'a. string -> string -> Xnc_property.properties ->
      (unit -> 'a) -> 'a =
    fun prefix name properties parse_value ->
      let qname =
        if String.length prefix > 0 then
          prefix ^ ":" ^ name
        else
          name in
      let rec skip_whitespaces () =
        let n = Option.get (Stream.peek stream) in
          match n # node_type with
              Pxp_document.T_data ->
                let data = n # data in
                  if Str.string_match whitespace_regexp data 0 then
                    begin
                      Stream.junk stream;
                      skip_whitespaces ()
                    end
                  else
                    ()
            | _ -> () in
      let xnc_node_type = Xnc_property.get_node_type properties in
      let _ = skip_whitespaces () in
      let node = Option.get (Stream.peek stream) in
        match xnc_node_type with
            Xnc_property.Element ->
              begin match node # node_type with
                  Pxp_document.T_element name when name = qname ->
                    let _ = Stream.next stream in
                      parse_value ()
                | _ ->
                    failwith ("Element \"" ^ name ^ "\" not found")
              end
          | Xnc_property.Attribute ->
              begin match node # node_type with
                  Pxp_document.T_attribute name when name = qname ->
                    parse_value ()
                | _ ->
                    failwith ("Attribute \"" ^ name ^ "\" not found")
              end
          | Xnc_property.Text
          | Xnc_property.Cdata ->
              let _ = Stream.next stream in
                parse_value ()

end;;

class string_stream_lexer stream : [string] lexical_analyzer =
object
  inherit stream_lexer ()

  method try_parse_null ?properties () =
    let token = Stream.peek stream in
      match token with
          None -> true
        | Some s when s = Xnc_scanning.eoe -> true
        | _ -> false

  method parse_homogeneous_enum : 'a. (unit -> 'a) -> 'a Enum.t =
    fun parse_element ->
      Enum.from
        (fun () ->
           let token = Stream.peek stream in
             match token with
                 None ->
                   raise Enum.No_more_elements
               | Some s when s = Xnc_scanning.eoe ->
                   raise Enum.No_more_elements
               | _ ->
                   parse_element ())

  method parse_value_by_name :
      'a. string -> string -> Xnc_property.properties ->
      (unit -> 'a) -> 'a =
    fun prefix name properties parse_value ->
      let qname =
        if String.length prefix > 0 then
          prefix ^ ":" ^ name
        else
          name in
      let check_name n node_type =
        if n <> qname then
          failwith ("Expected " ^ node_type ^ " named \"" ^
                    qname ^ "\" but got \"" ^ n ^ "\"")
        else
          () in
      let rec skip_whitespaces () =
        match Stream.peek stream with
          | Some token when (Str.string_match
                               whitespace_regexp token 0) ->
              Stream.junk stream;
              skip_whitespaces ()
          | _ -> () in
      let check_eoe name =
        skip_whitespaces ();
        match Stream.peek stream with
          | Some token when token = Xnc_scanning.eoe ->
              Stream.junk stream
          | _ ->
              failwith ("Missing end tag for element \"" ^ name) in
      let xnc_node_type = Xnc_property.get_node_type properties in
        match xnc_node_type with
            Xnc_property.Element ->
              let _ = skip_whitespaces () in
              let token = Stream.next stream in
              let _ = check_name token "element" in
              let value = parse_value () in
              let _ = check_eoe token in
                value
          | Xnc_property.Attribute ->
              let token = Stream.next stream in
                check_name token "attribute";
                parse_value ()
          | _ ->
              parse_value ()

end;;

