(** Basic lexing classes
  *
  * This module provides the classes needed to parse values from a
  * serialization support. The lexical analyzers implemented in this module use
  * scanning functions to obtain tokens from the serialization support and
  * evaluating functions to convert tokens to typed values.
  *)

class type ['a] lexical_analyzer =
object
  method try_parse_null : ?properties:Xnc_property.properties -> unit -> bool
  (** [try_parse_null ()] tries to parse a null value. Returns [true] if the the
    * operation succedes, otherwise returns [false] without consuming any input.
    *)

  method parse_value : (unit -> 'a) -> ('a -> 'b) -> 'b
  (** [parse_value s e] parses a value of type ['b], scanning it with [s] and
    * evaluating it with [e]. *)

  method parse_option : (unit -> bool) -> (unit -> 'c) -> 'c option
  (** [parse_option n p] parses an option value. [n] tests the input to check if
    * a [None] value is present, otherwise [p] is used to parse a value of type
    * ['c].*)

  method parse_object :
    'd. (#Xnc_serialization.serializable as 'd) -> (unit -> string) -> unit
  (** [parse_object o s] deserializes the serializable object [o], using [s] to
    * parse the object serialization. *)

  method parse_homogeneous_enum : (unit -> 'e) -> 'e Enum.t
  (** [parse_homogeneous_enum pe] parses a homogeneous enumeration, using [pe]
    * to parse each element. *)

  method parse_heterogeneous_enum : 'f Enum.t -> ('f -> 'f) -> 'f Enum.t
  (** [parse_heterogeneous_enum e pe] parses a heterogeneous enumeration, using
    * [e] as the enumeration descriptor (an enumeration that shares the same
    * structure of the one to be parsed) and [pe] to parse each element. *)

  method parse_value_by_name :
    string -> string -> Xnc_property.properties -> (unit -> 'g) -> 'g
  (** [parse_value_by_name p n prop pv] parses a field with prefix [p], name
    * [n], and properties [prop], using [pv] to parse a value of type ['g]. *)

  method close : unit -> unit
  (** [close ()] closes the serialization support. *)

end
(** This interface collects methods used to parse values from a serialization
  * support. The lexical analyzer that implements this interface uses scanning
  * functions to obtain tokens (of type ['a]) from the serialization support and
  * evaluating functions to convert tokens to typed values. *)

class virtual text_lexer :
  unit ->
object
  method parse_value :
    'a. (unit -> string) -> (string -> 'a) -> 'a
  (** See {!method: Xnc_lexing.lexical_analyzer.parse_value}. *)

  method parse_option :
    'a. (unit -> bool) -> (unit -> 'a) -> 'a option
  (** See {!method: Xnc_lexing.lexical_analyzer.parse_option}. *)

  method parse_object :
    'a. (#Xnc_serialization.serializable as 'a) -> (unit -> string) -> unit
  (** See {!method: Xnc_lexing.lexical_analyzer.parse_object}. *)

end
(** Base class for string (textual) lexers. *)

class xnc_scanbuf_lexer :
  ?config:Xnc_serialization.common_config ->
  Xnc_scanning.Scanning_buffer.t ->
  [string] lexical_analyzer
(** String lexer based on a scanbuffer.
  *
  * @param config serialization language configuration *)

class virtual stream_lexer :
  unit ->
object
  inherit text_lexer

  method parse_heterogeneous_enum : 'a. 'a Enum.t -> ('a -> 'a) -> 'a Enum.t
  (** See {!method: Xnc_lexing.lexical_analyzer.parse_heterogeneous_enum}. *)

  method close : unit -> unit
  (** See {!method: Xnc_lexing.lexical_analyzer.close}. *)

end

class ['a] pxp_node_stream_lexer :
  'a Pxp_document.node Stream.t ->
  [string] lexical_analyzer
(** String lexer based on PXP node streams. *)

class string_stream_lexer :
  string Stream.t ->
  [string] lexical_analyzer
(** String lexer based on string streams. *)

