(** Pipe chains
  *
  * This module provides functions to build a pipe chain. A pipe is an instance
  * of [Netchannels.pipe] and is a [Netchannel.io_obj_channel] that connects two
  * I/O resources with an intermediate buffer.
  *)

type t
(** The type of pipe chains. *)

val create : unit -> t
(** [create ()] creates an empty pipe chain. *)

val add : t -> Netchannels.pipe -> unit
(** [add c p] appends pipe [p] to chain [c]. *)

val build_output_filter : t ->
  Netchannels.out_obj_channel ->
  Netchannels.output_filter
(** [build_output_filter c out_ch] builds an output filter, that connects
  * [out_ch] to the pipe chain [c]. This way, the output written to [out_ch]
  * goes through the pipe chain. *)

val build_input_filter : t ->
  Netchannels.in_obj_channel ->
  Netchannels.input_filter
(** [build_input_filter c in_ch] builds an input filter, that connects [in_ch]
  * to the pipe chain [c]. This way, the input read from [in_ch] gets through
  * the pipe chain. *)

val close : t -> unit
(** [close c] closes the pipe chain [c], closing all the filters it contains.
  * Raises [Failure] if called before [build_output_filter] or
  * [build_input_filter]. *)

val with_output_filter :
  Netchannels.pipe list ->
  Netchannels.out_obj_channel ->
  (Netchannels.out_obj_channel -> 'a) ->
  'a
(** [with_output_filter l out_ch w] builds a pipe chain connecting all the
  * elements of [l] to [out_ch] and computes [w out_ch], returning the result.
  * Then the pipe chain is closed. *)

val with_input_filter :
  Netchannels.pipe list ->
  Netchannels.in_obj_channel ->
  (Netchannels.in_obj_channel -> 'a) ->
  'a
(** [with_input_filter l in_ch r] builds a pipe chain connecting all the
  * elements of [l] to [in_ch] and computes [r in_ch], returning the result.
  * Then the pipe chain is closed. *)

