class type ['a] printer =
object
  method print_value : 'b -> ('b -> 'a) -> unit
  method print_option : 'c option -> (unit -> unit) -> ('c -> unit) -> unit
  method print_object :
    #Xnc_serialization.serializable -> (string -> unit) -> unit
  method print_enum : 'd Enum.t -> ('d -> unit) -> unit
  method print_field :
    string -> string -> 'e -> Xnc_property.properties ->
    ('e -> unit) -> unit
  method flush : unit -> unit
  method close : unit -> unit
end;;

module type OUTPUT =
sig
  type t

  val output : t -> string -> int -> int -> unit
  val flush : t -> unit
  val close : t -> unit
end;;

module type PRETTY =
sig
  type t

  val create : (string -> int -> int -> unit) -> (unit -> unit) -> t
  val open_box : t -> int -> unit
  val close_box : t -> unit
  val print_string : t -> string -> unit
  val print_space : t -> unit
  val print_cut : t -> unit
  val print_flush : t -> unit
  val print_newline : t -> unit
end;;

module type TEXT_PRINTER =
sig
  type out_t
  type pretty_t

  class virtual text_printer :
    out_t ->
  object
    val output : out_t
    val pp : pretty_t

    method print_value : 'a. 'a -> ('a -> string) -> unit

    method print_option :
        'a. 'a option -> (unit -> unit) -> ('a -> unit) -> unit

    method print_object :
        'a. (#Xnc_serialization.serializable as 'a) ->
        (string -> unit) -> unit

    method print_enum : 'a. 'a Enum.t -> ('a -> unit) -> unit

    method close : unit -> unit

    method flush : unit -> unit

  end

  class default_printer :
    ?config:Xnc_serialization.common_config ->
    out_t ->
    [string] printer

  class xml_printer :
    out_t ->
    [string] printer

end;;

module Make_text_printer(Output : OUTPUT)(Pretty : PRETTY) =
struct
  type out_t = Output.t
  type pretty_t = Pretty.t

  let pretty_printer out =
    Pretty.create
      (Output.output out)
      (fun () -> Output.flush out)

  let print_qname prefix separator local print =
    if prefix <> "" then
      begin
        print prefix;
        print separator
      end
    else
      ();
    print local

  class virtual text_printer out =
  object
    val output = out
    val pp = pretty_printer out

    method print_value : 'a. 'a -> ('a -> string) -> unit =
      fun v format_value ->
        Pretty.print_string pp (format_value v)

    method print_option :
        'a. 'a option -> (unit -> unit) -> ('a -> unit) -> unit =
      fun o print_none print_value ->
        match o with
            None -> 
              print_none ()
          | Some v ->
              print_value v

    method print_object :
        'a. (#Xnc_serialization.serializable as 'a) ->
        (string -> unit) -> unit =
      fun o print_string ->
        print_string (o # to_string ())

    method print_enum : 'a. 'a Enum.t -> ('a -> unit) -> unit =
      fun enum print_element ->
        Enum.iter
          (fun e ->
             print_element e
          ) enum

    method flush () =
      Pretty.print_flush pp

    method close () =
      Pretty.print_newline pp;
      Output.close output

  end

  class default_printer ?(config = Xnc_serialization.default_config) out =
  object
    inherit text_printer out

    method print_enum : 'a. 'a Enum.t -> ('a -> unit) -> unit =
      fun enum print_element ->
        let print = Pretty.print_string pp in
        let last = (Enum.count enum) - 1 in
          Pretty.open_box pp 0;
          print config.Xnc_serialization.start_of_list;
          Pretty.open_box pp 0;
          Pretty.print_cut pp;
          Enum.iteri
            (fun i e ->
               print_element e;
               if i < last then
                 begin
                   print config.Xnc_serialization.list_separator;
                   Pretty.print_cut pp
                 end
               else
                 ();
            ) enum;
          Pretty.close_box pp;
          Pretty.print_cut pp;
          print config.Xnc_serialization.end_of_list;
          Pretty.close_box pp

    method print_field :
        'a. string -> string -> 'a -> Xnc_property.properties ->
        ('a -> unit) -> unit =
      fun prefix name value properties print_value ->
        let print = Pretty.print_string pp in
        if Xnc_property.is_attribute properties then
          print config.Xnc_serialization.attribute_marker
        else
          ();
        print_qname
          prefix config.Xnc_serialization.prefix_separator name print;
        print config.Xnc_serialization.name_value_separator;
        print_value value

  end

  type state =
      Printing_element
    | Printing_attributes
    | Printing_content

  class xml_printer out =
  object
    inherit text_printer out

    val mutable state = Printing_element
    val mutable first_element = true

    method print_value : 'a. 'a -> ('a -> string) -> unit =
      fun v format_value ->
        match state with
            Printing_content -> Pretty.print_string pp (format_value v)
          | _ -> ()

    method print_option :
        'a. 'a option -> (unit -> unit) -> ('a -> unit) -> unit =
      fun o print_none print_value ->
        match state with
            Printing_content -> Option.may print_value o
          | _ -> ()

    method print_object :
        'a. (#Xnc_serialization.serializable as 'a) ->
        (string -> unit) -> unit =
      fun o print_string ->
        match state with
            Printing_content -> print_string (o # to_string ())
          | _ -> ()

    method print_field :
        'a. string -> string -> 'a -> Xnc_property.properties ->
        ('a -> unit) -> unit =
      fun prefix name value properties print_value ->
        let print = Pretty.print_string pp in
        let print_name () = print_qname prefix ":" name print in
        let print_element () =
          Pretty.open_box pp 0;
          Pretty.open_box pp 2;
          print "<";
          print_name ();
          state <- Printing_attributes;
          Pretty.open_box pp 0;
          print_value value;
          Pretty.close_box pp;
          print ">";
          Pretty.print_cut pp;
          state <- Printing_content;
          first_element <- true;
          print_value value;
          Pretty.close_box pp;
          Pretty.print_cut pp;
          print "</";
          print_name ();
          print ">";
          Pretty.close_box pp in
        let print_attribute () =
          Pretty.print_space pp;
          print_name ();
          print "=\"";
          state <- Printing_content;
          print_value value;
          state <- Printing_attributes;
          print "\"" in
        let print_text () =
          print_value value in
        let print_cdata () =
          print "<![CDATA[";
          print_value value;
          print "]]" in
        let node_type = Xnc_property.get_node_type properties in
          match state with
              Printing_element ->
                begin match node_type with
                    Xnc_property.Attribute -> ()
                  | _ ->
                      print_element ();
                      state <- Printing_element
                end
            | Printing_attributes ->
                begin match node_type with
                    Xnc_property.Attribute -> print_attribute ()
                  | _ -> ()
                end
            | Printing_content ->
                begin match node_type with
                    Xnc_property.Element ->
                      if not first_element then
                        Pretty.print_cut pp
                      else
                        ();
                      print_element ();
                      first_element <- false
                  | Xnc_property.Attribute -> ()
                  | Xnc_property.Text -> print_text ()
                  | Xnc_property.Cdata -> print_cdata ()
                end

  end

end;;

module Buf =
struct
  type t = Buffer.t

  let output = Buffer.add_substring
  let flush buf = ()
  let close buf = ()

end;;

module Netch =
struct
  type t = Netchannels.out_obj_channel

  let output out_ch s pos len =
    ignore(out_ch # output s pos len)
  let flush out_ch =
    out_ch # flush ()
  let close out_ch =
    out_ch # close_out ()

end;;

module Non_pp =
struct
  type t =
      { output : string -> int -> int -> unit;
        flush : unit -> unit
      }

  let create out fl =
    { output = out;
      flush = fl
    }
  let open_box pp off = ()
  let close_box pp = ()
  let print_string pp s =
    pp.output s 0 (String.length s)
  let print_space pp =
    pp.output " " 0 1
  let print_cut pp = ()
  let print_flush pp =
    pp.flush ()
  let print_newline pp =
    pp.flush ()

end;;

module Format_pp =
struct
  type t = Format.formatter

  let create output flush =
    Format.make_formatter output flush
  let open_box = Format.pp_open_hvbox
  let close_box pp = Format.pp_close_box pp ()
  let print_string = Format.pp_print_string
  let print_space pp = Format.pp_print_space pp ()
  let print_cut pp = Format.pp_print_cut pp ()
  let print_flush pp = Format.pp_print_flush pp ()
  let print_newline pp = Format.pp_print_newline pp ()

end;;

module Buffer_printer = Make_text_printer(Buf)(Non_pp);;

module Netchannel_printer = Make_text_printer(Netch)(Non_pp);;

module Buffer_pretty_printer = Make_text_printer(Buf)(Format_pp);;

module Netchannel_pretty_printer = Make_text_printer(Netch)(Format_pp);;

