(** Basic printing classes
  *
  * This module provides the interface and classes needed to print values on a
  * serialization support. The printer implemented in this module use formatting
  * functions to format values before printing, and handles pretty printing.
  *)

class type ['a] printer =
object
  method print_value : 'b -> ('b -> 'a) -> unit
  (** [print_value v f] prints value [v] after formatting it with [f]. *)

  method print_option : 'c option -> (unit -> unit) -> ('c -> unit) -> unit
  (** [print_value o pn pv] prints option [o] using [pn] to print [None] and
    * [pv] to print [Some v]. *)

  method print_object :
    #Xnc_serialization.serializable -> (string -> unit) -> unit
  (** [print_object o p] prints serializable object [o] using [p] to print its
    * string representation. *)

  method print_enum : 'd Enum.t -> ('d -> unit) -> unit
  (** [print_enum e p] prints enumeration [e] using [p] to print each element .
    *)

  method print_field :
    string -> string -> 'e -> Xnc_property.properties ->
    ('e -> unit) -> unit
  (** [print_field p n v prop pv] prints a field
    * with prefix [p], name [n], value [v], and properties [prop], using [pv] to
    * print the value. *)

  method flush : unit -> unit
  (** [flush ()] flushes the underlying support. *)

  method close : unit -> unit
  (** [close ()] closes the underlying support. *)

end
(** This interface collects methods used to print values on a serialization
  * support. The serialization support is made of values of type ['a].
  *)

module type OUTPUT =
sig
  type t
  (** The type of serialization supports. *)

  val output : t -> string -> int -> int -> unit
  (** [output o s pos len] outputs [len] characters of [s] starting from [pos]
    * to [o]. *)

  val flush : t -> unit
  (** [flush o] flushes [o]. *)

  val close : t -> unit
  (** [close o] closes [o]. *)

end;;
(** Signature of serialization supports.
  *)

module type PRETTY =
sig
  type t
  (** The type of pretty printers. *)

  val create : (string -> int -> int -> unit) -> (unit -> unit) -> t
  (** [create o f] returns a pretty printer that writes according to the output
    * function [o], and the flushing function [f]. *)

  val open_box : t -> int -> unit
  (** [open_box pp d] opens a new pretty-printing box with offset [d]. *)

  val close_box : t -> unit
  (** [close_box pp] closes the most recently opened pretty-printing box. *)

  val print_string : t -> string -> unit
  (** [print_string pp s] prints [s] in the current box. *)

  val print_space : t -> unit
  (** [print_space pp] is used to separate items (typically to print a space
    * between two words). *)

  val print_cut : t -> unit
  (** [print_cut pp] is used to mark a good break position. *)

  val print_flush : t -> unit
  (** [print_flush pp] flushes the pretty printer: all opened boxes are closed,
    * and all pending text is displayed. *)

  val print_newline : t -> unit
  (** [print_newline pp] is equivalent to [print_flush] followed by a new line.
    *)

end;;
(** Signature of a pretty printers.
  *)

module type TEXT_PRINTER =
sig
  type out_t
  (** The type of serialization supports. *)
  type pretty_t
  (** The type of pretty printers. *)

  class virtual text_printer :
    out_t ->
  object
    val output : out_t
    (** Serialization support. *)
    val pp : pretty_t
    (** Pretty printer. *)

    method print_value : 'a. 'a -> ('a -> string) -> unit
    (** See {!method: Xnc_printing.printer.print_value}. *)

    method print_option :
        'a. 'a option -> (unit -> unit) -> ('a -> unit) -> unit
    (** See {!method: Xnc_printing.printer.print_option}. *)

    method print_object :
        'a. (#Xnc_serialization.serializable as 'a) ->
        (string -> unit) -> unit
    (** See {!method: Xnc_printing.printer.print_object}. *)

    method print_enum : 'a. 'a Enum.t -> ('a -> unit) -> unit
    (** See {!method: Xnc_printing.printer.print_enum}. *)

    method close : unit -> unit
    (** See {!method: Xnc_printing.printer.close}. *)

    method flush : unit -> unit
    (** See {!method: Xnc_printing.printer.flush}. *)

  end
  (** Base class for string (textual) printers. Initialized with a serialization
    * support, works with a string formatter. *)

  class default_printer :
    ?config:Xnc_serialization.common_config ->
    out_t ->
    [string] printer
  (** Default printer used by simple serializers. *)

  class xml_printer :
    out_t ->
    [string] printer
  (** XML printer used by XML serializers. *)

end

module Make_text_printer :
  functor (Output : OUTPUT) ->
  functor (Pretty : PRETTY) ->
sig
  include TEXT_PRINTER with
    type out_t = Output.t and
    type pretty_t = Pretty.t
                 
end
(** Functor building an implementation of a string (textual) printer. *)

module Buf :
sig
  type t = Buffer.t
  (** The type of string buffers. *)

  val output : t -> string -> int -> int -> unit
  (** See {!val: Xnc_printing.OUTPUT.output}. *)

  val flush : t -> unit
  (** See {!val: Xnc_printing.OUTPUT.flush}. *)

  val close : t -> unit
  (** See {!val: Xnc_printing.OUTPUT.close}. *)

end;;
(** This module implements a serialization support based on string buffers. *)

module Netch :
sig
  type t = Netchannels.out_obj_channel
  (** The type of output Netchannels. *)

  val output : t -> string -> int -> int -> unit
  (** See {!val: Xnc_printing.OUTPUT.output}. *)

  val flush : t -> unit
  (** See {!val: Xnc_printing.OUTPUT.flush}. *)

  val close : t -> unit
  (** See {!val: Xnc_printing.OUTPUT.close}. *)

end;;
(** This module implements a serialization support based on Netchannels. *)

module Non_pp : PRETTY
(** This module implements a pretty printer that does not break lines. *)

module Format_pp : PRETTY
(** This module implements a pretty printer based on [Format] module. *)

module Buffer_printer :
sig
  include TEXT_PRINTER with
    type out_t = Buf.t and
    type pretty_t = Non_pp.t

end
(** This module implements printers that print text on a string buffer without
  * pretty printing. *)

module Netchannel_printer :
sig
  include TEXT_PRINTER with
    type out_t = Netch.t and
    type pretty_t = Non_pp.t

end
(** This module implements printers that print text on a Netchannel without
  * pretty printing. *)

module Buffer_pretty_printer :
sig
  include TEXT_PRINTER with
    type out_t = Buf.t and
    type pretty_t = Format_pp.t

end
(** This module implements printers that pretty print text on a string buffer.
  *)

module Netchannel_pretty_printer :
sig
  include TEXT_PRINTER with
    type out_t = Netch.t and
    type pretty_t = Format_pp.t

end
(** This module implements printers that pretty print text on a Netchannel. *)

