(** Field properties (metadata)
  *
  * This module handles properties optionally associated to serializable fields.
  *)

type node_type =
    Element
    (** Specifies that current field is an element *)
  | Attribute
    (** Specifies that current field is an attribute *)
  | Text
    (** Specifies that current field is a text element *)
  | Cdata
    (** Specifies that current field is a CDATA element *)
(** This type enumerates all the node types handled by XML serializer.
  *)

type property =
    Node_type of node_type
    (** Specifies the [node_type] of the current field *)
  | Length of int
    (** Specifies the length in characters of the current field *)
  | Precision of int
    (** Specifies the precision (number of digits after the decimal point) of
      * the current field *)
(** This type enumerates all the properties that a field may have associated.
  *)

type properties
(** The type of sets of properties.
  *)

val empty : properties
(** Empty set of properties. *)

val singleton : property -> properties
(** [singleton p] creates the one-property set containing only [p]. *)

val from_list : property list -> properties
(** [from_list pl] creates a set of properties from a list of properties. *)

val element : property
(** Alias for [Node_type Element]. *)

val attribute : property
(** Alias for [Node_type Attribute]. *)

val text : property
(** Alias for [Node_type Text]. *)

val cdata : property
(** Alias for [Node_type Cdata]. *)

val get_node_type : properties -> node_type
(** Returns the node type from a list of properties. If a node type is not
  * found, defaults to [Element]. *)

val is_element : properties -> bool
(** Returns if there is an element node type in a list of properties. *)

val is_attribute : properties -> bool
(** Returns if there is an attribute node type in a list of properties. *)

val is_text : properties -> bool
(** Returns if there is a text node type in a list of properties. *)

val is_cdata : properties -> bool
(** Returns if there is a CDATA node type in a list of properties. *)

val get_length : properties -> int -> int
(** [get_length p d] returns the value of the length property. If not found,
  * returns [d]. *)

val get_precision : properties -> int -> int
(** [get_precision p d] returns the value of the precision property. If not
  * found, returns [d]. *)

val equal : properties -> properties -> bool
(** [equal p1 p2] tests whether the sets [p1] and [p2] are equal, that is,
  * contain equal properties. *)

val subset : properties -> properties -> bool
(** [subset p1 p2] tests whether the set [p1] is a subset of the set [p2]. *)

