type 'a t =
    { scanner : 'a Xnc_scanning.scanner;
      token_evaluator : 'a Xnc_token_eval.token_evaluator;
      lexer : 'a Xnc_lexing.lexical_analyzer
    }

let create s t l =
  { scanner = s;
    token_evaluator = t;
    lexer = l
  }
;;

let create_from_scanbuf
      ?(config = Xnc_serialization.default_config)
      scanbuf =
  let scanner =
    new Xnc_scanning.standard_scanner ~config scanbuf in
  let token_evaluator =
    new Xnc_token_eval.standard_evaluator () in
  let lexical_analyzer =
    new Xnc_lexing.xnc_scanbuf_lexer ~config scanbuf
  in
    create scanner token_evaluator lexical_analyzer
;;

let create_from_netchannel
      ?(config = Xnc_serialization.default_config)
      in_ch =
  let scanbuf =
    Xnc_scanning.Scanning_buffer.create_from_netchannel
      ~skip_whitespace:true in_ch
  in
    create_from_scanbuf ~config scanbuf
;;

let create_from_string
      ?(config = Xnc_serialization.default_config)
      s =
  let scanbuf =
    Xnc_scanning.Scanning_buffer.create_from_string
      ~skip_whitespace:true s
  in
    create_from_scanbuf ~config scanbuf
;;

let create_from_stream
      stream =
  let scanner =
    new Xnc_scanning.string_stream_scanner stream in
  let token_evaluator =
    new Xnc_token_eval.standard_evaluator () in
  let lexical_analyzer =
    new Xnc_lexing.string_stream_lexer stream
  in
    create scanner token_evaluator lexical_analyzer
;;

let create_from_pxp_node
      node =
  let stream =
    Xnc_scanning.pxp_node_to_stream node in
  let scanner =
    new Xnc_scanning.pxp_node_stream_scanner stream in
  let token_evaluator =
    new Xnc_token_eval.standard_evaluator () in
  let lexical_analyzer =
    new Xnc_lexing.pxp_node_stream_lexer stream
  in
    create scanner token_evaluator lexical_analyzer
;;

let try_parse_null ?properties reader =
  reader.lexer # try_parse_null ?properties ()
;;

let read_int ?properties reader =
  reader.lexer # parse_value
    (reader.scanner # scan_int ?properties)
    (reader.token_evaluator # get_int ?properties)
;;

let read_float ?properties reader =
  reader.lexer # parse_value
    (reader.scanner # scan_float ?properties)
    (reader.token_evaluator # get_float ?properties)
;;

let read_char ?properties reader =
  reader.lexer # parse_value
    (reader.scanner # scan_char ?properties)
    (reader.token_evaluator # get_char ?properties)
;;

let read_bool ?properties reader =
  reader.lexer # parse_value
    (reader.scanner # scan_bool ?properties)
    (reader.token_evaluator # get_bool ?properties)
;;

let read_string ?properties reader =
  reader.lexer # parse_value
    (reader.scanner # scan_string ?properties)
    (reader.token_evaluator # get_string ?properties)
;;

let read_date ?properties reader =
  reader.lexer # parse_value
    (reader.scanner # scan_date ?properties)
    (reader.token_evaluator # get_date ?properties)
;;

let read_value reader scan_token parse_token =
  reader.lexer # parse_value
    scan_token
    parse_token
;;

let read_option reader try_parse_null read_value =
  reader.lexer # parse_option
    try_parse_null
    read_value
;;

let read_object ?properties reader serializable =
  reader.lexer # parse_object
    serializable
    (fun () -> read_string ?properties reader)
;;

let read_homogeneous_enum reader read_element =
  reader.lexer # parse_homogeneous_enum
    read_element
;;

let read_homogeneous_collection reader read_element of_enum =
  let enum = read_homogeneous_enum reader read_element in
    of_enum enum
;;

let read_heterogeneous_enum reader enum_schema read_element =
  reader.lexer # parse_heterogeneous_enum
    enum_schema
    read_element
;;

let read_heterogeneous_collection reader schema to_enum read_element of_enum =
  let enum_schema = to_enum schema in
  let enum = read_heterogeneous_enum reader enum_schema read_element in
    of_enum enum
;;

let read_value_by_name reader prefix name properties read_value =
  reader.lexer # parse_value_by_name
    prefix name properties
    read_value
;;

let skip reader n =
  reader.scanner # skip n
;;

let close reader =
  reader.lexer # close ()
;;

