(** Readers
  *
  * This module provides the functions to read values scanned by a
  * [Xnc_scanning.scanner], evaluated by an [Xnc_token_eval.evaluator], and
  * parsed by a [Xnc_lexing.lexer].
  *)

type 'a t
(** The type of the readers that read from values of type ['a]. *)

val create :
  'a Xnc_scanning.scanner ->
  'a Xnc_token_eval.token_evaluator ->
  'a Xnc_lexing.lexical_analyzer ->
  'a t
(** [create s t l] creates a reader from scanner [s], token evaluator [t], and
  * lexer [l]. *)

val create_from_scanbuf :
  ?config:Xnc_serialization.common_config ->
  Xnc_scanning.Scanning_buffer.t ->
  string t
(** [create sb] creates a string reader from scanning buffer [sb].
  *
  * @param config serialization language configuration. *)

val create_from_netchannel :
  ?config:Xnc_serialization.common_config ->
  Netchannels.in_obj_channel ->
  string t
(** [create in_ch] creates a string reader from [in_ch].
  *
  * @param config serialization language configuration. *)

val create_from_string :
  ?config:Xnc_serialization.common_config ->
  string ->
  string t
(** [create s] creates a string reader from [s].
  *
  * @param config serialization language configuration. *)

val create_from_stream :
  string Stream.t ->
  string t
(** [create s] creates a string reader from [s].
  *
  * @param config serialization language configuration. *)

val create_from_pxp_node :
  'a Pxp_document.node ->
  string t
(** [create n] creates a string reader from [n]. *)

val try_parse_null :
  ?properties:Xnc_property.properties ->
  'a t -> bool
(** [try_parse_null r] tries to read a null value from reader [r]. Returns
  * [true] if the the operation succedes, otherwise returns [false] without
  * consuming any input. *)

val read_int :
  ?properties:Xnc_property.properties ->
  'a t -> int
(** [read_int r] reads an [int] value from reader [r]. *)

val read_float :
  ?properties:Xnc_property.properties ->
  'a t -> float
(** [read_float r] reads a [float] value from reader [r]. *)

val read_char :
  ?properties:Xnc_property.properties ->
  'a t -> char
(** [read_char r] reads a [char] value from reader [r]. *)

val read_bool :
  ?properties:Xnc_property.properties ->
  'a t -> bool
(** [read_bool r] reads a [bool] value from reader [r]. *)

val read_string :
  ?properties:Xnc_property.properties ->
  'a t -> string
(** [read_string r] reads a [string] value from reader [r]. *)

val read_date :
  ?properties:Xnc_property.properties ->
  'a t -> Netdate.t
(** [read_date r] reads a date value from reader [r]. *)

val read_value : 'a t -> (unit -> 'a) -> ('a -> 'b) -> 'b
(** [read_value r s e] reads a generic value from reader [r], using scanning
  * function [s] and evaluating function [e]. *)

val read_option : 'a t -> (unit -> bool) -> (unit -> 'b) -> 'b option
(** [read_option r n rv] reads an [option] value from reader [r], using [n] to
  * detect if the value to read is [None], and [rv] to read value of type ['b].
  *)

val read_object :
  ?properties:Xnc_property.properties ->
  'a t -> #Xnc_serialization.serializable -> unit
(** [read_object r o] deserializes the serializable object [o] with the
  * value read by reader [r]. *)

val read_homogeneous_enum : 'a t -> (unit -> 'b) -> 'b Enum.t
(** [read_homogeneous_enum r re] reads an enumeration from reader [r], using
  * [re] to read each element. *)

val read_homogeneous_collection :
  'a t -> (unit -> 'b) -> ('b Enum.t -> 'c) -> 'c
(** [read_homogeneous_collection r re c] reads a generic collection from reader
  * [r], using [re] to read each element, and [c] to convert an enumeration to
  * the collection type ['c]. *)

val read_heterogeneous_enum : 'a t -> 'b Enum.t -> ('b -> 'b) -> 'b Enum.t
(** [read_heterogeneous_enum r e re] reads an enumeration from reader [r]. [e]
  * is the enumeration descriptor (an enumeration that shares the same structure
  * of the one to be read). [re] is used to read each element. *)

val read_heterogeneous_collection :
  'a t -> 'b -> ('b -> 'c Enum.t) -> ('c -> 'c) -> ('c Enum.t -> 'b) -> 'b
(** [read_heterogeneous_collection r c to_enum re from_enum] reads a generic
  * collection from reader [r]. [c] is the collection descriptor (a collection
  * that shares the same structure of the one to be read). [to_enum] and
  * [from_enum] converts the collection of type ['b] to and from an enumeration.
  * [re] is used to read each element. *)

val read_value_by_name :
  'a t -> string -> string -> Xnc_property.properties ->
  (unit -> 'b) -> 'b
(** [read_value_by_name r p n prop rv] reads a field with prefix [p], name [n],
  * and properties [prop] from reader [r], using [rv] to read a value of type
  * ['b]. *)

val skip : 'a t -> int -> unit
(** [skip r n] advance the cursor of reader [r] by [n] positions. *)

val close : 'a t -> unit
(** [close r] closes reader [r]. *)

