exception Scan_failure of string

module Scanning_buffer =
struct
  (* TODO: better buffer handling *)
  type t =
      { refill_buffer : t -> unit;
        close : unit -> unit;
        mutable buffer : string;
        mutable position : int;
        mutable line_number : int;
        mutable eof : bool;
        mutable error: bool;
        skip_whitespace : bool;
      }

  let create_from_string ?(skip_whitespace = false) string_buffer =
    { refill_buffer = (fun scanbuf ->
                         scanbuf.eof <- true);
      close = (fun () -> ());
      buffer = string_buffer;
      position = 0;
      line_number = 0;
      eof = false;
      error = false;
      skip_whitespace = skip_whitespace
    }

  let create_from_netchannel ?(skip_whitespace = false) in_ch =
    let refill scanbuf =
      try
        scanbuf.buffer <- in_ch # input_line ();
        scanbuf.position <- 0;
        scanbuf.line_number <- succ scanbuf.line_number;
      with End_of_file ->
        scanbuf.eof <- true
    in
      { refill_buffer = refill;
        close = (fun () -> in_ch # close_in ());
        buffer = "";
        position = 0;
        line_number = -1;
        eof = false;
        error = false;
        skip_whitespace = skip_whitespace
      }

  let null_char = '\000'

  let buffer_underrun scanbuf =
    scanbuf.position >= String.length scanbuf.buffer

  let raise_eof description =
    failwith (description ^ ", EOF reached")

  let skip_whitespace scanbuf =
    let match_whitespace () =
      match scanbuf.buffer.[scanbuf.position] with
          ' '
        | '\t' ->
            scanbuf.position <- succ scanbuf.position;
            true
        | '\n' ->
            scanbuf.buffer <- Str.string_after
                                scanbuf.buffer (succ scanbuf.position);
            scanbuf.position <- 0;
            scanbuf.line_number <- succ scanbuf.line_number;
            true
        | _ ->
            false in
    let feed_buffer () =
      if buffer_underrun scanbuf then
        scanbuf.refill_buffer scanbuf
      else
        () in
    let rec loop () =
      feed_buffer ();
      if scanbuf.skip_whitespace && not scanbuf.eof && match_whitespace () then
        loop ()
      else
        ()
    in
      loop ()

  let peek_wrapper scanbuf peek_fun =
    skip_whitespace scanbuf;
    if not scanbuf.eof then
      peek_fun ()
    else
      ""

  let advance_position scanbuf n =
    scanbuf.position <- scanbuf.position + n

  let eof scanbuf =
    scanbuf.eof

  let peek_char scanbuf =
    skip_whitespace scanbuf;
    if not scanbuf.eof then
      scanbuf.buffer.[scanbuf.position]
    else
      null_char

  let next_char scanbuf =
    let c = peek_char scanbuf in
      advance_position scanbuf 1;
      c

  let raise_scan_failure scanbuf description =
    let (curr_char, unscanned) =
      if buffer_underrun scanbuf then
        ("", "")
      else
        (Std.string_of_char scanbuf.buffer.[scanbuf.position],
         Str.string_after scanbuf.buffer scanbuf.position)
    in
      raise (Scan_failure
               (Printf.sprintf 
                  "bad input at line %d, char %d: \
                   '%s' (unscanned buffer: \"%s\", EOF: %b), \
                   %s"
                  (succ scanbuf.line_number)
                  (succ scanbuf.position)
                  curr_char
                  unscanned
                  scanbuf.eof
                  description))

  let next_string_wrapper scanbuf peek_string_fun advance description =
    let s = peek_string_fun scanbuf in
      if scanbuf.error || scanbuf.eof then
        raise_scan_failure scanbuf description
      else
        begin
          advance ();
          s
        end

  let peek_string scanbuf len =
    peek_wrapper scanbuf
      (fun () ->
         try
           scanbuf.error <- false;
           String.sub scanbuf.buffer scanbuf.position len
         with _ ->
           scanbuf.error <- true;
           "")

  let next_string scanbuf len =
    next_string_wrapper scanbuf
      (fun scanbuf -> peek_string scanbuf len)
      (fun () -> advance_position scanbuf len)
      ("expecting " ^ (string_of_int len) ^ "-character long string")

  let peek_literal_string scanbuf literal =
    peek_wrapper scanbuf
      (fun () ->
         let len = String.length literal in
         let s = peek_string scanbuf len in
           if s = literal then
             begin
               scanbuf.error <- false;
               literal
             end
           else
             begin
               scanbuf.error <- true;
               ""
             end)

  let next_literal_string scanbuf literal =
    next_string_wrapper scanbuf
      (fun scanbuf -> peek_literal_string scanbuf literal)
      (fun () -> advance_position scanbuf (String.length literal))
      ("expecting \"" ^ literal ^ "\"")

  let parse_literal scanbuf literal =
    ignore(next_literal_string scanbuf literal)

  let match_literal scanbuf literal =
    if (peek_literal_string scanbuf literal) = literal then
      begin
        ignore(next_literal_string scanbuf literal);
        true
      end
    else
      false

  let peek_matching_string ?(get_match = Str.matched_string) scanbuf regexp =
    peek_wrapper scanbuf
      (fun () ->
         if Str.string_match regexp scanbuf.buffer scanbuf.position then
           begin
             scanbuf.error <- false;
             get_match scanbuf.buffer
           end
         else
           begin
             scanbuf.error <- true;
             ""
           end)

  let next_matching_string ?(get_match = Str.matched_string) scanbuf regexp =
    next_string_wrapper scanbuf
      (fun scanbuf -> peek_matching_string ~get_match scanbuf regexp)
      (fun () -> scanbuf.position <- Str.match_end ())
      "pattern not matched"

  let skip scanbuf n =
    ignore (next_string scanbuf n)

  let close scanbuf =
    scanbuf.close ()

end;;

class type ['a] scanner =
object
  method scan_int :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  method scan_float :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  method scan_char :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  method scan_bool :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  method scan_string :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  method scan_date :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  method skip : int -> unit
end

(* Constants *)
let int_regexp =
  Str.regexp "[+-]?[0-9]+"
;;
let float_regexp =
  Str.regexp_case_fold "[+-]?[0-9]+\\.\\([0-9]*\\)?\\(e[-+][0-9]+\\)?"
;;
let char_regexp =
  Str.regexp "."
;;
let bool_regexp =
  Str.regexp "true\\|false"
;;
let date_length = 19;;

class standard_scanner
  ?(config = Xnc_serialization.default_config)
  sb : [string] scanner =

(* Class variables *)
let quoted_delimiter =
  Str.quote (Std.string_of_char
               config.Xnc_serialization.delimiter_char) in
let quoted_escape =
  Str.quote (Std.string_of_char
               config.Xnc_serialization.escape_char) in
let escape_regexp = Str.regexp quoted_escape in

(* Class definition *)
object(self)
  val scanbuf = sb

  method scan_int ?properties () =
    self # scan_value
      int_regexp
      "Scanning int"

  method scan_float ?properties () =
    self # scan_value
      float_regexp
      "Scanning float"

  method scan_char ?properties () =
    self # scan_value
      char_regexp
      "Scanning char"

  method scan_bool ?properties () =
    self # scan_value
      bool_regexp
      "Scanning bool"

  method scan_string ?properties () =
    let qd = quoted_delimiter in
    let qe = quoted_escape in
    let string_regexp =
      Str.regexp (qd ^
                  "\\(\
                     \\([^" ^ qd ^ qe ^ "]*\
                        \\|\
                        [^" ^ qd ^ "]*" ^ qe ^ qd ^ "\\)+\
                   \\)" ^ qd) in
    let ds = self # scan_value
               ~get_match:(Str.matched_group 1)
               string_regexp
               ("Scanning '" ^ quoted_delimiter ^ "'-delimited string") in
      Str.global_replace escape_regexp "" ds

  method scan_date ?properties () =
    try
      Scanning_buffer.next_string scanbuf date_length
    with Scan_failure d ->
      raise (Scan_failure ("Scanning date: " ^ d))

  method skip n =
    Scanning_buffer.skip scanbuf n

  method private scan_value
        ?(get_match = Str.matched_string)
        regexp
        action =
    try
      Scanning_buffer.next_matching_string
        ~get_match
        scanbuf
        regexp
    with Scan_failure d ->
      raise (Scan_failure (action ^ ": " ^ d))

end;;

class virtual base_scanner () =
object(self)
  method scan_int ?properties () =
    self # scan_value
      ?properties
      "Scanning int"

  method scan_float ?properties () =
    self # scan_value
      ?properties
      "Scanning float"

  method scan_char ?properties () =
    self # scan_value
      ?properties
      "Scanning char"

  method scan_bool ?properties () =
    self # scan_value
      ?properties
      "Scanning bool"

  method scan_string ?properties () =
    self # scan_value
      ?properties
      "Scanning string"

  method scan_date ?properties () =
    self # scan_value
      ?properties
      "Scanning date"

  method virtual skip : int -> unit

  method virtual private scan_value :
    ?properties:Xnc_property.properties ->
    string -> string

end

let eoe = "/";;

class virtual ['a] stream_scanner (stream : 'a Stream.t) =
object
  inherit base_scanner ()

  method skip n =
    let rec skip_element acc =
      if (acc >= n) then
        ()
      else begin
        Stream.junk stream;
        skip_element (acc + 1)
      end
    in
      skip_element 0

end;;

let pxp_node_to_stream n =
  let from_node node =
    let s = Stack.create () in
      Stack.push node s;
      s
  in

  let next_node s =
    if Stack.is_empty s then
      None
    else begin
      let node = Stack.pop s in
      let children = List.rev (node # sub_nodes) in
      let _ = List.iter
                (fun e -> Stack.push e s)
                children in
      let _ =
        match node # node_type with
            Pxp_document.T_element _ ->
              let attributes = node # attributes_as_nodes in
                List.iter
                  (fun e -> Stack.push e s)
                  attributes
          | _ -> () in
        Some node
    end
  in

  let s = from_node n in
    Stream.from (fun _ -> next_node s)
;;

class ['ext] pxp_node_stream_scanner
  (stream : 'ext Pxp_document.node Stream.t) : [string] scanner =
object(self)
  inherit ['ext Pxp_document.node] stream_scanner stream

  method private scan_value ?(properties = Xnc_property.empty) action =
    try
      let node = Stream.next stream in
        match node # node_type with
            Pxp_document.T_data -> node # data
          | Pxp_document.T_attribute _ -> node # data
          | _ -> failwith "Node type not supported"
    with exc ->
      raise (Scan_failure (action ^ ": " ^ (Printexc.to_string exc)))

end;;

(* private *)
(* TODO:
 * - refactor
 * - write a direct Pxp_types.event scanner/lexer? *)
module Pxp_pull_parser =
struct
  type state =
      Element
    | Attribute
    | Data

  type t =
      { event_stream : Pxp_types.event Stream.t;
        mutable attribute_stream : string Stream.t;
        mutable data : string;
        mutable state : state
      }

  let empty_stream =
    Stream.from (fun _ -> None)

  let create config entity_manager =
    let next_event = Pxp_ev_parser.create_pull_parser
                       config
                       (`Entry_content [])
                       entity_manager in
    let stream = Stream.from (fun _ -> next_event ()) in
      { event_stream = stream;
        attribute_stream = empty_stream;
        data = "";
        state = Element
      }

  let rec next_string s =
    match s.state with
        Element ->
          begin match Stream.peek s.event_stream with
              None -> None
            | Some event ->
                Stream.junk s.event_stream;
                begin match event with
                    Pxp_types.E_start_tag (name, attlist, _, _) ->
                      let flattened = List.fold_left
                                        (fun accu (n, v) ->
                                           n :: v :: accu)
                                        []
                                        attlist in
                      let att_stream = Stream.of_list flattened in
                        s.attribute_stream <- att_stream;
                        s.state <- Attribute;
                        Some name
                  | Pxp_types.E_end_tag _ ->
                      Some eoe
                  | Pxp_types.E_char_data data ->
                      s.data <- s.data ^ data;
                      s.state <- Data;
                      next_string s
                  | Pxp_types.E_end_of_stream ->
                      None
                  | _ ->
                      next_string s
                end
          end
      | Attribute ->
          begin match Stream.peek s.attribute_stream with
              None ->
                s.state <- Element;
                next_string s
            | a ->
                Stream.junk s.attribute_stream;
                a
          end
      | Data ->
          begin match Stream.peek s.event_stream with
              None -> None
            | Some event ->
                begin match event with
                    Pxp_types.E_char_data data ->
                      Stream.junk s.event_stream;
                      s.data <- s.data ^ data;
                      next_string s
                  | _ ->
                      let data = s.data in
                        s.data <- "";
                        s.state <- Element;
                        Some data
                end
          end

  let to_string_stream config entity_manager =
    let s = create config entity_manager in
      Stream.from (fun _ -> next_string s)

end;;

let pull_parser_to_string_stream config entity_manager =
  Pxp_pull_parser.to_string_stream config entity_manager
;;

class string_stream_scanner (stream : string Stream.t) : [string] scanner =
object(self)
  inherit [string] stream_scanner stream

  method private scan_value ?(properties = Xnc_property.empty) action =
    try
      Stream.next stream
    with Stream.Failure ->
      raise (Scan_failure (action ^ ": stream failure"))

end;;

