(** Basic scanning classes
  *
  * This module provides the low level interface and classes used to scan
  * values.
  *)

exception Scan_failure of string
(** Raised by the scanner if it cannot scan the expected value. *)

module Scanning_buffer :
sig
  type t 
  (** The type of scanning buffers. *)

  val create_from_string :
    ?skip_whitespace:bool -> string -> t
  (** [create_from_string s] creates a scanning buffer from to scan [s].
    *
    * @param skip_whitespace Makes the scanner skips whitespaces or not.
    *)

  val create_from_netchannel :
    ?skip_whitespace:bool -> Netchannels.in_obj_channel -> t
  (** [create_from_netchannel in_ch] creates a scanning buffer to scan [in_ch].
    *
    * @param skip_whitespace Makes the scanner skips whitespaces or not.
    *)

  val eof : t -> bool
  (** [eof sb] tests [sb] to check if the scanner reached the end of the input.
    *)

  val peek_char : t -> char
  (** [peak_char sb] reads the next character from [sb] without consuming it. *)

  val next_char : t -> char
  (** [next_char sb] reads the next character from [sb]. *)

  val peek_string : t -> int -> string
  (** [peek_string sb n] reads the next [n] characters from [sb] without
    * consuming them. *)

  val next_string : t -> int -> string
  (** [next_string sb n] reads the next [n] characters from [sb]. *)

  val peek_literal_string : t -> string -> string
  (** [peek_literal_string sb s] reads [s] from [sb] without consuming any
    * character. *)

  val next_literal_string : t -> string -> string
  (** [next_literal_string sb s] reads [s] from [sb]. *)

  val parse_literal : t -> string -> unit
  (** [parse_literal sb s] reads and discards [s] from [sb]. *)

  val match_literal : t -> string -> bool
  (** [match_literal sb s] tests if [s] is the next literal from [sb]. *)

  val peek_matching_string :
    ?get_match:(string -> string) -> t -> Str.regexp -> string
  (** [peek_matching_string sb r] reads a string that matches [r] from [sb]
    * without consuming any character.
    *
    * @param get_match returns the matched substring *)

  val next_matching_string :
    ?get_match:(string -> string) -> t -> Str.regexp -> string
  (** [next_matching_string sb r] reads a string that matches [r] from [sb].
    *
    * @param get_match returns the matched substring *)

  val skip : t -> int -> unit
  (** [skip sb n] discards [n] characters from [sb]. *)

  val close : t -> unit
  (** [close sb] closes [sb]. *)

end
(** This module implements a scanning buffer. *)

class type ['a] scanner =
object
  method scan_int :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  (** [scan_int ()] returns an [int] token. *)

  method scan_float :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  (** [scan_float ()] returns a [float] token. *)

  method scan_char :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  (** [scan_char ()] returns a [char] token. *)

  method scan_bool :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  (** [scan_bool ()] returns a [bool] token. *)

  method scan_string :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  (** [scan_string ()] returns a [string] token. *)

  method scan_date :
    ?properties:Xnc_property.properties ->
    unit -> 'a
  (** [scan_string ()] returns a date token. *)

  method skip : int -> unit
  (** [skip n] discard the next [n] characters. *)

end
(** This interface defines a scanner that reads tokens of type ['a]. All the
  * methods to scan tokens share a [properties] parameter that is used to pass
  * optional metadata.
  *)

class standard_scanner :
  ?config:Xnc_serialization.common_config ->
  Scanning_buffer.t ->
  [string] scanner
(** String scanner used by simple serializers that reads from a scanning buffer.
  * This scanner is based on regular expressions.
  *
  * @param config serialization language configuration *)

class virtual base_scanner : unit ->
object
  method scan_int :
    ?properties:Xnc_property.properties ->
    unit -> string
  (** See {!method: Xnc_scanning.scanner.scan_int}. *)

  method scan_float :
    ?properties:Xnc_property.properties ->
    unit -> string
  (** See {!method: Xnc_scanning.scanner.scan_float}. *)

  method scan_char :
    ?properties:Xnc_property.properties ->
    unit -> string
  (** See {!method: Xnc_scanning.scanner.scan_char}. *)

  method scan_bool :
    ?properties:Xnc_property.properties ->
    unit -> string
  (** See {!method: Xnc_scanning.scanner.scan_bool}. *)

  method scan_string :
    ?properties:Xnc_property.properties ->
    unit -> string
  (** See {!method: Xnc_scanning.scanner.scan_string}. *)

  method scan_date :
    ?properties:Xnc_property.properties ->
    unit -> string
  (** See {!method: Xnc_scanning.scanner.scan_date}. *)

  method virtual skip : int -> unit
  (** See {!method: Xnc_scanning.scanner.skip}. *)

  method virtual private scan_value :
    ?properties:Xnc_property.properties ->
    string -> string
  (** [scan_value action] scans next token. Raises [Scan_failure] if scan fails
    * and uses [action] to detail error description. *)

end
(** String scanner that delegates all scanning logic to method [scan_value]. *)

val eoe : string
(** String constant specifying the end of an element in a string stream. *)

class virtual ['a] stream_scanner :
  'a Stream.t ->
object
  inherit base_scanner

  method skip : int -> unit
  (** [skip n] junks the first [n] elements of the stream. *)

end
(** Virtual string scanner used as base for stream scanners. *)

val pxp_node_to_stream :
  'a Pxp_document.node -> ('a Pxp_document.node) Stream.t
(** [pxp_node_to_stream n] creates a PXP node stream from PXP node [d]. *)

class ['a] pxp_node_stream_scanner :
  'a Pxp_document.node Stream.t ->
  [string] scanner
(** String scanner based on PXP node stream, used by validating XML serializers.
  *)

val pull_parser_to_string_stream :
  Pxp_types.config ->
  Pxp_entity_manager.entity_manager ->
  string Stream.t
(** [pull_parser_to_string_stream c e] creates a string stream from PXP pull
  * parser made of configuration [c] and entity manager [e]. *)

class string_stream_scanner :
  string Stream.t ->
  [string] scanner
(** String scanner based on string streams. Used by non validating XML
  * serializers. *)

