exception Invalid_serializable_type

type 'a serializable_value =
    Int of int
  | Float of float
  | Char of char
  | Bool of bool
  | String of string
  | Date of Netdate.t
  | Option of 'a serializable_value option
  | Object of Xnc_serialization.serializable
  | Custom of ('a value_writer * 'a value_reader * 'a serializable_value)
  | Array of 'a field DynArray.t
  | List of 'a field list
and 'a value_writer =
  ?properties:Xnc_property.properties ->
  'a Xnc_writer.t -> 'a serializable_value -> unit
and 'a value_reader =
  ?properties:Xnc_property.properties ->
  'a Xnc_reader.t -> 'a serializable_value -> 'a serializable_value
and 'a field =
  { prefix : string;
    name : string;
    mutable value : 'a serializable_value;
    properties : Xnc_property.properties
  }
;;

(* Constants *)
let default_int =
  Int 0
;;

let default_float =
  Float 0.0
;;

let default_char =
  Char '?'
;;

let default_bool =
  Bool false
;;

let default_string =
  String ""
;;

let default_date =
  Date { Netdate.year = 1970;
         Netdate.month = 1;
         Netdate.day = 1;
         Netdate.hour = 0;
         Netdate.minute = 0;
         Netdate.second = 0;
         Netdate.zone = 0;
         Netdate.week_day = -1
       }
;;

let default_option specific_default =
  Option (Some specific_default)
;;

let default_array () =
  Array (DynArray.create ())
;;

let default_list =
  List []
;;

(* Generic constructor *)
let create ?(properties = Xnc_property.empty) ?(prefix = "") name value =
  { name = name;
    value = value;
    properties = properties;
    prefix = prefix
  }
;;

(* Value constructors *)
let wrap_int value =
  Int value
;;

let wrap_float value =
  Float value
;;

let wrap_char value =
  Char value
;;

let wrap_bool value =
  Bool value
;;

let wrap_string value =
  String value
;;

let wrap_date value =
  Date value
;;

let wrap_option value =
  Option value
;;

let wrap_object value =
  Object (value :> Xnc_serialization.serializable)
;;

let wrap_dynarray value =
  Array value
;;

let wrap_array value =
  Array (DynArray.of_array value)
;;

let wrap_list value =
  List value
;;

(* Field constructors *)
let new_field
      ?properties
      ?prefix
      name
      value
      wrapper =
  create ?properties ?prefix name (wrapper value)
;;

let create_custom
      ?properties
      ?prefix
      name
      write_value
      read_value
      value =
  create ?properties ?prefix name (Custom (write_value, read_value, value))
;;
                 
(* Convenience contructors *)
let create_empty_array ?properties ?prefix name =
  create ?properties ?prefix name (default_array ())
;;

let create_empty_list ?properties ?prefix name =
  create ?properties ?prefix name default_list
;;

(* Accessors *)
let get_prefix field =
  field.prefix
;;

let get_name field =
  field.name
;;

let get_value field =
  field.value
;;

let get_properties field =
  field.properties
;;

(* Mutators *)
let set_value field new_value =
  field.value <- new_value
;;

(* Value accessors *)
let unwrap_int value =
  match value with
      Int i -> i
    | _ -> raise Invalid_serializable_type
;;

let unwrap_float value =
  match value with
      Float f -> f
    | _ -> raise Invalid_serializable_type
;;

let unwrap_char value =
  match value with
      Char c -> c
    | _ -> raise Invalid_serializable_type
;;

let unwrap_bool value =
  match value with
      Bool b -> b
    | _ -> raise Invalid_serializable_type
;;

let unwrap_string value =
  match value with
      String s -> s
    | _ -> raise Invalid_serializable_type
;;

let unwrap_date value =
  match value with
      Date d -> d
    | _ -> raise Invalid_serializable_type
;;

let unwrap_option value =
  match value with
      Option o -> o
    | _ -> raise Invalid_serializable_type
;;

let unwrap_object value =
  match value with
      Object o -> o
    | _ -> raise Invalid_serializable_type
;;

let unwrap_custom value =
  match value with
      Custom (_, _, c) -> c
    | _ -> raise Invalid_serializable_type
;;

let unwrap_dynarray value =
  match value with
      Array a -> a
    | _ -> raise Invalid_serializable_type
;;

let unwrap_array value =
  DynArray.to_array (unwrap_dynarray value)
;;

let unwrap_list value =
  match value with
      List l -> l
    | _ -> raise Invalid_serializable_type
;;

(* Field accessors *)
let get_int field =
  unwrap_int (get_value field)
;;

let get_float field =
  unwrap_float (get_value field)
;;

let get_char field =
  unwrap_char (get_value field)
;;

let get_bool field =
  unwrap_bool (get_value field)
;;

let get_string field =
  unwrap_string (get_value field)
;;

let get_date field =
  unwrap_date (get_value field)
;;

let get_option field =
  unwrap_option (get_value field)
;;

let get_object field =
  unwrap_object (get_value field)
;;

let get_custom field =
  unwrap_custom (get_value field)
;;

let get_dynarray field =
  unwrap_dynarray (get_value field)
;;

let get_array field =
  unwrap_array (get_value field)
;;

let get_list field =
  unwrap_list (get_value field)
;;

(* Tree visitors *)
type traversal = Pre_order | Post_order | Leaves_only;;

let rec iter ~traversal f field =
  let value = get_value field in
    begin match traversal with
        Pre_order -> f field
      | _ -> ()
    end;
    begin match value with
        List l -> List.iter (fun e -> iter ~traversal f e) l
      | Array a -> DynArray.iter (fun e -> iter ~traversal f e) a
      | _ ->
          begin match traversal with
              Leaves_only -> f field
            | _ -> ()
          end
    end;
    begin match traversal with
        Post_order -> f field
      | _ -> ()
    end
;;

let rec fold ~traversal f accu field =
  let value = get_value field in
  let accu =
    match traversal with
        Pre_order -> f accu field
      | _ -> accu in
  let accu =
    match value with
        List l ->
          List.fold_left (fun r e -> fold ~traversal f r e) accu l
      | Array a ->
          DynArray.fold_left (fun r e -> fold ~traversal f r e) accu a
      | _ ->
          begin match traversal with
              Leaves_only -> f accu field
            | _ -> accu
          end
  in
    match traversal with
        Post_order -> f accu field
      | _ -> accu
;;

let rec map ~traversal f field =
  let value = get_value field in
  let field =
    match traversal with
        Pre_order -> f field
      | _ -> field in
  let field =
    match value with
        List l ->
          let new_list =
            ExtList.List.map (fun e -> map ~traversal f e) l
          in
            { field with
                  value = List new_list
            }
      | Array a ->
          let new_array =
            DynArray.map (fun e -> map ~traversal f e) a
          in
            { field with
                  value = Array new_array
            }
      | _ ->
          begin match traversal with
              Leaves_only -> f field
            | _ -> field
          end
  in
    match traversal with
        Post_order -> f field
      | _ -> field
;;

let to_stream field =
  let stack = Stack.create () in
  let () = Stack.push field stack in
  let next _ =
    let get_next_field () =
      let next_field = Stack.pop stack in
      let value = get_value next_field in
      let children =
        match value with
            List l ->
              List.rev l
          | Array a ->
              List.rev (DynArray.to_list a)
          | _ -> []
      in
        List.iter (fun e -> Stack.push e stack) children;
        next_field
    in
      if Stack.is_empty stack then
        None
      else
        Some (get_next_field ())
  in
    Stream.from next
;;

let to_name_value_stream field =
  let get_full_name field =
    let prefix = get_prefix field in
    let name = get_name field in
    let properties = get_properties field in
    let attribute_prefix =
      if Xnc_property.is_attribute properties then "@" else "" in
    let prefix_name_separator = if prefix <> "" then ":" else "" in
      attribute_prefix ^ prefix ^ prefix_name_separator ^ name
  in
    
  let push_name_value parent field stack =
    let full_name = parent ^ "/" ^ (get_full_name field) in
    let value = get_value field in
      Stack.push (full_name, value) stack
  in
    
  let stack = Stack.create () in
  let () = push_name_value "" field stack in
  let next _ =
    let get_next_field () =
      let (full_name, value) = Stack.pop stack in
      let children =
        match value with
            List l ->
              List.rev l
          | Array a ->
              List.rev (DynArray.to_list a)
          | _ -> []
      in
        List.iter
          (fun e ->
             push_name_value full_name e stack)
          children;
        (full_name, value)
    in
      if Stack.is_empty stack then
        None
      else
        Some (get_next_field ())
  in
    Stream.from next
;;

let rec custom_value_equal ~obj_eq value1 value2 =
  let enum_eq enum1 enum2 =
    if (Enum.count enum1) = (Enum.count enum2) then
      Enum.fold2
        (fun e1 e2 accu ->
           accu && (custom_equal ~obj_eq e1 e2))
        true
        enum1
        enum2
    else
      false
  in
    try
      match value1 with
          Int i -> i = (unwrap_int value2)
        | Float f -> f = (unwrap_float value2)
        | Char c -> c = (unwrap_char value2)
        | Bool b -> b = (unwrap_bool value2)
        | String s -> s = (unwrap_string value2)
        | Date d -> d = (unwrap_date value2)
        | Option o -> o = (unwrap_option value2)
        | Object o -> obj_eq o (unwrap_object value2)
        | Custom (_, _, c) -> c = (unwrap_custom value2)
        | Array a ->
            let enum1 = DynArray.enum a in
            let enum2 = DynArray.enum (unwrap_dynarray value2) in
              enum_eq enum1 enum2
        | List l ->
            let enum1 = ExtList.List.enum l in
            let enum2 = ExtList.List.enum (unwrap_list value2) in
              enum_eq enum1 enum2
    with Invalid_serializable_type -> false
and custom_equal ~obj_eq field1 field2 =
  let prefix1 = get_prefix field1 in
  let name1 = get_name field1 in
  let value1 = get_value field1 in
  let properties1 = get_properties field1 in
  let prefix2 = get_prefix field2 in
  let name2 = get_name field2 in
  let value2 = get_value field2 in
  let properties2 = get_properties field2 in
    prefix1 = prefix2 &&
    name1 = name2 &&
    (Xnc_property.equal properties1 properties2) &&
    (custom_value_equal ~obj_eq value1 value2)
;;

let value_equal value1 value2 =
  custom_value_equal ~obj_eq:(fun o1 o2 -> false) value1 value2
;;

let equal field1 field2 =
  custom_equal ~obj_eq:(fun o1 o2 -> false) field1 field2
;;

let rec clone_value value =
  match value with
      Int i -> Int i
    | Float f -> Float f
    | Char c -> Char c
    | Bool b -> Bool b
    | String s -> String s
    | Date d -> Date d
    | Option o -> Option o
    | Object o -> Object (Oo.copy o)
    | Custom (r, w, c) -> Custom (r, w, c)
    | Array a -> Array (DynArray.copy a)
    | List l -> List (ExtList.List.map
                        (fun e ->
                           clone_field e)
                        l)
and clone_field field =
  let prefix = get_prefix field in
  let name = get_name field in
  let value = get_value field in
  let properties = get_properties field in
    create ~prefix ~properties name (clone_value value)
;;

(* private *)
let test_meta
      ?(properties = Xnc_property.empty)
      ?(prefix = "")
      field name =
    let field_prefix = get_prefix field in
    let field_name = get_name field in
    let field_properties = get_properties field in
      field_prefix = prefix &&
      field_name = name &&
      Xnc_property.subset properties field_properties
;;
(* [test_meta ?pr ?ps f n] tests if field [f] has name [n], prefix [pr] and
 * properties [ps]. *)

let value_present ?properties ?prefix field name =
  try
    Xnc_stream.fold
      ~f:(fun field accu ->
            let found = test_meta ?properties ?prefix field name in
              (accu || found, found))
      false
      (to_stream field)
  with Stream.Failure -> false
;;

let find_all ?properties ?prefix field name get_element =
  let rec find_all_elements name accu field =
    if test_meta ?properties ?prefix field name then
      let elem = get_element field in
        elem :: accu
    else
      let value = get_value field in
      let found_list =
        match value with
            Array a ->
              DynArray.fold_left (find_all_elements name) [] a
          | List l ->
              List.fold_left (find_all_elements name) [] l
          | _ -> []
      in
        List.rev_append (List.rev found_list) accu in
  let rev_list = find_all_elements name [] field in
    List.rev rev_list
;;

let find ?properties ?prefix field name get_element =
  try
    let found_field =
      Xnc_stream.fold
        ~f:(fun field _ ->
              let found = test_meta ?properties ?prefix field name in
                (field, found))
        field
        (to_stream field)
    in
      get_element found_field
  with Stream.Failure -> raise Not_found
;;

let find_all_fields ?properties ?prefix field name =
  find_all ?properties ?prefix field name Std.identity
;;

let find_field ?properties ?prefix field name =
  find ?properties ?prefix field name Std.identity
;;

let find_all_values ?properties ?prefix field name =
  find_all ?properties ?prefix field name get_value
;;

let find_value ?properties ?prefix field name =
  find ?properties ?prefix field name get_value
;;

let count_leaves field =
  fold ~traversal:Leaves_only (fun accu e -> succ accu) 0 field
;;

let rec write_value ?properties writer value =
  match value with
      Int i -> Xnc_writer.write_int ?properties writer i
    | Float f -> Xnc_writer.write_float ?properties writer f
    | Char c -> Xnc_writer.write_char ?properties writer c
    | Bool b -> Xnc_writer.write_bool ?properties writer b
    | String s -> Xnc_writer.write_string ?properties writer s
    | Date d -> Xnc_writer.write_date ?properties writer d
    | Option o -> Xnc_writer.write_option writer o
                    (Xnc_writer.write_null ?properties writer)
                    (write_value ?properties writer)
    | Object o -> Xnc_writer.write_object ?properties writer o
    | Custom (write, _, c) -> write ?properties writer c
    | Array a -> Xnc_writer.write_collection writer a DynArray.enum
                   (write writer)
    | List l -> Xnc_writer.write_collection writer l ExtList.List.enum
                  (write writer)
and write writer field =
  let prefix = get_prefix field in
  let name = get_name field in
  let value = get_value field in
  let properties = get_properties field in
    Xnc_writer.write_field writer
      prefix name value properties (write_value ~properties writer)
;;

let string_of_value value =
  let buf = Buffer.create 16 in
  let writer = Xnc_writer.create_from_buffer buf in
    write_value writer value;
    Buffer.contents buf
;;

let string_of_field field =
  let buf = Buffer.create 128 in
  let writer = Xnc_writer.create_from_buffer buf in
    write writer field;
    Buffer.contents buf
;;

let rec read_value ?properties reader prototype_value =
  match prototype_value with
      Int _ -> Int (Xnc_reader.read_int ?properties reader)
    | Float _ -> Float (Xnc_reader.read_float ?properties reader)
    | Char _ -> Char (Xnc_reader.read_char ?properties reader)
    | Bool _ -> Bool (Xnc_reader.read_bool ?properties reader)
    | String _ -> String (Xnc_reader.read_string ?properties reader)
    | Date _ -> Date (Xnc_reader.read_date ?properties reader)
    | Option o -> Option (Xnc_reader.read_option reader
                            (fun () -> Xnc_reader.try_parse_null ?properties
                                         reader)
                            (fun () -> read_value ?properties
                                         reader (Option.get o)))
    | (Object o) as obj -> Xnc_reader.read_object ?properties reader o;
                           obj
    | Custom (write, read, c) -> Custom (write, read,
                                         (read ?properties reader c))
    | Array a -> Array (Xnc_reader.read_heterogeneous_collection reader
                          a DynArray.enum
                          (fun p ->
                             read reader p;
                             p) DynArray.of_enum)
    | List l -> List (Xnc_reader.read_heterogeneous_collection reader
                        l ExtList.List.enum
                        (fun p ->
                           read reader p;
                           p) ExtList.List.of_enum)
and read reader prototype_field =
  let prefix = get_prefix prototype_field in
  let name = get_name prototype_field in
  let properties = get_properties prototype_field in
  let prototype_value = get_value prototype_field in
  let value = Xnc_reader.read_value_by_name reader
                prefix name properties
                (fun () -> read_value ~properties reader prototype_value)
  in
    prototype_field.value <- value
;;

let value_of_string s prototype_value =
  let reader = Xnc_reader.create_from_string s in
    read_value reader prototype_value
;;

let field_of_string s prototype_field =
  let reader = Xnc_reader.create_from_string s in
    read reader prototype_field;
    prototype_field
;;

(* Container functions *)
let append_field field_container field =
  let old_value = get_value field_container in
  let new_value =
    match old_value with
        (Array a) as arr ->
          DynArray.add a field;
          arr
      | List l ->
          let new_list = List.rev (field :: (List.rev l)) in
            List new_list
      | _ ->
          invalid_arg "Xnc_serializable.append_field is not applicable \
                       to atomic fields"
  in
    field_container.value <- new_value
;;

let replace_field field_container new_field =
  let old_value = get_value field_container in
  let new_value =
    match old_value with
        (Array a) as arr ->
          DynArray.iteri
            (fun i elem ->
               if elem.prefix = new_field.prefix &&
                  elem.name = new_field.name then
                 DynArray.set a i new_field
               else
                 ())
            a;
            arr
      | List l ->
          let new_list =
            List.map
              (fun elem ->
                 if elem.prefix = new_field.prefix &&
                    elem.name = new_field.name then
                   new_field
                 else
                   elem)
              l
          in
            List new_list
      | _ ->
          invalid_arg "Xnc_serializable.replace_field is not applicable \
                       to atomic fields"
  in
    field_container.value <- new_value
;;

(* TODO: rewrite this helper function using a custom writer?
 * (e.g.: HTML table writer)
 *)
let to_name_value_string_list field_container =
  let format_field field =
    let name = get_name field in
    let value = get_value field in
    let formatted_value = string_of_value value in
      (name, formatted_value) in
  let value = get_value field_container in
    match value with
        Array a -> DynArray.fold_right
                     (fun field l ->
                        (format_field field) :: l
                     ) a []
      | List l -> List.map format_field l
      | _ -> invalid_arg "Xnc_serializable.to_name_value_string_list \
                          is not applicable to atomic fields"
;;

