(** Serializable fields
  * 
  * This module provides serializable fields. A serializable field is made of a
  * name that may have a prefix to support namespaces a serializable value and
  * some properties that represent additional metadata. A serializable value
  * contains the value to be serialized.
  *)

exception Invalid_serializable_type
(** Raised when incorrect type is used. *)

type 'a serializable_value =
    Int of int
    (** [int] atom *)
  | Float of float
    (** [float] atom *)
  | Char of char
    (** [char] atom *)
  | Bool of bool
    (** [bool] atom *)
  | String of string
    (** [string] atom *)
  | Date of Netdate.t
    (** date atom *)
  | Option of 'a serializable_value option
    (** [option] value *)
  | Object of Xnc_serialization.serializable
    (** serializable object *)
  | Custom of ('a value_writer * 'a value_reader * 'a serializable_value)
    (** custom value with its custom reader and custom writer *)
  | Array of 'a field DynArray.t
    (** dynamic array of serializable fields *)
  | List of 'a field list
    (** [list] of serializable fields *)
(** The type of serializable values. Leaves and nodes of the serialization tree
  *)
and 'a value_writer =
  ?properties:Xnc_property.properties ->
  'a Xnc_writer.t -> 'a serializable_value -> unit
(** The type of writers that writes serializable values to some kind of
  * serialization support. *)
and 'a value_reader =
  ?properties:Xnc_property.properties ->
  'a Xnc_reader.t -> 'a serializable_value -> 'a serializable_value
(** The type of readers that read serializable values, provinding a prototype
  * value that shares the same structure. *)
and 'a field
(** The type of serializable fields. It stores value and metadata. Root of the
  * serialization tree. *)

val default_int : 'a serializable_value
(** Default [int] value ([0]). *)

val default_float : 'a serializable_value
(** Default [float] value ([0.0]). *)

val default_char : 'a serializable_value
(** Default [char] value (['?']). *)

val default_bool : 'a serializable_value
(** Default [bool] value ([false]). *)

val default_string : 'a serializable_value
(** Default [string] value ([""]). *)

val default_date : 'a serializable_value
(** Default [Netdate.t] value (the Epoch). *)

val default_option : 'a serializable_value -> 'a serializable_value
(** [default_option d] wraps the default value [d] to produce an serializable
  * value of kind [Option]. *)

val default_array : unit -> 'a serializable_value
(** [default_array ()] returns a default array (empty). *)

val default_list : 'a serializable_value
(** Default list (empty). *)

val create :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  string -> 'a serializable_value -> 'a field
(** [create n v] creates a serializable field with name [n] and value [v].
  * Optionally a prefix and some properties may be specified. *)

val wrap_int : int -> 'a serializable_value
(** [wrap_int v] wraps value [v]. *)

val wrap_float : float -> 'a serializable_value
(** [wrap_float v] wraps value [v]. *)

val wrap_char : char -> 'a serializable_value
(** [wrap_char v] wraps value [v]. *)

val wrap_bool : bool -> 'a serializable_value
(** [wrap_bool v] wraps value [v]. *)

val wrap_string : string -> 'a serializable_value
(** [wrap_string v] wraps value [v]. *)

val wrap_date : Netdate.t -> 'a serializable_value
(** [wrap_date v] wraps value [v]. *)

val wrap_option :
  'a serializable_value option -> 'a serializable_value
(** [wrap_option v] wraps value [v]. *)

val wrap_object :
  #Xnc_serialization.serializable -> 'a serializable_value
(** [wrap_object v] wraps value [v]. *)

val wrap_dynarray : 'a field DynArray.t -> 'a serializable_value
(** [wrap_dynarray v] wraps value [v]. *)

val wrap_array : 'a field array -> 'a serializable_value
(** [wrap_array v] wraps value [v]. *)

val wrap_list : 'a field list -> 'a serializable_value
(** [wrap_list v] wraps value [v]. *)

val new_field :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  string ->
  'a ->
  ('a -> 'b serializable_value) ->
  'b field
(** [new_field n v wrap] creates a serializable field with name [n] and value
  * [v] wrapped by [wrap]. Optionally a name prefix and some properties may be
  * supplied. *)

val create_custom :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  string ->
  'a value_writer ->
  'a value_reader ->
  'a serializable_value -> 'a field
(** [create_custom n w r v] creates a custom field with name [n], value [v],
  * custom writer [w], and custom reader [r]. Optionally a name prefix and some
  * properties may be supplied. *)

val create_empty_array :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  string -> 'a field
(** [create_array n] creates an empty array field with name [n]. Optionally a
  * name prefix and some properties may be supplied. *)

val create_empty_list :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  string -> 'a field
(** [create_list n] creates an empty list field with name [n]. Optionally a
  * name prefix and some properties may be supplied. *)

val get_prefix : 'a field -> string
(** [get_prefix f] returns the name prefix of [f]. *)

val get_name : 'a field -> string
(** [get_name f] returns the name of [f]. *)

val get_value : 'a field -> 'a serializable_value
(** [get_value f] returns the serializable value of [f]. *)

val get_properties : 'a field -> Xnc_property.properties
(** [get_properties f] returns the properties of [f]. *)

val set_value : 'a field -> 'a serializable_value -> unit
(** [set_value f v] replaces the current serializable value of [f] with [v]. *)

val unwrap_int : 'a serializable_value -> int
(** [unwrap_int v] unwraps value [v]. *)

val unwrap_float : 'a serializable_value -> float
(** [unwrap_float v] unwraps value [v]. *)

val unwrap_char : 'a serializable_value -> char
(** [unwrap_char v] unwraps value [v]. *)

val unwrap_bool : 'a serializable_value -> bool
(** [unwrap_bool v] unwraps value [v]. *)

val unwrap_string : 'a serializable_value -> string
(** [unwrap_string v] unwraps value [v]. *)

val unwrap_date : 'a serializable_value -> Netdate.t
(** [unwrap_date v] unwraps value [v]. *)

val unwrap_option : 'a serializable_value -> 'a serializable_value option
(** [unwrap_option o] unwraps the option [o] returning the serializable value
  * contained. *)

val unwrap_object :
  'a serializable_value -> Xnc_serialization.serializable
(** [unwrap_object o] unwraps object [o] coercing it to
  * [Xnc_serialization.serializable] interface. *)

val unwrap_custom : 'a serializable_value -> 'a serializable_value
(** [unwrap_custom c] unwraps the custom value [c] returning the serializable
  * value contained. *)

val unwrap_dynarray : 'a serializable_value -> 'a field DynArray.t
(** [unwrap_dynarray d] unwraps the array value [d] returning the array of
  * serializable fields contained. *)

val unwrap_array : 'a serializable_value -> 'a field array
(** [unwrap_array a] unwraps the array value [a] returning the array of
  * serializable fields contained. *)

val unwrap_list : 'a serializable_value -> 'a field list
(** [unwrap_list l] unwraps the list value [l] returning the list of
  * serializable fields contained. *)

val get_int : 'a field -> int
(** [get_int f] returns the [int] value contained in [f]. *)

val get_float : 'a field -> float
(** [get_float f] returns the [float] value contained in [f]. *)

val get_char : 'a field -> char
(** [get_char f] returns the [char] value contained in [f]. *)

val get_bool : 'a field -> bool
(** [get_bool f] returns the [bool] value contained in [f]. *)

val get_string : 'a field -> string
(** [get_string f] returns the [string] value contained in [f]. *)

val get_date : 'a field -> Netdate.t
(** [get_date f] returns the date value contained in [f]. *)

val get_option : 'a field -> 'a serializable_value option
(** [get_date f] returns the [option] serializable value contained in [f]. *)

val get_object : 'a field -> Xnc_serialization.serializable
(** [get_object f] returns the object contained in [f], coercing it to
  * [Xnc_serialization.serializable]. *)

val get_custom : 'a field -> 'a serializable_value
(** [get_custom f] returns the custom serializable value contained in [f]. *)

val get_dynarray : 'a field -> 'a field DynArray.t
(** [get_dynarray f] returns the [DynArray.t] of serializable fields contained
  * in [f]. *)

val get_array : 'a field -> 'a field array
(** [get_array f] returns the [array] of serializable fields contained in [f].
  *)

val get_list : 'a field -> 'a field list
(** [get_list f] returns the [list] of serializable fields contained in [f]. *)

type traversal =
    Pre_order
    (** depth first, VLR (Visiting Left Right) *)
  | Post_order
    (** depth first, LRV (Left Right Visiting) *)
  | Leaves_only
    (** depth first, ignores nodes with children *)
(** The type of tree traversals. *)

val iter : traversal:traversal -> ('a field -> unit) -> 'a field -> unit
(** [iter traversal f field] visits all the nodes of [field] applying [f] to
  * each visited node, using [traversal] method. *)

val fold :
  traversal:traversal -> ('b -> 'a field -> 'b) -> 'b -> 'a field -> 'b
(** [fold traversal f a field] folds [f] between the nodes of [field] in the
  * order specified by [traversal], starting from value [a], producing a value
  * of type ['b]. *)

val map : traversal:traversal -> ('a field -> 'a field) -> 'a field -> 'a field
(** [map traversal f field] applies [f] to each node of [field] visited in the
  * order specified by [traversal], producing a new serializable field. *)

val to_stream : 'a field -> 'a field Stream.t
(** [to_stream traversal field] flattens the tree in [field] and returns a
  * stream made up of the nodes of the tree in depth-first LTR order. *)

val to_name_value_stream : 'a field -> (string * 'a serializable_value) Stream.t
(** [to_name_value_stream traversal field] flattens the tree in [field] and
  * returns a stream made up of node absolute path and its value. *)

val custom_value_equal :
  obj_eq:(Xnc_serialization.serializable ->
            Xnc_serialization.serializable -> bool) ->
  'a serializable_value -> 'a serializable_value -> bool
(** [custom_value_equal obj_eq v1 v2] compares [v1] and [v2], using [obj_eq] as
  * the comparison function for serializable objects. *)

val custom_equal :
  obj_eq:(Xnc_serialization.serializable ->
            Xnc_serialization.serializable -> bool) ->
  'a field -> 'a field -> bool
(** [custom_equal obj_eq f1 f2] compares [f1] and [f2], using [obj_eq] as
  * the comparison function for serializable objects. The comparison is made on
  * prefix, name, value, and properties of each field. *)

val value_equal : 'a serializable_value -> 'a serializable_value -> bool
(** [value_equal v1 v2] compares [v1] and [v2]. If serializable objects are
  * present, the comparison always returns [false]. *)

val equal : 'a field -> 'a field -> bool
(** [equal f1 f2] compares [f1] and [f2]. If serializable objects are present,
  * the comparison always returns [false]. *)

val clone_value : 'a serializable_value -> 'a serializable_value
(** [clone_value v] returns a new serializable value that is a copy of [v]. *)

val clone_field : 'a field -> 'a field
(** [clone_field f] returns a new serializable field that is a copy of [f]. *)

val value_present :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  'a field -> string -> bool
(** [value_present f n] tests [f] for the presence of a field with name [n].
  *
  * @param properties Optional field properties that must be a subset of the
  * properties of the field to test.
  * @param prefix Optional name prefix to match. Defaults to [""].
  *)

val find_all_fields :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  'a field -> string -> 'a field list
(** [find_all_fields f n] returns a list of all the fields of [f] with name [n].
  *
  * @param properties Optional field properties that must be a subset of the
  * properties of the field to test.
  * @param prefix Optional name prefix to match. Defaults to [""].
  *)

val find_field :
  ?properties:Xnc_property.properties ->
  ?prefix:string -> 'a field -> string -> 'a field
(** [find_field f n] finds the first field of [f] with name [n].
  *
  * @param properties Optional field properties that must be a subset of the
  * properties of the field to test.
  * @param prefix Optional name prefix to match. Defaults to [""].
  *)

val find_all_values :
  ?properties:Xnc_property.properties ->
  ?prefix:string -> 'a field -> string -> 'a serializable_value list
(** [find_all_values f n] returns a list of all the values of [f] with name [n].
  *
  * @param properties Optional field properties that must be a subset of the
  * properties of the field to test.
  * @param prefix Optional name prefix to match. Defaults to [""].
  *)

val find_value :
  ?properties:Xnc_property.properties ->
  ?prefix:string ->
  'a field -> string -> 'a serializable_value
(** [find_value f n] finds the first field of [f] with name [n].
  *
  * @param properties Optional field properties that must be a subset of the
  * properties of the field to test.
  * @param prefix Optional name prefix to match. Defaults to [""].
  *)

val count_leaves : 'a field -> int
(** [count_leaves f] returns the number of leaves of [f]. *)

val write_value : 'a value_writer
(** [write_value w v] writes [v] using writer [w]. *)

val write : 'a Xnc_writer.t -> 'a field -> unit
(** [write w f] writes [f] using writer [w]. *)

val string_of_value : string serializable_value -> string
(** [string_of_value v] serializes [v] to a string. *)

val string_of_field : string field -> string
(** [string_of_value f] serializes [f] to a string. *)

val read_value : 'a value_reader
(** [read_value r v] reads a serializable value from reader [r] using [v] as
  * descriptor (a value that shares the same structure of the value to be read).
  *)

val read : 'a Xnc_reader.t -> 'a field -> unit
(** [read r f] reads a serializable field from reader [r] using [f] as
  * descriptor (a field that shares the same structure of the value to be read).
  *)

val value_of_string :
  string -> string serializable_value -> string serializable_value
(** [value_of_string s v] deserializes a serializable value from [s] using [v]
  * as descriptor (a value that shares the same structure of the value to be
  * read).
  *)

val field_of_string : string -> string field -> string field
(** [field_of_string s f] deserializes a serializable field from [s] using [f]
  * as descriptor (a field that shares the same structure of the value to be
  * read).
  *)

val append_field : 'a field -> 'a field -> unit
(** [append_field c f] appends field [f] to the container [c]. Raises
  * [Invalid_arg] is [c] is an atomic field. *)

val replace_field : 'a field -> 'a field -> unit
(** [replace_field c f] replaces a field with the same name and prefix of [f] in
  * the container [c]. Raises [Invalid_arg] is [c] is an atomic field. *)

val to_name_value_string_list : string field -> (string * string) list
(** [to_name_value_string_list f] returns a name/value list representing [f]. *)

