(** Common types shared by other modules
  *)

class type serializable =
object
  method to_string : unit -> string
  (** Serialize the state of the object to a string value. *)

  method from_string : string -> unit
  (** Rebuilds the state of the object from its string serialization. *)

end
(** This interface is used to serialize objects.
  *)

class ['a] generic_serializer :
  'a -> ('a -> string) -> (string -> 'a) ->
object
  method get_value : 'a
  (** Returns the value to be serialized/deserialized. *)

  method to_string : unit -> string
  (** Calls the serialization function passed in the constructor. *)

  method from_string : string -> unit
  (** Calls the deserialization function passed in the constructor. After
    * calling this method the deserialized value can be retrieved calling
    * [get_value] *)

end
(** This class can be used to serialize any type ['a]. The constructor receives
  * the value to serialize, the function that serializes it to a string and the
  * function that deserialize it.
  *)

type common_config =
    { delimiter_char : char;
      (** Quote character. *)
      escape_char : char;
      (** Escape character. *)
      none_string : string;
      (** String representing the [None] value. *)
      start_of_list : string;
      (** String that starts a list of values. *)
      end_of_list : string;
      (** String that ends a list of values. *)
      list_separator : string;
      (** String that separates values in a list. *)
      name_value_separator : string;
      (** String that separates the name of a field from its value. *)
      attribute_marker : string;
      (** String that marks attributes. *)
      prefix_separator : string
      (** String that separates the namespace from the name of a field. *)
    }
(** This type defines the configuration needed to parametrize the serialization
  * language used by simple serializers.
  *)

val default_config : common_config
(** The default configuration of the serialization language:
  * - String delimiter character: ['"']
  * - Escape character: ['\']
  * - None string: ["None"]
  * - Start of list: ["\["]
  * - End of list: ["\]"]
  * - List separator: [";"]
  * - Name/value separator: ["="]
  * - Attribute marker: ["@"]
  * - Prefix/name separator: [":"]
  *)

