(** Serializers
  *
  * This module provides classes to perform serialization and deserialization of
  * serializable fields.
  *)

class type ['a] serializer =
object
  method serialize :
    'a Xnc_serializable.field -> 'a
  (** [serialize f] serializes field [f] to a value of type ['a]. *)

  method deserialize :
    'a Xnc_serializable.field -> 'a -> unit
  (** [deserialize f a] deserializes a serializable field from [a], using [f] as
    * descriptor (a field that shares the same structure. *)

  method to_netchannel :
    'a Xnc_serializable.field -> Netchannels.out_obj_channel -> unit
  (** [to_netchannel f] serializes field [f] to a Netchannel. *)

  method from_netchannel :
    'a Xnc_serializable.field -> Netchannels.in_obj_channel -> unit
  (** [from_netchannel f in_ch] deserializes a serializable field from [in_ch],
    * using [f] as descriptor (a field that shares the same structure. *)

end;;
(** This interface provides the methods to serialize and deserialize a
  * serializable field. The serialization target is value of type ['a] or a
  * Netchannel. *)

class virtual netchannel_serializer :
  ?buf_size:int ->
  unit ->
object
  method serialize :
    string Xnc_serializable.field -> string
  (** See {!method: Xnc_serializing.serializer.serialize}. *)

  method deserialize :
    string Xnc_serializable.field -> string -> unit
  (** See {!method: Xnc_serializing.serializer.deserialize}. *)

  method virtual to_netchannel :
    string Xnc_serializable.field -> Netchannels.out_obj_channel -> unit
  (** See {!method: Xnc_serializing.serializer.to_netchannel}. *)

  method virtual from_netchannel :
    string Xnc_serializable.field -> Netchannels.in_obj_channel -> unit
  (** See {!method: Xnc_serializing.serializer.from_netchannel}. *)


end;;
(** Netchannel based string serializer. *)

class null_serializer :
  unit ->
  [string] serializer
(** No-op serializer. *)

class simple_serializer :
  ?config:Xnc_serialization.common_config ->
  unit ->
  [string] serializer
(** Simple string serializer.
  *
  * @param config serialization language configuration. *)

class xml_serializer :
  ?config:Pxp_types.config ->
  unit ->
  [string] serializer
(** XML string serializer. The serialization methods produce pretty printed XML,
  * the deserialization methods are based on PXP pull parser.
  *
  * @param config PXP parser configuration. *)

class validating_xml_serializer :
  ?config:Pxp_types.config ->
  ?alt:Pxp_reader.resolver list ->
  ?system_id:string ->
  ?fixenc:Pxp_types.encoding ->
  ?spec:('a Pxp_document.node Pxp_document.extension as 'a) Pxp_document.spec ->
  unit ->
  [string] serializer
(** XML string serializer with DTD validation. The serialization methods produce
  * pretty printed XML, the deserialization methods are based on PXP document.
  *
  * @param config PXP parser configuration.
  * @param alt list of further PXP resolvers
  * @param system_id assigns the channel a SYSTEM ID that is only used to
  * resolve further relative SYSTEM IDs.
  * @param fixenc forces a certain character encoding
  * @param spec PXP object specification.
*)

