let xor_transform k s =
  let rec get_key k len =
    let klen = String.length k in
      if klen = len then
        k
      else if klen > len then
        String.sub k 0 len
      else
        get_key (k ^ k) len in
  let result = String.copy s in
  let slen = String.length s in
  let key = get_key k slen in
    Cryptokit.xor_string key 0 result 0 slen;
    result
;;

let base64_encode s =
  Netencoding.Base64.encode s
;;

let base64_decode s =
  Netencoding.Base64.decode s
;;

let base16_encode s =
  let len = String.length s in
  let buffer = Buffer.create (len * 2) in
    String.iter
      (fun c ->
         let code = Char.code c in
         let hex = Printf.sprintf "%02x" code in
           Buffer.add_string buffer hex
      )
      s;
    Buffer.contents buffer
;;

let base16_decode s =
  let len = String.length s in
  let buffer = Buffer.create (len / 2) in
  let rec loop is =
    if String.length is > 0 then
      begin
        let hex = String.sub is 0 2 in
        let code = Scanf.sscanf hex "%x" Std.identity in
        let c = Char.chr code in
          Buffer.add_char buffer c;
          loop (Str.string_after is 2)
      end
    else
      ()
  in
    loop s;
    Buffer.contents buffer
;;

let escape_xml s =
  Str.global_substitute
    (Str.regexp "<\\|>\\|&\\|\"\\|'")
    (fun s ->
       match Str.matched_string s with
           "<" -> "&lt;"
         | ">" -> "&gt;"
         | "&" -> "&amp;"
         | "\"" -> "&quot;"
         | "'" -> "&apos;"
         | _ -> assert false)
    s
;;

let unescape_xml s =
  Str.global_substitute
    (Str.regexp "&\\(lt\\|gt\\|amp\\|quot\\|apos\\);")
    (fun s ->
       match Str.matched_group 1 s with
           "lt" -> "<"
         | "gt" -> ">"
         | "amp" -> "&"
         | "quot" -> "\""
         | "apos" -> "'"
         | _ -> assert false)
    s
;;

module Cipher =
struct
  type cipher_type =
      Aes
    | Des
    | Triple_des

  let transform c key mode s =
    let padding = Cryptokit.Padding.length in
    let cipher =
      match c with
          Aes ->
            Cryptokit.Cipher.aes ~pad:padding
        | Des ->
            Cryptokit.Cipher.des ~pad:padding
        | Triple_des ->
            Cryptokit.Cipher.triple_des ~pad:padding in
    let transform = cipher key mode in
      Cryptokit.transform_string transform s

  let encrypt c key s =
    transform c key Cryptokit.Cipher.Encrypt s

  let decrypt c key s =
    transform c key Cryptokit.Cipher.Decrypt s

end;;

module Mac =
struct
  type mac_type =
      Md5
    | Sha1
    | Aes
    | Des
    | Triple_des
    | Des_final_triple_des

  let digest m key s =
    let padding = Cryptokit.Padding.length in
    let mac_algorithm =
      match m with
          Md5 ->
            Cryptokit.MAC.hmac_md5
        | Sha1 ->
            Cryptokit.MAC.hmac_sha1
        | Aes ->
            (fun s -> Cryptokit.MAC.aes ~pad:padding s)
        | Des ->
            (fun s -> Cryptokit.MAC.des ~pad:padding s)
        | Triple_des ->
            (fun s -> Cryptokit.MAC.triple_des ~pad:padding s)
        | Des_final_triple_des ->
            (fun s -> Cryptokit.MAC.des_final_triple_des ~pad:padding s) in
    let hash = mac_algorithm key in
      Cryptokit.hash_string hash s

end;;

let zlib_compress s =
  let transform = Cryptokit.Zlib.compress () in
    Cryptokit.transform_string transform s
;;

let zlib_uncompress s =
  let transform = Cryptokit.Zlib.uncompress () in
    Cryptokit.transform_string transform s
;;

let compose_cryptokit_transform transform_list =
  List.fold_left
    (fun result transform ->
       Cryptokit.compose result transform)
    (List.hd transform_list)
    (List.tl transform_list)
;;

