(** String encoding/decoding
  *
  * This module handles string encoding, encryption, hashing, and compression.
  *)

val xor_transform : string -> string -> string
(** [xor_transform s k] performs the XOR (exclusive OR) of characters s{_ 0},
  * s{_ 1}, ..., s{_ len - 1} of [s], with characters k{_ 0}, k{_ 1}, ..., k{_
  * len - 1} of [k]. If length of [k] is lesser than length of [s], the key is
  * padded appending copies of [k]: e.g., [xor_transform "test string" "key"] is
  * equivalent to [xor_transform "test string" "keykeykeyke"].
  *)

val base64_encode : string -> string
(** [base64_encode s] performs base64 encoding on [s]. *)

val base64_decode : string -> string
(** [base64_decode s] performs base64 decoding on [s]. *)

val base16_encode : string -> string
(** [base16_encode s] performs base16 encoding on [s]. *)

val base16_decode : string -> string
(** [base16_decode s] performs base16 decoding on [s]. *)

val escape_xml : string -> string
(** [escape_xml s] encodes characters of [s] that need protection in XML
  * documents by converting them to entity references.
  *)

val unescape_xml : string -> string
(** [unescape_xml s] decodes characters of [s] converted to entity references in
  * XML documents.
  *)

module Cipher :
sig
  type cipher_type =
      Aes
      (** Advanced Encryption Standard, also known as Rijndael. *)
    | Des
      (** Data Encryption Standard. *)
    | Triple_des
      (** Triple DES. *)
  (** Enumerates available ciphers. *)

  val encrypt : cipher_type -> string -> string -> string
  (** [encrypt c k s] encrypts [s] with key [k] using cipher [c].
    *)

  val decrypt : cipher_type -> string -> string -> string
  (** [decrypt c k s] dencrypts [s] encrypted with key [k] using cipher [c].
    *)
end
(** This modules provides string encryption and decryption.
  *)

module Mac :
sig
  type mac_type =
      Md5
      (** HMAC MD5. *)
    | Sha1
      (** HMAC SHA1. *)
    | Aes
      (** HMAC AES (CBC mode). *)
    | Des
      (** HMAC DES (CBC mode). *)
    | Triple_des
      (** HMAC Triple DES (CBC mode). *)
    | Des_final_triple_des
      (** Fast HMAC Triple DES (CBC mode). *)
  (** Enumerates available MAC algorithms. *)

  val digest : mac_type -> string -> string -> string
  (** [digest m k s] digests [s] with key [k] using MAC algorithm [m].
    *)
end
(** This modules provides message authentication codes.
  *)

val zlib_compress : string -> string
(** [zlib_compress s] compresses [s] using [zlib] library.
  *)

val zlib_uncompress : string -> string
(** [zlib_compress s] decompresses [s] using [zlib] library.
  *)

val compose_cryptokit_transform :
  Cryptokit.transform list -> Cryptokit.transform
(** Compose a list of transforms, feeding the output of each transform to the
  * input of the next transform in the list.
  *)

