(** Writers
  *
  * This module provides the functions to write values formatted by
  * [Xnc_formatting.formatter], and printed to a support by
  * [Xnc_printing.printer].
  *)

type 'a t
(** The type of writers. *)

val create :
  'a Xnc_formatting.formatter ->
  'a Xnc_printing.printer ->
  'a t
(** [create f p] creates a writer from formatter [f] and printer [p]. *)

val create_from_netchannel :
  ?config:Xnc_serialization.common_config ->
  Netchannels.out_obj_channel ->
  string t
(** [create_from_netchannel out_ch] creates a string writer that writes to
  * [out_ch].
  *
  * @param config serialization language configuration. *)

val create_from_buffer :
  ?config:Xnc_serialization.common_config ->
  Buffer.t ->
  string t
(** [create_from_buffer b] creates a string writer that writes to [b].
  *
  * @param config serialization language configuration. *)

val write_null :
  ?properties:Xnc_property.properties ->
  'a t -> unit -> unit
(** [write_null w] writes a null value to writer [w]. *)

val write_int :
  ?properties:Xnc_property.properties ->
  'a t -> int -> unit
(** [write_int w v] writes value [v] to writer [w]. *)

val write_float :
  ?properties:Xnc_property.properties ->
  'a t -> float -> unit
(** [write_float w v] writes value [v] to writer [w]. *)

val write_char :
  ?properties:Xnc_property.properties ->
  'a t -> char -> unit
(** [write_char w v] writes value [v] to writer [w]. *)

val write_bool :
  ?properties:Xnc_property.properties ->
  'a t -> bool -> unit
(** [write_bool w v] writes value [v] to writer [w]. *)

val write_raw_string :
  ?properties:Xnc_property.properties ->
  'a t -> string -> unit
(** [write_raw_string w s] writes string [s] to writer [w] without
  * modifications. *)

val write_string :
  ?properties:Xnc_property.properties ->
  'a t -> string -> unit
(** [write_string w v] writes value [v] to writer [w]. *)

val write_date :
  ?properties:Xnc_property.properties ->
  'a t -> Netdate.t -> unit
(** [write_date w v] writes value [v] to writer [w]. *)

val write_value :
  'a t -> 'b -> ('b -> 'a) -> unit
(** [write_value w v f] writes value [v] formatted by [f] to writer [w]. *)

val write_option :
  'a t -> 'b option -> (unit -> unit) -> ('b -> unit) -> unit
(** [write_option w o wn ws] writes option [o] to writer [w], using [wn] to
  * write [None], and [ws] to write [Some v]. *)

val write_object :
  ?properties:Xnc_property.properties ->
  'a t -> #Xnc_serialization.serializable -> unit
(** [write_object w o] writes serializable object [o] to writer [w]. *)

val write_enum :
  'a t -> 'b Enum.t -> ('b -> unit) -> unit
(** [write_enum w e we] writes enumeration [e] to writer [w], using [we] to
  * write each element. *)

val write_collection :
  'a t -> 'b -> ('b -> 'c Enum.t) -> ('c -> unit) -> unit
(** [write_collection w c to_enum we] writes collection [c] to writer [w], using
  * [we] to write each element. [to_enum] is used to convert the collection to
  * an enumeration. *)

val write_field :
  'a t -> string -> string -> 'b -> Xnc_property.properties ->
  ('b -> unit) -> unit
(** [write_field w p n v prop wv] writes a field with prefix [p], name [n],
  * value [v], and properties [prop] to writer [w], using [wv] to write the
  * value. *)

val flush : 'a t -> unit
(** [flush w] flushes writer [w]. *)

val close : 'a t -> unit
(** [close w] closes writer [w]. *)

