open OUnit

module Test_cache =
  Xnc_cache.Make(Xnc_core.Null_mutex)
;;

module Test_safe_cache =
  Xnc_cache.Make(Mutex)
;;

(* tests *)
let test_insert _ =
  let cache = Test_cache.create () in
    Test_cache.insert cache "value" 10;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected cache content"
      10
      (Test_cache.get cache "value");
    Test_cache.insert cache "value" 20;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected cache content"
      20
      (Test_cache.get cache "value");
;;

let test_mem _ =
  let cache = Test_cache.create () in
  let () = Test_cache.insert cache "v1" "a" in
  let () = Test_cache.insert cache "v2" "a" in
    OUnit.assert_bool
      "v1 should be present"
      (Test_cache.mem cache "v1");
    OUnit.assert_bool
      "v2 should be present"
      (Test_cache.mem cache "v2");
    OUnit.assert_bool
      "v3 should not be present"
      (not (Test_cache.mem cache "v3"));
;;

let test_peek _ =
  let cache = Test_cache.create () in
  let () = Test_cache.insert cache "v1" "a" in
  let () = Test_cache.insert cache "v2" "b" in
    OUnit.assert_equal
      ~msg:"Unexpected cache content: v1 should be Some \"a\""
      (Some "a")
      (Test_cache.peek cache "v1");
    OUnit.assert_equal
      ~msg:"Unexpected cache content: v2 should be Some \"b\""
      (Some "b")
      (Test_cache.peek cache "v2");
    OUnit.assert_equal
      ~msg:"Unexpected cache content: v3 should be None"
      (None)
      (Test_cache.peek cache "v3");
;;

let test_get _ =
  let cache = Test_cache.create () in
  let () = Test_cache.insert cache "value" "a" in
  let v = Test_cache.get cache "value" in
  let v' = Test_cache.get cache "value" in
  let () = Test_cache.insert cache "value" "a" in
  let v'' = Test_cache.get cache "value" in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected cache content"
      "a"
      v;
    OUnit.assert_bool
      "v should be physically equal to v'"
      (v == v');
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected cache content (replaced element)"
      "a"
      v'';
    OUnit.assert_bool
      "v should not be physically equal to v''"
      (v != v'');
    OUnit.assert_raises
      ~msg:"Unexpected cache content"
      Not_found
      (fun () -> Test_cache.get cache "not_present");
;;

let test_remove _ =
  let cache = Test_cache.create () in
    Test_cache.insert cache "value" 10;
    Test_cache.remove cache "value";
    OUnit.assert_raises
      ~msg:"Unexpected cache content"
      Not_found
      (fun () -> Test_cache.get cache "value");
;;

let test_relative_expiration _ =
  let cache = Test_cache.create () in
    Test_cache.insert ~expiration:(Xnc_cache.Relative 0.05) cache "value" 10;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected cache content"
      10
      (Test_cache.get cache "value");
    Core_test_helper.sleep_float 0.1;
    OUnit.assert_raises
      ~msg:"Unexpected cache content after 0.1 seconds"
      Not_found
      (fun () -> Test_cache.get cache "value");
;;

let test_absolute_expiration _ =
  let cache = Test_cache.create () in
    Test_cache.insert
      ~expiration:(Xnc_cache.Absolute
                     { Netdate.year = 2030;
                       Netdate.month = 1;
                       Netdate.day = 2;
                       Netdate.hour = 18;
                       Netdate.minute = 47;
                       Netdate.second = 5;
                       Netdate.zone = 120;
                       Netdate.week_day = 0 
                     })
      cache "value" 10;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected cache content (absolute expiration in the future)"
      10
      (Test_cache.get cache "value");
    Test_cache.insert
      ~expiration:(Xnc_cache.Absolute
                     { Netdate.year = 2006;
                       Netdate.month = 10;
                       Netdate.day = 22;
                       Netdate.hour = 15;
                       Netdate.minute = 47;
                       Netdate.second = 5;
                       Netdate.zone = 120;
                       Netdate.week_day = 0 
                     })
      cache "value" 10;
    OUnit.assert_raises
      ~msg:"Unexpected cache content (absolute expiration in the past)"
      Not_found
      (fun () -> Test_cache.get cache "value");
;;

let test_file_expiration _ =
  let cache = Test_cache.create () in
  let filename = Printf.sprintf
                   "/tmp/Test_cache.%d.test"
                   (Unix.getpid ())
  in
    close_out (open_out filename);
    Test_cache.insert
      ~expiration:(Xnc_cache.File filename)
      cache "value" 10;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected cache content"
      10
      (Test_cache.get cache "value");
    (* mtime is rounded at seconds, so we must wait at least a second *)
    Core_test_helper.sleep_float 1.05;
    close_out (open_out filename);
    OUnit.assert_raises
      ~msg:"Unexpected cache content after file modification"
      Not_found
      (fun () -> Test_cache.get cache "value");
    Test_cache.insert
      ~expiration:(Xnc_cache.File filename)
      cache "value" 10;
    Sys.remove filename;
    OUnit.assert_raises
      ~msg:"Unexpected cache content after file deletion"
      Not_found
      (fun () -> Test_cache.get cache "value");
;;

let test_sliding_expiration _ =
  let cache = Test_cache.create () in
    Test_cache.insert
      ~expiration:(Xnc_cache.Sliding 0.15)
      cache "value" 10;
    Core_test_helper.sleep_float 0.1;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected cache content after 0.1 seconds"
      10
      (Test_cache.get cache "value");
    Core_test_helper.sleep_float 0.1;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected cache content after 0.2 seconds"
      10
      (Test_cache.get cache "value");
    Core_test_helper.sleep_float 0.2;
    OUnit.assert_raises
      ~msg:"Unexpected cache content after 0.4 seconds"
      Not_found
      (fun () -> Test_cache.get cache "value");
;;

let test_on_remove _ =
  let cache = Test_cache.create () in
  let rec on_remove key value reason =
    match reason with
        Xnc_cache.Expired ->
          Test_cache.insert ~on_remove cache key (value + 1)
      | Xnc_cache.Removed ->
          Test_cache.insert cache key (value * 2)
  in
    Test_cache.insert
      ~expiration:(Xnc_cache.Sliding 0.05)
      ~on_remove
      cache "value" 10;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected cache content"
      10
      (Test_cache.get cache "value");
    Core_test_helper.sleep_float 0.1;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected cache content after 0.1 seconds"
      11
      (Test_cache.get cache "value");
    Test_cache.remove cache "value";
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected cache content after remove call"
      22
      (Test_cache.get cache "value");
;;

let test_expire _ =
  let cache = Test_cache.create () in
    Test_cache.insert
      ~expiration:(Xnc_cache.Sliding 0.15)
      cache "v1" 10;
    Test_cache.insert
      ~expiration:(Xnc_cache.Relative 0.05)
      cache "v2" 1;
    Test_cache.insert
      ~expiration:(Xnc_cache.Sliding 0.05)
      cache "v3" 0;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of items"
      3
      (Test_cache.count cache);
    Core_test_helper.sleep_float 0.1;
    Test_cache.expire cache;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of items after expire"
      1
      (Test_cache.count cache);
;;

let test_clear _ =
  let cache = Test_cache.create () in
    Test_cache.insert
      ~expiration:(Xnc_cache.Sliding 0.15)
      cache "v1" 10;
    Test_cache.insert
      ~expiration:(Xnc_cache.Relative 0.05)
      cache "v2" 1;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of items"
      2
      (Test_cache.count cache);
    Test_cache.clear cache;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of items after clear"
      0
      (Test_cache.count cache);
;;

let test_thread_safe_cache _ =
  let cache = Test_safe_cache.create () in
  let () = Test_safe_cache.insert cache "counter" (ref 0) in
  let t1 = Thread.create
             (fun _ ->
                let () = Test_safe_cache.lock cache in
                let cref = Test_safe_cache.get cache "counter" in
                let cval = !cref in
                  Thread.delay 0.5;
                  cref := cval + 1;
                  Test_safe_cache.unlock cache)
             () in
  let t2 = Thread.create
             (fun _ ->
                Test_safe_cache.with_cache_lock cache
                  (fun () ->
                     let cref = Test_safe_cache.get cache "counter" in
                     let cval = !cref in
                       Thread.delay 0.25;
                       cref := cval + 1))
             () in
    Thread.join t1;
    Thread.join t2;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of items"
      2
      (!(Test_safe_cache.get cache "counter"));
;;

(* test suite *)
let suite = "Cache test" >:::
  ["test_insert" >:: test_insert;
   "test_mem" >:: test_mem;
   "test_peek" >:: test_peek;
   "test_get" >:: test_get;
   "test_remove" >:: test_remove;
   "test_relative_expiration" >:: test_relative_expiration;
   "test_absolute_expiration" >:: test_absolute_expiration;
   "test_file_expiration" >:: test_file_expiration;
   "test_sliding_expiration" >:: test_sliding_expiration;
   "test_on_remove" >:: test_on_remove;
   "test_expire" >:: test_expire;
   "test_clear" >:: test_clear;
   "test_thread_safe_cache" >:: test_thread_safe_cache]
;;

