open Xnc_stdlib
open Xnc_functional
open OUnit

module U = Unix;;
module Db = Xnc_dbm;;

let sprintf = Printf.sprintf;;
let pf = Printf.printf;;
let norm_path = Neturl.split_path >> Neturl.norm_path >> Neturl.join_path;;
let (//) p1 p2 = Filename.concat p1 p2 |> norm_path;;

let mkdir dirname =
  try U.mkdir dirname 0o755
  with | U.Unix_error (U.EEXIST, "mkdir", dirname) -> ()
;;

let create_test_dir () =
  Random.self_init ();
  let random_string = Random.int 1000000 |> sprintf "%d" |>
                        Digest.string |> Digest.to_hex in
  let test_base = "/tmp/ex_nunc_temp_test" in
    mkdir test_base;
    let random_dir = test_base//random_string in
      mkdir random_dir;
      random_dir
;;  (* END of create_test_dir *)

let test_dir = create_test_dir ();;

let db_open_no_checks () =
  let db_name = test_dir//"db_open_no_checks.db" in
  let db = Db.db_open ~check_schema:false db_name in
    "db_close failed" @? (Db.db_close db);
    let size = (U.stat db_name).U.st_size in
      assert_equal ~msg:"Just created DB MUST have size = 0" size 0
;;  (* END of db_open_no_checks *)

let db_open_sqlite_error1 () =
  let db_name = "/impossible_dir"//"db_open_sqlite_error1.db" in
    assert_raises
      ~msg:"db_open must fail to create this file!"
      (Db.Error("error opening database: unable to open database file"))
      (fun () -> Db.db_open ~check_schema:false db_name |> ignore)
;;  (* END of db_open_sqlite_error *)

let db_open_sqlite_error2 () =
  let db_name = test_dir//"db_open_sqlite_error2.db" in
  let tested_function () =
    let db = Db.db_open ~check_schema:true db_name in
      Db.db_close db |> ignore in

  let oc = open_out db_name in
    output_string oc "this is not a valid Sqlite3 database!";
    close_out oc;
    assert_raises
      ~msg:"db_open must fail to open this invalid file!"
      (Db.Error("NOTADB"))
      tested_function
;;  (* END of db_open_sqlite_error *)

let db_open_sqlite_error3 () =
  let db_name = test_dir//"db_open_sqlite_error3.db" in

  let db = Sqlite3.db_open db_name in
    Sqlite3.exec_not_null_no_headers db ~cb:ignore
      "CREATE TABLE data
       (
         key   INTEGER PRIMARY KEY NOT NULL,
         value TEXT NULL
       );" |> ignore;
    Sqlite3.db_close db |> ignore;

  let tested_function () =
    let db = Db.db_open ~check_schema:true db_name in
      Db.db_close db |> ignore in

    assert_raises
      ~msg:"db_open must fail opening a DB with an unexpected data table"
      (Db.Error("This SQLite database has an incorrect schema"))
      tested_function
;;  (* END of db_open_sqlite_error *)

let db_close_close_invalid_handle () =
  let db_name = test_dir//"db_close_close_invalid_handle.db" in
  let db = Db.db_open ~check_schema:true db_name in
    Db.db_close db |> ignore;
    
    assert_raises
      ~msg:"db_open must fail closing twice a DB handle"
      (Db.Error("Sqlite3.close called with closed database"))
      (fun () -> Db.db_close db |> ignore)
;;  (* END of db_close_close_invalid_handle *)

(* test suite *)
let suite =
  "DBM test" >::: [
    (
      "db_open" >::: [
        "no_checks" >:: db_open_no_checks;
        "sqlite_error1" >:: db_open_sqlite_error1;
        "sqlite_error2" >:: db_open_sqlite_error2;
        "sqlite_error3" >:: db_open_sqlite_error3;
      ]
    );
    (
      "db_close" >::: [
        "close_invalid_handle" >:: db_close_close_invalid_handle;
      ]
    );
  ]
;;

