open OUnit

(* tests *)
let test_record_count _ =
  let logger = Xnc_logger.create () in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Empty log buffer should have 0 records"
      0
      (Xnc_logger.record_count logger);
    Xnc_logger.warn logger "Warning!";
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected record count"
      1
      (Xnc_logger.record_count logger);
;;
 
let test_log _ =
  let logger = Xnc_logger.create () in
    Xnc_logger.log logger Xnc_logger.Info "Info";
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected log output"
      "[info] Info\n"
      (Xnc_logger.to_string logger);
;;
 
let test_warn _ =
  let logger = Xnc_logger.create () in
    Xnc_logger.warn logger "Warning!";
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected log output"
      "[warn] Warning!\n"
      (Xnc_logger.to_string logger);
;;
 
let test_write_to_netchannel _ =
  let buf = Buffer.create 16 in
  let out_ch = new Netchannels.output_buffer buf in
  let logger = Xnc_logger.create () in
    Xnc_logger.debug logger "Debug";
    Xnc_logger.write_to_netchannel logger out_ch;
    out_ch # close_out ();
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected log output"
      "[debug] Debug\n"
      (Buffer.contents buf);
;;

let test_write_to_buffer _ =
  let buf = Buffer.create 16 in
  let logger = Xnc_logger.create () in
    Xnc_logger.debug logger "Debug";
    Xnc_logger.write_to_buffer logger buf;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected log output"
      "[debug] Debug\n"
      (Buffer.contents buf);
;;

let test_to_string_simple_layout _ =
  let logger = Xnc_logger.create () in
    Xnc_logger.debug logger "Debug";
    Xnc_logger.info logger "Info";
    Xnc_logger.warn logger "Warning";
    Xnc_logger.error logger "Error";
    Xnc_logger.fatal logger "Fatal";
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected log output"
      "[debug] Debug\n\
       [info] Info\n\
       [warn] Warning\n\
       [error] Error\n\
       [fatal] Fatal\n"
      (Xnc_logger.to_string logger);
;;

let test_to_string_full_layout _ =
  let logger = Xnc_logger.create ~layout:Xnc_logger.Full () in
    Xnc_logger.info logger "Info";
    OUnit.assert_bool
      "Log output does not match full layout pattern"
      (Str.string_match
         (Str.regexp "^[0-9][0-9][0-9][0-9]-[0-9][0-9]-[0-9][0-9] \
                       [0-9][0-9]:[0-9][0-9]:[0-9][0-9] \
                       [0-9][0-9][0-9]\\.[0-9][0-9][0-9][0-9][0-9][0-9] \
                       \\[info\\] Info\n$")
         (Xnc_logger.to_string logger)
         0
      );
;;

let test_to_string_custom_layout _ =
  let format_record level message timestamp start_time =
    Printf.sprintf "level=%s message=%s\n"
      (Xnc_logger.string_of_level level)
      message in
  let logger = Xnc_logger.create
                 ~layout:(Xnc_logger.Custom format_record) () in
    Xnc_logger.info logger "Info";
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected log output"
      "level=info message=Info\n"
      (Xnc_logger.to_string logger);
;;

let test_clear _ =
  let logger = Xnc_logger.create () in
    Xnc_logger.warn logger "Warning!";
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected record count"
      1
      (Xnc_logger.record_count logger);
    Xnc_logger.clear logger;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Empty log buffer should have 0 records"
      0
      (Xnc_logger.record_count logger);
;;
 
(* Commented out to avoid messing with OUnit output
let test_flush_to_stderr _ =
  let logger = Xnc_logger.create () in
    Xnc_logger.info logger "Info";
    Xnc_logger.error logger "Error";
    Xnc_logger.fatal logger "Fatal";
    Xnc_logger.flush_to_stderr logger;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Log buffer should be empty after a flush"
      0
      (Xnc_logger.record_count logger);
;;
 *)

let test_min_level _ =
  let logger = Xnc_logger.create ~level:Xnc_logger.Error () in
    Xnc_logger.fatal logger "Fatal";
    Xnc_logger.debug logger "Debug";
    Xnc_logger.info logger "Info";
    Xnc_logger.error logger "Error";
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected log output"
      "[fatal] Fatal\n[error] Error\n"
      (Xnc_logger.to_string logger);
;;

let test_is_debug _ =
  let logger = Xnc_logger.create () in
    OUnit.assert_bool
      "Debug log level should be enabled"
      (Xnc_logger.is_debug logger);
  let logger = Xnc_logger.create ~level:Xnc_logger.Info () in
    OUnit.assert_bool
      "Debug log level should be disabled"
      (not (Xnc_logger.is_debug logger));
;;

let test_is_info _ =
  let logger = Xnc_logger.create ~level:Xnc_logger.Info () in
    OUnit.assert_bool
      "Info log level should be enabled"
      (Xnc_logger.is_info logger);
  let logger = Xnc_logger.create ~level:Xnc_logger.Warn () in
    OUnit.assert_bool
      "Info log level should be disabled"
      (not (Xnc_logger.is_info logger));
;;

let test_is_warn _ =
  let logger = Xnc_logger.create ~level:Xnc_logger.Warn () in
    OUnit.assert_bool
      "Warn log level should be enabled"
      (Xnc_logger.is_warn logger);
  let logger = Xnc_logger.create ~level:Xnc_logger.Error () in
    OUnit.assert_bool
      "Warn log level should be disabled"
      (not (Xnc_logger.is_warn logger));
;;

let test_is_error _ =
  let logger = Xnc_logger.create ~level:Xnc_logger.Error () in
    OUnit.assert_bool
      "Error log level should be enabled"
      (Xnc_logger.is_error logger);
  let logger = Xnc_logger.create ~level:Xnc_logger.Fatal () in
    OUnit.assert_bool
      "Error log level should be disabled"
      (not (Xnc_logger.is_error logger));
;;

let test_is_fatal _ =
  let logger = Xnc_logger.create ~level:Xnc_logger.Error () in
    OUnit.assert_bool
      "Fatal log level should be enabled"
      (Xnc_logger.is_fatal logger);
  let logger = Xnc_logger.create ~level:Xnc_logger.Fatal () in
    OUnit.assert_bool
      "Fatal log level should be enabled"
      (Xnc_logger.is_fatal logger);
;;

let test_is_level _ =
  let logger = Xnc_logger.create ~level:Xnc_logger.Debug () in
    OUnit.assert_bool
      "Warn log level should be enabled"
      (Xnc_logger.is_level logger Xnc_logger.Warn);
  let logger = Xnc_logger.create ~level:Xnc_logger.Info () in
    OUnit.assert_bool
      "Warn log level should be enabled"
      (Xnc_logger.is_level logger Xnc_logger.Warn);
  let logger = Xnc_logger.create ~level:Xnc_logger.Fatal () in
    OUnit.assert_bool
      "Warn log level should be disabled"
      (not (Xnc_logger.is_level logger Xnc_logger.Warn));
;;

(* test suite *)
let suite = "Logger test" >:::
  ["test_record_count" >:: test_record_count;
   "test_log" >:: test_log;
   "test_warn" >:: test_warn;
   "test_write_to_netchannel" >:: test_write_to_netchannel;
   "test_write_to_buffer" >:: test_write_to_buffer;
   "test_to_string_simple_layout" >:: test_to_string_simple_layout;
   "test_to_string_full_layout" >:: test_to_string_full_layout;
   "test_to_string_custom_layout" >:: test_to_string_custom_layout;
   "test_clear" >:: test_clear;
(*   "test_flush_to_stderr" >:: test_flush_to_stderr; *)
   "test_min_level" >:: test_min_level;
   "test_is_debug" >:: test_is_debug;
   "test_is_info" >:: test_is_info;
   "test_is_warn" >:: test_is_warn;
   "test_is_error" >:: test_is_error;
   "test_is_fatal" >:: test_is_fatal;
   "test_is_level" >:: test_is_level]
;;

