open OUnit

let counter = ref 0;;

module Thread_timer : Xnc_scheduler.SCHEDULER =
struct
  type t =
      { callback : unit -> unit;
        interval : float;
        mutable thread_id : int
      }

  let stop_table = Hashtbl.create 8
  let table_mutex = Mutex.create ()

  let create c i =
    { callback = c;
      interval = i;
      thread_id = -1
    }

  let start s =
    let thread_stop id =
      let _ = Mutex.lock table_mutex in
      let result = Hashtbl.find stop_table id in
        Mutex.unlock table_mutex;
        result in
    let t = Thread.create
              (fun s ->
                 let id = Thread.id (Thread.self ()) in
                   Mutex.lock table_mutex;
                   Hashtbl.add stop_table id false;
                   Mutex.unlock table_mutex;
                   while not (thread_stop id)
                   do
                     Thread.delay s.interval;
                     if not (thread_stop id) then
                       s.callback ()
                     else
                       ()
                   done)
              s in
    let id = Thread.id t in
      s.thread_id <- id

  let stop s =
    Mutex.lock table_mutex;
    Hashtbl.replace stop_table s.thread_id true;
    Mutex.unlock table_mutex

end;;

(* fixture setup and teardown *)
let set_up _ =
  counter := 0
;;

let tear_down _ =
  ()
;;

(* tests *)
let test_itimer_start _ =
  let scheduler = Xnc_scheduler.Itimer.create
                    (fun () ->
                       incr counter)
                    0.5 in
  let _ = Xnc_scheduler.Itimer.start scheduler in
    Unix.pause ();
    Unix.pause ();
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected counter value after 1 second"
      2
      !counter;
    Xnc_scheduler.Itimer.stop scheduler;
;;

let test_itimer_stop _ =
  let scheduler = Xnc_scheduler.Itimer.create
                    (fun () ->
                       incr counter)
                    0.5 in
  let _ = Xnc_scheduler.Itimer.start scheduler in
    Unix.pause ();
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected counter value after 0.5 seconds"
      1
      !counter;
    Xnc_scheduler.Itimer.stop scheduler;
    OUnit.assert_equal
      ~msg:"Unexpected interval timer status"
      { Unix.it_interval = 0.;
        Unix.it_value = 0.
      }
      (Unix.getitimer Unix.ITIMER_REAL);
;;

let test_thread_timer_start _ =
  let scheduler = Thread_timer.create
                    (fun () ->
                       incr counter)
                    0.075 in
  let _ = Thread_timer.start scheduler in
    Core_test_helper.sleep_float 0.2;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected counter value after 0.2 second"
      2
      !counter;
    Thread_timer.stop scheduler;
;;

let test_thread_timer_stop _ =
  let scheduler = Thread_timer.create
                    (fun () ->
                       incr counter)
                    0.075 in
  let _ = Thread_timer.start scheduler in
    Core_test_helper.sleep_float 0.2;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected counter value after 2 second"
      2
      !counter;
    Thread_timer.stop scheduler;
    Core_test_helper.sleep_float 0.2;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected counter value after stopping timer"
      2
      !counter;
;;

(* test suite *)
let suite = "Scheduler test" >:::
  ["test_itimer_start" >:: (bracket set_up test_itimer_start tear_down);
   "test_itimer_stop" >:: (bracket set_up test_itimer_stop tear_down);
   "test_thread_timer_start"
     >:: (bracket set_up test_thread_timer_start tear_down);
   "test_thread_timer_stop"
     >:: (bracket set_up test_thread_timer_stop tear_down)]
;;

