type expiration =
    No_expiration
  | Relative of float
  | Absolute of Netdate.t
  | Sliding of float
  | File of string
;;

type removal_reason =
    Removed
  | Expired
;;

module type CACHE =
sig
  type ('a, 'b) t

  val create : ?initial_size:int -> unit -> ('a, 'b) t
  val count : ('a, 'b) t -> int
  val insert :
    ?expiration:expiration ->
    ?on_remove:('a -> 'b -> removal_reason -> unit) ->
    ('a, 'b) t -> 'a -> 'b -> unit
  val mem : ('a, 'b) t -> 'a -> bool
  val peek : ('a, 'b) t -> 'a -> 'b option
  val get : ('a, 'b) t -> 'a -> 'b
  val remove : ('a, 'b) t -> 'a -> unit
  val expire : ('a, 'b) t -> unit
  val clear : ('a, 'b) t -> unit
  val lock : ('a, 'b) t -> unit
  val unlock : ('a, 'b) t -> unit
  val with_cache_lock : ('a, 'b) t -> (unit -> 'c) -> 'c
end

module Make(Lockable : Xnc_core.LOCKABLE) =
struct
  type ('a, 'b) item =
      { value : 'b;
        creation_time : float;
        mutable last_access : float;
        expiration : expiration;
        on_remove : 'a -> 'b -> removal_reason -> unit
      }

  type ('a, 'b) t =
      { table : ('a, ('a, 'b) item) Hashtbl.t;
        master_lock : Lockable.t;
        client_lock : Lockable.t
      }

  let create ?(initial_size = 16) () =
    { table = Hashtbl.create initial_size;
      master_lock = Lockable.create ();
      client_lock = Lockable.create ()
    }

  (* private *)
  let with_lock l f =
    try
      let () = Lockable.lock l in
      let result = f () in
        Lockable.unlock l;
        result
    with e ->
      Lockable.unlock l;
      raise e
      
  let count cache =
    with_lock cache.master_lock
      (fun () -> Hashtbl.length cache.table)

  let get_mtime filename =
    let stats = Unix.stat filename in
      stats.Unix.st_mtime

  let insert
        ?(expiration = No_expiration)
        ?(on_remove = (fun _ _ _ -> ()))
        cache key value =
    let current_time =
      match expiration with
          File filename ->
            get_mtime filename
        | _ ->
            Unix.gettimeofday () in
    let item =
      { value = value;
        creation_time = current_time;
        last_access = current_time;
        expiration = expiration;
        on_remove = on_remove
      }
    in
      with_lock cache.master_lock
        (fun () -> Hashtbl.replace cache.table key item)

  (* private *)
  let is_expired item =
    let current_time = Unix.gettimeofday () in
      match item.expiration with
          No_expiration ->
            false
        | Relative interval -> 
            current_time > (item.creation_time +. interval)
        | Absolute date ->
            let timestamp = Netdate.since_epoch date in
              current_time > timestamp
        | Sliding interval -> 
            current_time > (item.last_access +. interval)
        | File filename ->
            try
              let mtime = get_mtime filename in
                mtime > item.creation_time
            with Unix.Unix_error (Unix.ENOENT, _, _) -> true

  (* private *)
  (* this function must be called while holding a lock. *)
  let remove_item cache key item reason =
    Hashtbl.remove cache.table key;
    (* Release the lock to avoid deadlocks, that could happen using
     * non-recursive mutexes. *)
    Lockable.unlock cache.master_lock;
    item.on_remove key item.value reason;
    (* Re-acquire the lock starting a new atomic operation. *)
    Lockable.lock cache.master_lock

  (* private *)
  (* this function must be called while holding a lock. *)
  let remove_if_expired cache key =
    if Hashtbl.mem cache.table key then
      begin
        let item = Hashtbl.find cache.table key in
          if is_expired item then
            remove_item cache key item Expired
          else
            ()
      end
    else
      ()

  let mem cache key =
    with_lock cache.master_lock
      (fun () ->
         remove_if_expired cache key;
         Hashtbl.mem cache.table key)

  let peek cache key =
    with_lock cache.master_lock
      (fun () ->
         remove_if_expired cache key;
         if Hashtbl.mem cache.table key then
           begin
             let item = Hashtbl.find cache.table key in
               item.last_access <- Unix.gettimeofday ();
               Some item.value
           end
         else
           None)

  let get cache key =
    match peek cache key with
        None -> raise Not_found
      | Some value -> value

  let remove cache key =
    with_lock cache.master_lock
      (fun () ->
         if Hashtbl.mem cache.table key then
           begin
             let item = Hashtbl.find cache.table key in
               remove_item cache key item Removed
           end
         else
           ())

  let expire cache =
    with_lock cache.master_lock
      (fun () ->
         let cache_copy = Hashtbl.copy cache.table in
           Hashtbl.iter
             (fun key item ->
                remove_if_expired cache key)
             cache_copy)

  let clear cache =
    with_lock cache.master_lock
      (fun () ->
         Hashtbl.clear cache.table)

  let lock cache =
    Lockable.lock cache.client_lock

  let unlock cache =
    Lockable.unlock cache.client_lock

  let with_cache_lock cache f =
    with_lock cache.client_lock f

end;;

