(** Application cache.
  *
  * This module implements an in-memory cache.
  *)

type expiration =
    No_expiration (** The item never expires. *)
  | Relative of float (** The item expires when the specified
                        * number of seconds is over. *)
  | Absolute of Netdate.t (** The item expires when the specified date
                            * is reached. *)
  | Sliding of float (** The item expires if not accessed for more
                       * than the number of seconds specified. *)
  | File of string (** The item expires when the file is modified. *)
(** Expiration policies. *)

type removal_reason =
    Removed (** The item has been removed by [remove] function. *)
  | Expired (** The item has been removed because it is expired. *)
(** Item removal reason. *)

module type CACHE =
sig
  type ('a, 'b) t
  (** The type of a cache. *)

  val create : ?initial_size:int -> unit -> ('a, 'b) t
  (** [create ()] creates a new cache.
    *
    * @param initial_size Initial number of items allocated. Defaults to [16].
    *)

  val count : ('a, 'b) t -> int
  (** [count c] returns the number of items stored in cache [c]. *)

  val insert :
    ?expiration:expiration ->
    ?on_remove:('a -> 'b -> removal_reason -> unit) ->
    ('a, 'b) t -> 'a -> 'b -> unit
  (** [insert c k v] inserts the item [v] into cache [c] with key [k],
    * overwriting previously inserted items with the same key [k].
    *
    * @param expiration Expiration term. Defaults to [No_expiration].
    * @param on_remove [on_remove k v r] is invoked when the item [v] associated
    * with key [k] is removed from the cache because of reason [r]. The default
    * callback does nothing. *)

  val mem : ('a, 'b) t -> 'a -> bool
  (** [mem c k] tests the presence of an item with key [k] in cache [c]. *)

  val peek : ('a, 'b) t -> 'a -> 'b option
  (** [peek c k] returns [Some] of the value assiciated with key [k], or [None]
    * if such item is not found. *)

  val get : ('a, 'b) t -> 'a -> 'b
  (** [get c k] returns the item associated with key [k] from cache [c]. Raises
    * [Not_found] if there is no item with key [k]. *)

  val remove : ('a, 'b) t -> 'a -> unit
  (** [remove c k] removes the item associated with key [k] from cache [c]. *)

  val expire : ('a, 'b) t -> unit
  (** [expire c] removes any expired item from cache [c]. *)

  val clear : ('a, 'b) t -> unit
  (** [clear c] removes the contents of cache [c]. *)

  val lock : ('a, 'b) t -> unit
  (** [lock c] acquires a lock on cache [c] to allow atomic operations. *)

  val unlock : ('a, 'b) t -> unit
  (** [unlock c] releases a lock previously acquired on cache [c]. *)

  val with_cache_lock : ('a, 'b) t -> (unit -> 'c) -> 'c
  (** [with_cache_lock c f] computes [f ()] protected by a lock on [c]. *)
end
(** Signature of cache module. *)

module Make :
  functor (Lockable : Xnc_core.LOCKABLE) -> CACHE
(** Functor building an implementation of the cache given a module to handle
  * locking. *)

