open Xnc_stdlib
open Xnc_functional

module Sq3 = Sqlite3
module Pf = Printf

let p = Pf.printf

exception Error of string

exception Key_already_exists of string

type db = Sq3.db

type key = string

type value = string
      
let raise_wrap e = raise (Error (Sq3.Rc.to_string e));;

let raise_error s = raise (Error s);;

let single_quote = Str.regexp_string "'";;

let escape_sql s = Str.global_replace single_quote "''" s;;

let get_option v =
  match v with
    | Some s -> s
    | None -> raise Not_found
;;

let has_table db =
  let table_ok = ref false in
  let callback row = table_ok := true in
  let res = Sq3.exec_not_null_no_headers db ~cb:callback
      "SELECT name FROM sqlite_master
       WHERE type='table' AND name = 'data';" in
    match res with
      | Sq3.Rc.OK -> !table_ok
      | e -> Sq3.db_close db |> ignore; raise_wrap e
;;

let create_table db =
  let res = Sq3.exec_not_null_no_headers db ~cb:ignore
    "CREATE TABLE data
     (
       key   TEXT PRIMARY KEY NOT NULL,
       value TEXT NOT NULL
     );" in
    match res with
      | Sq3.Rc.OK -> ()
      | e -> Sq3.db_close db |> ignore; raise_wrap e
;;

let table_is_ok db =
  let key_column_ok = ref false in
  let value_column_ok = ref false in

  let cb row headers =
    let name_idx = ref (-1) in
    let type_idx = ref (-1) in
    let notnull_idx = ref (-1) in
    let pk_idx = ref (-1) in
      for i = 0 to ((Array.length headers) - 1)
      do
        if headers.(i) = "name" then name_idx := i;
        if headers.(i) = "type" then type_idx := i;
        if headers.(i) = "notnull" then notnull_idx := i;
        if headers.(i) = "pk" then pk_idx := i;
      done;

      if !name_idx = (-1) || !type_idx = (-1) || !notnull_idx = (-1) ||
         !pk_idx = (-1)
      then (
          Sq3.db_close db |> ignore;
          raise_error "Cannot access SQLite3 table metadata"
        )
      else (
          match row.(!name_idx) with
            | Some "key" ->
                begin
                  if get_option (row.(!type_idx)) = "TEXT" &&
                     get_option (row.(!notnull_idx)) = "99" &&
                     get_option (row.(!pk_idx)) = "1"
                  then key_column_ok := true;
                end
            | Some "value" ->
                begin
                  if get_option(row.(!type_idx)) = "TEXT" &&
                     get_option(row.(!notnull_idx)) = "99" &&
                     get_option(row.(!pk_idx)) = "0"
                  then value_column_ok := true;
                end
            | _ -> ()
        ) in

  let res = Sq3.exec db ~cb:cb "PRAGMA TABLE_INFO(data);" in
    match res with
      | Sq3.Rc.OK -> !key_column_ok && !value_column_ok
      | err -> Sq3.db_close db |> ignore; raise_wrap err
;;

let db_open ?(check_schema = true) filename : db =
  let db = try Sq3.db_open filename
  with | Sq3.Error msg -> raise_error msg in
  
    if check_schema
    then
      if not (has_table db)
      then (create_table db; db)
      else
        if (table_is_ok db)
        then db
        else raise_error "This SQLite database has an incorrect schema"
    else
      db
;;

let db_close db =
  try Sq3.db_close db
  with | Sq3.Error msg -> raise_error msg
;;

let find (db : db) (key : key) : value option =
  let value = ref None in
  let callback row = value := Some (row.(0)) in
  let query = "SELECT value FROM data WHERE key = '" ^ (escape_sql key) ^ "';" in
  let res = try Sq3.exec_not_null_no_headers db ~cb:callback query
    with | Sq3.Error msg -> raise_error msg in

    match res with
      | Sq3.Rc.OK -> !value
      | e -> raise_wrap e
;;

let add (db : db) (key : key) (value : value) : unit =
  let query =
    "INSERT INTO data (key, value) VALUES ('" ^ (escape_sql key) ^
    "', '" ^ (escape_sql value) ^ "');" in

  let res = try Sq3.exec_not_null_no_headers db ~cb:ignore query
    with | Sq3.Error msg -> raise_error msg in

    match res with
      | Sq3.Rc.OK -> ()
      | Sq3.Rc.CONSTRAINT -> raise (Key_already_exists key)
      | e -> raise_wrap e
;;

let remove (db : db) (key : key) =
  let query = "DELETE FROM data WHERE key = '" ^ (escape_sql key) ^ "'" in

  let res = try Sq3.exec_not_null_no_headers db ~cb:ignore query
    with | Sq3.Error msg -> raise_error msg in

    match res with
      | Sq3.Rc.OK -> ()
      | e -> raise_wrap e
;;

let replace (db : db) (key : key) (value : value) : unit =
  try add db key value
  with
    | Key_already_exists _ ->
      remove db key;
      add db key value
;;

let iter (f : key -> value -> unit) (db : db) =
  let apply_f row = f (row.(0)) (row.(1)) in
  let res =
    try
      Sq3.exec_not_null_no_headers db ~cb:apply_f "SELECT key, value FROM data"
    with
      | Sq3.Error msg -> raise_error msg in

    match res with
      | Sq3.Rc.OK -> ()
      | e -> raise_wrap e
;;

let fold (f : key -> value -> 'a -> 'a) (db : db) (acc : 'a) : 'a =
  let current_acc = ref acc in
  let apply_f row = current_acc := f (row.(0)) (row.(1)) !current_acc in
  
  let res =
    try
      Sq3.exec_not_null_no_headers db ~cb:apply_f "SELECT key, value FROM data"
    with
      | Sq3.Error msg -> raise_error msg in

    match res with
      | Sq3.Rc.OK -> !current_acc
      | e -> raise_wrap e
;;

let keys (db : db) : key list =
  fold (fun k v a -> k::a) db []
;;

let values (db : db) : value list =
  fold (fun k v a -> v::a) db []
;;

let dump (db : db) : (key * value) list =
  fold (fun k v a -> (k, v)::a) db []
;;

let reorganize db =
  let res = try Sq3.exec_not_null_no_headers db ~cb:ignore "VACUUM"
    with | Sq3.Error msg -> raise_error msg in

    match res with
      | Sq3.Rc.OK -> ()
      | e -> raise_wrap e
;;

