(** Simple SQLite3 interface module suitable for storing client sessions
  * on disk.
  * 
  * This module provides an interface to store and retreive key/value pairs
  * of strings.
  *
  * @author Paolo Donadeo
  * @see <http://www.ex-nunc.org> http://www.ex-nunc.org
  *)

exception Error of string
(** This exception is raised whenever an error occurs. *)

exception Key_already_exists of string
(** This exception is specifically raised when you try to store an already
  * present key. *)

type db
(** The type of the database, used as handle by other functions.
  *
  * {b From SQLite3 documentation}:
  *
  *  - DO NOT USE THIS HANDLE WITHIN THREADS OTHER THAN THE ONE THAT
  *    CREATED IT!!!
  *
  *  - database handles are closed (see Sqlite3.db_close) automatically
  *    when they are reclaimed by the GC unless they have already been closed
  *    earlier by the user. It is good practice to manually close database
  *    handles to free resources as quickly as possible.
  *)

type key = string
(** The type of the map keys. *)

type value = string
(** The type of the map values. *)

val db_open : ?check_schema:bool -> string -> db
(** [db_open filename] opens the database file [filename], and returns a
  * database handle. The optional parameter [check_schema] set to [true]
  * executes some checks on the schema and in case the data table is not
  * present, it will create it. It is useful the first time you open a session
  * db, but if you are sure that the given filename is a valid session db, leave
  * this parameter as false.
  * 
  * Parameter [check_schema] default is [true]
  *)

val db_close : db -> bool
(** [db_close db] closes database [db] and invalidates the handle. *)

val find : db -> key -> value option
(** [find db key] returns the data associated with the given key in the database
  * opened for the descriptor [db]. Returns [None] if the key has no associated
  * data. *)

val add : db -> key -> value -> unit
(** [add db key data] inserts the pair [(key, data)] in the database [db].
  * If the database already contains data associated with [key],
  * raise [Xnc_dbm.Key_already_exists]. *)

val remove : db -> key -> unit
(** [remove db key] removes the data associated with [key] in [db]. If key has
  * no associated data, no exception is raised. *)

val replace : db -> key -> value -> unit
(** [replace db key data] inserts the pair [(key, data)] in the database [db].
  * If the database already contains data associated with [key], that data is
  * discarded and silently replaced by the new [data]. *)

val iter : (key -> value -> unit) -> db-> unit
(** [iter f db] applies [f] to each [(key, data)] pair in the database db. [f]
  * receives [key] as first argument and [data] as second argument. *)

val fold : (key -> value -> 'a -> 'a) -> db -> 'a -> 'a
(** [fold f db acc] computes [(f kN dN ... (f k1 d1 acc)...)], where [k1 ... kN]
  * are the keys of all bindings in db (no particular order), and [d1 ... dN]
  * are the associated data. *)

val keys : db -> key list
(** [keys db] returns the list of all keys in [db] *)

val values : db -> value list
(** [values db] returns the list of all values in [db] *)

val dump : db -> (key * value) list
(** [dump db] returns a list of all [(key, value)] pairs in [db] *)

val reorganize : db -> unit
(** Reoganize the underling SQLite3 database using the VACUUM special query. *)

