open Xnc_stdlib

let module_name = "Xnc_functional"

let (|>) x f = f x

let (>>) f g x = g(f x)

let rec iter n f x = if n = 0 then x else iter (n-1) f (f x);;

let rec loop p f x = if (p x) then x else loop p f (f x)

let operate_on_n f l index =
  let rec operate_aux resid count acc_list =
    match resid with
      | hd :: tl ->
          if count = index
          then f acc_list hd tl
          else operate_aux tl (count+1) (hd::acc_list)
      | [] -> invalid_arg (module_name ^ ".operate_on_n") in
    operate_aux l 0 []

let memoize f =
  let m = Hashtbl.create 1 in
    fun x ->
      try Hashtbl.find m x 
      with Not_found ->
        let f_x = f x in
          Hashtbl.add m x f_x;
          f_x

let memoize_rec f =
  let m = Hashtbl.create 1 in
  let rec f' x =
    try Hashtbl.find m x
    with Not_found ->
      let f_x = f f' x in
        Hashtbl.add m x f_x;
        f_x in
    f';;

let fold_pair f acc lst =
  let rec inner_aux f acc lst =
    match lst with
      | p0::((p1::tl) as l) -> inner_aux f (f p0 p1 acc) l
      | [p0] -> acc
      | [] -> failwith "Impossible pattern!" in
    inner_aux f acc lst

let fold_pair_wrap f acc lst =
  let first = List.hd lst in
  let rec inner_aux f acc lst =
    match lst with
      | p0 :: ((p1 :: tl) as l) -> inner_aux f (f p0 p1 acc) l
      | [p0] -> inner_aux f (f p0 first acc) []
      | [] -> acc in
    inner_aux f acc lst

(*
 * Utilities from common experience
 *)

(* Fold over an input channel line by line *)
let rec in_channel_fold f acc ic =
  let test, line =
    try true, input_line ic
    with | End_of_file -> false, "" in
    if test
    then in_channel_fold f (f acc line) ic
    else acc

(* Iter over an input channel line by line with line number *)
let in_channel_iteri ?(i=1) f ic =                                        
  in_channel_fold (fun acc l -> let () = f acc l in acc + 1) i ic |> ignore

(* Iter over an input channel line by line *)
let in_channel_iter f ic = in_channel_iteri (fun _ l -> f l) ic

let print_timings oc start_space start_wtime start_ptime =
  let end_ptime = Unix.times () in      (* process time *)
  let self_u = end_ptime.Unix.tms_utime -. start_ptime.Unix.tms_utime in
  let self_s = end_ptime.Unix.tms_stime -. start_ptime.Unix.tms_stime in
  let self = self_u +. self_s in
  let child_u = end_ptime.Unix.tms_cutime -. start_ptime.Unix.tms_cutime in
  let child_s = end_ptime.Unix.tms_cstime -. start_ptime.Unix.tms_cstime in
  let child = child_u +. child_s in
  let run_time = self +. child in
  let real_time = Unix.time () -. start_wtime in (* wall time *)
  let real_time_sec = int_of_float real_time in
  let real_time_min = real_time_sec / 60 in
  let real_time_sec = real_time_sec mod 60 in
  let real_time_hrs = real_time_min / 60 in
  let real_time_min = real_time_min mod 60 in
  (* we would need a compaction if we were reporting memory usage *)
  (* Gc.compact (); *)
  let end_space = Gc.quick_stat () in
    Printf.fprintf oc "Run time: %f sec\n%!" run_time;
    Printf.fprintf oc "Self:     %f sec\n%!" self;
    Printf.fprintf oc "     sys: %f sec\n%!" self_u;
    Printf.fprintf oc "    user: %f sec\n%!" self_s;
    Printf.fprintf oc "Children: %f sec\n%!" child;
    Printf.fprintf oc "     sys: %f sec\n%!" child_u;
    Printf.fprintf oc "    user: %f sec\n%!" child_s;
    Printf.fprintf oc "GC:     minor: %d\n%!"
      (end_space.Gc.minor_collections - start_space.Gc.minor_collections);
    Printf.fprintf oc "        major: %d\n%!"
      (end_space.Gc.major_collections - start_space.Gc.major_collections);
    Printf.fprintf oc "  compactions: %d\n%!"
      (end_space.Gc.compactions - start_space.Gc.compactions);
    Printf.fprintf oc "Allocated:  %.1f words\n%!"
      (end_space.Gc.minor_words +. end_space.Gc.major_words
       -. start_space.Gc.minor_words -. start_space.Gc.major_words
       -. end_space.Gc.promoted_words +. start_space.Gc.promoted_words);
    Printf.fprintf oc "Wall clock:  %.0f sec (%02d:%02d:%02d)\n%!"
      real_time real_time_hrs real_time_min real_time_sec;
    if real_time > 0. then
      Printf.fprintf oc "Load:  %.2f%%\n%!" (run_time *. 100.0 /. real_time)
;;


(* Run function, then report time and space usage *)
let run ?(channel = Pervasives.stdout) func args =
  Gc.compact (); (* so that prior execution does not skew the results *)
  let start_space = Gc.quick_stat () in
  let start_wtime = Unix.time () in     (* wall time *)
  let start_ptime = Unix.times () in    (* process time *)
  let ret =
    try func args
    with e -> print_timings channel start_space start_wtime start_ptime;
              raise e in
    print_timings channel start_space start_wtime start_ptime;
    ret
;;


let using resource finalize process =
  try
    let result = process resource in
      finalize resource;
      result
  with e ->
    finalize resource;
    raise e
;;


let with_open_in filename = using (open_in filename) close_in;;
let with_open_in_bin filename = using (open_in_bin filename) close_in;;
let with_open_out filename = using (open_out filename) close_out;;
let with_open_out_bin filename = using (open_out_bin filename) close_out;;


let fst3 (a, b, c) = a;;
let snd3 (a, b, c) = b;;
let trd3 (a, b, c) = c;;

let fst4 (a, b, c, d) = a;;
let snd4 (a, b, c, d) = b;;
let trd4 (a, b, c, d) = c;;
let fth4 (a, b, c, d) = d;;


module Ordered_strings =
struct
  type t = string
  let compare = String.compare
end


module String_map = Map.Make(Ordered_strings)

