(** Higher-order functions and common functional operators
  *
  * @author Paolo Donadeo
  * @see <http://www.ex-nunc.org> http://www.ex-nunc.org
  *)


val ( |> ) : 'a -> ('a -> 'b) -> 'b
(** Pipeline operator *)

val ( >> ) : ('a -> 'b) -> ('b -> 'c) -> 'a -> 'c
(** Function composition operator *)

val iter : int -> ('a -> 'a) -> 'a -> 'a
(** Computes the nth iterate of the function [f]. [iter n f x0] computes
  * [f(...f(f(f(x0))))] [n] times *)

val loop : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a
(** [loop pred f x0] computes [f(...f(f(f(x0))))] until the predicate
  * is false *)

val operate_on_n :
  ('a list -> 'a -> 'a list -> 'b) -> 'a list -> int -> 'b
(** [operate_on_n f l n] splits the list [l] in three parts: the (reversed)
  * list of elements bebore the element [n], the element [n] itself, and the
  * list of elements after the element [n]. Then executes f passing the three
  * parts.
  *
  * [operate_on_n (fun hd el tl -> print_int el) [1; 2; 3; 4; 5; 6; 7] 3], for
  * example, executes [f [3; 2; 1] 4 [5; 6; 7]] and so prints the numeber 4.
  *)

val memoize : ('a -> 'b) -> 'a -> 'b
(** [memoize f] returns the memoized version of a function.
  *
  * It uses an hash table from the Hashtbl standard module to memoize binding
  * from arguments and result, so be careful and use [memoize] only with
  * functions handling types compatible with Hashtbl. *)

val memoize_rec : (('a -> 'b) -> 'a -> 'b) -> 'a -> 'b
(** [memoize_rec f] returns the memoized version of a recursive function.
  * 
  * Same caveats of [memoize].
  *)

val fold_pair : ('a -> 'a -> 'b -> 'b) -> 'b -> 'a list -> 'b
(** [fold_pair f acc l] iteratively calls [f] passing it {b two} elements
  * of the list [l] each time, and the accumulator. If [l] is [ [l1; l2; ...;
  * l10] ], then [f] is called with the following arguments:
  * [(f l1 l2 acc) |> (f l2 l3) |> ... |> (f l9 l10)]
  *)

val fold_pair_wrap : ('a -> 'a -> 'b -> 'b) -> 'b -> 'a list -> 'b
(** [fold_pair_wrap f acc l] iteratively calls [f] passing it {b two} elements
  * of the list [l] each time, and the accumulator. If [l] is [ [l1; l2; ...;
  * l10] ], then [f] is called with the following arguments:
  * [(f l1 l2 acc) |> (f l2 l3) |> ... |> (f l9 l10) |> f l10 l1]
  *
  * While [fold_pair] iters over the list, [fold_pair_wrap] calls [f] another
  * time passing the last and first element of [l].
  *)

val in_channel_fold : ('a -> string -> 'a) -> 'a -> in_channel -> 'a
(** Fold over an input channel line by line.
  *
  * [in_channel_fold f acc ic] folds the function [f] over the open input
  * channel [ic], passing to [f] each line and the accumulator.
  *)

val in_channel_iter : (string -> unit) -> in_channel -> unit
(** Iter over an input channel line by line.
  *
  * [in_channel_iter f ic] iters the function [f] over the open input
  * channel [ic], passing to [f] each line.
  *)

val in_channel_iteri : ?i:int -> (int -> string -> unit) -> in_channel -> unit
(** Iter over an input channel line by line with line counter.
  *
  * [in_channel_iteri f ic] iters the function [f] over the open input
  * channel [ic], passing to [f] for each line the line number and line contents.
  *)

val run : ?channel:out_channel -> ('a -> 'b) -> 'a -> 'b
(** Prints time and memory statistics.
  *
  * [run f] returns a wrapper of the function [f] that, after the execution,
  * prints to the passed [out_channel] (default is [Pervasives.stdout]) time and
  * memory statistics about the execution of [f].
  *)

val with_open_in : string -> (in_channel -> 'a) -> 'a
(** [with_open_in fname f] opens the named file for reading and executes [f] and
  * returns the result of the computation.
  * If [f] raises an exception [e], then [e] is raised again.
  * 
  * The file will be {i automatically} closed {b in any case}.
  *)

val with_open_in_bin : string -> (in_channel -> 'a) -> 'a
(** [with_open_in_bin fname f] opens the named file for reading in bynary mode
  * and executes [f] and returns the result of the computation.
  * If [f] raises an exception [e], then [e] is raised again.
  * 
  * The file will be {i automatically} closed {b in any case}.
  *)

val with_open_out : string -> (out_channel -> 'a) -> 'a
(** [with_open_out fname f] opens the named file for writing and executes [f]
  * and returns the result of the computation.
  * If [f] raises an exception [e], then [e] is raised again.
  * 
  * The file will be {i automatically} closed {b in any case}.
  *)

val with_open_out_bin : string -> (out_channel -> 'a) -> 'a
(** [with_open_out_bin fname f] opens the named file for writing in bynary mode
  * and executes [f] and returns the result of the computation.
  * If [f] raises an exception [e], then [e] is raised again.
  * 
  * The file will be {i automatically} closed {b in any case}.
  *)

val fst3 : 'a * 'b * 'c -> 'a
(** Return the first component of a triple. *)

val snd3 : 'a * 'b * 'c -> 'b
(** Return the second component of a triple. *)

val trd3 : 'a * 'b * 'c -> 'c
(** Return the third component of a triple. *)

val fst4 : 'a * 'b * 'c * 'd -> 'a
(** Return the first component of a quadruple. *)

val snd4 : 'a * 'b * 'c * 'd -> 'b
(** Return the second component of a quadruple. *)

val trd4 : 'a * 'b * 'c * 'd -> 'c
(** Return the third component of a quadruple. *)

val fth4 : 'a * 'b * 'c * 'd -> 'd
(** Return the fourth component of a quadruple. *)


(** Total ordered strings module
  *
  * Module containing a type (string) and a total ordering function
  * used by the {!module: String_map} module
  *)
module Ordered_strings :
  sig
    type t = string
      (** The type of the map keys. *)
    val compare : String.t -> String.t -> int
      (** A total ordering function over strings. *)
  end


(** Association tables over strings.
  *
  * This module implements a dictionary over strings. The signature of this
  * module is exactly the signature of {!module: Xnc_stdlib.Map.MAP}, with
  * the type [key] equal to [string].
  *)
module String_map :
  sig
    type key = Ordered_strings.t
    type 'a t = 'a Xnc_stdlib.Map.Make(Ordered_strings).t
    val empty : 'a t
    val is_empty : 'a t -> bool
    val add : key -> 'a -> 'a t -> 'a t
    val find : key -> 'a t -> 'a
    val remove : key -> 'a t -> 'a t
    val mem : key -> 'a t -> bool
    val iter : (key -> 'a -> unit) -> 'a t -> unit
    val map : ('a -> 'b) -> 'a t -> 'b t
    val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
    val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val compare : ('a -> 'a -> int) -> 'a t -> 'a t -> int
    val equal : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool
    val find_opt : key -> 'a t -> 'a option
    val keys : 'a t -> key list
    val values : 'a t -> 'a list
    val filter_key : (key -> bool) -> 'a t -> 'a t
    val filter_value : ('a -> bool) -> 'a t -> 'a t
  end

