type log_level =
    Debug
  | Info
  | Warn
  | Error
  | Fatal
;;

type log_layout =
    Simple
  | Full
  | Custom of (log_level -> string -> float -> float -> string)
;;

type log_record =
    { level : log_level;
      message : string;
      timestamp : float
    }
;;

type t =
    { log_buffer : log_record Queue.t;
      min_level : log_level;
      layout : log_layout;
      start_time : float;
      buffered : bool
    }
;;

let create
      ?(level = Debug)
      ?(layout = Simple)
      ?(buffered = true) () =
  { log_buffer = Queue.create ();
    min_level = level;
    layout = layout;
    start_time = Unix.gettimeofday ();
    buffered = buffered
  }
;;

let record_count logger =
  Queue.length logger.log_buffer
;;

let string_of_level level =
  match level with
      Debug -> "debug"
    | Info -> "info"
    | Warn -> "warn"
    | Error -> "error"
    | Fatal -> "fatal"
;;

let string_of_time time =
  let t = Unix.localtime time in
    Printf.sprintf "%d-%02d-%02d %02d:%02d:%02d"
      (t.Unix.tm_year + 1900)
      t.Unix.tm_mon
      t.Unix.tm_mday
      t.Unix.tm_hour
      t.Unix.tm_min
      t.Unix.tm_sec
;;

let record_to_string record layout start_time =
  match layout with
      Simple ->
        Printf.sprintf "[%s] %s\n"
          (string_of_level record.level)
          record.message
    | Full ->
        Printf.sprintf "%s %010.6f [%s] %s\n"
          (string_of_time record.timestamp)
          (record.timestamp -. start_time)
          (string_of_level record.level)
          record.message
    | Custom format_record ->
        format_record
          record.level
          record.message
          record.timestamp
          start_time
;;

let write_to_netchannel logger out_ch =
  Queue.iter
    (fun log_record ->
       out_ch # output_string
         (record_to_string log_record logger.layout logger.start_time)
    )
    logger.log_buffer
;;

let write_to_buffer logger buf =
  let out_ch = new Netchannels.output_buffer buf in
    write_to_netchannel logger out_ch
;;

let to_string logger =
  let buf = Buffer.create 128 in
    write_to_buffer logger buf;
    Buffer.contents buf
;;

let clear logger =
  Queue.clear logger.log_buffer
;;

let flush_to_stderr logger =
  prerr_string (to_string logger);
  flush stderr;
  clear logger
;;

let level_to_int level =
  match level with
      Debug -> 0
    | Info -> 1
    | Warn -> 2
    | Error -> 3
    | Fatal -> 4
;;

let (>=) l1 l2 =
  (level_to_int l1) >= (level_to_int l2)
;;

let log logger level message =
  if level >= logger.min_level then
    begin
      Queue.add
        { level = level;
                  message = message;
                  timestamp = Unix.gettimeofday ()
        }
        logger.log_buffer; 
      if (not logger.buffered) then
        flush_to_stderr logger
    end
  else
    ()
;;

let debug logger message =
  log logger Debug message
;;

let info logger message =
  log logger Info message
;;

let warn logger message =
  log logger Warn message
;;

let error logger message =
  log logger Error message
;;

let fatal logger message =
  log logger Fatal message
;;

let is_level logger level =
  level >= logger.min_level
;;

let is_debug logger =
  Debug >= logger.min_level
;;

let is_info logger =
  Info >= logger.min_level
;;

let is_warn logger =
  Warn >= logger.min_level
;;

let is_error logger =
  Error >= logger.min_level
;;

let is_fatal logger =
  Fatal >= logger.min_level
;;

