(** Logging facility.
  *
  * This module implements a logger that collects and writes tracing messages to
  * [stderr].
  *)

type log_level =
    Debug
  | Info
  | Warn
  | Error
  | Fatal
(** Level of logging. *)

type log_layout =
    Simple
  | Full
  | Custom of (log_level -> string -> float -> float -> string)
(** Layout mode. *)

type t
(** The type of loggers. *)

val create :
  ?level:log_level -> ?layout:log_layout -> ?buffered:bool -> unit -> t
(** [create ()] creates a logger.
  *
  * @param level The minimum log level to trace. Defaults to [Debug].
  * @param layout The layout mode. Defaults to [Simple].
  * @param buffered When buffered the log records are collected until flushed,
  * otherwise every log record is immediately written to [stderr] and discarded.
  * Defaults to [true].
  *)

val record_count : t -> int
(** [record_count l] returns the number of currently buffered records in logger
  * [l]. *)

val string_of_level : log_level -> string
(** [string_of_level l] returns the string description of log level [l]. *)

val string_of_time : float -> string
(** [string_of_time t] returns the string description of time stamp [t]. *)

val write_to_netchannel : t -> #Netchannels.out_obj_channel -> unit
(** [write_to_netchannel l o] writes the buffered records of [l] to [o]. The
  * buffer is not flushed. *)

val write_to_buffer : t -> Buffer.t -> unit
(** [write_to_buffer l b] writes the buffered records of [l] to [b]. The buffer
  * is not flushed. *)

val to_string : t -> string
(** [to_string l b] returns the string representation of the buffered records of
  * [l]. The buffer is not flushed. *)

val clear : t -> unit
(** [clear l] discards all buffered records of [l]. *)

val flush_to_stderr : t -> unit
(** [flush_to_stderr l] writes the buffered records of [l] to [stderr] and
  * flushed the buffer. *)

val log : t -> log_level -> string -> unit
(** [log l ll m] logs message [m] with log level [ll] to logger [l]. *)

val debug : t -> string -> unit
(** [debug l m] logs message [m] with log level [Debug] to logger [l]. *)

val info : t -> string -> unit
(** [info l m] logs message [m] with log level [Info] to logger [l]. *)

val warn : t -> string -> unit
(** [warn l m] logs message [m] with log level [Warn] to logger [l]. *)

val error : t -> string -> unit
(** [error l m] logs message [m] with log level [Error] to logger [l]. *)

val fatal : t -> string -> unit
(** [fatal l m] logs message [m] with log level [Fatal] to logger [l]. *)

val is_level : t -> log_level -> bool
(** [is_level l ll] tests if log level [ll] messages are traced by logger [l].
  *)

val is_debug : t -> bool
(** [is_debug l] tests if [Debug] messages are traced by logger [l]. *)

val is_info : t -> bool
(** [is_info l] tests if [Info] messages are traced by logger [l]. *)

val is_warn : t -> bool
(** [is_warn l] tests if [Warn] messages are traced by logger [l]. *)

val is_error : t -> bool
(** [is_error l] tests if [Error] messages are traced by logger [l]. *)

val is_fatal : t -> bool
(** [is_fatal l] tests if [Fatal] messages are traced by logger [l]. *)

