(** Singleton.
  *
  * This module implements the singleton design pattern. It is used to cache a
  * global variable lazily created at first access.
  *)

module Make :
  functor (Initializable : Xnc_core.INITIALIZABLE) ->
  functor (Lockable : Xnc_core.LOCKABLE) ->
sig
  type t = Initializable.t
  (** The type of the value to store. *)

  val get_instance : unit -> t
  (** [get_instance ()] returns the stored instance. The instance is created at
    * first invocation. *)

  val update_instance : t -> unit
  (** [update_instance v] updates the stored instance with value [v]. *)

  val clear_instance : unit -> unit
  (** [clear_instance ()] resets the stored instance to an undefined value. The
    * next time [get_instance] gets called, the instance is recreated. *)

end
(** Functor building an implementation of the singleton given an initializable
  * type. *)

