(** Pervasives and Objective Caml Standard Library extension module
  *
  * This module extends the Pervasives module and other modules in the Objective
  * Caml Standard Library. Opening [Xnc_stdlib] binds standard names with the
  * extended version of the module.
  *
  * Only functions non present in the standard library modules are
  * documented here.
  *
  * @author Paolo Donadeo
  * @see <http://www.ex-nunc.org> http://www.ex-nunc.org
  *)


val char_of_string : string -> char
(** [char_of_string s] returns the first and only character of [s].
  * Raises [Invalid_argument "Xnc_stdlib.char_of_string"] if
  * the length of [s] is not equal to 1. *) 

module Map :
  sig
    module type MAP =
      sig
        type key
        type +'a t
        val empty : 'a t
        val is_empty : 'a t -> bool
        val add : key -> 'a -> 'a t -> 'a t
        val find : key -> 'a t -> 'a
        val remove : key -> 'a t -> 'a t
        val mem : key -> 'a t -> bool
        val iter : (key -> 'a -> unit) -> 'a t -> unit
        val map : ('a -> 'b) -> 'a t -> 'b t
        val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
        val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
        val compare : ('a -> 'a -> int) -> 'a t -> 'a t -> int
        val equal : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool

        val find_opt : key -> 'a t -> 'a option
        (** [find_opt x m] returns the current binding of [x] in [m]
          * as [Some x], or returns [None] if no such binding exists. *)
        
        val keys : 'a t -> key list
        (** [keys m] returns the list of all keys currently in [m] *)
        
        val values : 'a t -> 'a list
        (** [keys m] returns the list of all values currently in [m] *)
        
        val filter_key : (key -> bool) -> 'a t -> 'a t
        (** [filter p m] returns a map with bindings of the map [m], and
            in which the keys satisfy the predicate [p]. *)
        
        val filter_value : ('a -> bool) -> 'a t -> 'a t
        (** [filter p m] returns a map with bindings of the map [m], and
            in which the values satisfy the predicate [p]. *)

        val map_of_assoc : (key * 'a) list -> 'a t
        (** [map_of_assoc l] returns a map containing bindings from the
            association list [l] *)
      end
    (** Output signature of the functor {!module: Xnc_stdlib.Map.Make}. *)

    module Make :
      functor (Ord : Map.OrderedType) ->
        sig
          type key = Ord.t
          type +'a t
          val empty : 'a t
          val is_empty : 'a t -> bool
          val add : key -> 'a -> 'a t -> 'a t
          val find : key -> 'a t -> 'a
          val remove : key -> 'a t -> 'a t
          val mem : key -> 'a t -> bool
          val iter : (key -> 'a -> unit) -> 'a t -> unit
          val map : ('a -> 'b) -> 'a t -> 'b t
          val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t
          val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
          val compare : ('a -> 'a -> int) -> 'a t -> 'a t -> int
          val equal : ('a -> 'a -> bool) -> 'a t -> 'a t -> bool

          val find_opt : key -> 'a t -> 'a option
          (** [find_opt x m] returns the current binding of [x] in [m]
            * as [Some x], or returns [None] if no such binding exists. *)

          val keys : 'a t -> key list
          (** [keys m] returns the list of all keys currently in [m] *)

          val values : 'a t -> 'a list
          (** [keys m] returns the list of all values currently in [m] *)

          val filter_key : (key -> bool) -> 'a t -> 'a t
          (** [filter p m] returns a map with bindings of the map [m], and
              in which the keys satisfy the predicate [p]. *)

          val filter_value : ('a -> bool) -> 'a t -> 'a t
          (** [filter p m] returns a map with bindings of the map [m], and
              in which the values satisfy the predicate [p]. *)

          val map_of_assoc : (key * 'a) list -> 'a t
          (** [map_of_assoc l] returns a map containing bindings from the
              association list [l] *)
        end
    (** Functor building an implementation of the map structure
        given a totally ordered type. *)
  end
(** Association tables over ordered types.

   This module implements applicative association tables, also known as
   finite maps or dictionaries, given a total ordering function
   over the keys.
   All operations over maps are purely applicative (no side-effects).
   The implementation uses balanced binary trees, and therefore searching
   and insertion take time logarithmic in the size of the map.
*)


module Set :
  sig
    module type SET =
      sig
        type elt
        type t
        val empty : t
        val is_empty : t -> bool
        val mem : elt -> t -> bool
        val add : elt -> t -> t
        val singleton : elt -> t
        val remove : elt -> t -> t
        val union : t -> t -> t
        val inter : t -> t -> t
        val diff : t -> t -> t
        val compare : t -> t -> int
        val equal : t -> t -> bool
        val subset : t -> t -> bool
        val iter : (elt -> unit) -> t -> unit
        val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
        val for_all : (elt -> bool) -> t -> bool
        val exists : (elt -> bool) -> t -> bool
        val filter : (elt -> bool) -> t -> t
        val partition : (elt -> bool) -> t -> t * t
        val cardinal : t -> int
        val elements : t -> elt list
        val min_elt : t -> elt
        val max_elt : t -> elt
        val choose : t -> elt
        val split : elt -> t -> t * bool * t
        val set_of_list : elt list -> t
        (**[set_of_list l] creates a set from a list of elements *)
      end
    (** Output signature of the functor {!module: Xnc_stdlib.Set.Make}. *)

    module Make :
      functor (Ord : Set.OrderedType) ->
        sig
          type elt = Ord.t
          type t
          val empty : t
          val is_empty : t -> bool
          val mem : elt -> t -> bool
          val add : elt -> t -> t
          val singleton : elt -> t
          val remove : elt -> t -> t
          val union : t -> t -> t
          val inter : t -> t -> t
          val diff : t -> t -> t
          val compare : t -> t -> int
          val equal : t -> t -> bool
          val subset : t -> t -> bool
          val iter : (elt -> unit) -> t -> unit
          val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
          val for_all : (elt -> bool) -> t -> bool
          val exists : (elt -> bool) -> t -> bool
          val filter : (elt -> bool) -> t -> t
          val partition : (elt -> bool) -> t -> t * t
          val cardinal : t -> int
          val elements : t -> elt list
          val min_elt : t -> elt
          val max_elt : t -> elt
          val choose : t -> elt
          val split : elt -> t -> t * bool * t
          val set_of_list : elt list -> t
          (**[set_of_list l] creates a set from a list of elements *)
        end
  end
(** Sets over ordered types.

   This module implements the set data structure, given a total ordering
   function over the set elements. All operations over sets
   are purely applicative (no side-effects).
   The implementation uses balanced binary trees, and is therefore
   reasonably efficient: insertion and membership take time
   logarithmic in the size of the set, for instance.
*)



module List :
  sig
    val length : 'a list -> int
    val hd : 'a list -> 'a
    val tl : 'a list -> 'a list
    val nth : 'a list -> int -> 'a
    val rev : 'a list -> 'a list
    val append : 'a list -> 'a list -> 'a list
    val rev_append : 'a list -> 'a list -> 'a list
    val concat : 'a list list -> 'a list
    val flatten : 'a list list -> 'a list
    val iter : ('a -> unit) -> 'a list -> unit
    val map : ('a -> 'b) -> 'a list -> 'b list
    val rev_map : ('a -> 'b) -> 'a list -> 'b list
    val fold_left : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a
    val fold_right : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
    val iter2 : ('a -> 'b -> unit) -> 'a list -> 'b list -> unit
    val map2 : ('a -> 'b -> 'c) -> 'a list -> 'b list -> 'c list
    val rev_map2 : ('a -> 'b -> 'c) -> 'a list -> 'b list -> 'c list
    val fold_left2 : ('a -> 'b -> 'c -> 'a) -> 'a -> 'b list -> 'c list -> 'a
    val fold_right2 :
      ('a -> 'b -> 'c -> 'c) -> 'a list -> 'b list -> 'c -> 'c
    val for_all : ('a -> bool) -> 'a list -> bool
    val exists : ('a -> bool) -> 'a list -> bool
    val for_all2 : ('a -> 'b -> bool) -> 'a list -> 'b list -> bool
    val exists2 : ('a -> 'b -> bool) -> 'a list -> 'b list -> bool
    val mem : 'a -> 'a list -> bool
    val memq : 'a -> 'a list -> bool
    val find : ('a -> bool) -> 'a list -> 'a
    val filter : ('a -> bool) -> 'a list -> 'a list
    val find_all : ('a -> bool) -> 'a list -> 'a list
    val partition : ('a -> bool) -> 'a list -> 'a list * 'a list
    val assoc : 'a -> ('a * 'b) list -> 'b
    val assq : 'a -> ('a * 'b) list -> 'b
    val mem_assoc : 'a -> ('a * 'b) list -> bool
    val mem_assq : 'a -> ('a * 'b) list -> bool
    val remove_assoc : 'a -> ('a * 'b) list -> ('a * 'b) list
    val remove_assq : 'a -> ('a * 'b) list -> ('a * 'b) list
    val split : ('a * 'b) list -> 'a list * 'b list
    val combine : 'a list -> 'b list -> ('a * 'b) list
    val sort : ('a -> 'a -> int) -> 'a list -> 'a list
    val stable_sort : ('a -> 'a -> int) -> 'a list -> 'a list
    val fast_sort : ('a -> 'a -> int) -> 'a list -> 'a list
    val merge : ('a -> 'a -> int) -> 'a list -> 'a list -> 'a list
    
    val str_sort : string list -> string list
    (**
      * Sort a list of strings in increasing order according to the comparison
      * function [String.compare].
      *)

    val hd_opt : 'a list -> 'a option
    (** Return [Some f], the first element of the given list.
        Return [None] if the list is empty. *)
    
    val find_opt : ('a -> bool) -> 'a list -> 'a option
    (**
      * [find_opt p l] returns the first element of the list [l] that satisfies
      * the predicate [p]. Returns None if there is no value that satisfies [p]
      * in the list [l].
      *)

    val range : ?start:int -> ?step:int -> int -> int list
    (** Returns a list of integers from 0 to [stop] - 1 inclusives.
     
        If [start] and/or [step] are specified the sequence starts from [start]
        with step [step]. *)
  end
(** List operations. *)


module Array :
  sig
    external length : 'a array -> int = "%array_length"
    external get : 'a array -> int -> 'a = "%array_safe_get"
    external set : 'a array -> int -> 'a -> unit = "%array_safe_set"
    external make : int -> 'a -> 'a array = "caml_make_vect"
    external create : int -> 'a -> 'a array = "caml_make_vect"
    val init : int -> (int -> 'a) -> 'a array
    val make_matrix : int -> int -> 'a -> 'a array array
    val create_matrix : int -> int -> 'a -> 'a array array
    val append : 'a array -> 'a array -> 'a array
    val concat : 'a array list -> 'a array
    val sub : 'a array -> int -> int -> 'a array
    val copy : 'a array -> 'a array
    val fill : 'a array -> int -> int -> 'a -> unit
    val blit : 'a array -> int -> 'a array -> int -> int -> unit
    val to_list : 'a array -> 'a list
    val of_list : 'a list -> 'a array
    val iter : ('a -> unit) -> 'a array -> unit
    val map : ('a -> 'b) -> 'a array -> 'b array
    val iteri : (int -> 'a -> unit) -> 'a array -> unit
    val mapi : (int -> 'a -> 'b) -> 'a array -> 'b array
    val fold_left : ('a -> 'b -> 'a) -> 'a -> 'b array -> 'a
    val fold_right : ('a -> 'b -> 'b) -> 'a array -> 'b -> 'b
    val sort : ('a -> 'a -> int) -> 'a array -> unit
    val stable_sort : ('a -> 'a -> int) -> 'a array -> unit
    val fast_sort : ('a -> 'a -> int) -> 'a array -> unit
    external unsafe_get : 'a array -> int -> 'a = "%array_unsafe_get"
    external unsafe_set : 'a array -> int -> 'a -> unit = "%array_unsafe_set"

    val range : ?start:int -> ?step:int -> int -> int array
    (** Returns an array of integers from 0 to [stop] - 1 inclusives.
        If [start] and/or [step] are specified the sequence starts from [start]
        with step [step] *)
  end
(** Array operations. *)

