module S = Xnc_serializable;;
module P = Xnc_property;;

type connector_type =
    Cgi
  | Fcgi
  | Ajp
  | Scgi
  | Netplex
;;

let connector_type_of_string s =
  match s with
      "cgi" -> Cgi
    | "fcgi" -> Fcgi
    | "ajp" -> Ajp
    | "scgi" -> Scgi
    | "netplex" -> Netplex
    | _ -> failwith "connector_type_of_string"
;;

type socket_type =
    No_socket
  | Local
  | Internet
;;

let socket_type_of_string s =
  match s with
      "local" -> Local
    | "internet" -> Internet
    | _ -> failwith "socket_type_of_string"
;;

type netplex_configuration_type =
    Embedded
  | External
;;

let netplex_configuration_type_of_string s =
  match s with
      "embedded" -> Embedded
    | "external" -> External
    | _ -> failwith "netplex_configuration_type_of_string"
;;

type t =
    { start_page : string;
      error_page : string;
      not_found_page : string;
      session_manager_mode : Srv_types.manager_mode;
      session_persistence_mode : Srv_types.state_persistence_mode;
      session_timeout : int;
      page_parameter_manager_mode : Srv_types.manager_mode;
      page_parameter_persistence_mode : Srv_types.state_persistence_mode;
      page_parameter_timeout : int;
      sqlite3_path : string;
      current_page_field_name : string;
      page_parameters_field_name : string;
      session_field_name : string;
      connector_type : connector_type;
      socket_type : socket_type;
      bind_address : string;
      port : int;
      socket_path : string;
      netplex_configuration_type : netplex_configuration_type;
      netplex_configuration_path : string;
      netplex_enable_netcgi : bool;
      netplex_enable_nethttpd : bool;
      netplex_nethttpd_handler : string;
      netplex_nethttpd_path : string;
      netplex_configuration_content : string;
    }

let default_conf =
  { start_page = "start_page.xnunc";
    error_page = "standard_error_page.xnunc";
    not_found_page = "standard_not_found_page.xnunc";
    session_manager_mode = Srv_types.Client;
    session_persistence_mode = Srv_types.Hidden_field;
    session_timeout = 30;
    page_parameter_manager_mode = Srv_types.Client;
    page_parameter_persistence_mode = Srv_types.Hidden_field;
    page_parameter_timeout = 30;
    sqlite3_path = "/tmp/state.db";
    current_page_field_name = "xnunc_page";
    page_parameters_field_name = "xnunc_page_parameters";
    session_field_name = "xnunc_session";
    connector_type = Cgi;
    socket_type = No_socket;
    bind_address = "localhost";
    port = 9876;
    socket_path = "/tmp/ex-nunc";
    netplex_configuration_type = Embedded;
    netplex_configuration_path = "/tmp/netplex.cfg";
    netplex_enable_netcgi = true;
    netplex_enable_nethttpd = false;
    netplex_nethttpd_handler = "ex-nunc";
    netplex_nethttpd_path = "/ex-nunc";
    netplex_configuration_content = "";
  }
;;

let create
      ?(session_manager_mode = default_conf.session_manager_mode)
      ?(session_persistence_mode = default_conf.session_persistence_mode)
      () =
  { default_conf with
        session_manager_mode = session_manager_mode;
        session_persistence_mode = session_persistence_mode
  }
;;

(* private *)
let rec from_name_value_stream conf stream =
  let from_name_value name value =
    match name with
        "/application"
      | "/application/@id"
      | "/application/runtime"
      | "/application/runtime/state-manager"
      | "/application/runtime/hidden-fields"
      | "/application/connector"
      | "/application/connector/socket"
      | "/application/connector/netplex" ->
          (* ignore container elements *)
          conf
      | "/application/runtime/start-page" ->
          let start_page = S.unwrap_string value in
            { conf with start_page = start_page }
      | "/application/runtime/error-page" ->
          let error_page = S.unwrap_string value in
            { conf with error_page = error_page }
      | "/application/runtime/not-found-page" ->
          let not_found_page = S.unwrap_string value in
            { conf with not_found_page = not_found_page }
      | "/application/runtime/state-manager/@page-parameter-storage" ->
          let page_parameter_manager_mode =
            Srv_types.manager_mode_of_string (S.unwrap_string value)
          in
            { conf with
                  page_parameter_manager_mode = page_parameter_manager_mode }
      | "/application/runtime/state-manager/@page-parameter-persistence" ->
          let page_parameter_persistence_mode =
            Srv_types.state_persistence_mode_of_string (S.unwrap_string value)
          in
            { conf with
                  page_parameter_persistence_mode =
                    page_parameter_persistence_mode }
      | "/application/runtime/state-manager/@page-parameter-timeout" ->
          let page_parameter_timeout = S.unwrap_int value in
            { conf with page_parameter_timeout = page_parameter_timeout }
      | "/application/runtime/state-manager/@session-storage" ->
          let session_manager_mode =
            Srv_types.manager_mode_of_string (S.unwrap_string value)
          in
            { conf with
                  session_manager_mode = session_manager_mode }
      | "/application/runtime/state-manager/@session-persistence" ->
          let session_persistence_mode =
            Srv_types.state_persistence_mode_of_string (S.unwrap_string value)
          in
            { conf with
                  session_persistence_mode = session_persistence_mode }
      | "/application/runtime/state-manager/@session-timeout" ->
          let session_timeout = S.unwrap_int value in
            { conf with session_timeout = session_timeout }
      | "/application/runtime/state-manager/@sqlite3-path" ->
          let sqlite3_path = S.unwrap_string value in
            { conf with sqlite3_path = sqlite3_path }
      | "/application/runtime/hidden-fields/@current-page-field" ->
          let current_page_field_name = S.unwrap_string value in
            { conf with current_page_field_name = current_page_field_name }
      | "/application/runtime/hidden-fields/@page-parameter-field" ->
          let page_parameters_field_name = S.unwrap_string value in
            { conf with
                  page_parameters_field_name = page_parameters_field_name }
      | "/application/runtime/hidden-fields/@session-field" ->
          let session_field_name = S.unwrap_string value in
            { conf with session_field_name = session_field_name }
      | "/application/connector/@type" ->
          let conn_type = connector_type_of_string (S.unwrap_string value) in
            { conf with connector_type = conn_type }
      | "/application/connector/socket/@type" ->
          let socket_type = socket_type_of_string (S.unwrap_string value) in
            { conf with socket_type = socket_type }
      | "/application/connector/socket/file-name" ->
          let file_name = S.unwrap_string value in
            { conf with socket_path = file_name }
      | "/application/connector/socket/address" ->
          let address = S.unwrap_string value in
            { conf with bind_address = address }
      | "/application/connector/socket/port" ->
          let port = S.unwrap_int value in
            { conf with port = port }
      | "/application/connector/netplex/@configuration" ->
          let configuration = netplex_configuration_type_of_string
                                (S.unwrap_string value) in
            { conf with netplex_configuration_type = configuration }
      | "/application/connector/netplex/@path" ->
          let path = S.unwrap_string value in
          let absolute_path =
            begin match String.get path 0 with
                '/' -> path
              | _ -> (Sys.getcwd ()) ^ path
            end
          in
            { conf with netplex_configuration_path = absolute_path }
      | "/application/connector/netplex/@enable-netcgi" ->
          let enable_netcgi = S.unwrap_bool value in
            { conf with netplex_enable_netcgi = enable_netcgi }
      | "/application/connector/netplex/@enable-nethttpd" ->
          let enable_nethttpd = S.unwrap_bool value in
            { conf with netplex_enable_nethttpd = enable_nethttpd }
      | "/application/connector/netplex/@nethttpd-handler" ->
          let nethttpd_handler = S.unwrap_string value in
            { conf with netplex_nethttpd_handler = nethttpd_handler }
      | "/application/connector/netplex/@nethttpd-path" ->
          let nethttpd_path = S.unwrap_string value in
            { conf with netplex_nethttpd_path = nethttpd_path }
      | "/application/connector/netplex/content" ->
          let content = S.unwrap_string value in
            { conf with netplex_configuration_content = content }
      | _ -> failwith ("Unexpected element: " ^ name)
  in

    match Stream.peek stream with
        None ->
          conf
      | Some (name, value) ->
          let () = Stream.junk stream in
          let conf = from_name_value name value in
            from_name_value_stream conf stream
;;

let load application_values command_line_args =
  let file = open_in command_line_args.Srv_command_line.configuration_path in
  let in_ch = new Netchannels.input_channel file in
  let resolver =
    new Pxp_reader.lookup_id_as_file
      ~fixenc:`Enc_iso88591
      [
        Pxp_types.Public ("-//EXNUNC//Configuration File//EN",
                               "ex-nunc-config.dtd"),
        command_line_args.Srv_command_line.configuration_dtd;
      ] in
  let data = Srv_configuration_loader.load ~resolvers:[resolver] in_ch in
  let app_list = S.find_all_fields data "application" in
  let app = List.find
              (fun f ->
                 let id = S.unwrap_string
                            (S.find_value
                               ~properties:(P.singleton P.attribute)
                               f "id") in
                   id = command_line_args.Srv_command_line.application_name
              )
              app_list in
  let stream = S.to_name_value_stream app in
    from_name_value_stream default_conf stream
;;

let start_page conf =
  conf.start_page
;;

let error_page conf =
  conf.error_page
;;

let not_found_page conf =
  conf.not_found_page
;;

let session_manager_mode conf =
  conf.session_manager_mode
;;

let session_persistence_mode conf =
  conf.session_persistence_mode
;;

let session_timeout conf =
  conf.session_timeout
;;

let page_parameter_manager_mode conf =
  conf.page_parameter_manager_mode
;;

let page_parameter_persistence_mode conf =
  conf.page_parameter_persistence_mode
;;

let page_parameter_timeout conf =
  conf.page_parameter_timeout
;;

let sqlite3_path conf =
  conf.sqlite3_path
;;

let current_page_field_name conf =
  conf.current_page_field_name
;;

let page_parameters_field_name conf =
  conf.page_parameters_field_name
;;

let session_field_name conf =
  conf.session_field_name
;;

let connector_type conf =
  conf.connector_type
;;

let socket_type conf =
  conf.socket_type
;;

let bind_address conf =
  match conf.socket_type with
      Internet -> conf.bind_address
    | _ -> failwith "No socket address available for non-Internet sockets"
;;

let port conf =
  match conf.socket_type with
      Internet -> conf.port
    | _ -> failwith "No socket port available for non-Internet sockets"
;;

let socket_path conf =
  match conf.socket_type with
      Local -> conf.socket_path
    | _ -> failwith "No socket path available for non-Local sockets"
;;

let socket_address conf =
  match conf.socket_type with
      Local ->
        Unix.ADDR_UNIX conf.socket_path
    | Internet ->
        Unix.ADDR_INET (Unix.inet_addr_of_string conf.bind_address, conf.port)
    | No_socket ->
        failwith "No socket_address available"
;;

let netplex_configuration_path conf =
  conf.netplex_configuration_path
;;

let netplex_enable_netcgi conf =
  conf.netplex_enable_netcgi
;;

let netplex_enable_nethttpd conf =
  conf.netplex_enable_nethttpd
;;

let netplex_nethttpd_handler conf =
  conf.netplex_nethttpd_handler
;;

let netplex_nethttpd_path conf =
  conf.netplex_nethttpd_path
;;

let prepare_netplex_configuration conf =
  match conf.netplex_configuration_type with
      Embedded ->
        let out_ch = open_out_gen
                       [Open_creat; Open_excl; Open_wronly]
                       0o600
                       conf.netplex_configuration_path in
          output_string out_ch conf.netplex_configuration_content;
          close_out out_ch
    | _ -> ()
;;

let clean_netplex_configuration conf =
  match conf.netplex_configuration_type with
      Embedded when Sys.file_exists conf.netplex_configuration_path ->
        Sys.remove conf.netplex_configuration_path
    | _ -> ()
;;

