(** This module provides the server configuration.
  *)

type connector_type =
    Cgi (** CGI connector *)
  | Fcgi (** FastCGI connector *)
  | Ajp (** AJP 1.3 connector *)
  | Scgi (** SCGI connector *)
  | Netplex (** Netplex connector (Netcgi2-plex or Nethttpd-plex) *)
(** The type of connectors. *)

type socket_type =
    No_socket (** Connector does not need a socket (e.g., CGI) *)
  | Local (** Local socket (UNIX domain) *)
  | Internet (** Internet socket *)
(** The type of sockets used by connectors. *)

type t
(** The type of server configuration. *)

val default_conf : t
(** Default configuration. *)

val create :
  ?session_manager_mode:Srv_types.manager_mode ->
  ?session_persistence_mode:Srv_types.state_persistence_mode ->
  unit ->
  t
(** [create ()] creates a new configuration value.
  * @param session_manager_mode specifies a session manager mode
  * @param session_persistence_mode specifies a session persistence mode *)

val load :
  Srv_values.application_values -> Srv_command_line.args -> t
(** [load v a] loads the configuration file specified in [a], using application
  * values stored in [v]. *)

val start_page : t -> string
(** [start_page c] returns the name of the start page from configuration [c].
  *)

val error_page : t -> string
(** [error_page c] returns the name of the error page from configuration [c].
  *)

val not_found_page : t -> string
(** [not_found_page c] returns the name of the not found (404) page from
  * configuration [c]. *)

val session_manager_mode : t -> Srv_types.manager_mode
(** [session_manager c] returns the mode of the session manager from
  * configuration [c]. *)

val session_persistence_mode : t -> Srv_types.state_persistence_mode
(** [session_persistence_mode c] returns the session persistence mode from
  * configuration [c]. *)

val session_timeout : t -> int
(** [session_persistence_mode c] returns the session timeout in minutes from
  * configuration [c]. *)

val page_parameter_manager_mode : t -> Srv_types.manager_mode
(** [page_parameter_manager_mode c] returns the mode of the page signature
  * manager from configuration [c]. *)

val page_parameter_persistence_mode : t -> Srv_types.state_persistence_mode
(** [page_parameter_persistence_mode c] returns the persistence mode of the page
  * signature manager from configuration [c]. *)

val page_parameter_timeout : t -> int
(** [page_parameter_persistence_mode c] returns the page parameter timeout in
  * minutes from configuration [c]. *)

val sqlite3_path : t -> string
(** [sqlite3_path c] returns the Sqlite3 file path from configuration [c]. *)

val current_page_field_name : t -> string
(** [current_page_field_name c] returns the hidden field name containing the
  * current page from configuration [c]. *)

val page_parameters_field_name : t -> string
(** [page_parameters_field_name c] returns the hidden field name containing the
  * page parameters from configuration [c]. *)

val session_field_name : t -> string
(** [session_field_name c] returns the hidden field name containing the
  * session data from configuration [c]. *)

val connector_type : t -> connector_type
(** [connector_type c] returns the type of the connector from configuration [c].
  *)

val socket_type : t -> socket_type
(** [socket_type c] returns the type of the socket used by connector from
  * configuration [c]. *)

val bind_address : t -> string
(** [bind_address c] returns the bind address of the internet socket from
  * configuration [c]. *)

val port : t -> int
(** [port c] returns the port number of the internet socket from configuration
  * [c]. *)

val socket_path : t -> string
(** [socket_path c] returns the path of the local socket from configuration [c].
  *)

val socket_address : t -> Unix.sockaddr
(** [socket_address c] returns the socket address of the control channel from
  * configuration [c]. *)

val netplex_configuration_path : t -> string
(** [netplex_configuration_path c] returns the path of the configuration file
  * for Netplex from configuration [c]. *)

val netplex_enable_netcgi : t -> bool
(** [netplex_enable_netcgi c] returns if the Netcgi2-plex connector is enabled
  * from configuration [c]. *)

val netplex_enable_nethttpd : t -> bool
(** [netplex_enable_nethttpd c] returns if the Nethttpd-plex connector is
  * enabled from configuration [c]. *)

val netplex_nethttpd_handler : t -> string
(** [netplex_nethttpd_handler c] returns the name of the Nethttpd handler
  * from configuration [c]. *)

val netplex_nethttpd_path : t -> string
(** [netplex_nethttpd_path c] returns the path of the Nethttpd handler
  * from configuration [c]. *)

val prepare_netplex_configuration : t -> unit
(** [prepare_netplex_configuration c] prepare the Netplex configuration file
  * specified in configuration [c]. If the Netplex configuration is embedded, it
  * saves the content into the file specified by [netplex_configuration_path].
  * If the Netplex configuration is external, it does nothing. *)

val clean_netplex_configuration : t -> unit
(** [clean_netplex_configuration c] cleans the Netplex configuration file
  * specified in configuration [c]. If the Netplex configuration is embedded, it
  * deletes the file specified by [netplex_configuration_path]. If the Netplex
  * configuration is external, it does nothing. *)

