module S = Xnc_serializable;;
module P = Xnc_property;;

let load ?resolvers (in_ch : Netchannels.in_obj_channel) =
  let create_attribute name default =
      S.create ~properties:(P.singleton P.attribute) name default
  in

  let load_attribute name =
      create_attribute name S.default_string
  in

  (* TODO: add "connector" validation: e.g., reject configuration files
   * containing a "socket" element under a socket of type "netplex" *)
  let load_element name node =
    match name with
        "configuration"
      | "application"
      | "runtime"
      | "hidden-fields"
      | "connector" ->
          let field = S.create_empty_array name in
          let attributes = List.rev (node # attributes) in
            List.iter
              (fun (n, _) ->
                 let attribute = load_attribute n in
                   S.append_field field attribute)
              attributes;
            (field, true)
      | "state-manager" ->
          let field =
            S.new_field "state-manager"
              [create_attribute "page-parameter-storage" S.default_string;
               create_attribute "page-parameter-persistence" S.default_string;
               create_attribute "page-parameter-timeout" S.default_int;
               create_attribute "session-storage" S.default_string;
               create_attribute "session-persistence" S.default_string;
               create_attribute "session-timeout" S.default_int;
               create_attribute "sqlite3-path" S.default_string;
              ] S.wrap_list
          in
            (field, false)
      | "socket" ->
          let socket_type = node # required_string_attribute "type" in
          let socket_inner_field_list =
            match socket_type with
                "local" ->
                  [ S.create "file-name" S.default_string ]
              | "internet" ->
                  [ S.create "address" S.default_string;
                    S.create "port" S.default_int ]
              | _ ->
                  failwith ("Unexpected socket type: " ^ socket_type) in
          let field = 
            S.new_field "socket"
              ((create_attribute "type" S.default_string) ::
               socket_inner_field_list) S.wrap_list;
          in
            (field, false)
      | "netplex" ->
          let netplex_configuration =
            node # required_string_attribute "configuration" in
          let netplex_content_field_list =
            match netplex_configuration with
                "embedded" ->
                  [ S.create "content" S.default_string ]
              | "external" ->
                  []
              | _ ->
                  failwith ("Unexpected netplex configuration: " ^
                            netplex_configuration) in
          let field =
            S.new_field "netplex"
              ((create_attribute "configuration" S.default_string) ::
               (create_attribute "path" S.default_string) ::
               (create_attribute "enable-netcgi" S.default_bool) ::
               (create_attribute "enable-nethttpd" S.default_bool) ::
               (create_attribute "nethttpd-handler" S.default_string) ::
               (create_attribute "nethttpd-path" S.default_string) ::
                netplex_content_field_list
              ) S.wrap_list
          in
            (field, false)
      | "start-page"
      | "error-page"
      | "not-found-page" ->
          let field = S.create name S.default_string in
            (field, false)
      | _ ->
          failwith ("Unexpected element: " ^ name)
  in

  let rec load_node top node =
    let node_type = node # node_type in
      match node_type with
          Pxp_document.T_element name ->
            let (field, load_children) = load_element name node in
              S.append_field top field;
              if load_children then
                node # iter_nodes
                  (fun n ->
                     load_node field n)
              else
                ()
        | _ ->
            failwith "load_node: cannot load non-element nodes"
  in

  let doc =
    Pxp_tree_parser.parse_document_entity
      Pxp_types.default_config
      (Pxp_types.from_obj_channel
         ?alt:resolvers
         in_ch)
      Pxp_tree_parser.default_spec in
  let root_field = S.create_empty_array "root" in
  let root_node = doc # root in
  let () = load_node root_field root_node in
  let configuration_field = DynArray.get (S.get_dynarray root_field) 0 in
  let reader = Xnc_reader.create_from_pxp_node root_node in
  let () = S.read reader configuration_field in
    configuration_field
;;

