(** Input validators.
  *
  * This module provides the interface of input validator classes and some
  * common implementations. *)

exception Invalid_type_name
(** Raised when passing the wrong type name to [type_validator]. *)

class type input_validator =
object
  method get_error_message : string
  (** [get_error_message] returns the error message. *)

  method get_value : string
  (** [get_value] returns the value to validate. *)

  method is_valid : unit -> bool
  (** [is_valid ()] tests if the value is valid. *)

end
(** Interface of input validators. *)

class virtual input_validator_base :
  string ->
  string ->
object
  method get_error_message : string
  (** See {!method: Srv_input_validator.input_validator.get_error_message}. *)

  method get_value : string
  (** See {!method: Srv_input_validator.input_validator.get_value}. *)

  method virtual is_valid : unit -> bool
  (** See {!method: Srv_input_validator.input_validator.is_valid}. *)

end
(** [input_validator v e]: Base class for validators, where [v] is the value to
  * validate and [e] is the error message to display if the validation fails.
  *)

class custom_validator :
  string ->
  (string -> bool) ->
  string ->
  input_validator
(** [new custom_validator v f e] returns a validator that validates value [v]
  * using function [f] and uses error message [e] if the validation fails. *)

class required_field_validator :
  string ->
  string ->
  input_validator
(** [new required_field_validator v e] returns a validator that checks that
  * value [v] is not the empty string and uses error message [e] if the
  * validation fails. *)

class type_validator :
  string ->
  string ->
  string ->
  input_validator
(** [new type_validator v t e] returns a validator that checks that value [v] is
  * convertible to type [t] and uses error message [e] if the validation fails.
  *)

