module Make
  (Request_processor : Srv_request_processor.REQUEST_PROCESSOR) =
struct
  let select_server context =
    let logger = context.Srv_runtime_context.logger in
    let configuration = context.Srv_runtime_context.configuration in

    let start_server run =
      Request_processor.register_signal_handlers context;
      try
        run ()
      with
          Srv_page_behavior.Server_shutdown ->
            Xnc_logger.info logger "Shutting down the server"
        | err ->
            Xnc_logger.fatal logger
              ("Unhandled exception: " ^ Printexc.to_string err)
    in

    let netcgi_config =
      { Netcgi.default_config with
            Netcgi.default_exn_handler = false }
    in

    let buffered_transactional_optype () =
      let buffered _ ch = new Netchannels.buffered_trans_channel ch in
        `Transactional buffered
    in

    let start_netcgi run =
        start_server
          (fun () ->
             let output_type = buffered_transactional_optype () in
               run netcgi_config output_type
          )
    in

    let process (cgi : Netcgi.cgi_activation) =
      let process_request =
        Request_processor.select_request_processor context cgi
      in
        process_request cgi
    in

    let start_fcgi () =
      let sockaddr = Srv_configuration.socket_address configuration in
        start_netcgi
          (fun config output_type ->
             Netcgi_fcgi.run
               ~config
               ~output_type
               ~sockaddr
               (fun cgi -> process (cgi :> Netcgi.cgi_activation))
          )
    in

    let start_cgi () =
      start_netcgi
        (fun config output_type ->
           Netcgi_cgi.run
             ~config
             ~output_type
             process
        )
    in

    let start_ajp () =
      let port = Srv_configuration.port configuration in
        start_netcgi
          (fun config output_type ->
             Netcgi_ajp.run
               ~config
               ~output_type
               ~port
               process
          )
    in

    let start_scgi () =
      let port = Srv_configuration.port configuration in
        start_netcgi
          (fun config output_type ->
             Netcgi_scgi.run
               ~config
               ~output_type
               ~port
               process
          )
    in

    let start_netplex () =
      let nethttpd_factory () =
        let process_compat (cgi : Netcgi1_compat.Netcgi_types.cgi_activation) =
          let cgi' = Netcgi1_compat.Netcgi_types.of_compat_activation cgi in
            process cgi'
        in

        let handler_data =
          { Nethttpd_services.dyn_handler = (fun _ -> process_compat);
            dyn_activation =
              Nethttpd_services.std_activation `Std_activation_buffered;
            dyn_uri = None;
            dyn_translator = (fun _ -> "");
            dyn_accept_all_conditionals = false;
          } in
        let handler_name =
          Srv_configuration.netplex_nethttpd_handler configuration
        in
          Nethttpd_plex.nethttpd_factory
            ~handlers:[handler_name, handler_data]
            ()
      in

      let netcgi_factory () =
        let output_type = buffered_transactional_optype () in
          Netcgi_plex.factory
            ~config:netcgi_config
            ~output_type
            (fun _ cgi -> process cgi)
      in

      let () = Srv_configuration.prepare_netplex_configuration configuration in
      let config_filename =
        Srv_configuration.netplex_configuration_path configuration in
      let cmdline_cfg =
        Netplex_main.create
          ~config_filename
          ~foreground:true () in
      (* TODO: make it configurable *)
      let parallelizer = Netplex_mp.mp () in
      let enable_netcgi =
        Srv_configuration.netplex_enable_netcgi configuration in
      let enable_nethttpd =
        Srv_configuration.netplex_enable_nethttpd configuration in
      let factories =
        match (enable_netcgi, enable_nethttpd) with
            (true, true) -> [netcgi_factory (); nethttpd_factory ()]
          | (true, false) -> [netcgi_factory ()]
          | (false, true) -> [nethttpd_factory ()]
          | (false, false) -> [] in
      let run () =
        Netplex_main.startup
          parallelizer
          Netplex_log.logger_factories
          Netplex_workload.workload_manager_factories
          factories
          cmdline_cfg
      in
        start_server run
    in

      match Srv_configuration.connector_type configuration with
          Srv_configuration.Fcgi -> start_fcgi
        | Srv_configuration.Cgi -> start_cgi
        | Srv_configuration.Ajp -> start_ajp
        | Srv_configuration.Scgi -> start_scgi
        | Srv_configuration.Netplex -> start_netplex

  let start () =
    let context = Request_processor.set_up () in
    let server = select_server context in
      server ()

end

