exception Server_shutdown

(* TODO: implement page life cycle:
 * 1. Control initialization
 * 2. Session loading
 * 3. State restoration
 * 4. Event processing
 * 5. Next page selection
 * 6. Session saving
 * 7. XHTML generation
 * 8. Output transaction completion
 * 9. Resource deallocation
 *)
module type PAGE_BEHAVIOR =
sig
  type session_data_type
  type page_parameter_type
  type page_state_type

  type page_prototype = (string * page_parameter_type)
  type target_page = Self | Goto of page_prototype
  type targets =
      { standard : target_page;
        invalid_input : target_page;
        redirect_mode : Srv_types.redirect_mode;
      }

  class type page_behavior =
  object
    method get_name : string
    method get_page_parameters : page_parameter_type
    method get_cgi_context : Netcgi.cgi_activation
    method is_self_submitted : bool
    method get_session_data : session_data_type
    method get_page_structure : Srv_page_structure.t
    method get_targets : targets
    method get_validator_collection : Srv_validator_collection.t
    method get_logger : Xnc_logger.t
    method set_target_page :
      ?redirect_mode:Srv_types.redirect_mode ->
      ?on_invalid_input:target_page ->
      target_page -> unit
    method set_page_parameters : page_parameter_type -> unit
    method register_callback :
      Srv_types.lifecycle_hook -> Xnc_callback_repository.handler -> unit
    method invoke_callbacks : Srv_types.lifecycle_hook -> unit
    method add_validator : string -> Srv_input_validator.input_validator -> unit
    method is_valid : unit -> bool
    method trace : ?level:Xnc_logger.log_level -> string -> unit
    method add_debug_info : unit -> unit
    method build_url :
      ?fragment:string ->
      string -> page_parameter_type -> Srv_types.url_rendering_mode -> string
    method init : unit -> unit
    method load_session : unit -> unit
    method load_page_parameters : unit -> unit
    method build_validator_collection : unit -> unit
    method read_user_input : unit -> unit
    method build_page_structure : unit -> unit
    method handle_ui_events : unit -> unit
    method save_state : unit -> unit
    method render : unit -> unit
    method complete : unit -> unit

  end

  class virtual page_behavior_base :
    string ->
    page_parameter_type ->
    Srv_page_state.page_context ->
  object
    (* Accessors *)
    method get_name : string
    method get_page_parameters : page_parameter_type
    method get_cgi_context : Netcgi.cgi_activation
    method is_self_submitted : bool
    method get_session_data : session_data_type
    method get_page_structure : Srv_page_structure.page_structure
    method get_targets : targets
    method get_validator_collection : Srv_validator_collection.t
    method get_logger : Xnc_logger.t

    (* Mutators *)
    method set_page_parameters : page_parameter_type -> unit
    method set_target_page :
      ?redirect_mode:Srv_types.redirect_mode ->
      ?on_invalid_input:target_page ->
      target_page -> unit

    (* Callback management *)
    method register_callback :
      Srv_types.lifecycle_hook -> Xnc_callback_repository.handler -> unit
    method invoke_callbacks : Srv_types.lifecycle_hook -> unit

    (* Validator management *)
    method add_validator :
      string -> Srv_input_validator.input_validator -> unit
    method is_valid : unit -> bool

    (* Tracing and debugging *)
    method trace : ?level:Xnc_logger.log_level -> string -> unit
    method add_debug_info : unit -> unit

    (* URL rendering *)
    method build_url :
      ?fragment:string ->
      string -> page_parameter_type -> Srv_types.url_rendering_mode -> string

    (* Lifecycle methods *)
    method init : unit -> unit
    method load_session : unit -> unit
    method load_page_parameters : unit -> unit
    method virtual build_validator_collection : unit -> unit
    method virtual read_user_input : unit -> unit
    method virtual build_page_structure : unit -> unit
    method handle_ui_events : unit -> unit
    method save_state : unit -> unit
    method render : unit -> unit
    method complete : unit -> unit

    (* Private methods *)
    method private save_page_name : unit -> unit
    method private save_page_parameters : unit -> unit
    method private save_session : unit -> unit

  end

  module Factory :
  sig
    type page_constructor = (Srv_page_state.page_context -> page_behavior)
    type t

    val create : unit -> t
    val add_page_constructor : t -> string -> page_constructor -> t
    val count : t -> int
    val find_page_constructor : t -> string -> page_constructor

  end

end;;

module Make
  (State : Srv_page_state.PAGE_STATE) =
struct
  type session_data_type = State.session_data_type
  type page_parameter_type = State.page_parameter_type
  type page_state_type = State.t

  type page_prototype = string * page_parameter_type
  type target_page =
      Self
    | Goto of page_prototype
  type targets =
      { standard : target_page;
        invalid_input : target_page;
        redirect_mode : Srv_types.redirect_mode;
      }

  class type page_behavior =
  object
    method get_name : string
    method get_page_parameters : page_parameter_type
    method get_cgi_context : Netcgi.cgi_activation
    method is_self_submitted : bool
    method get_session_data : session_data_type
    method get_page_structure : Srv_page_structure.t
    method get_targets : targets
    method get_validator_collection : Srv_validator_collection.t
    method get_logger : Xnc_logger.t
    method set_target_page :
      ?redirect_mode:Srv_types.redirect_mode ->
      ?on_invalid_input:target_page ->
      target_page -> unit
    method set_page_parameters : page_parameter_type -> unit
    method register_callback :
      Srv_types.lifecycle_hook -> Xnc_callback_repository.handler -> unit
    method invoke_callbacks : Srv_types.lifecycle_hook -> unit
    method add_validator : string -> Srv_input_validator.input_validator -> unit
    method is_valid : unit -> bool
    method trace : ?level:Xnc_logger.log_level -> string -> unit
    method add_debug_info : unit -> unit
    method build_url :
      ?fragment:string ->
      string -> page_parameter_type -> Srv_types.url_rendering_mode -> string
    method init : unit -> unit
    method load_session : unit -> unit
    method load_page_parameters : unit -> unit
    method build_validator_collection : unit -> unit
    method read_user_input : unit -> unit
    method build_page_structure : unit -> unit
    method handle_ui_events : unit -> unit
    method save_state : unit -> unit
    method render : unit -> unit
    method complete : unit -> unit

  end

  class virtual page_behavior_base
    (page_name : string)
    (page_params : page_parameter_type)
    (page_ctx : Srv_page_state.page_context) =
  object(self)
    val state =
      State.create
        page_name
        page_params
        page_ctx
    val mutable targets =
      { standard = Self;
        invalid_input = Self;
        redirect_mode = Srv_types.Server_side
      }

    (* Accessors *)
    method get_name =
      page_name
    method get_page_parameters =
      State.get_page_parameters state
    method get_cgi_context =
      State.get_cgi_activation state
    method is_self_submitted =
      State.is_self_submitted state
    method get_session_data =
      State.get_session_data state
    method get_page_structure =
      State.get_page_structure state
    method get_targets =
      targets
    method get_validator_collection =
      State.get_validator_collection state
    method get_logger =
      State.get_logger state

    (* Mutators *)
    method set_page_parameters parameters =
      State.set_page_parameters state parameters
    method set_target_page
        ?(redirect_mode = Srv_types.Server_side)
        ?(on_invalid_input = Self)
        tp =
      targets <-
          { standard = tp;
            invalid_input = on_invalid_input;
            redirect_mode = redirect_mode
          }

    (* Callback management *)
    method register_callback event handler =
      State.register_callback state event handler

    method invoke_callbacks event =
      State.invoke_callbacks state event

    (* Validator management *)
    method add_validator label validator =
      State.add_validator state label validator

    method is_valid () =
      State.is_valid state

    (* Tracing and debugging *)
    method trace ?(level = Xnc_logger.Debug) message =
      State.log state level message

    method add_debug_info () =
      State.add_debug_info state

    method build_url ?fragment name parameters mode =
      let url = State.build_url state name parameters mode in
        match fragment with
            None -> url
          | Some f -> url ^ "#" ^ f

    (* Lifecycle methods *)
    method init () =
      ()

    method load_session () =
      State.retrieve_session state

    method load_page_parameters () =
      State.retrieve_page_parameters state

    method virtual build_validator_collection :
      unit -> unit

    method virtual read_user_input :
      unit -> unit

    method virtual build_page_structure :
      unit -> unit

    method handle_ui_events () =
      State.handle_ui_events state

    method save_state () =
      self # save_session ();
      self # save_page_name ();
      self # save_page_parameters ()

    method render () =
      State.render state

    method complete () =
      let cgi = State.get_cgi_activation state in
        cgi # out_channel # commit_work ()

    (* Private methods *)
    method private save_page_name () =
      State.save_page_name state

    method private save_page_parameters () =
      State.save_page_parameters state

    method private save_session () =
      State.save_session state

  end

  module Factory =
  struct
    type page_constructor =
        (Srv_page_state.page_context -> page_behavior)

    type t = (string * page_constructor) list

    let create () : t =
      []

    let add_page_constructor t page_name page_constructor =
      (page_name, page_constructor) :: t

    let count t =
      List.length t

    let find_page_constructor t page_name =
      List.assoc page_name t

  end

end;;

