(** Page behavior.
  *
  * This module implements page lifecycle. *)

exception Server_shutdown
(** Raised when a requesting a server shutdown. *)

module type PAGE_BEHAVIOR =
sig
  type session_data_type
  (** The type of session data. *)

  type page_parameter_type
  (** The type of page parameters. *)

  type page_state_type
  (** The type of page states. *)

  type page_prototype = string * page_parameter_type
  (** The type of page prototype. *)

  type target_page =
      Self
    | Goto of page_prototype
  (** The type of target page. *)

  type targets =
      { standard : target_page;
        invalid_input : target_page;
        redirect_mode : Srv_types.redirect_mode;
      }
  (** The type that collects target pages for the current page. *)

  class type page_behavior =
  object
    method get_name : string
    (** [get_name] returns the page name. *)

    method get_page_parameters : page_parameter_type
    (** [get_page_parameters] returns the page parameters. *)

    method get_cgi_context : Netcgi.cgi_activation
    (** [get_cgi_context] returns the CGI activation. *)

    method is_self_submitted : bool
    (** [is_self_submitted] returns [true] if current request resulted from self
      * submitting the form. *)

    method get_session_data : session_data_type
    (** [get_session_data] returns the session data. *)

    method get_page_structure : Srv_page_structure.t
    (** [get_page_structure] returns the page structure. *)

    method get_targets : targets
    (** [get_targets] returns the target pages for the current page. *)

    method get_validator_collection : Srv_validator_collection.t
    (** [get_validator_collection] returns the validator collection. *)

    method get_logger : Xnc_logger.t
    (** [get_logger] returns the page logger. *)

    method set_target_page :
      ?redirect_mode:Srv_types.redirect_mode ->
      ?on_invalid_input:target_page ->
      target_page -> unit
    (** [set_target_page v] sets the target page to [v].
      *
      * @param redirect_mode Page redirect mode.
      * @param on_invalid_input Target page in case of invalid user input. *)

    method set_page_parameters : page_parameter_type -> unit
    (** [set_page_parameters v] sets the page parameters to [v]. *)

    method register_callback :
      Srv_types.lifecycle_hook -> Xnc_callback_repository.handler -> unit
    (** [register_callback e h] registers callback [h] to event [e]. *)

    method invoke_callbacks : Srv_types.lifecycle_hook -> unit
    (** [invoke_callbacks e] invokes registered callbacks for event [e]. *)

    method add_validator :
      string -> Srv_input_validator.input_validator -> unit
    (** [add_validator n v] adds validator [v] with label [n]. *)

    method is_valid : unit -> bool
    (** [is_valid ()] tests input validators. *)

    method trace : ?level:Xnc_logger.log_level -> string -> unit
    (** [trace m] traces message [m] using page logger.
      *
      * @param level Log level. *)

    method add_debug_info : unit -> unit
    (** [add_debug_info ()] adds debug information to page structure. *)

    method build_url :
      ?fragment:string ->
      string -> page_parameter_type -> Srv_types.url_rendering_mode -> string
    (** [build_url n p m] returns the URL corresponding to page with name [n],
      * and parameters [p], using URL rendering mode [m].
      *
      * @param fragment Optional URL fragment. *)

    method init : unit -> unit
    (** [init ()] initializes the page. *)

    method load_session : unit -> unit
    (** [load_session ()] loads session. *)

    method load_page_parameters : unit -> unit
    (** [load_page_parameters ()] loads page parameters. *)

    method build_validator_collection : unit -> unit
    (** [build_validator_collection ()] builds validator collection. *)

    method read_user_input : unit -> unit
    (** [read_user_input ()] reads use input. *)

    method build_page_structure : unit -> unit
    (** [build_page_structure ()] builds page structure. *)

    method handle_ui_events : unit -> unit
    (** [handle_ui_events ()] handles UI events. *)

    method save_state : unit -> unit
    (** [save_state ()] saves page state (page name, page parameters, session).
      *)

    method render : unit -> unit
    (** [render ()] renders the page. *)

    method complete : unit -> unit
    (** [complete ()] commits the work. *)

  end
  (** Interface of pages. *)

  class virtual page_behavior_base :
    string ->
    page_parameter_type ->
    Srv_page_state.page_context ->
  object
    method get_name : string
    (** See {!method: Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.get_name}.
      *)

    method get_page_parameters : page_parameter_type
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.get_page_parameters}. *)

    method get_cgi_context : Netcgi.cgi_activation
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.get_cgi_context}. *)

    method is_self_submitted : bool
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.is_self_submitted}. *)

    method get_session_data : session_data_type
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.get_session_data}. *)

    method get_page_structure : Srv_page_structure.page_structure
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.get_page_structure}. *)

    method get_targets : targets
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.get_targets}. *)

    method get_validator_collection : Srv_validator_collection.t
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.get_validator_collection}.
      *)

    method get_logger : Xnc_logger.t
    (** See {!method: Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.get_logger}.
      *)

    method set_page_parameters : page_parameter_type -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.set_page_parameters}. *)

    method set_target_page :
      ?redirect_mode:Srv_types.redirect_mode ->
      ?on_invalid_input:target_page ->
      target_page -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.set_target_page}. *)

    method register_callback :
      Srv_types.lifecycle_hook -> Xnc_callback_repository.handler -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.register_callback}. *)

    method invoke_callbacks : Srv_types.lifecycle_hook -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.invoke_callbacks}. *)

    method add_validator :
      string -> Srv_input_validator.input_validator -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.add_validator}. *)

    method is_valid : unit -> bool
    (** See {!method: Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.is_valid}.
      *)

    method trace : ?level:Xnc_logger.log_level -> string -> unit
    (** See {!method: Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.trace}. *)

    method add_debug_info : unit -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.add_debug_info}. *)

    method build_url :
      ?fragment:string ->
      string -> page_parameter_type -> Srv_types.url_rendering_mode -> string
    (** See {!method: Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.build_url}.
      *)

    method init : unit -> unit
    (** See {!method: Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.init}. *)

    method load_session : unit -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.load_session}. *)

    method load_page_parameters : unit -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.load_page_parameters}. *)

    method virtual build_validator_collection : unit -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.build_validator_collection}.
      * *)

    method virtual read_user_input : unit -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.read_user_input}. *)

    method virtual build_page_structure : unit -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.build_page_structure}. *)

    method handle_ui_events : unit -> unit
    (** See {!method:
      * Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.handle_ui_events}. *)

    method save_state : unit -> unit
    (** See {!method: Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.save_state}.
      *)

    method render : unit -> unit
    (** See {!method: Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.render}. *)

    method complete : unit -> unit
    (** See {!method: Srv_page_behavior.PAGE_BEHAVIOR.page_behavior.complete}.
      *)

    method private save_page_name : unit -> unit
    (** [save_page_name ()] saves page name. *)

    method private save_page_parameters : unit -> unit
    (** [save_page_name ()] saves page parameters. *)

    method private save_session : unit -> unit
    (** [save_page_name ()] saves session. *)

  end
  (** [page_behavior_base n p c]: Base class of all pages, constructed with
    * name [n], parameters [p], and page context [c]. *)

  module Factory :
  sig
    type page_constructor = (Srv_page_state.page_context -> page_behavior)
    (** The type of page constructors. *)

    type t
    (** The type of page factory. *)

    val create : unit -> t
    (** [create ()] returns a new page factory. *)

    val add_page_constructor : t -> string -> page_constructor -> t
    (** [add_page_constructor f n c] adds page constructor [c] with name [n] to
      * factory [f]. *)

    val count : t -> int
    (** [count f] returns the number of page constructors of factory [f]. *)

    val find_page_constructor : t -> string -> page_constructor
    (** [find_page_constructor f n] returns page constructor stored in factory
      * [f] with name [n]. *)

  end
  (** This module implements page factory. *)

end
(** Module signature of page behavior. *)

module Make :
  functor (State : Srv_page_state.PAGE_STATE) ->
sig
  include PAGE_BEHAVIOR
    with type session_data_type = State.session_data_type and
      type page_parameter_type = State.page_parameter_type

end
(** This functor builds concrete page behavior modules. *)

