module type PAGE_DISPATCHER =
sig
  type page_factory_type

  type page_parameter_type

  val serve_page :
    page_factory_type ->
    Netcgi.cgi_activation ->
    Srv_configuration.t ->
    Xnc_logger.t ->
    Srv_state_persisting.Serialized.t ->
    unit

  val serve_error_page :
    page_factory_type ->
    Netcgi.cgi_activation ->
    Srv_configuration.t ->
    Xnc_logger.t ->
    Srv_state_persisting.Serialized.t ->
    exn ->
    unit

end;;

module Make
  (Page_parameters : Srv_types.PAGE_PARAMETERS)
  (Page_state_machine : Srv_page_state_machine.PAGE_STATE_MACHINE
     with type page_parameter_type = Page_parameters.t) =
struct
  type page_parameter_type = Page_parameters.t

  type page_factory_type = Page_state_machine.page_factory_type

  let call_page
        page_factory
        (cgi : Netcgi.cgi_activation)
        config
        logger
        serialized_state
        page_name
        page_parameters
        is_self_submitted =
    let run_machine state_machine =
      let rec run machine =
        run (Page_state_machine.next_step machine)
      in
        try
          run state_machine
        with Srv_page_state_machine.Final_state_reached ->
          Xnc_logger.debug logger "State machine reached end state"
    in

    let session_part = Srv_state_persisting.Serialized.get_state_part
                         serialized_state Srv_state_persisting.Session in
    let session_to_load =
      match session_part.Srv_state_persisting.Serialized.location with
          Srv_state_persisting.Not_present -> false
        | _ -> true in
    let state_machine = Page_state_machine.create
                          page_factory
                          logger
                          cgi
                          serialized_state
                          page_name
                          page_parameters
                          is_self_submitted
                          session_to_load
    in
      run_machine state_machine

  let serve_page
        page_factory
        (cgi: Netcgi.cgi_activation)
        config
        logger
        serialized_state =
    let page_name_part =
      Srv_state_persisting.Serialized.get_state_part
        serialized_state Srv_state_persisting.Page_name in
    let page_parameter_part =
      Srv_state_persisting.Serialized.get_state_part
        serialized_state Srv_state_persisting.Page_parameters
    in
      match page_name_part.Srv_state_persisting.Serialized.location with
          Srv_state_persisting.Not_present ->
            begin match page_parameter_part.Srv_state_persisting.Serialized.location with
                Srv_state_persisting.Not_present ->
                  call_page
                    page_factory
                    cgi
                    config
                    logger
                    serialized_state
                    (Srv_configuration.start_page config)
                    (Some Page_parameters.default_start_page_parameters)
                    false
              | _ ->
                  call_page
                    page_factory
                    cgi
                    config
                    logger
                    serialized_state
                    (Srv_configuration.not_found_page config)
                    (Some (Page_parameters.create_not_found_page_parameters page_parameter_part.Srv_state_persisting.Serialized.content))
                    false
            end
        | _ ->
            call_page
              page_factory
              cgi
              config
              logger
              serialized_state
              page_name_part.Srv_state_persisting.Serialized.content
              None
              (page_name_part.Srv_state_persisting.Serialized.location = Srv_state_persisting.Argument_value)

  let serve_error_page
        page_factory
        (cgi: Netcgi.cgi_activation)
        config
        logger
        serialized_state
        error =
    let error_descr = Printexc.to_string error in
    let backtrace =
      if Printexc.backtrace_status () then
        Printexc.get_backtrace ()
      else
        "" in
    let error_page_name = Srv_configuration.error_page config in
      Xnc_logger.error logger ("Unhandled exception: " ^
                               error_descr);
      Xnc_logger.error logger ("Backtrace: " ^
                               backtrace);
      Xnc_logger.debug logger ("Calling error page " ^
                               error_page_name);
      call_page
        page_factory
        cgi
        config
        logger
        serialized_state
        error_page_name
        (Some (Page_parameters.create_error_page_parameters
                 error_descr backtrace))
        false
end;;

