type page_context =
    { cgi : Netcgi.cgi_activation;
      state_serialization : Srv_state_persisting.Serialized.t;
      is_self_submitted : bool
    }
;;

module type PAGE_STATE =
sig
  type t

  type session_data_type

  type page_parameter_type

  val create :
    string ->
    page_parameter_type ->
    page_context ->
    t

  val get_page_name : t -> string

  val get_page_parameters : t -> page_parameter_type

  val set_page_parameters : t -> page_parameter_type -> unit

  val get_cgi_activation : t -> Netcgi.cgi_activation

  val is_self_submitted : t -> bool

  val get_session_data : t -> session_data_type

  val retrieve_session : t -> unit 

  val retrieve_page_parameters : t -> unit

  val get_page_structure : t -> Srv_page_structure.t

  val get_logger : t -> Xnc_logger.t

  val get_validator_collection : t -> Srv_validator_collection.t

  val add_debug_info : t -> unit

  val log : t -> Xnc_logger.log_level -> string -> unit

  val handle_ui_events : t -> unit

  val render : t -> unit

  val add_simple_hidden_field : t -> string -> string -> unit

  val add_complex_hidden_field :
    ?serializer:string Xnc_serializing.serializer ->
    t -> string -> Xnc_field_collection.t -> unit

  val add_simple_cookie :
    ?max_age:int ->
    t -> string -> string -> unit

  val add_complex_cookie :
    ?serializer:string Xnc_serializing.serializer ->
    ?max_age:int ->
    t -> string -> Xnc_field_collection.t -> unit

  val register_callback :
    t -> Srv_types.lifecycle_hook -> Xnc_callback_repository.handler -> unit

  val invoke_callbacks : t -> Srv_types.lifecycle_hook -> unit

  val add_validator : t -> string -> Srv_input_validator.input_validator -> unit

  val is_valid : t -> bool

  val session_to_field_collection : t -> Xnc_field_collection.t

  val page_parameters_to_field_collection : t -> Xnc_field_collection.t

  val build_url :
    t ->
    string ->
    page_parameter_type ->
    Srv_types.url_rendering_mode ->
    string

  val save_session : t -> unit

  val save_page_name : t -> unit

  val save_page_parameters : t -> unit

end;;

module Make
  (Context : Srv_runtime_context.RUNTIME_CONTEXT)
  (Session_data : Srv_types.SESSION_DATA)
  (Page_parameters : Srv_types.PAGE_PARAMETERS) =
struct
  module Session =
    Srv_state.Make
      (struct
         type t = Session_data.t

         let to_field_collection =
           Session_data.to_field_collection

         let from_field_collection _ f =
           Session_data.from_field_collection f
       end)

  module Parameters =
    Srv_state.Make
      (Page_parameters)

  (* TODO: refactor with Page_parameter_manager_singleton *)
  module Session_manager_singleton =
    Context.Make_singleton(
      struct
        type t = Session.Manager.t

        let create () =
          let context = Context.get_current () in
          let configuration = context.Srv_runtime_context.configuration in
          let session_manager_mode =
            Srv_configuration.session_manager_mode configuration
          in
            Session.Manager.create session_manager_mode configuration

      end
    )

  module Page_parameter_manager_singleton =
    Context.Make_singleton(
      struct
        type t = Parameters.Manager.t

        let create () =
          let context = Context.get_current () in
          let configuration = context.Srv_runtime_context.configuration in
          let page_parameter_manager_mode =
            Srv_configuration.page_parameter_manager_mode configuration
          in
            Parameters.Manager.create page_parameter_manager_mode configuration

      end
    )

  type t =
      { page_name : string;
        mutable parameters : Parameters.State.t;
        parameter_manager : Parameters.Manager.t;
        cgi_activation : Netcgi.cgi_activation;
        serialized_state : Srv_state_persisting.Serialized.t;
        self_submitted : bool;
        page_structure : Srv_page_structure.t;
        mutable session : Session.State.t;
        session_manager : Session.Manager.t;
        callbacks : Srv_types.lifecycle_hook Xnc_callback_repository.t;
        validators : Srv_validator_collection.t;
        logger : Xnc_logger.t;
        configuration : Srv_configuration.t
      }

  type session_data_type = Session_data.t

  type page_parameter_type = Page_parameters.t

  let create
        page_name
        page_parameters
        page_ctx =
    let context = Context.get_current () in
    let conf = context.Srv_runtime_context.configuration in
    let client_ip = page_ctx.cgi # environment # cgi_remote_addr in
    let user_agent = page_ctx.cgi # environment # input_header_field
                       ~default:"N/A" "User-Agent" in
      { page_name = page_name;
        parameters =
          Parameters.State.create page_parameters client_ip user_agent;
        parameter_manager = Page_parameter_manager_singleton.get_instance ();
        cgi_activation = page_ctx.cgi;
        serialized_state = page_ctx.state_serialization;
        self_submitted = page_ctx.is_self_submitted;
        page_structure = Srv_page_structure.create conf;
        session =
          Session.State.create (Session_data.create ()) client_ip user_agent;
        session_manager = Session_manager_singleton.get_instance ();
        callbacks = Xnc_callback_repository.create ();
        validators = Srv_validator_collection.create ();
        logger = Srv_page_helper.create_page_logger ();
        configuration = conf
      }

  let get_page_name state =
    state.page_name

  let get_page_parameters state =
    Parameters.State.get_data state.parameters

  let set_page_parameters state page_parameters =
    state.parameters <- Parameters.State.set_data
                          state.parameters
                          page_parameters

  let get_cgi_activation state =
    state.cgi_activation

  let is_self_submitted state =
    state.self_submitted

  let get_session_data state =
    Session.State.get_data state.session

  let retrieve_session state =
    let timeout = Srv_configuration.session_timeout state.configuration in
    let s = Srv_state_persisting.Serialized.get_state_part
              state.serialized_state Srv_state_persisting.Session in
    let get_from_repository get_descriptor =
      Context.Field_collection_repository.get get_descriptor "session"
    in
      state.session <- Session.Manager.load_from_serialized_state
                         ~get_from_repository
                         state.session_manager
                         state.session
                         s
                         state.configuration
                         timeout

  let retrieve_page_parameters state =
    let timeout =
      Srv_configuration.page_parameter_timeout state.configuration in
    let s = Srv_state_persisting.Serialized.get_state_part
              state.serialized_state Srv_state_persisting.Page_parameters in
    let get_from_repository get_descriptor =
      Context.Field_collection_repository.get
        get_descriptor
        ("page_parameters_" ^ state.page_name)
    in
      state.parameters <- Parameters.Manager.load_from_serialized_state
                            ~get_from_repository
                            state.parameter_manager
                            state.parameters
                            s
                            state.configuration
                            timeout

  let get_page_structure state =
    state.page_structure

  let get_logger state =
    state.logger

  let get_validator_collection state =
    state.validators

  let add_debug_info state =
    let session_data = get_session_data state in
    let page_parameters = get_page_parameters state in
    let context = Context.get_current () in
    let div = new Srv_page_element.div_element in
      div # set_id (Some "debug_info");
      div # add_element
        (new Srv_page_helper.misc_info context state.cgi_activation);
      div # add_element
        (new Srv_page_helper.name_value_table
           "Header fields"
           (fun () ->
              state.cgi_activation # environment # input_header_fields));
      div # add_element
        (new Srv_page_helper.name_value_table
           "Arguments"
           (fun () ->
              (List.map
                 (fun argument -> (argument # name, argument # value))
                 state.cgi_activation # arguments)
           ));
      div # add_element
        (new Srv_page_helper.name_value_table
           "Cookies"
           (fun () ->
              (List.map
                 (fun cookie -> (Netcgi.Cookie.name cookie,
                                 Netcgi.Cookie.value cookie))
                 state.cgi_activation # environment # cookies)
           ));
      (* TODO: refactor name/value list generation from field collection *)
      div # add_element
        (new Srv_page_helper.name_value_table
           "Session"
           (fun () ->
              let fc = Session.Data.to_field_collection session_data in
                Xnc_field_collection.to_name_value_string_list fc
           ));
      div # add_element
        (new Srv_page_helper.name_value_table
           "Page parameters"
           (fun () ->
              let fc = Parameters.Data.to_field_collection
                         page_parameters in
                Xnc_field_collection.to_name_value_string_list fc
           ));
      div # add_element
        (new Srv_page_helper.trace_table state.cgi_activation state.logger);
      Srv_page_structure.append_to_body state.page_structure div

  let log state level message =
    Xnc_logger.log state.logger level message

  let handle_ui_events state =
    Srv_page_structure.handle_ui_events
      state.page_structure
      state.cgi_activation

  let render state =
    Srv_page_structure.render_header
      state.page_structure
      state.cgi_activation;
    Srv_page_structure.render 
      state.page_structure
      state.cgi_activation # out_channel

  let add_simple_hidden_field state name value =
    Srv_page_structure.add_simple_hidden_field
      state.page_structure
      name
      value

  let add_complex_hidden_field ?serializer state name value =
    Srv_page_structure.add_complex_hidden_field
      ?serializer
      state.page_structure
      name
      value

  let add_simple_cookie ?max_age state name value =
    Srv_page_structure.add_simple_cookie
      ?max_age
      state.page_structure
      name
      value
      state.cgi_activation

  let add_complex_cookie ?serializer ?max_age state name value =
    Srv_page_structure.add_complex_cookie
      ?serializer
      ?max_age
      state.page_structure
      name
      value
      state.cgi_activation

  let register_callback state event handler =
    Xnc_callback_repository.add_handler state.callbacks event handler

  let invoke_callbacks state event =
    Xnc_callback_repository.invoke state.callbacks event

  let add_validator state label validator =
    Srv_validator_collection.add_validator
      state.validators label validator

  let is_valid state =
    Srv_validator_collection.is_valid state.validators

  let session_to_field_collection state =
    Session.Manager.state_to_field_collection 
      state.session_manager 
      state.session

  let page_parameters_to_field_collection state =
    Parameters.Manager.state_to_field_collection 
      state.parameter_manager 
      state.parameters

  let build_url state name parameters mode =
    let create_state_part part manager_mode persistence_mode location fc =
      { Srv_state_persisting.Deserialized.part =
          part;
        Srv_state_persisting.Deserialized.location =
          if manager_mode = Srv_types.Off ||
             persistence_mode = Srv_types.Cookie then
            Srv_state_persisting.Not_present
          else
            location;
        Srv_state_persisting.Deserialized.content =
          if manager_mode = Srv_types.Off ||
             persistence_mode = Srv_types.Cookie then
            Srv_state_persisting.Deserialized.String ""
          else
            Srv_state_persisting.Deserialized.Field_collection fc;
      }
    in

    let session_fc = session_to_field_collection state in
    let page_parameter_fc = Parameters.Manager.state_to_field_collection 
                              state.parameter_manager 
                              (Parameters.State.set_data
                                 state.parameters
                                 parameters) in
    let location =
      match mode with
          Srv_types.Cgi_style ->
            Srv_state_persisting.Query_string
        | Srv_types.Plain_url ->
            Srv_state_persisting.Path_info in
    let deserialized_state =
      Srv_state_persisting.Deserialized.create
        [ { Srv_state_persisting.Deserialized.part =
              Srv_state_persisting.Page_name;
            Srv_state_persisting.Deserialized.location =
              Srv_state_persisting.Path_info;
            Srv_state_persisting.Deserialized.content =
              Srv_state_persisting.Deserialized.String name
          };
          create_state_part
            Srv_state_persisting.Session
            (Srv_configuration.session_manager_mode state.configuration)
            (Srv_configuration.session_persistence_mode state.configuration)
            location
            session_fc;
          create_state_part
            Srv_state_persisting.Page_parameters
            (Srv_configuration.page_parameter_manager_mode state.configuration)
            (Srv_configuration.page_parameter_persistence_mode
               state.configuration)
            location
            page_parameter_fc;
        ] in
    let serialized_state = Srv_state_persisting.Deserialized.serialize_state
                             deserialized_state
                             state.configuration in
      Srv_state_persisting.Serialized.build_url
        serialized_state
        state.cgi_activation
        state.configuration

  (* TODO: refactor save_session and save_page_parameters *)
  let save_session state =
    let fc = session_to_field_collection state in
    let session_field_name =
      Srv_configuration.session_field_name state.configuration in
    let max_age =
      (Srv_configuration.session_timeout state.configuration) * 60
    in
      Session.Manager.persist state.session_manager state.session;
      match Srv_configuration.session_persistence_mode state.configuration with
          Srv_types.Hidden_field ->
            add_complex_hidden_field state session_field_name fc
        | Srv_types.Cookie ->
            add_complex_cookie ~max_age state session_field_name fc

  let save_page_name state =
    add_simple_hidden_field
      state
      (Srv_configuration.current_page_field_name state.configuration) 
      state.page_name

  let save_page_parameters state =
    let fc = page_parameters_to_field_collection state in
    let page_parameters_field_name =
      Srv_configuration.page_parameters_field_name state.configuration in
    let max_age =
      (Srv_configuration.page_parameter_timeout state.configuration) * 60
    in
      Parameters.Manager.persist state.parameter_manager state.parameters;
      match Srv_configuration.page_parameter_persistence_mode
              state.configuration with
          Srv_types.Hidden_field ->
            add_complex_hidden_field state page_parameters_field_name fc
        | Srv_types.Cookie ->
            add_complex_cookie ~max_age state page_parameters_field_name fc

end;;

