(** Page state.
  *
  * This module implements page state. *)

type page_context =
    { cgi : Netcgi.cgi_activation;
      state_serialization : Srv_state_persisting.Serialized.t;
      is_self_submitted : bool
    }
(** The type of page contexts. *)

module type PAGE_STATE =
sig
  type t
  (** The type of page state. *)

  type session_data_type
  (** The type of session data. *)

  type page_parameter_type
  (** The type of page parameters. *)

  val create :
    string ->
    page_parameter_type ->
    page_context ->
    t
  (** [create n p c] creates page state with name [n], page parameters [p],
    * and page context [c]. *)

  val get_page_name : t -> string
  (** [get_page_name p] returns page name from [p]. *)

  val get_page_parameters : t -> page_parameter_type
  (** [get_page_parameters p] returns page parameters from [p]. *)

  val set_page_parameters : t -> page_parameter_type -> unit
  (** [set_page_parameters p v] sets page parameters of [p] to [v]. *)

  val get_cgi_activation : t -> Netcgi.cgi_activation
  (** [get_cgi_activation p] returns CGI activation from [p]. *)

  val is_self_submitted : t -> bool
  (** [is_self_submitted] returns [true] if current request resulted from self
    * submitting the form. *)

  val get_session_data : t -> session_data_type
  (** [get_session_data p] returns session data from [p]. *)

  val retrieve_session : t -> unit 
  (** [retrieve_session p] retrieves session state and stores it in [p]. *)

  val retrieve_page_parameters : t -> unit
  (** [retrieve_page_parameters p] retrieves page parameters and stores it in
    * [p]. *)

  val get_page_structure : t -> Srv_page_structure.t
  (** [get_page_structure p] returns page structure from [p]. *)

  val get_logger : t -> Xnc_logger.t
  (** [get_logger p] returns page logger from [p]. *)

  val get_validator_collection : t -> Srv_validator_collection.t
  (** [get_validator_collection p] returns validator collection from [p]. *)

  val add_debug_info : t -> unit
  (** [add_debug_info p] adds debug information to page structure in [p]. *)

  val log : t -> Xnc_logger.log_level -> string -> unit
  (** [log p l m] traces message [m] with log level [l] using page logger of
    * [p]. *)

  val handle_ui_events : t -> unit
  (** [handle_ui_events p] invokes UI event handlers of page structure in [p].
    *)

  val render : t -> unit
  (** [render p] renders page structure in [p]. *)

  val add_simple_hidden_field : t -> string -> string -> unit
  (** [add_simple_hidden_field p n v] adds a hidden field with name [n] and
    * value [v] to page structure in [p]. *)

  val add_complex_hidden_field :
    ?serializer:string Xnc_serializing.serializer ->
    t -> string -> Xnc_field_collection.t -> unit
  (** [add_complex_hidden_field p n f] adds a hidden field with name [n] and
    * value obtained from [f] to page structure in [p].
    *
    * @param serializer Optional custom serializer. *)

  val add_simple_cookie :
    ?max_age:int ->
    t -> string -> string -> unit
  (** [add_simple_cookie p n v] adds a cookie with name [n] and value [v] to
    * page structure in [p].
    *
    * @param max_age sets the lifetime of the cookie to [max_age] seconds. *)

  val add_complex_cookie :
    ?serializer:string Xnc_serializing.serializer ->
    ?max_age:int ->
    t -> string -> Xnc_field_collection.t -> unit
  (** [add_complex_cookie p n f] adds a cookie with name [n] and value obtained
    * from [f] to page structure in [p].
    *
    * @param serializer Optional custom serializer.
    * @param max_age sets the lifetime of the cookie to [max_age] seconds. *)

  val register_callback :
    t -> Srv_types.lifecycle_hook -> Xnc_callback_repository.handler -> unit
  (** [register_callback p e h] registers callback [h] to event [e] in [p]. *)

  val invoke_callbacks : t -> Srv_types.lifecycle_hook -> unit
  (** [invoke_callbacks p e] invokes registered callbacks for event [e] in [p].
    *)

  val add_validator :
    t -> string -> Srv_input_validator.input_validator -> unit
  (** [add_validator p n v] adds validator [v] with label [n] to [p]. *)

  val is_valid : t -> bool
  (** [is_valid p] tests input validators in [p]. *)

  val session_to_field_collection : t -> Xnc_field_collection.t
  (** [session_to_field_collection s] returns the field colleciton storing
    * session contained in page state [s]. *)

  val page_parameters_to_field_collection : t -> Xnc_field_collection.t
  (** [page_parameters_to_field_collection s] returns the field colleciton
    * storing page parameters contained in page state [s]. *)

  val build_url :
    t ->
    string ->
    page_parameter_type ->
    Srv_types.url_rendering_mode ->
    string
  (** [build_url s n p m] returns the URL corresponding to page with name [n],
    * and parameters [p] using URL rendering mode [u]. *)

  val save_session : t -> unit
  (** [save_session s] persists session contained in page state [s]. *)

  val save_page_name : t -> unit
  (** [save_page_name s] persists page name contained in page state [s]. *)

  val save_page_parameters : t -> unit
  (** [save_page_parameters s] persists page parameters contained in page state
    * [s]. *)

end
(** Module signature of page state. *)

module Make :
  functor (Context : Srv_runtime_context.RUNTIME_CONTEXT) ->
  functor (Session_data : Srv_types.SESSION_DATA) ->
  functor (Page_parameters : Srv_types.PAGE_PARAMETERS) ->
sig
  include PAGE_STATE
    with type session_data_type = Session_data.t and
      type page_parameter_type = Page_parameters.t

end
(** This functor build concrete page state modules. *)

