(* TODO: review tracing code *)
exception Invalid_transition
exception Final_state_reached

module type PAGE_STATE_MACHINE =
sig
  type state
  type event

  type page_parameter_type
  type page_factory_type
  type t

  val create :
        page_factory_type ->
        Xnc_logger.t ->
        Netcgi.cgi_activation ->
        Srv_state_persisting.Serialized.t ->
        string ->
        page_parameter_type option ->
        bool ->
        bool ->
        t
  val current_state : t -> state
  val do_transition : t -> event -> t
  val next_step : t -> t

end

module Make
  (Page_behavior : Srv_page_behavior.PAGE_BEHAVIOR) =
struct
  type _state =
      Uninitialized
    | Ready_to_load_session
    | Initialized
    | Ready_to_validate_user_input
    | Ready_to_read_user_input
    | State_loaded
    | Ready_to_handle_ui_events
    | Ready_to_save_state
    | State_saved
    | Ready_to_render
    | Rendered
    | Completed

  type _event =
      Initialize
    | Load_session
    | Init_page_parameters
    | Load_page_parameters
    | Validate_user_input
    | Read_user_input
    | Build_page_structure
    | Handle_ui_events
    | Handle_page_change
    | Save_state
    | Render
    | Complete

  type state = _state
  type event = _event
  type page_parameter_type = Page_behavior.page_parameter_type
  type page_factory_type = Page_behavior.Factory.t
  type t = 
      { current_state : state;
        page_factory : page_factory_type;
        logger : Xnc_logger.t;
        cgi_context : Netcgi.cgi_activation;
        serialized_state : Srv_state_persisting.Serialized.t;
        page_name : string;
        page_parameters : page_parameter_type option;
        is_self_submitted : bool;
        session_to_load : bool;
        page_behavior : Page_behavior.page_behavior option;
        valid : bool
      }

  let create
        page_factory
        logger
        cgi_context
        serialized_state
        page_name
        page_parameters
        is_self_submitted
        session_to_load =
    { current_state = Uninitialized;
      page_factory = page_factory;
      logger = logger;
      cgi_context = cgi_context;
      serialized_state = serialized_state;
      page_name = page_name;
      page_parameters = page_parameters;
      is_self_submitted = is_self_submitted;
      session_to_load = session_to_load;
      page_behavior = None;
      valid = true
    }

  let current_state m =
    m.current_state

  let parameters_to_load m =
    Option.is_none m.page_parameters

  let state_to_string state =
    match state with
        Uninitialized -> "Uninitialized"
      | Ready_to_load_session -> "Ready_to_load_session"
      | Initialized -> "Initialized"
      | Ready_to_validate_user_input -> "Ready_to_validate_user_input"
      | Ready_to_read_user_input -> "Ready_to_read_user_input"
      | State_loaded -> "State_loaded"
      | Ready_to_handle_ui_events -> "Ready_to_handle_ui_events"
      | Ready_to_save_state -> "Ready_to_save_state"
      | State_saved -> "State_saved"
      | Ready_to_render -> "Ready_to_render"
      | Rendered -> "Rendered"
      | Completed -> "Completed"

  let event_to_string event =
    match event with
        Initialize -> "Initialize"
      | Load_session -> "Load_session"
      | Init_page_parameters -> "Init_page_parameters"
      | Load_page_parameters -> "Load_page_parameters"
      | Validate_user_input -> "Validate_user_input"
      | Read_user_input -> "Read_user_input"
      | Build_page_structure -> "Build_page_structure"
      | Handle_ui_events -> "Handle_ui_events"
      | Handle_page_change -> "Handle_page_change"
      | Save_state -> "Save_state"
      | Render -> "Render"
      | Complete -> "Complete"

  (* TODO: refactor actions *)
  let do_transition m event =
    Xnc_logger.debug m.logger ("Current state: " ^
                           (state_to_string m.current_state));
    Xnc_logger.debug m.logger ("Event received: " ^
                           (event_to_string event));
    match m.current_state with
        Uninitialized ->
          (match event with
               Initialize ->
                 let () = Xnc_logger.debug m.logger
                            ("Initializing page " ^ m.page_name) in
                 let constructor =
                   Page_behavior.Factory.find_page_constructor
                     m.page_factory
                     m.page_name in
                 let () = Xnc_logger.debug m.logger
                            ("Constructor for page " ^
                             m.page_name ^ " found") in
                 let page_behavior =
                   constructor
                     { Srv_page_state.cgi = m.cgi_context;
                       Srv_page_state.state_serialization = m.serialized_state;
                       Srv_page_state.is_self_submitted = m.is_self_submitted
                     } in
                 let () = Xnc_logger.debug m.logger
                            (Printf.sprintf "Page constructed \
                                             (parameters to load=%B, \
                                             session to load=%B, \
                                             self submitted=%B)"
                               (parameters_to_load m)
                               m.session_to_load
                               m.is_self_submitted) in
                 let () = page_behavior # trace "init" in
                 let () = page_behavior # init () in
                 let () = Xnc_logger.debug m.logger "Page initialized" in
                 let target_state =
                   if m.session_to_load then
                     Ready_to_load_session
                   else
                     Initialized
                 in
                   { m with current_state = target_state;
                            page_behavior = Some page_behavior
                   }
             | _ -> raise Invalid_transition)

      | Ready_to_load_session ->
          (match event with
               Load_session ->
                 let pb = Option.get m.page_behavior in
                   pb # invoke_callbacks Srv_types.Pre_session_loading;
                   Xnc_logger.debug m.logger
                     "Pre_session_loading callbacks invoked";
                   pb # trace "load_session";
                   pb # load_session ();
                   Xnc_logger.debug m.logger "Session loaded";
                   pb # invoke_callbacks Srv_types.Post_session_loading;
                   Xnc_logger.debug m.logger
                     "Post_session_loading callbacks invoked";
                   { m with current_state = Initialized }
             | _ -> raise Invalid_transition)

      | Initialized ->
          (match event with
               Init_page_parameters ->
                 let pb = Option.get m.page_behavior in
                   pb # invoke_callbacks Srv_types.Pre_parameter_init;
                   Xnc_logger.debug m.logger
                     "Pre_parameter_init callbacks invoked";
                   pb # trace "set_page_parameters";
                   pb # set_page_parameters (Option.get m.page_parameters);
                   Xnc_logger.debug m.logger "Page parameters set";
                   pb # invoke_callbacks Srv_types.Post_parameter_init;
                   Xnc_logger.debug m.logger
                     "Post_parameter_init callbacks invoked";
                   { m with current_state = State_loaded }
             | Load_page_parameters ->
                 let pb = Option.get m.page_behavior in
                   pb # invoke_callbacks Srv_types.Pre_parameter_init;
                   Xnc_logger.debug m.logger
                     "Pre_parameter_init callbacks invoked";
                   pb # trace "load_page_parameters";
                   pb # load_page_parameters ();
                   Xnc_logger.debug m.logger "Page parameters loaded";
                   pb # invoke_callbacks Srv_types.Post_parameter_init;
                   Xnc_logger.debug m.logger
                     "Post_parameter_init callbacks invoked";
                   let target_state =
                     if m.is_self_submitted then
                       Ready_to_validate_user_input 
                     else
                       State_loaded
                   in
                     { m with current_state = target_state }
             | _ -> raise Invalid_transition)

      | Ready_to_validate_user_input ->
          (match event with
               Validate_user_input ->
                 let pb = Option.get m.page_behavior in
                   pb # invoke_callbacks Srv_types.Pre_user_input_validation;
                   Xnc_logger.debug m.logger
                     "Pre_user_input_validation callbacks invoked";
                   pb # trace "build_validator_collection";
                   pb # build_validator_collection ();
                   Xnc_logger.debug m.logger "Validator collection built";
                   pb # invoke_callbacks Srv_types.Post_user_input_validation;
                   Xnc_logger.debug m.logger
                     "Post_user_input_validation callbacks invoked";
                   let is_valid = pb # is_valid () in
                   Xnc_logger.debug m.logger ("Page valid: " ^
                                          (string_of_bool is_valid));
                   let target_state =
                     if is_valid then
                       Ready_to_read_user_input 
                     else
                       State_loaded
                   in
                     { m with current_state = target_state;
                              valid = is_valid
                     }
             | _ -> raise Invalid_transition)

      | Ready_to_read_user_input ->
          (match event with
               Read_user_input ->
                 let pb = Option.get m.page_behavior in
                   pb # invoke_callbacks Srv_types.Pre_user_input_reading;
                   Xnc_logger.debug m.logger
                     "Pre_user_input_reading callbacks invoked";
                   pb # trace "read_user_input";
                   pb # read_user_input ();
                   Xnc_logger.debug m.logger "User input read";
                   pb # invoke_callbacks Srv_types.Post_user_input_reading;
                   Xnc_logger.debug m.logger
                     "Post_user_input_reading callbacks invoked";
                   { m with current_state = State_loaded }
             | _ -> raise Invalid_transition)

      | State_loaded ->
          (match event with
               Build_page_structure ->
                 let pb = Option.get m.page_behavior in
                 let ps = pb # get_page_structure in
                   pb # invoke_callbacks Srv_types.Pre_structure_building;
                   Xnc_logger.debug m.logger
                     "Pre_structure_building callbacks invoked";
                   pb # trace "build_page_structure";
                   pb # build_page_structure ();
                   if ps # get_debug then
                     begin
                       pb # add_debug_info ();
                       Xnc_logger.debug m.logger "Debug info added";
                     end
                   else
                     ();
                   Xnc_logger.debug m.logger "Page structure built";
                   pb # invoke_callbacks Srv_types.Post_structure_building;
                   Xnc_logger.debug m.logger
                     "Post_structure_building callbacks invoked";
                   let target_state =
                     if m.is_self_submitted then
                       Ready_to_handle_ui_events
                     else
                       Ready_to_save_state
                   in
                     { m with current_state = target_state }
             | _ -> raise Invalid_transition)

      | Ready_to_handle_ui_events ->
          (match event with
               Handle_ui_events ->
                 let pb = Option.get m.page_behavior in
                   pb # invoke_callbacks Srv_types.Pre_ui_event_handling;
                   Xnc_logger.debug m.logger
                     "Pre_ui_event_handling callbacks invoked";
                   pb # trace "handle_ui_events";
                   pb # handle_ui_events ();
                   Xnc_logger.debug m.logger "UI events handled";
                   pb # invoke_callbacks Srv_types.Post_ui_event_handling;
                   Xnc_logger.debug m.logger
                     "Post_ui_event_handling callbacks invoked";
                   { m with current_state = Ready_to_save_state }
             | _ -> raise Invalid_transition)

      | Ready_to_save_state ->
          (match event with
               Save_state ->
                 let pb = Option.get m.page_behavior in
                   pb # invoke_callbacks Srv_types.Pre_state_saving;
                   Xnc_logger.debug m.logger
                     "Pre_state_saving callbacks invoked";
                   pb # trace "save_state";
                   pb # save_state ();
                   Xnc_logger.debug m.logger "Page state saved";
                   pb # invoke_callbacks Srv_types.Post_state_saving;
                   Xnc_logger.debug m.logger
                     "Post_state_saving callbacks invoked";
                   { m with current_state = State_saved }
             | _ -> raise Invalid_transition)

      | State_saved ->
          (match event with
               Handle_page_change ->
                 let pb = Option.get m.page_behavior in
                 let targets = pb # get_targets in
                 let _ =
                   Xnc_logger.debug m.logger
                     ("Page valid flag: " ^ (string_of_bool m.valid)) in
                 let target_page =
                   if m.valid then
                     targets.Page_behavior.standard
                   else
                     targets.Page_behavior.invalid_input
                 in
                   begin match target_page with
                        Page_behavior.Self ->
                          Xnc_logger.debug m.logger "Target page: Self";
                          { m with current_state = Ready_to_render }
                      | Page_behavior.Goto (target_name,
                                            target_parameters) ->
                          Xnc_logger.debug m.logger
                            ("Target page: " ^ target_name);
                          begin match targets.Page_behavior.redirect_mode with
                              Srv_types.Server_side ->
                                { m with current_state = Uninitialized;
                                         page_name = target_name;
                                         page_parameters =
                                           (Some target_parameters);
                                         is_self_submitted = false
                                }
                            | Srv_types.Client_side url_rendering_mode ->
                                let url = pb # build_url
                                            target_name
                                            target_parameters
                                            url_rendering_mode
                                in
                                  Xnc_logger.debug m.logger
                                    ("Target URL: " ^ url);
                                  (* Does not work with nethttpd in Ocamlnet
                                   * 2.2.9
                                  m.cgi_context # set_redirection_header url; *)
                                  (* Workaround *)
                                  let env = m.cgi_context # environment in
                                  env # set_output_header_fields [];
                                  env # set_output_header_field "Location" url;
                                  env # set_status `Found;
                                  { m with current_state = Rendered }
                          end
                   end
             | _ -> raise Invalid_transition)

      | Ready_to_render ->
          (match event with
               Render ->
                 let pb = Option.get m.page_behavior in
                   pb # invoke_callbacks Srv_types.Pre_render;
                   Xnc_logger.debug m.logger "Pre_render callbacks invoked";
                   pb # trace "render";
                   pb # render ();
                   Xnc_logger.debug m.logger "Page rendered";
                   pb # invoke_callbacks Srv_types.Post_render;
                   Xnc_logger.debug m.logger "Post_render callbacks invoked";
                   { m with current_state = Rendered }
             | _ -> raise Invalid_transition)

      | Rendered ->
          (match event with
               Complete ->
                 let pb = Option.get m.page_behavior in
                   pb # invoke_callbacks Srv_types.Pre_completion;
                   Xnc_logger.debug m.logger "Pre_completion callbacks invoked";
                   pb # trace "complete";
                   pb # complete ();
                   Xnc_logger.debug m.logger "Page completed";
                   pb # invoke_callbacks Srv_types.Post_completion;
                   Xnc_logger.debug m.logger
                     "Post_completion callbacks invoked";
                   { m with current_state = Completed }
             | _ -> raise Invalid_transition)

      | _ -> raise Invalid_transition

  let next_step m =
    match m.current_state with
        Uninitialized ->
          do_transition m Initialize
      | Ready_to_load_session ->
          do_transition m Load_session
      | Initialized ->
          if parameters_to_load m then
            do_transition m Load_page_parameters
          else
            do_transition m Init_page_parameters
      | Ready_to_validate_user_input ->
          do_transition m Validate_user_input
      | Ready_to_read_user_input ->
          do_transition m Read_user_input
      | State_loaded ->
          do_transition m Build_page_structure
      | Ready_to_handle_ui_events ->
          do_transition m Handle_ui_events
      | Ready_to_save_state ->
          do_transition m Save_state
      | State_saved ->
          do_transition m Handle_page_change
      | Ready_to_render ->
          do_transition m Render
      | Rendered ->
          do_transition m Complete
      | Completed ->
          (* TODO: 
          if Xnc_logger.is_debug m.logger then
            begin
              Xnc_logger.debug m.logger "-- Page trace --";
              Xnc_logger.flush_to_stderr
                ((Option.get m.page_behavior) # get_logger);
              Xnc_logger.debug m.logger "-- Page trace end --";
            end;
           *)
          raise Final_state_reached

end

