(** Page FSM.
  *
  * This module implements page FSM. *)

exception Invalid_transition
(** Raised when an unexpected event is received in a specific state. *)

exception Final_state_reached
(** Raised when the FSM reaches the termination state. *)

module type PAGE_STATE_MACHINE =
sig
  type state
  (** The type of the states. *)

  type event
  (** The type of the events. *)

  type page_parameter_type
  (** The type of page parameters. *)

  type page_factory_type
  (** The type of page factory. *)

  type t
  (** The type of page FSM. *)

  val create :
    page_factory_type ->
    Xnc_logger.t ->
    Netcgi.cgi_activation ->
    Srv_state_persisting.Serialized.t ->
    string ->
    page_parameter_type option ->
    bool ->
    bool ->
    t
  (** [create f l c n p lp ls] creates a page FSM with page factory [f], logger
    * [l], CGI activation [c], page name [n], page parameters [p], if parameters
    * are to be loaded [lp], and if session is to be loaded [ls]. *)

  val current_state : t -> state
  (** [current_state m] returns the current state of FSM [m]. *)

  val do_transition : t -> event -> t
  (** [do_transition m e] sends event [e] to FSM [m] and executes the proper
    * state transition. *)

  val next_step : t -> t
  (** [next_step m] advances the FSM [m] of one step and returns the new state.
    *)

end
(** Module signature of page FSM. *)

module Make :
  functor (Page_behavior : Srv_page_behavior.PAGE_BEHAVIOR) ->
sig
  type _state =
      Uninitialized
    | Ready_to_load_session
    | Initialized
    | Ready_to_validate_user_input
    | Ready_to_read_user_input
    | State_loaded
    | Ready_to_handle_ui_events
    | Ready_to_save_state
    | State_saved
    | Ready_to_render
    | Rendered
    | Completed

  type _event =
      Initialize
    | Load_session
    | Init_page_parameters
    | Load_page_parameters
    | Validate_user_input
    | Read_user_input
    | Build_page_structure
    | Handle_ui_events
    | Handle_page_change
    | Save_state
    | Render
    | Complete

  include PAGE_STATE_MACHINE
    with type state = _state and
      type event = _event and
      type page_parameter_type = Page_behavior.page_parameter_type and 
      type page_factory_type = Page_behavior.Factory.t

end
(** This functor builds concrete page FSM modules. *)

