(* private *)
let doctype dtd =
  match dtd with
      "xhtml11" ->
        "<!DOCTYPE html \
              PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \
              \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">"
    | "xhtml1-strict" ->
        "<!DOCTYPE html \
              PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \
              \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">"
    | "xhtml1-transitional" ->
        "<!DOCTYPE html \
              PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \
              \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">"
    | "xhtml1-frameset" ->
        "<!DOCTYPE html \
              PUBLIC \"-//W3C//DTD XHTML 1.0 Frameset//EN\" \
              \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd\">"
    | _ -> assert false
;;

(* TODO: Set up tree data structure to contain page controls
 *       Style management
 *       Other HEAD fields (e.g. META)
 *       Encoding *)
class page_structure (conf : Srv_configuration.t) =
object(self)
  inherit Srv_page_element.container_element as super
  inherit Srv_page_element.Attributes.i18n as i18n

  (* Attributes *)
  val mutable xml_version = "1.0"
  val mutable encoding = "ISO-8859-1"
  val mutable dtd = "xhtml11"
  val mutable debug = false
  (* Special elements *)
  val mutable body_element : Srv_page_element.body_element option = None
  val mutable form_element : Srv_page_element.form_element option = None
  val mutable cookies : Netcgi.Cookie.t list = []
  val mutable http_status : Nethttp.http_status = `Ok

  (* Getters *)
  method get_xml_version = xml_version
  method get_encoding = encoding
  method get_dtd = dtd
  method get_debug = debug
  method get_http_status = http_status

  (* Setters *)
  method set_xml_version value = xml_version <- value
  method set_encoding value = encoding <- value
  method set_dtd value = dtd <- value
  method set_debug value = debug <- value
  method set_http_status value = http_status <- value

  method create_body_element () =
    let body = new Srv_page_element.body_element in
      self # add_element (body :> Srv_page_element.page_element);
      body_element <- (Some body);
      body

  method create_form_element (cgi : Netcgi.cgi_activation) =
    let action =
      if Srv_types.is_nethttpd cgi then
        Srv_configuration.netplex_nethttpd_path conf
      else
        cgi # environment # cgi_script_name in
    let form = new Srv_page_element.form_element action in
      self # append_to_body (form :> Srv_page_element.page_element);
      form_element <- (Some form);
      form

  method add_hidden_field hidden_field =
    Option.may
      (fun form -> form # add_hidden_field hidden_field)
      form_element

  method add_cookie cookie =
    cookies <- cookie :: cookies
      
  method append_to_body element =
    let body = Option.get body_element in
      body # add_element element

  method render_header (cgi : Netcgi.cgi_activation) =
    cgi # set_header 
      ~status:http_status
      ~cache:`No_cache 
      ~content_type:("text/html; charset=\"" ^ encoding ^ "\"")
      ~set_cookies:cookies
      ()

  method render (out : Netchannels.trans_out_obj_channel) =
    out # output_string "<?xml version=\"";
    out # output_string xml_version;
    out # output_string "\" encoding=\"";
    out # output_string encoding;
    out # output_string "\"?>";
    out # output_string (doctype dtd);
    out # output_string "<html xmlns=\"http://www.w3.org/1999/xhtml\"";
    i18n # render_attributes out;
    out # output_string ">";
    super # render out;
    out # output_string "</html>";

end;;

type t = page_structure;;

let create conf =
  new page_structure conf
;;

let render_header p cgi =
  p # render_header cgi
;;

let render p (out : Netchannels.trans_out_obj_channel) =
  p # render out
;;

let handle_ui_events p (cgi : Netcgi.cgi_activation) =
  p # handle_ui_events cgi
;;

let add_element e element =
  e # add_element (element :> Srv_page_element.page_element)
;;

let create_body_element p =
  p # create_body_element ()
;;

let create_form_element p (cgi : Netcgi.cgi_activation) =
  p # create_form_element cgi
;;

let add_hidden_field p hidden_field =
  p # add_hidden_field (hidden_field :> Srv_page_element.hidden_field)
;;

let add_simple_hidden_field p name value =
  let hidden_field = new Srv_page_element.simple_hidden_field name value in
    add_hidden_field p hidden_field
;;

let add_complex_hidden_field ?serializer p name field_collection =
  let hidden_field = new Srv_page_element.complex_hidden_field
                       ?serializer
                       name field_collection in
    add_hidden_field p hidden_field
;;

let add_cookie p cookie cgi =
  try
    let old_cookie = cgi # environment # cookie (Netcgi.Cookie.name cookie) in
    let old_value = Netcgi.Cookie.value old_cookie in
      if old_value <> (Netcgi.Cookie.value cookie) then
        p # add_cookie cookie
      else
        ()
  with Not_found ->
    p # add_cookie cookie
;;

let add_simple_cookie ?max_age p name value cgi =
  let cookie = Netcgi.Cookie.make ?max_age name value in
    add_cookie p cookie cgi
;;

let add_complex_cookie ?serializer ?max_age p name field_collection cgi =
  let value = Xnc_field_collection.to_string
                ?serializer
                field_collection in
    add_simple_cookie ?max_age p name value cgi
;;

let append_to_body p element =
  p # append_to_body (element :> Srv_page_element.page_element)
;;

let set_http_status p v =
  p # set_http_status v
;;

