(** Page structure.
  *
  * This module implements the tree structure of pages. *)

class page_structure :
  Srv_configuration.t ->
object
  inherit Srv_page_element.container_element
  inherit Srv_page_element.Attributes.i18n

  method get_xml_version : string
  (** [get_xml_version] returns the XML version of the page. *)

  method get_encoding : string
  (** [get_encoding] returns the encoding of the page. *)

  method get_dtd : string
  (** [get_dtd] returns the DTD of the page. *)

  method get_debug : bool
  (** [get_debug] returns the debug flag of the page. *)

  method get_http_status : Nethttp.http_status
  (** [get_http_status] returns the current HTTP status that will be rendered
    * in the header. *)

  method set_xml_version : string -> unit
  (** [set_xml_version v] sets the XML version of the page to [v]. *)

  method set_encoding : string -> unit
  (** [set_encoding v] sets the encoding of the page to [v]. *)

  method set_dtd : string -> unit
  (** [set_encoding v] sets the DTD of the page to [v]. *)

  method set_debug : bool -> unit
  (** [set_debug v] sets the debug flag of the page to [v]. *)

  method set_http_status : Nethttp.http_status -> unit
  (** [set_http_status v] sets the current HTTP status of the page to [v]. *)

  method create_body_element : unit -> Srv_page_element.body_element
  (** [create_body_element ()] creates a body element, appends it to page
    * structure, and returns it to the caller. *)

  method create_form_element :
    Netcgi.cgi_activation -> Srv_page_element.form_element
  (** [create_form_element ()] creates a form element, appends it to page
    * structure, and returns it to the caller. *)

  method add_hidden_field : Srv_page_element.hidden_field -> unit
  (** [add_hidden_field h] appends hidden field [h] to the form element if
    * previously created, otherwise it does nothing. *)

  method add_cookie : Netcgi.Cookie.t -> unit
  (** [add_cookie c] appends cookie [c] to the cookie list. *)

  method append_to_body : Srv_page_element.page_element -> unit
  (** [append_to_body e] appends page element [e] to the body element if
    * previously created, otherwise it raises [No_value]. *)

  method render_header : Netcgi.cgi_activation -> unit
  (** [render_header c] sets HTTP header of CGI activation [c]. *)

end
(** This class implements the page element that represents the whole page. *)

type t = page_structure
(** Alias used by non-OO interface. *)

val create : Srv_configuration.t -> t
(** [create c] returns a new page structure using configuration [c]. *)

val render_header : t -> Netcgi.cgi_activation -> unit
(** [render_header p c] sets HTTP header of [c] using data of [p]. *)

val render : t -> Netchannels.trans_out_obj_channel -> unit
(** [render p o] renders page structure [p] to output channel [o]. *)

val handle_ui_events : t -> Netcgi.cgi_activation -> unit
(** [handle_ui_events p c] invokes UI event handlers of [p] generated from [c].
  *)

val add_element :
  #Srv_page_element.page_element -> #Srv_page_element.page_element -> unit
(** [add_element f c] appends page element [c] to page element [f]. *)

val create_body_element : t -> Srv_page_element.body_element
(** [create_body_element p] creates body element, appends it to [p], and returns
  * it to the caller. *)

val create_form_element :
  t -> Netcgi.cgi_activation -> Srv_page_element.form_element
(** [create_form_element p c] creates form element using [c], appends it to [p],
  * and returns it to the caller. *)

val add_hidden_field : t -> #Srv_page_element.hidden_field -> unit
(** [add_hidden_field p h] appends hidden field [h] to form element of page
  * structure [p] if present, otherwise it does nothing. *)

val add_simple_hidden_field : t -> string -> string -> unit
(** [add_simple_hidden_field p n v] adds a hidden field with name [n] and value
  * [v] to page structure [p]. *)

val add_complex_hidden_field :
  ?serializer:string Xnc_serializing.serializer ->
  t -> string -> Xnc_field_collection.t -> unit
(** [add_complex_hidden_field p n f] adds a hidden field with name [n] and
  * value obtained from [f] to page structure [p].
  *
  * @param serializer Optional custom serializer. *)

val add_cookie : t -> Netcgi.Cookie.t -> Netcgi.cgi_activation -> unit
(** [add_cookie p c cgi] appends cookie [c] to page structure [p], using cgi
  * context [cgi] to check if cookie [c] is already present. If it is, the
  * cookie will not be added. *)

val add_simple_cookie :
  ?max_age:int ->
  t -> string -> string -> Netcgi.cgi_activation -> unit
(** [add_simple_cookie p n v c] adds a cookie with name [n] and value [v] to
  * page structure [p], using cgi context [c].
  *
  * @param max_age sets the lifetime of the cookie to [max_age] seconds *)

val add_complex_cookie :
  ?serializer:string Xnc_serializing.serializer ->
  ?max_age:int ->
  t -> string -> Xnc_field_collection.t -> Netcgi.cgi_activation -> unit
(** [add_complex_cookie p n f c] adds a cookie with name [n] and value obtained
  * from [f] to page structure [p], using cgi context [c].
  *
  * @param serializer Optional custom serializer.
  * @param max_age sets the lifetime of the cookie to [max_age] seconds. *)

val append_to_body : t -> #Srv_page_element.page_element -> unit
(** [append_to_body p e] appends page element [e] to body element of page
  * structure [p]. Raises [No_value] if body element is not present in [p]. *)

val set_http_status : t -> Nethttp.http_status -> unit
(** [set_http_status p v] sets the current HTTP status of page [p] to [v]. *)

