module type REQUEST_PROCESSOR =
sig
  val set_up : unit -> Srv_runtime_context.context

  val register_signal_handlers :
    Srv_runtime_context.context -> unit

  val select_request_processor :
    Srv_runtime_context.context ->
    Netcgi.cgi_activation ->
    (Netcgi.cgi_activation -> unit)

end;;

module Make
  (Builder : Srv_runtime_context.CONTEXT_BUILDER)
  (Context : Srv_runtime_context.RUNTIME_CONTEXT)
  (Page_dispatcher : Srv_page_dispatcher.PAGE_DISPATCHER)
  (Page_factory_builder : Xnc_core.INITIALIZABLE
     with type t = Page_dispatcher.page_factory_type)
  (Resource_list_builder : Xnc_core.INITIALIZABLE
     with type t = Srv_resource_list.t) =
struct
  let set_up () =
    Context.update_context
      (fun _ ->
         Builder.create ())

  let register_signal_handlers context =
    let logger = context.Srv_runtime_context.logger in
    let config = context.Srv_runtime_context.configuration in

    let system_signal_handler signal_number = 
      (* TODO: log signal name, instead of system independent signal number *)
      Xnc_logger.info logger
        ("Received signal " ^ (string_of_int signal_number));
      (* Handle SIGHUP, SIGINT, SIGQUIT, SIGTERM, SIGUSR1 shutting down
       * the server *)
      Srv_configuration.clean_netplex_configuration config;
      exit 0
    in

    let user_signal_handler signal_number =
      (* TODO: log signal name, instead of system independent signal number *)
      Xnc_logger.info logger
        ("Received signal " ^ (string_of_int signal_number));
      (* Handle SIGUSR2 reloading the configuration *)
      ignore (Context.update_context
                (fun context ->
                   Builder.update context))
    in

      List.iter
        (fun n ->
           Sys.set_signal n (Sys.Signal_handle system_signal_handler))
        [Sys.sighup; Sys.sigquit; Sys.sigint; Sys.sigterm; Sys.sigusr1];
      Sys.set_signal Sys.sigusr2 (Sys.Signal_handle user_signal_handler);
      Sys.set_signal Sys.sigpipe Sys.Signal_ignore

  let select_request_processor
        (context : Srv_runtime_context.context)
        (cgi : Netcgi.cgi_activation) =
    let process_page_request
          (cgi : Netcgi.cgi_activation) =
      let page_factory = Page_factory_builder.create () in
      let resource_list = Resource_list_builder.create () in
      let context = Context.get_current () in
      let logger = context.Srv_runtime_context.logger in
      let config = context.Srv_runtime_context.configuration in
      let update_statistics increment =
        let new_context =
          Context.update_context
            (fun context ->
               let statistics = context.Srv_runtime_context.statistics in
               let new_statistics = increment statistics in
                 { context with
                       Srv_runtime_context.statistics = new_statistics
                 })
        in
          new_context.Srv_runtime_context.statistics
      in

      let serialized_state = Srv_state_persisting.Serialized.retrieve_state
                               cgi
                               config
                               resource_list
      in
        try
          let new_statistics =
            update_statistics Srv_statistics.increment_total_requests
          in
            Xnc_logger.debug logger "-- MARK --";
            Xnc_logger.debug logger
              ("Serving request number " ^
               (string_of_int new_statistics.Srv_statistics.total_requests));
            Page_dispatcher.serve_page
              page_factory
              cgi
              config
              logger
              serialized_state
        with
            Srv_page_behavior.Server_shutdown as e ->
              raise e
          | error ->
              let new_statistics =
                update_statistics Srv_statistics.increment_total_errors in
              let () =
                Xnc_logger.debug logger
                  ("Unhandled exception number " ^
                   (string_of_int
                      new_statistics.Srv_statistics.total_errors))
              in
                Page_dispatcher.serve_error_page
                  page_factory
                  cgi
                  config
                  logger
                  serialized_state
                  error
    in
      process_page_request

end;;

