type context =
    { logger : Xnc_logger.t;
      application_values : Srv_values.application_values;
      command_line : Srv_command_line.args;
      configuration : Srv_configuration.t;
      statistics : Srv_statistics.counters
    }

let default_context =
  { logger = Xnc_logger.create ~buffered:false ();
    application_values = Srv_values.default_values;
    command_line = Srv_command_line.default_args;
    configuration = Srv_configuration.default_conf;
    statistics = Srv_statistics.default_counters
  }
;;

module type CONTEXT_BUILDER =
sig
  val create : unit -> context
  val update : context -> context

end;;

module Default_context_builder =
struct
  let create () =
    let context = default_context in
    let logger = context.logger in
    (** TODO: pass logger to parse function *)
    let command_line = Srv_command_line.parse
                         context.application_values in
    let application_name = command_line.Srv_command_line.application_name in
    let () = Xnc_logger.info logger
               (Printf.sprintf "Starting server %s (pid = %d)"
                  application_name
                  (Unix.getpid ())) in
    let () = Xnc_logger.debug logger "Command line processed" in
    (** TODO: pass logger to load function *)
    let conf = Srv_configuration.load
                 context.application_values
                 command_line in
    let () = Xnc_logger.debug logger "Server configuration loaded" in
      { context with
            command_line = command_line;
            configuration = conf
      }

  let update context =
    let logger = context.logger in
    let conf = Srv_configuration.load
                 context.application_values
                 context.command_line in
    let () = Xnc_logger.debug logger "Server configuration reloaded" in
      { context with
            configuration = conf
      }

end;;

module type RUNTIME_CONTEXT =
sig
  module Make_singleton :
    functor (Initializable : Xnc_core.INITIALIZABLE) ->
  sig
    type t = Initializable.t

    val get_instance : unit -> t
    val update_instance : t -> unit
    val clear_instance : unit -> unit

  end

  module Make_repository :
    functor (Conf : Xnc_cache_manager.CACHE_CONF) ->
  sig
    type key = Conf.key
    type item = Conf.item

    val get : (unit -> item) -> key -> item
    val update : key -> item -> unit
    val clear : unit -> unit
    val with_lock : (unit -> 'a) -> 'a

  end

  val get_current : unit -> context
  val update_current : context -> unit
  val update_context : (context -> context) -> context
  val clear : unit -> unit
  val with_lock : (unit -> 'a) -> 'a

  module Field_collection_repository :
  sig
    val get :
      (unit -> Xnc_field_collection.t) -> string -> Xnc_field_collection.t
    val update : string -> Xnc_field_collection.t -> unit

  end

end;;

module Make
  (Lockable : Xnc_core.LOCKABLE)
  (Scheduler : Xnc_scheduler.SCHEDULER) =
struct
  module Make_singleton
    (Initializable : Xnc_core.INITIALIZABLE) =
  struct
    type t = Initializable.t

    module Singleton =
      Xnc_singleton.Make(Initializable)(Lockable)

    let get_instance = Singleton.get_instance

    let update_instance = Singleton.update_instance

    let clear_instance = Singleton.clear_instance

  end

  module Make_repository
    (Conf : Xnc_cache_manager.CACHE_CONF) =
  struct
    type key = Conf.key
    type item = Conf.item

    module Cache_manager =
        Xnc_cache_manager.Make(Conf)(Lockable)(Scheduler)

    let get create key =
      let cache = Cache_manager.get_cache () in
        match Cache_manager.Cache.peek cache key with
            None ->
              let value = create () in
                Cache_manager.Cache.insert cache key value;
                value
          | Some value ->
              value

    let update key item =
      let cache = Cache_manager.get_cache () in
        Cache_manager.Cache.insert cache key item
    
    let clear () =
      let cache = Cache_manager.get_cache () in
        Cache_manager.Cache.clear cache

    let with_lock f =
      let cache = Cache_manager.get_cache () in
        Cache_manager.Cache.with_cache_lock cache f

  end

  module Cache_conf =
  struct
    type key = string
    type item =
        Context of context
      | Field_collection of Xnc_field_collection.t
    let interval = 1.

  end

  module Repository = Make_repository(Cache_conf)

  let context_key = "context"

  let get_current () =
    let value =
      Repository.get
        (fun () -> Cache_conf.Context default_context)
        context_key
    in
      match value with
          Cache_conf.Context c -> c
        | _ -> assert false

  let update_current context =
    Repository.update context_key (Cache_conf.Context context)
  
  let update_context update =
    let context = get_current () in
    let new_context = update context in
      update_current new_context;
      new_context

  let clear = Repository.clear

  let with_lock = Repository.with_lock

  module Field_collection_repository =
  struct
    let get create key =
      let value =
        Repository.get
          (fun () -> Cache_conf.Field_collection (create ()))
          key
    in
      match value with
          Cache_conf.Field_collection fc -> fc
        | _ -> assert false

    let update key value =
      Repository.update key (Cache_conf.Field_collection value)
    
  end

end;;

