(** Runtime context.
  *)

type context =
    { logger : Xnc_logger.t;
      application_values : Srv_values.application_values;
      command_line : Srv_command_line.args;
      configuration : Srv_configuration.t;
      statistics : Srv_statistics.counters
    }
(** Type of runtime context. *)

val default_context : context
(** Default context. *)

module type CONTEXT_BUILDER =
sig
  val create : unit -> context
  (** [create ()] creates the runtime context. It is called at startup time. *)

  val update : context -> context
  (** [update c] returns an updated version of the runtime context. It is
    * called, for example, when the application needs to read an updated version
    * of the configuration file. *)

end
(** Signature of context builder module. *)

module Default_context_builder :
sig
  val create : unit -> context
  (** [create ()] initializes the server context. This function is called when
    * the server starts. *)

  val update : context -> context
  (** [update c] updates the context. This function is called when the user
    * requests a context update. *)

end
(** This module provides the default implementation of the context builder. *)

module type RUNTIME_CONTEXT =
sig
  module Make_singleton :
    functor (Initializable : Xnc_core.INITIALIZABLE) ->
  sig
    type t = Initializable.t

    val get_instance : unit -> t
    (** [get_instance ()] returns the stored instance. The instance is created
      * at first invocation. *)

    val update_instance : t -> unit
    (** [update_instance v] updates the stored instance with value [v]. *)

    val clear_instance : unit -> unit
    (** [clear_instance ()] resets the stored instance to an undefined value. *)

  end
  (** This functor builds singletons. *)

  module Make_repository :
    functor (Conf : Xnc_cache_manager.CACHE_CONF) ->
  sig
    type key = Conf.key
    type item = Conf.item

    val get : (unit -> item) -> key -> item
    (** [get c k] returns an item stored with key [k] into the global
      * repository, using [c] to create the item, if it is not already present
      * in the repository. *)

    val update : key -> item -> unit
    (** [update k v] updates the item stored with key [k] using value [v]. *)

    val clear : unit -> unit
    (** [clear ()] clears the global repository. *)

    val with_lock : (unit -> 'a) -> 'a
    (** [with_lock f] computes [f ()] protected by a lock on the global
      * repository. *)

  end
  (** This functor builds repositories to cache items. *)

  val get_current : unit -> context
  (** [get_current ()] returns the current instance of runtime context. If the
    * runtime context is not initialized, returns [default_context]. *)

  val update_current : context -> unit
  (** [update_current c] updates the current instance of runtime context with
    * [c]. *)

  val update_context : (context -> context) -> context
  (** [update_context f] updates the current instance of runtime context with
    * the return value of function [f], and returns the new context. *)

  val clear : unit -> unit
  (** [clear ()] clears the current runtime context. *)

  val with_lock : (unit -> 'a) -> 'a
  (** [with_lock f] applies [f] wrapped by a lock on the current context. *)

  module Field_collection_repository :
  sig
    val get :
      (unit -> Xnc_field_collection.t) -> string -> Xnc_field_collection.t
    (** [get c k] returns a field collection stored with key [k] into the global
      * repository, using [c] to create the field collection, if it is not
      * already present in the repository. *)

    val update : string -> Xnc_field_collection.t -> unit
    (** [update k v] updates the field collection stored with key [k] using
      * value [v]. *)

  end
  (** This module implements a repository to cache field collections. *)

end
(** Output signature. *)

module Make :
  functor (Lockable : Xnc_core.LOCKABLE) ->
  functor (Scheduler : Xnc_scheduler.SCHEDULER) ->
  RUNTIME_CONTEXT
(** Functor that builds a runtime context module. *)

