open Xnc_stdlib
open Xnc_functional
open Xnc_string_transform

module S3 = Sqlite3
module Pf = Printf

module SMap = Map.Make(struct type t = string let compare = String.compare end)
let (->>) map key = SMap.find key map
let (<<-) map (key, value) = SMap.add key value map

let index_from s pos c =
  try Some (String.index_from s pos c)
  with Not_found -> None
;;

exception Error of string
exception Invalid_string

let escape_sql s =
  if String.contains s '\000' then
    raise Invalid_string
  else
    let start = ref 0 in
    let size = String.length s in
    let b = Buffer.create 1024 in
      try
        while true do
          let pos = String.index_from s !start '\'' in
            Buffer.add_string b (String.sub s !start (pos - !start));
            Buffer.add_string b "''";
            start := pos + 1;
        done;
        Buffer.contents b;
      with
        | Not_found -> begin
          Buffer.add_string b (String.sub s !start (size - !start));
          Buffer.contents b
        end
;;

let escape_sql_quote s = Pf.sprintf "'%s'" (escape_sql s);;

type state_manager = S3.db

type sort_order =
  | No_order
  | By_ctime
  | By_ctime_desc
  | By_mtime
  | By_mtime_desc

type state =
    {
      key : string;
      data : string;
      ctime : float;
      mtime : float;
      client_ip : string option;
      client_user_agent : string option;
    }

let raise_wrap e = raise (Error (S3.Rc.to_string e));;

let (!!) opt = match opt with Some v -> v | None -> raise Not_found;;

let rec row_map ?(i=0) ?(map=SMap.empty) row headers =
  if i = Array.length row
  then map
  else
    let new_map = map <<- (headers.(i), row.(i)) in
      row_map ~i:(i+1) ~map:(new_map) row headers
;;

let wrap (f : string option SMap.t -> unit) =
  fun row headers ->
    let map = row_map row headers in
      f map
;;

let has_table db =
  let table_ok = ref false in
  let callback row = table_ok := true in
  let res = S3.exec_not_null_no_headers db ~cb:callback
    "SELECT name FROM sqlite_master
     WHERE  type='table' AND name = 'xnc_states';" in
    match res with
      | S3.Rc.OK -> !table_ok
      | e -> S3.db_close db |> ignore; raise_wrap e
;;

let rec create_index res idx_names db =
  match idx_names with
    | [] -> res
    | hd::tl ->
        let query =
          Pf.sprintf "CREATE INDEX %s_idx ON xnc_states (%s ASC)" hd hd in
        let qres = S3.exec_not_null_no_headers db ~cb:ignore query in
          match qres with
            | S3.Rc.OK -> create_index res tl db
            | e -> S3.db_close db |> ignore; raise_wrap e
;;

let create_table db =
  let res = S3.exec_not_null_no_headers db ~cb:ignore
    "CREATE TABLE xnc_states
     (
        key               TEXT    PRIMARY KEY NOT NULL,
        data              TEXT                NOT NULL,
        ctime             REAL                NOT NULL,
        mtime             REAL                NOT NULL,
        client_ip         TEXT                NULL,
        client_user_agent TEXT                NULL
     );" in
    match res with
      | S3.Rc.OK -> create_index db ["ctime"; "mtime"] db
      | e -> S3.db_close db |> ignore; raise_wrap e
;;

let connect ?(check_schema=true) filename =
  let db = S3.db_open filename in
    if check_schema
    then if has_table db then db else create_table db
    else db
;;

let disconnect state_mgr =
  if not (S3.db_close state_mgr)
  then raise (Error "Sqlite3.close called with busy database: retry later")
;;

let decode_string_option s =
  match s with
    | None -> None
    | Some s -> Some s
;;

let state_of_row row =
  {
    key = !!(row->>"key");
    data = !!(row->>"data");
    ctime = !!(row->>"ctime") |> float_of_string;
    mtime = !!(row->>"mtime") |> float_of_string;
    client_ip = row->>"client_ip" |> decode_string_option;
    client_user_agent = row->>"client_user_agent" |> decode_string_option;
  }
;;

let retrieve state_mgr key =
  let value = ref None in
  let callback row = value := Some (state_of_row row) in

  let query = "SELECT key, data, ctime, mtime, client_ip, client_user_agent
               FROM xnc_states WHERE key = '" ^ (escape_sql key) ^ "';" in
  let res = S3.exec state_mgr ~cb:(callback |> wrap) query in
    match res with
      | S3.Rc.OK -> !value
      | e -> raise_wrap e
;;

let remove state_mgr key =
  let query =
    "DELETE FROM xnc_states WHERE key = '" ^ (escape_sql key) ^ "'" in
  let res = S3.exec_not_null_no_headers state_mgr ~cb:ignore query in
    match res with
      | S3.Rc.OK -> ()
      | e -> raise_wrap e
;;

let insert_query key data ctime mtime client_ip client_user_agent =
  let key = escape_sql_quote key in
  let data = escape_sql_quote data in
  let ctime = Pf.sprintf "%0.17g" ctime in
  let mtime = Pf.sprintf "%0.17g" mtime in
  let client_ip = match client_ip with
    | Some s -> escape_sql_quote s
    | None -> "NULL" in
  let client_user_agent = match client_user_agent with
    | Some s -> escape_sql_quote s
    | None -> "NULL" in

    Pf.sprintf "INSERT INTO xnc_states (key, data, ctime, mtime,
                                        client_ip, client_user_agent)
                VALUES (%s, %s, %s, %s, %s, %s);"
      key data ctime mtime client_ip client_user_agent
;;

let update_query key data ctime mtime client_ip client_user_agent =
  let ctime = match ctime with
    | None -> ""
    | Some t -> Pf.sprintf ", ctime = %0.17g" t in
  let mtime = match mtime with
    | None -> Pf.sprintf ", mtime = %0.17g" (Unix.time ())
    | Some t -> Pf.sprintf ", mtime = %0.17g" t in
  let client_ip = match client_ip with
    | None -> ""
    | Some cl -> Pf.sprintf ", client_ip = %s" (escape_sql_quote cl) in
  let client_user_agent = match client_user_agent with
    | None -> ""
    | Some cl -> Pf.sprintf ", client_user_agent = %s" (escape_sql_quote cl) in

    "UPDATE xnc_states SET
   data = " ^ (escape_sql_quote data) ^ ctime ^ mtime ^ client_ip ^
      client_user_agent ^ "
   WHERE key = " ^ (escape_sql_quote key) ^ ";"
;;

let replace state_mgr client_ip client_user_agent ctime mtime key data =
  let q = update_query key data ctime mtime client_ip client_user_agent in
  let res = S3.exec_not_null_no_headers state_mgr ~cb:ignore q in
    match res with
      | S3.Rc.OK -> ()
      | e -> raise_wrap e
;;

let add_or_replace
    state_mgr
    ?(client_ip)
    ?(client_user_agent)
    ?(ctime)
    ?(mtime)
    key
    data =
  let ct, mt =
    match ctime, mtime with
      | None, None -> let t = Unix.time () in t, t
      | Some ct, None -> ct, Unix.time ()
      | None, Some mt -> Unix.time (), mt
      | Some ct, Some mt -> ct, mt in
  let q = insert_query key data ct mt client_ip client_user_agent in
  let res = S3.exec_not_null_no_headers state_mgr ~cb:ignore q in
    match res with
      | S3.Rc.OK -> ()
      | S3.Rc.CONSTRAINT ->
          replace state_mgr client_ip client_user_agent ctime mtime key data
      | e -> raise_wrap e
;;

let order_clause c =
  match c with
    | No_order -> ""
    | By_ctime -> " ORDER BY ctime ASC"
    | By_ctime_desc -> " ORDER BY ctime DESC"
    | By_mtime -> " ORDER BY mtime ASC"
    | By_mtime_desc -> " ORDER BY mtime DESC"
;;

let fold ?(order=No_order) f acc state_mgr =
  let current_acc = ref acc in
  let apply_f row = current_acc := f !current_acc (state_of_row row) in
  let res = S3.exec state_mgr ~cb:(apply_f |> wrap)
    ("SELECT * FROM xnc_states" ^ (order_clause order) ^ ";") in
    match res with
      | S3.Rc.OK -> !current_acc
      | e -> raise_wrap e
;;

let iter ?(order=No_order) f state_mgr =
  fold ~order:order (fun acc -> f) () state_mgr
;;

let dump ?(order=No_order) state_mgr =
  fold ~order:order (fun acc s -> s::acc) [] state_mgr |> List.rev
;;

let default_pred timeout state =
  if state.mtime +. timeout < Unix.time ()
  then true else false
;;

let garbage_collect ?(pred=(default_pred 600.0)) state_mgr =
  fold
    (
      fun acc state ->
        if pred state then (remove state_mgr state.key; acc + 1)
        else acc
    )
    0
    state_mgr
;;

