(** SQLite3 backend to store Ex-nunc sessions
  *
  * This module provides a simple and effective interface to store sessions into
  * a SQLite3 database. Session are stored as simple key/value pairs, both of
  * which are strings, plus ctime, mtime and optionally the client IP and user
  * agent informations.
  *)

exception Error of string
(** Raised when
  *   - the db schema is incorrect or the db is busy;
  *   - in case of errors in the query issued to SQLite3.
  *)

exception Invalid_string
(** Raised when keys or values contains ['\000'] characters, which are invalid
  * for SQLite3. *)

type state_manager
(** The type of the connection handler. *)

type sort_order =
  | No_order      (** No order *)
  | By_ctime      (** Order by creation time, ascending *)
  | By_ctime_desc (** Order by creation time, descending *)
  | By_mtime      (** Order by modification time, ascending *)
  | By_mtime_desc (** Order by modification time, descending *)
(** Enumeration used in iterations to sort pairs by ctime or mtime. *)

type state =
    {
      key : string;
      data : string;
      ctime : float;
      mtime : float;
      client_ip : string option;
      client_user_agent : string option;
    }
(** This record represents a "state", i.e. a session for Ex-nunc. *)

val connect : ?check_schema:bool -> string -> state_manager
(** [connect "filename"] connects to the SQLite3 DB and
  * creates it if it doesn't exist. The [check_schema] flag (default [true])
  * controls if the function have to check the schema. *)

val disconnect : state_manager -> unit
(** Disconnect the database. *)

val retrieve : state_manager -> string -> state option
(** [retrieve handler "key"] returns [Some session] if a binding with "key"
  * exists in the database, or [None] if such binding doesn't exist. *)

val remove : state_manager -> string -> unit
(** Removes an entry, given the key. If the key is not present, no errors are
  * issued. *)

val add_or_replace : state_manager -> ?client_ip:string ->
  ?client_user_agent:string -> ?ctime:float -> ?mtime:float -> string ->
  string -> unit
  (** [add_or_replace handler key value] adds the [key]/[value] binding into the
    * database. If the binding is present, the new [value] is binded to [key].
    * [client_ip] and [client_user_agent] defaults to [None]. [ctime] and
    * [mtime], if not present {b and} the binding is not present, take the
    * current time value, like in [Unix.time]. If the binding is present,
    * [ctime] is unchanged, while [mtime] is updated to the present moment. The
    * user can explicitly sets the two value and in this case no controls are
    * performed. *)

val fold : ?order:sort_order -> ('a -> state -> 'a) -> 'a ->
  state_manager -> 'a
(** Fold over the key/value collection in the desired order. [~order] defaults
  * to [No_order]. *)

val iter : ?order:sort_order -> (state -> unit) -> state_manager -> unit
(** Iter over the key/value collection in the desired order. [~order] defaults
  * to [No_order]. *)

val dump : ?order:sort_order -> state_manager -> state list
(** Dump the entire collection of states. [~order] defaults to [No_order]. *)

val default_pred : float -> state -> bool
(** The default predicate for garbage collection (see below).
  * The standard behaviour is to delete all sessions with a [mtime] older than
  * the passed timeout, with respect to now (e.g. [Unix.time ()]). *)

val garbage_collect : ?pred:(state -> bool) -> state_manager -> int
(** [garbage_collect ~pred:f handler] deletes all the states for which [f]
  * returns [true]. [~pred] defaults to [default_pred 600], i.e. deletes all
  * sessions older than 5 minutes. *)

