module type STATE =
sig
  module Data : Srv_types.SERIALIZABLE_DATA

  module State :
  sig
    type t

    val create :
      ?id:string option ->
      ?creation_time:float ->
      ?last_access:float ->
      Data.t ->
      string ->
      string ->
      t
    val get_id : t -> string option
    val get_data : t -> Data.t
    val get_client_ip : t -> string
    val get_user_agent : t -> string
    val get_creation_time : t -> float
    val get_last_access : t -> float
    val set_data : t -> Data.t -> t
    val generate_id : t -> unit
    val generate_id_once : t -> unit
  end

  module Manager :
  sig
    type t

    val create :
      Srv_types.manager_mode -> Srv_configuration.t -> t
    val state_to_field_collection :
      t -> State.t -> Xnc_field_collection.t
    val state_from_field_collection :
      t -> State.t -> Xnc_field_collection.t -> int -> State.t
    val persist :
      t -> State.t -> unit
    val load_from_serialized_state :
      ?get_from_repository:((unit -> Xnc_field_collection.t) ->
        Xnc_field_collection.t) ->
      ?with_lock:((unit -> State.t) -> State.t) ->
      t ->
      State.t ->
      Srv_state_persisting.Serialized.state_part ->
      Srv_configuration.t ->
      int ->
      State.t

  end

end

module Make
  (Serializable_data: Srv_types.SERIALIZABLE_DATA) =
struct
  module Data = Serializable_data

  (* TODO: use Context.Make_singleton functor *)
  module Random_generator_singleton =
    Xnc_singleton.Make(
      struct
        type t = Cryptokit.Random.rng

        let create () =
          Cryptokit.Random.pseudo_rng
            (Cryptokit.Random.string
               (* TODO: use secure_rng, avoiding Cryptokit error
                * No_entropy_source *)
               (Cryptokit.Random.device_rng "/dev/urandom")
               Srv_values.seed_length)

      end)
      (Xnc_core.Null_mutex)

  module State =
  struct
    (* TODO: store state access timestamp, to purge unused state *)
    type t =
        { mutable id : string option;
          data : Data.t;
          client_ip : string;
          user_agent : string;
          creation_time : float;
          mutable last_access : float
        }

    let create
          ?(id = None)
          ?creation_time
          ?last_access
          data
          client_ip
          user_agent =
      let current_time = Unix.gettimeofday () in
        { id = id;
          data = data;
          client_ip = client_ip;
          user_agent = user_agent;
          creation_time =
            begin match creation_time with
                None -> current_time
              | Some t -> t
            end;
          last_access =
            begin match last_access with
                None -> current_time
              | Some t -> t
            end;
        }

    let get_id state =
      state.id

    let get_data state =
      state.data

    let get_client_ip state =
      state.client_ip

    let get_user_agent state =
      state.user_agent

    let get_creation_time state =
      state.creation_time

    let get_last_access state =
      state.last_access

    let set_data state data =
      { state with
            data = data;
            last_access = Unix.gettimeofday ()
      }

    let generate_id state =
      let rng = Random_generator_singleton.get_instance () in
      (* TODO: check for uniqueness *)
      let random_id = Cryptokit.Random.string rng Srv_values.state_id_length in
        state.id <- Some (Xnc_string_transform.base16_encode random_id);
        state.last_access <- Unix.gettimeofday ()

    let generate_id_once state =
      match state.id with
          None ->
            generate_id state
        | Some _ ->
            ()

  end

  module Manager =
  struct
    type state_storage =
        Internal of (string option, State.t) Hashtbl.t
      | External
      | Sqlite3 of string

    type t =
        { mode : Srv_types.manager_mode;
          storage : state_storage
        }

    let label = "State"

    let create mode conf =
      let storage =
        match mode with
            Srv_types.Off
          | Srv_types.Client ->
              External
          | Srv_types.In_memory ->
              (* TODO: make initial size configurable *)
              let hash = Hashtbl.create 10 in
                (Internal hash)
          | Srv_types.Sqlite3 ->
              let sqlite3_path = Srv_configuration.sqlite3_path conf in
                (Sqlite3 sqlite3_path)
      in
        { mode = mode;
          storage = storage
        }

    (* private *)
    let get_descriptor manager data =
      match manager.mode with
          Srv_types.Off
        | Srv_types.Client ->
            Data.to_field_collection data
        | Srv_types.In_memory
        | Srv_types.Sqlite3 ->
            let fc = Xnc_field_collection.create label in
              Xnc_field_collection.register_serializable
                fc "id" (Xnc_serializable.wrap_option
                           (Some Xnc_serializable.default_string));
              fc

    (* private *)
    let get_internal_hashtable manager =
      match manager.storage with
          Internal hash ->
            hash
        | External
        | Sqlite3 _ ->
            failwith "Cannot retrieve state hashtable from non internal storage"

    (* private *)
    let get_sqlite3_handle manager =
      match manager.storage with
          Internal _
        | External ->
            failwith "Cannot retrieve state db from non DBM storage"
        | Sqlite3 db ->
            (* TODO: use check_schema? *)
            Srv_sqlite3_state.connect db

    (* private *)
    let check_state current_state loaded_state timeout =
      if State.get_client_ip current_state
           = State.get_client_ip loaded_state &&
         State.get_user_agent current_state
           = State.get_user_agent loaded_state then
        begin
          let current_time = Unix.gettimeofday () in
          let expiration_time = State.get_creation_time loaded_state +.
                                (float_of_int timeout) *. 60.
          in
            if current_time > expiration_time then
              (* TODO: log session expired (raise an exception?) *)
              current_state
            else
              loaded_state
        end
      else
        (* TODO: log non matching IP or User-Agent (raise an exception?) *)
        current_state

    let state_to_field_collection manager state =
      match manager.mode with
          Srv_types.Off
        | Srv_types.Client ->
            let data = State.get_data state in
              Data.to_field_collection data
        | Srv_types.In_memory
        | Srv_types.Sqlite3 ->
            let () = State.generate_id_once state in
            let id = State.get_id state in
            let fc = Xnc_field_collection.create label in
              Xnc_field_collection.register_option fc
                "id" id Xnc_serializable.wrap_string;
              fc

    let state_from_field_collection manager state field_collection timeout =
      try
        let loaded_state =
          match manager.mode with
              Srv_types.Off
            | Srv_types.Client ->
                let data = Data.from_field_collection
                             (State.get_data state)
                             field_collection in
                  State.set_data state data
            | Srv_types.In_memory ->
                let id = Xnc_field_collection.get_option
                           field_collection "id"
                           Xnc_serializable.unwrap_string in
                let hash = get_internal_hashtable manager in
                  Hashtbl.find hash id
            | Srv_types.Sqlite3 ->
                let id = Xnc_field_collection.get_option
                           field_collection "id"
                           Xnc_serializable.unwrap_string in
                let db = get_sqlite3_handle manager in
                let data = State.get_data state in
                let data_fc = Data.to_field_collection data in
                let db_data = Option.get
                                (Srv_sqlite3_state.retrieve
                                   db
                                   (Option.get id)) in
                let data_string = db_data.Srv_sqlite3_state.data in
                let () = Xnc_field_collection.from_string
                           data_fc
                           data_string in
                let data = Data.from_field_collection
                             data
                             data_fc in
                (* TODO: log if disconnect returns false *)
                let _ = Srv_sqlite3_state.disconnect db in
                  (* TODO: load client_ip, user_agent, ecc. from DB *)
                  State.create
                    ~id
                    ~creation_time:db_data.Srv_sqlite3_state.ctime
                    ~last_access:db_data.Srv_sqlite3_state.mtime
                    data
                    (Option.get db_data.Srv_sqlite3_state.client_ip)
                    (Option.get db_data.Srv_sqlite3_state.client_user_agent)
        in
          check_state state loaded_state timeout
      with
          Not_found
        | Option.No_value ->
            (* TODO:
             * - log that session was not found
             * - raise another exception to signal FSM that session was not
             * loadable*)
            state

    let persist manager state =
      match manager.mode with
          Srv_types.Off
        | Srv_types.Client ->
            ()
        | Srv_types.In_memory ->
            let hash = get_internal_hashtable manager in
            let () = State.generate_id_once state in
            let id = State.get_id state in
              Hashtbl.replace hash id state
        | Srv_types.Sqlite3 ->
            let db = get_sqlite3_handle manager in
            let () = State.generate_id_once state in
            let id = State.get_id state in
            let data = State.get_data state in
            let data_fc = Data.to_field_collection data in
            let data_string = Xnc_field_collection.to_string data_fc in
            let () = Srv_sqlite3_state.add_or_replace db
                       ~client_ip:(State.get_client_ip state)
                       ~client_user_agent:(State.get_user_agent state)
                       ~ctime:(State.get_creation_time state)
                       ~mtime:(State.get_last_access state)
                       (Option.get id)
                       data_string in
              (* TODO: log if disconnect returns false *)
              ignore (Srv_sqlite3_state.disconnect db)

    let load_from_serialized_state
          ?(get_from_repository = fun get_descriptor -> get_descriptor ())
          ?(with_lock = fun f -> f ())
          manager
          state
          serialized_state_part
          conf
          timeout =
      let fc = get_from_repository
                 (fun () ->
                    get_descriptor manager (State.get_data state)) in
      let serializer =
        Srv_state_persisting.get_serializer
          serialized_state_part.Srv_state_persisting.Serialized.location
          serialized_state_part.Srv_state_persisting.Serialized.part
          conf
      in
        with_lock
          (fun () ->
            Xnc_field_collection.from_string
              ~serializer
              fc
              serialized_state_part.Srv_state_persisting.Serialized.content;
            state_from_field_collection manager state fc timeout)

  end

end;;

