(** Server state.
  *
  * This module implements server-side state management. *)

module type STATE =
sig
  module Data : Srv_types.SERIALIZABLE_DATA
  (** This module contains the actual data. *)

  module State :
  sig
    type t
    (** The type of server state. *)

    val create :
      ?id:string option ->
      ?creation_time:float ->
      ?last_access:float ->
      Data.t ->
      string ->
      string ->
      t
    (** [create d ip ua] creates server state from data [d], storing also client
      * IP address [ip] and client user agent [ua].
      * @param id state ID
      * @param creation_time state creation time
      * @param last_access state last access time *)

    val get_id : t -> string option
    (** [get_id s] returns the state ID contained in [s]. *)

    val get_data : t -> Data.t
    (** [get_data s] returns the data contained in state [s]. *)

    val get_client_ip : t -> string
    (** [get_client_ip s] returns the client IP address in state [s]. *)

    val get_user_agent : t -> string
    (** [get_user_agent s] returns the client user agent in state [s]. *)

    val get_creation_time : t -> float
    (** [get_user_agent s] returns the creation timestamp of state [s]. *)

    val get_last_access : t -> float
    (** [get_last_access s] returns the last access timestamp of state [s]. *)

    val set_data : t -> Data.t -> t
    (** [set_data s d] returns the updated state replacing data contained in
      * state [s] with [d]. *)

    val generate_id : t -> unit
    (** [generate_id s] updates state [s] generating a random ID. *)

    val generate_id_once : t -> unit
    (** [generate_id_once s] updates state [s] generating a random ID, if and
      * only if [s] does not contain a valid ID. *)

  end
  (** This module implements a wrapper for state data to store additional
    * housekeeping information (e.g., ID). *)

  module Manager :
  sig
    type t
    (** The type of state managers. *)

    val create :
      Srv_types.manager_mode -> Srv_configuration.t -> t
    (** [create s c] creates a session manager with mode [s] using configuration
      * [c]. *)

    val state_to_field_collection :
      t -> State.t -> Xnc_field_collection.t
    (** [state_to_field_collection m s] returns the data structure used to
     * serialize state [s] using state manager [m]. *)

    val state_from_field_collection :
      t -> State.t -> Xnc_field_collection.t -> int -> State.t
    (** [state_from_field_collection m s f t] updates state [s] restoring data
      * from serialization format [f] using state manager [m] and timeout [t].
      *)

    val persist :
      t -> State.t -> unit
    (** [persist m s] saves state [s] to the storage. *)

    val load_from_serialized_state :
      ?get_from_repository:((unit -> Xnc_field_collection.t) ->
        Xnc_field_collection.t) ->
      ?with_lock:((unit -> State.t) -> State.t) ->
      t ->
      State.t ->
      Srv_state_persisting.Serialized.state_part ->
      Srv_configuration.t ->
      int ->
      State.t
    (** [load_from_serialized_state m s ss c t] updates state [s] restoring data
      * from serialized state [ss], using state manager [m], configuration [c],
      * and timeout [t].
      *
      * @param get_from_repository Retrieve state descriptor from a global
      * repository.
      * @param with_lock Wrapper to handle concurrent access. *)

  end
  (** This module implements a state manager that saves and restores state from
    * a specific storage. *)

end
(** Module signature of server session. *)

module Make :
  functor (Serializable_data : Srv_types.SERIALIZABLE_DATA) ->
sig
  include STATE
    with module Data = Serializable_data

end
(** This functor builds the implementation for a specific session data module.
  *)

