type part =
    Page_name
  | Session
  | Page_parameters
;;

type location =
    Argument_value
  | Query_string
  | Path_info
  | Cookie
  | Not_present
;;

(* private *)
let path_separator = "/";;
let null_string = "None";;
let session_prefix = "S";;
let session_element_separator = ",";;

(* private *)
class url_formatter () :
  [string] Xnc_formatting.formatter =
object(self)
  inherit Xnc_formatting.basic_formatter ()

  method format_null ?properties () =
    null_string

  method format_char ?properties c =
    self # format_string ?properties (Std.string_of_char c)

  method format_string ?properties s =
    let url_encoded = Netencoding.Url.encode s in
      self # format_raw_string ?properties url_encoded

end;;

(* private *)
class url_printer out_ch element_separator :
  [string] Xnc_printing.printer =
object
  inherit Xnc_printing.Netchannel_printer.text_printer out_ch

  method print_enum : 'a. 'a Enum.t -> ('a -> unit) -> unit =
    fun enum print_element ->
      Enum.iteri
        (fun i e ->
           if i > 0 then
             Xnc_printing.Non_pp.print_string pp element_separator
           else
             ();
           print_element e
        ) enum

  method print_field :
      'a. string -> string -> 'a -> Xnc_property.properties ->
      ('a -> unit) -> unit =
    fun prefix name value properties print_value ->
      print_value value

end

(* private *)
class url_scanner scanbuf element_separator :
  [string] Xnc_scanning.scanner =

let element_separator_regexp =
  Str.regexp ("[^" ^ element_separator ^ "]+") in

object(self)
  inherit Xnc_scanning.base_scanner ()

  method skip n =
    Xnc_scanning.Scanning_buffer.skip scanbuf n

  method private scan_value ?(properties = Xnc_property.empty) action =
    try
      let url_encoded =
        Xnc_scanning.Scanning_buffer.next_matching_string scanbuf
          element_separator_regexp
      in
        Netencoding.Url.decode url_encoded
    with Xnc_scanning.Scan_failure d ->
      raise (Xnc_scanning.Scan_failure (action ^ ": " ^ d))

end;;

(* private *)
class url_lexer scanbuf element_separator :
  [string] Xnc_lexing.lexical_analyzer =
object
  inherit Xnc_lexing.text_lexer ()

  method try_parse_null ?(properties = Xnc_property.empty) () =
    Xnc_scanning.Scanning_buffer.match_literal scanbuf null_string

  method parse_homogeneous_enum : 'a. (unit -> 'a) -> 'a Enum.t =
    fun parse_element ->
      failwith "Not supported method"

  method parse_heterogeneous_enum : 'a. 'a Enum.t -> ('a -> 'a) -> 'a Enum.t =
    fun enum_schema parse_element ->
      Enum.mapi 
        (fun i schema_element ->
           if i > 0 then
             Xnc_scanning.Scanning_buffer.parse_literal scanbuf
               element_separator
           else
             ();
           parse_element schema_element)
        enum_schema

  method parse_value_by_name :
      'a. string -> string -> Xnc_property.properties ->
      (unit -> 'a) -> 'a =
    fun prefix name properties parse_value ->
      parse_value ()

  method close () =
    Xnc_scanning.Scanning_buffer.close scanbuf

end;;

(* private *)
class url_serializer element_separator :
  [string] Xnc_serializing.serializer =
object(self)
  inherit Xnc_serializing.netchannel_serializer ()

  method to_netchannel serializable_field out_ch =
    let formatter = new url_formatter () in
    let printer = new url_printer out_ch element_separator in
    let writer = Xnc_writer.create formatter printer in
      Xnc_serializable.write writer serializable_field;
      Xnc_writer.flush writer

  method from_netchannel prototype in_ch =
    let scanbuf = Xnc_scanning.Scanning_buffer.create_from_netchannel in_ch in
    let scanner = new url_scanner scanbuf element_separator in
    let token_evaluator = new Xnc_token_eval.standard_evaluator () in
    let lexer = new url_lexer scanbuf element_separator in
    let reader = Xnc_reader.create scanner token_evaluator lexer in
      Xnc_serializable.read reader prototype

end;;

(* TODO:
 * - cache serializer creation into a data structure
 * - make custom serializer based on part and configuration
 * - make a custom query-string serializer ?name1=value1&name2=value2...
 * - fix special cases when "," or "/" is included in a string using
 * url_serializer *)
let get_serializer location part config =
  match location with
      Argument_value
    | Query_string
    | Cookie ->
        new Xnc_serializing.simple_serializer ()
    | Path_info ->
        begin match part with
            Page_name ->
              new Xnc_serializing.null_serializer ()
          | Session ->
              begin match Srv_configuration.session_manager_mode
                            config with
                  Srv_types.Off ->
                    new Xnc_serializing.null_serializer ()
                | _ ->
                    new url_serializer session_element_separator
              end
          | Page_parameters ->
              begin match Srv_configuration.page_parameter_manager_mode
                            config with
                  Srv_types.Off ->
                    new Xnc_serializing.null_serializer ()
                | _ ->
                    new url_serializer path_separator
              end
        end
    | Not_present ->
        new Xnc_serializing.null_serializer ()
;;

module Serialized =
struct
  type state_part =
      { part : part;
        location : location;
        content : string;
      }

  (* TODO: use a set or a map or a hash *)
  type t = state_part list

  let create l =
    l

  let get_state_part l p =
    List.find
      (fun s -> s.part = p)
      l

  (* private *)
  let retrieve_page_name
        (cgi : Netcgi.cgi_activation)
        conf
        resource_list =
    let scan_name_list path_info =
      let page_name_list =
        Srv_resource_list.get_page_name_list resource_list
      in
        List.fold_left
          (fun accu name ->
             let regexp =
               Str.regexp (path_separator ^ Str.quote name)
             in
               if Str.string_match regexp path_info 0 then
                 (name,
                  Path_info,
                  Str.string_after path_info (Str.match_end ()))
               else
                 accu
          )
          ("", Not_present, path_info)
          page_name_list
    in

    let path_info =
      match cgi # environment # cgi_path_info with
          "" ->
            let nethttpd_path_regexp =
              Str.regexp (Srv_configuration.netplex_nethttpd_path conf) in
            let script_name = cgi # environment # cgi_script_name in
              if Srv_types.is_nethttpd cgi &&
                 Str.string_match nethttpd_path_regexp script_name 0 then
                Str.string_after script_name (Str.match_end ())
              else
                script_name
        | s -> s in
    let current_page_field_name =
      Srv_configuration.current_page_field_name conf
    in
      match cgi # argument_value current_page_field_name with
          "" ->
            scan_name_list path_info
        | value ->
            (value, Argument_value, path_info)

  (* private *)
  let get_cookie cgi name =
    try
      Netcgi.Cookie.value (cgi # environment # cookie name)
    with Not_found -> ""

  (* private *)
  let retrieve_session
        remaining_path_info
        (cgi : Netcgi.cgi_activation)
        conf =
    let parse_path_info ?(prefix = "") path_info =
      let regexp = Str.regexp ("/" ^ prefix ^ "\\([^/]+\\)") in
        if Str.string_match regexp path_info 0 then
          begin
            let match_end = Str.match_end () in
              (Str.matched_group 1 path_info,
               Path_info,
               Str.string_after path_info match_end)
          end
        else
          ("", Not_present, path_info)
    in

    let session_field_name =
      Srv_configuration.session_field_name conf
    in
      match cgi # argument_value session_field_name with
          "" ->
            begin match get_cookie cgi session_field_name with
                "" ->
                  parse_path_info ~prefix:session_prefix remaining_path_info
              | value ->
                  (value, Cookie, remaining_path_info)
            end
        | value ->
            (value, Argument_value, remaining_path_info)

  (* private *)
  let retrieve_page_parameters
        remaining_path_info
        (cgi : Netcgi.cgi_activation)
        conf =
    let page_parameters_field_name =
      Srv_configuration.page_parameters_field_name conf
    in
      match cgi # argument_value page_parameters_field_name with
          "" ->
            begin match get_cookie cgi page_parameters_field_name with
                "" ->
                  begin match cgi # environment # cgi_query_string with
                      "" ->
                        begin match remaining_path_info with
                            "" ->
                              ("",
                               Not_present,
                               "")
                          | value ->
                              (Str.string_after value 1,
                               Path_info,
                               "")
                        end
                    | value ->
                        (value, Query_string, remaining_path_info)
                  end
              | value ->
                  (value, Cookie, remaining_path_info)
            end
        | value ->
            (value, Argument_value, remaining_path_info)

  let retrieve_state
        (cgi : Netcgi.cgi_activation)
        conf
        resource_list =
    let (page_name, page_name_location, remaining_path_info) =
      retrieve_page_name cgi conf resource_list in
    let (session, session_location, remaining_path_info) =
      match Srv_configuration.session_manager_mode conf with
          Srv_types.Off ->
            ("", Not_present, remaining_path_info)
        | _ ->
            retrieve_session remaining_path_info cgi conf in
    let (page_parameters, parameter_location, remaining_path_info) =
      match Srv_configuration.page_parameter_manager_mode conf with
          Srv_types.Off ->
            ("", Not_present, remaining_path_info)
        | _ ->
            retrieve_page_parameters remaining_path_info cgi conf
    in
      [
        { part = Page_name;
          location = page_name_location;
          content = page_name
        };
        { part = Session;
          location = session_location;
          content = session
        };
        { part = Page_parameters;
          location = parameter_location;
          content = page_parameters
        }
      ]

  (* TODO: refactor *)
  let build_url
        serialized_state
        (cgi : Netcgi.cgi_activation)
        conf =
    (* TODO: sort list 1-page_name, 2-session, 3-page_parameters *)
    let path_info =
      List.fold_left
        (fun path state_part ->
           let prefix =
             match state_part.part with
                 Session -> session_prefix
               | _ -> ""
           in
             path ^ path_separator ^ prefix ^ state_part.content)
        ""
        (List.filter
           (fun state_part ->
              state_part.location = Path_info)
           serialized_state) in
    let cgi_arguments =
      List.map
        (fun state_part ->
           let name =
             match state_part.part with
                 Page_name ->
                   Srv_configuration.current_page_field_name conf
               | Session ->
                   Srv_configuration.session_field_name conf
               | Page_parameters ->
                   Srv_configuration.page_parameters_field_name conf
           in
             Netcgi.Argument.simple name state_part.content
        )
        (List.filter
           (fun state_part ->
              state_part.location = Query_string)
           serialized_state)
    in
    let script_name =
      if Srv_types.is_nethttpd cgi then
        Srv_configuration.netplex_nethttpd_path conf
      else
        cgi # environment # cgi_script_name in

      cgi # url
        ~with_script_name:(`This script_name)
        ~with_path_info:(`This path_info)
        ~with_query_string:(`This cgi_arguments) ()

end;;

module Deserialized =
struct
  type deserialization =
      String of string
    | Field_collection of Xnc_field_collection.t

  type state_part =
      { part : part;
        location : location;
        content : deserialization
      }

  type t = state_part list

  let create l =
    l

  let get_state_part l p =
    List.find
      (fun s -> s.part = p)
      l

  (* TODO: refactor *)
  let serialize_state deserialized_state conf =
    List.map
      (fun state_part ->
         match state_part.part with
             Page_name ->
               { Serialized.part = state_part.part;
                 Serialized.location = state_part.location; 
                 Serialized.content =
                   match state_part.content with
                       String s ->
                         s
                     | Field_collection _ ->
                         failwith "Page name/unparsed path info cannot \
                                   be deserialized from a field collection"
               }
           | Page_parameters
           | Session ->
               let serializer = get_serializer
                                  state_part.location
                                  state_part.part
                                  conf in
                 { Serialized.part = state_part.part;
                   Serialized.location = state_part.location; 
                   Serialized.content =
                     match state_part.content with
                        String _ ->
                          begin match state_part.location with
                              Not_present ->
                                ""
                            | _ ->
                                failwith "Page parameters or session cannot be \
                                          deserialized from a string"
                          end
                       | Field_collection fc ->
                           Xnc_field_collection.to_string ~serializer fc
                 }
      )
      deserialized_state

end

