(** State persisting.
  *
  * This module provides functions to perform state serialization and
  * deserialization. *)

type part =
    Page_name
  | Session
  | Page_parameters
(** State part. *)

type location =
    Argument_value
  | Query_string
  | Path_info
  | Cookie
  | Not_present
(** Location of the state part. *)

val get_serializer :
  location ->
  part ->
  Srv_configuration.t ->
  string Xnc_serializing.serializer
(** [get_serializer l c] returns the serializer to use with location [l]. *)

module Serialized :
sig
  type state_part =
      { part : part;
        location : location;
        content : string
      }
  (** The type of a serialized state part. *)

  type t
  (** The type of serialized state. *)

  val create :
    state_part list -> t
  (** [create l] creates serialized state starting from state part list [l].
    *)

  val get_state_part :
    t -> part -> state_part
  (** [get_state_part s p] returns part [p] of serialized state [s]. *)

  val retrieve_state :
    Netcgi.cgi_activation ->
    Srv_configuration.t ->
    Srv_resource_list.t ->
    t
  (** [retrieve_state a c l] loads state from CGI activation [a] with
    * configuration [c] and resource list [l]. *)

  val build_url :
    t ->
    Netcgi.cgi_activation ->
    Srv_configuration.t ->
    string
  (** [build_url s a c] returns the URL formed from serialized state [s] using
    * CGI activation [a] and configuration [c]. *)

end
(** This module implement serialized state handling. *)

module Deserialized :
sig
  type deserialization =
      String of string
    | Field_collection of Xnc_field_collection.t
  (** Type of deserialized content. *)

  type state_part =
      { part : part;
        location : location;
        content : deserialization
      }
  (** The type of a deserialized state part. *)

  type t
  (** The type of deserialized state. *)

  val create :
    state_part list -> t
  (** [create l] creates deserialized state starting from state part list [l].
    *)

  val get_state_part :
    t -> part -> state_part
  (** [get_state_part s p] returns part [p] of serialized state [s]. *)

  val serialize_state :
    t ->
    Srv_configuration.t ->
    Serialized.t
  (** [serialize_state d c] serialize state from deserialization [d] with
    * configuration [c]. *)

end

