(** Application server types.
  *
  * This module collects module signatures that are implemented by the modules
  * generated by the page compiler.
  *)

exception Invalid_page_parameters
(** Raised when page parameters are passed to the wrong page. *)

type lifecycle_hook =
    Pre_session_loading
  | Post_session_loading
  | Pre_parameter_init
  | Post_parameter_init
  | Pre_user_input_validation
  | Post_user_input_validation
  | Pre_user_input_reading
  | Post_user_input_reading
  | Pre_structure_building
  | Post_structure_building
  | Pre_ui_event_handling
  | Post_ui_event_handling
  | Pre_state_saving
  | Post_state_saving
  | Pre_render
  | Post_render
  | Pre_completion
  | Post_completion
(** Page lifecycle hooks. *)

type manager_mode =
    Off (** Does not store state. *)
  | Client (** Stores state client-side (in a hidden field or a cookie). *)
  | In_memory (** Stores state in server memory. *)
  | Sqlite3 (** Stores state in SQLite3 DB. *)
(** Enumerates the state storage modes. *)

val manager_mode_of_string : string -> manager_mode
(** [manager_mode_of_string s] converts string [s] to a value of type
  * [manager_mode]. Raises [Failure "manager_mode_of_string"] if [s] is not
  * convertible. *)

type state_persistence_mode =
    Hidden_field (** Stores state or state ID in a hidden field. *)
  | Cookie (** Stores state or state ID in a client-side cookie. *)
(** Enumerates the state persistence modes. *)

val state_persistence_mode_of_string : string -> state_persistence_mode
(** [state_persistence_mode_of_string s] converts string [s] to a value of type
  * [state_persistence_mode]. Raises [Failure "id_storage_mode_of_string"] if
  * [s] is not convertible. *)

type url_rendering_mode =
    Cgi_style
  | Plain_url
(** URL rendering mode. *)

type redirect_mode =
    Server_side (** Standard server-side mode: the server performs the
                  * redirection, and it is invisible for the client. *)
  | Client_side of url_rendering_mode (** Client-side mode: the redirection
                                        * directive is passed back
                                        * to the client, and the client will
                                        * repeat the request for
                                        * the new location. *)
(** Page change mode. *)

module type SERIALIZABLE_DATA =
sig
  type t
  (** The type of serializable data. *)

  val to_field_collection : t -> Xnc_field_collection.t
  (** [to_field_collection s] converts [s] to a field collection, to
    * serialize it. *)

  val from_field_collection : t -> Xnc_field_collection.t -> t
  (** [from_field_collection d f] deserializes serializable data [s] from field
    * collection [f]. *)

end
(** Signature that defines serializable data. *)

module type PAGE_PARAMETERS =
sig
  type t
  (** The type of page parameters. *)

  val default_start_page_parameters : t
  (** Default start page parameters, used to dispatch start page first time.
    *)

  val create_error_page_parameters : string -> string -> t
  (** [create_error_page_parameters d b] creates page parameters for the error
    * page using description [d], and backtrace [b]. *)

  val default_error_page_parameters : t
  (** Default error page parameters, with empty description and empty backtrace.
    *)

  val get_error_description : t -> string
  (** [get_error_description p] returns the error description contained in [p].
    * Raises [Invalid_page_parameters] if [p] does not contain error page
    * parameters. *)

  val get_backtrace : t -> string
  (** [get_backtrace p] returns the error backtrace contained in [p].
    * Raises [Invalid_page_parameters] if [p] does not contain error page
    * parameters. *)

  val create_not_found_page_parameters : string -> t
  (** [create_not_found_page_parameters m] creates page parameters for the not
    * found (404) page passing missing resource [m]. *)

  val default_not_found_page_parameters : t
  (** Default not found (404) page parameters, with empty missing resource. *)

  val get_missing_resource : t -> string
  (** [get_missing_resource p] returns the missing resource contained in [p].
    * Raises [Invalid_page_parameters] if [p] does not contain not found (404)
    * page parameters. *)

  val to_field_collection : t -> Xnc_field_collection.t
  (** [to_field_collection p] converts [p] to a field collection, to
    * serialize it. *)

  val from_field_collection : t -> Xnc_field_collection.t -> t
  (** [from_field_collection p f] fills [p] with data contained in [f]. *)

end
(** Signature that provides the common functionality for page parameters. *)

module type SESSION_DATA =
sig
  type t
  (** The type of session data. *)

  val create : unit -> t
  (** [create ()] creates the default session data. *)

  val to_field_collection : t -> Xnc_field_collection.t
  (** [to_field_collection s] converts [s] to a field collection, to
    * serialize it. *)

  val from_field_collection : Xnc_field_collection.t -> t
  (** [from_field_collection f] returns session with data contained in [f]. *)

end
(** Signature that provides the common functionality for session data. *)

(* TODO: move this function to a proper module *)
val is_nethttpd : Netcgi.cgi_activation -> bool
(** [is_nethttpd c] determines if the request is processed by Nethttpd.  *)

