open OUnit

module Concrete_context =
  Test_runtime_context.Runtime_context
;;

module Concrete_session_data =
  Test_types.Concrete_session_data
;;

module Concrete_page_parameters =
  Test_types.Concrete_page_parameters
;;

module Concrete_page_state =
  Test_page_state.Concrete_page_state
;;

module Concrete_page_behavior =
  Srv_page_behavior.Make
    (Concrete_page_state)
;;

module Form_data =
struct
  type t =
      { test_value : int
      }
  ;;

  let create () =
    { test_value = 42
    }
  ;;

  let from_cgi_context (cgi : Netcgi.cgi_activation) =
    let test_value = int_of_string (cgi # argument_value "test_value") in
      { test_value = test_value
      }
  ;;

end;;

class test_page
  page_context =
object(self)
  inherit Concrete_page_behavior.page_behavior_base
        "test_page.xnunc"
        (Concrete_page_parameters.Test_page_parameters (42, true))
        page_context

  val mutable form_data = Form_data.create ()

  method build_validator_collection () =
    self # add_validator
      "test_value"
      (new Srv_input_validator.type_validator
         (self # get_cgi_context # argument_value "test_value")
         "int"
         "Error during type validation")

  method read_user_input () =
    form_data <- Form_data.from_cgi_context (self # get_cgi_context)

  method build_page_structure () =
    let _ = Srv_page_structure.create_body_element
              (self # get_page_structure) in
    let form = Srv_page_structure.create_form_element
                 (self # get_page_structure)
                 (self # get_cgi_context) in
      Srv_page_structure.add_element
        form
        (new Srv_page_element.literal_element
           ("<!-- " ^ (string_of_int
                         form_data.Form_data.test_value) ^ " -->"))

end;;

(* fixture setup and teardown *)
let set_up _ =
  let context = Concrete_context.get_current () in
  let config = context.Srv_runtime_context.configuration in
  let page_name = Netcgi.Argument.simple
                    (Srv_configuration.current_page_field_name config)
                    "test_page.xnunc" in
  let page_parameters = Netcgi.Argument.simple
                          (Srv_configuration.page_parameters_field_name config)
                          "Parameters=[integer=42;boolean=true]" in
  let session = Netcgi.Argument.simple
                  (Srv_configuration.session_field_name config)
                  "Session=[a_int=42;\
                            a_float=3.14;\
                            a_char=a;\
                            a_string=&quot;test&quot;;\
                            a_bool=true]" in
  let form_data = Netcgi.Argument.simple
                    "test_value"
                    "100" in
  let cgi = As_test_helper.create_test_activation
              [page_name;
               page_parameters;
               session;
               form_data] in
  let resource_list = Srv_resource_list.create
                        [
                          { Srv_resource_list.name = "test_page.xnunc";
                            Srv_resource_list.resource_type = "xnc:page"
                          }
                        ] in
  let serialized_state = Srv_state_persisting.Serialized.retrieve_state
                           cgi
                           config
                           resource_list in
    (new test_page
       { Srv_page_state.cgi = cgi;
         Srv_page_state.state_serialization = serialized_state;
         Srv_page_state.is_self_submitted = false
       })
;;

let tear_down _ =
  Concrete_context.clear ()
;;

(* tests *)
let test_invoke_callbacks (tp) =
  let i = ref 0 in
    tp # register_callback Srv_types.Pre_render (fun () -> incr i);
    tp # register_callback Srv_types.Pre_render (fun () -> i := !i * 5);
    tp # invoke_callbacks Srv_types.Pre_render;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected result after callback invocation"
      5
      !i
;;

let test_is_valid (tp) =
  tp # build_validator_collection ();
  OUnit.assert_bool
    "Test page should be valid"
    (tp # is_valid ());
  tp # add_validator
    "test_invalid"
    (new Srv_input_validator.required_field_validator
       ""
       "Error during validation");
  OUnit.assert_bool
    "Test page should not be valid"
    (not (tp # is_valid ()))
;;

let test_build_page_structure (tp) =
  tp # build_page_structure ();
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected rendered output"
    "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\
     <!DOCTYPE html PUBLIC \
          \"-//W3C//DTD XHTML 1.1//EN\" \
          \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\
     <html xmlns=\"http://www.w3.org/1999/xhtml\">\
       <body>\
         <form method=\"post\" action=\"\">\
           <!-- 42 -->\
         </form>\
       </body>\
     </html>"
    (As_test_helper.render_page (tp # get_page_structure))
;;

let test_read_user_input (tp) =
  tp # read_user_input ();
  tp # build_page_structure ();
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected rendered output"
    "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\
     <!DOCTYPE html \
          PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \
          \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\
     <html xmlns=\"http://www.w3.org/1999/xhtml\">\
       <body>\
         <form method=\"post\" action=\"\">\
           <!-- 100 -->\
         </form>\
       </body>\
     </html>"
    (As_test_helper.render_page (tp # get_page_structure))
;;

let test_load_page_parameters (tp) =
  tp # load_page_parameters ();
  OUnit.assert_equal
    ~msg:"Unexpected parameters loaded"
    (Concrete_page_parameters.Test_page_parameters (42, true))
    (tp # get_page_parameters)
;;

let test_load_session (tp) =
  let () = tp # load_session () in
  let session_data = tp # get_session_data in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Error during session deserialization"
      42
      session_data.Concrete_session_data.a_int
;;

let test_save_state (tp) =
  tp # build_page_structure ();
  tp # save_state ();
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Error during session deserialization"
    "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\
     <!DOCTYPE html \
          PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \
          \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\
     <html xmlns=\"http://www.w3.org/1999/xhtml\">\
       <body>\
         <form method=\"post\" action=\"\">\
           <!-- 42 -->\
           <div>\
             <input type=\"hidden\" name=\"xnunc_session\" value=\"Session=[a_int=42;a_float=3.14;a_char=a;a_string=&quot;test&quot;;a_bool=true]\" />\
             <input type=\"hidden\" name=\"xnunc_page\" value=\"test_page.xnunc\" />\
             <input type=\"hidden\" name=\"xnunc_page_parameters\" value=\"Parameters=[integer=42;boolean=true]\" />\
           </div>\
         </form>\
       </body>\
     </html>"
    (As_test_helper.render_page (tp # get_page_structure))
;;

(* test suite *)
let suite = "Page behavior" >:::
  ["test_invoke_callbacks"
     >:: (bracket set_up test_invoke_callbacks tear_down);
   "test_is_valid" >:: (bracket set_up test_is_valid tear_down);
   "test_build_page_structure"
     >:: (bracket set_up test_build_page_structure tear_down);
   "test_read_user_input" >:: (bracket set_up test_read_user_input tear_down);
   "test_load_page_parameters"
     >:: (bracket set_up test_load_page_parameters tear_down);
   "test_load_session" >:: (bracket set_up test_load_session tear_down);
   "test_save_state" >:: (bracket set_up test_save_state tear_down)]
;;

