open OUnit

(* tests *)
let test_render_literal _ =
  let e = new Srv_page_element.literal_element "<h1>This is a fragment</h1>" in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected literal element rendered output"
      "<h1>This is a fragment</h1>"
      (As_test_helper.render_element e)
;;

let test_render_button _ =
  let e = (new Srv_page_element.button_element
             "Quit" :> Srv_page_element.page_element) in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected button element rendered output"
      "<input type=\"submit\" name=\"Quit\" value=\"Quit\" />"
      (As_test_helper.render_element e)
;;

let test_render_form _ =
  let e = (new Srv_page_element.form_element ""
             :> Srv_page_element.page_element) in
    e # add_element
      (new Srv_page_element.literal_element "<h1>Hello, world!</h1>");
    e # add_element
      (new Srv_page_element.button_element "OK" :> Srv_page_element.page_element);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected form element rendered output"
      "<form method=\"post\" action=\"\">\
         <h1>Hello, world!</h1>\
         <input type=\"submit\" name=\"OK\" value=\"OK\" />\
       </form>"
      (As_test_helper.render_element e)
;;

let test_render_simple_hidden_field _ =
  let e = new Srv_page_element.simple_hidden_field "field1" "value1" in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected simple hidden field rendered output"
      "<input type=\"hidden\" name=\"field1\" value=\"value1\" />"
      (As_test_helper.render_element e)
;;

let test_render_complex_hidden_field _ =
  let fc = Xnc_field_collection.create "Test" in
  let _ = Xnc_field_collection.register_int fc "int_field" 10 in
  let _ = Xnc_field_collection.register_float fc "float_field" 0.51 in
  let e = new Srv_page_element.complex_hidden_field "field1" fc in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected simple hidden field rendered output"
      "<input type=\"hidden\" \
              name=\"field1\" \
              value=\"Test=[int_field=10;float_field=0.51]\" />"
      (As_test_helper.render_element e)
;;

let test_add_hidden_field _ =
  let hf = new Srv_page_element.simple_hidden_field "field1" "value1" in
  let fc = Xnc_field_collection.create "Test" in
  let _ = Xnc_field_collection.register_bool fc "bool_field" true in
  let _ = Xnc_field_collection.register_char fc "char_field" 'x' in
  let hf' = new Srv_page_element.complex_hidden_field "field2" fc in
  let e = new Srv_page_element.form_element "" in
    e # add_hidden_field (hf :> Srv_page_element.hidden_field);
    e # add_hidden_field (hf' :> Srv_page_element.hidden_field);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected form element rendered output"
      "<form method=\"post\" action=\"\">\
         <div>\
           <input type=\"hidden\" \
                  name=\"field1\" \
                  value=\"value1\" />\
           <input type=\"hidden\" \
                  name=\"field2\" \
                  value=\"Test=[bool_field=true;char_field=x]\" />\
         </div>\
       </form>"
      (As_test_helper.render_element e)
;;

let test_render_container _ =
  let inner_cont = (new Srv_page_element.container_element :>
                      Srv_page_element.page_element) in
  let cont = (new Srv_page_element.container_element :>
                Srv_page_element.page_element) in
    inner_cont # add_element
      (new Srv_page_element.literal_element "<p>Hello, world!</p>");
    inner_cont # add_element
      (new Srv_page_element.button_element "Cancel" :> Srv_page_element.page_element);
    cont # add_element
      (new Srv_page_element.literal_element "<h1>Hello, world!</h1>");
    cont # add_element inner_cont;
    cont # add_element
      (new Srv_page_element.button_element "OK" :> Srv_page_element.page_element);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected XHTML string"
      "<h1>Hello, world!</h1>\
       <p>Hello, world!</p>\
       <input type=\"submit\" name=\"Cancel\" value=\"Cancel\" />\
       <input type=\"submit\" name=\"OK\" value=\"OK\" />"
      (As_test_helper.render_element cont)
;;

let test_handle_ui_events _ =
  let arg_ok = Netcgi.Argument.simple "OK" "OK" in
  let arg_ko = Netcgi.Argument.simple "KO" "KO" in
  let cgi = As_test_helper.create_test_activation [arg_ok; arg_ko] in
  let inner_cont =
    (new Srv_page_element.container_element :> Srv_page_element.page_element) in
  let cont =
    (new Srv_page_element.container_element :> Srv_page_element.page_element) in
  let b_ko = new Srv_page_element.button_element "KO" in
  let b_ok = new Srv_page_element.button_element "OK" in
  let i = ref 1 in
    b_ko # add_click_handler (fun () -> incr i);
    b_ok # add_click_handler (fun () -> i := !i * 2);
    inner_cont # add_element (b_ko :> Srv_page_element.page_element);
    cont # add_element inner_cont;
    cont # add_element (b_ok :> Srv_page_element.page_element);
    cont # handle_ui_events cgi;
    OUnit.assert_equal
      ~printer:(fun x -> string_of_int x)
      4
      !i
;;

let test_container_event_handling _ =
  let arg = Netcgi.Argument.simple "Quit" "Quit" in
  let cgi = As_test_helper.create_test_activation [arg] in
  let i = ref 1 in
  let e = new Srv_page_element.button_element "Quit" in
    e # add_click_handler (fun () -> incr i);
    e # add_click_handler (fun () -> i := !i * 2);
    e # handle_ui_events cgi;
    OUnit.assert_equal
      ~printer:(fun x -> string_of_int x)
      4
      !i
;;

let test_render_validator_valid _ =
  let v = new Srv_input_validator.type_validator
            "true"
            "bool"
            "Error during type validation" in
  let e = (new Srv_page_element.validator_element
             v :> Srv_page_element.page_element) in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected validator element rendered output"
      ""
      (As_test_helper.render_element e)
;;

let test_render_validator_invalid _ =
  let v = new Srv_input_validator.type_validator
            "test"
            "bool"
            "Error during type validation" in
  let e = (new Srv_page_element.validator_element
             v :> Srv_page_element.page_element) in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected validator element rendered output"
      "<span>Error during type validation</span>"
      (As_test_helper.render_element e)
;;

let test_render_validation_summary _ =
  let v = Srv_validator_collection.create () in
  let () = Srv_validator_collection.add_validator v
             "bool_field"
             (new Srv_input_validator.type_validator
                "true"
                "bool"
                "Error during bool validation") in
  let () = Srv_validator_collection.add_validator v
             "int_field"
             (new Srv_input_validator.type_validator
                "abc"
                "int"
                "Error during int validation") in
  let () = Srv_validator_collection.add_validator v
             "string_field"
             (new Srv_input_validator.required_field_validator
                ""
                "Error during required field validation") in
  let e = (new Srv_page_element.validation_summary
             v "Error summary:" :> Srv_page_element.page_element) in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected validator element rendered output"
      "<span>Error summary:<br />\
             - Error during int validation<br />\
             - Error during required field validation<br />\
       </span>"
      (As_test_helper.render_element e)
;;

let test_render_head_element _ =
  let e = new Srv_page_element.head_element in
    e # set_lang (Some "it-IT");
    e # set_dir (Some "ltr");
    e # set_id (Some "head");
    e # set_profile (Some "http://www.acme.com/profiles/core");
    e # add_element
      (new Srv_page_element.literal_element "<title>Title</title>");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected head element rendered output"
      "<head xml:lang=\"it-IT\" \
             dir=\"ltr\" \
             id=\"head\" \
             profile=\"http://www.acme.com/profiles/core\">\
         <title>Title</title>\
       </head>"
      (As_test_helper.render_element e)
;;

let test_render_body_element _ =
  let e = new Srv_page_element.body_element in
    e # set_id (Some "body");
    e # set_class (Some "main");
    e # set_style (Some "background-color: white");
    e # set_title (Some "This is the page body");
    e # set_lang (Some "en");
    e # set_dir (Some "ltr");
    e # set_onclick (Some "onclick()");
    e # set_ondblclick (Some "ondlbclick()");
    e # set_onmousedown (Some "onmousedown()");
    e # set_onmouseup (Some "onmouseup()");
    e # set_onmouseover (Some "onmouseover()");
    e # set_onmousemove (Some "onmousemove()");
    e # set_onmouseout (Some "onmouseout()");
    e # set_onkeypress (Some "onkeypress()");
    e # set_onkeydown (Some "onkeydown()");
    e # set_onkeyup (Some "onkeyup()");
    e # set_onload (Some "onload()");
    e # set_onunload (Some "onunload()");
    e # add_element
      (new Srv_page_element.literal_element "<h1>Title</h1>");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected body element rendered output"
      "<body id=\"body\" \
             class=\"main\" \
             style=\"background-color: white\" \
             title=\"This is the page body\" \
             xml:lang=\"en\" \
             dir=\"ltr\" \
             onclick=\"onclick()\" \
             ondblclick=\"ondlbclick()\" \
             onmousedown=\"onmousedown()\" \
             onmouseup=\"onmouseup()\" \
             onmouseover=\"onmouseover()\" \
             onmousemove=\"onmousemove()\" \
             onmouseout=\"onmouseout()\" \
             onkeypress=\"onkeypress()\" \
             onkeydown=\"onkeydown()\" \
             onkeyup=\"onkeyup()\" \
             onload=\"onload()\" \
             onunload=\"onunload()\">\
         <h1>Title</h1>\
       </body>"
      (As_test_helper.render_element e)
;;

let test_render_div_element _ =
  let e = new Srv_page_element.div_element in
    e # set_id (Some "div");
    e # add_element
      (new Srv_page_element.literal_element "<p>Content</p>");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected div element rendered output"
      "<div id=\"div\">\
         <p>Content</p>\
       </div>"
      (As_test_helper.render_element e)
;;

let test_render_input_element _ =
  let e = new Srv_page_element.input_element in
    e # set_name (Some "input");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected input element rendered output"
      "<input type=\"text\" name=\"input\" />"
      (As_test_helper.render_element e)
;;

(* test suite *)
let suite = "Page element test" >:::
  ["test_render_literal" >:: test_render_literal;
   "test_render_button" >:: test_render_button;
   "test_render_container" >:: test_render_container;
   "test_render_form" >:: test_render_form;
   "test_render_simple_hidden_field" >:: test_render_simple_hidden_field;
   "test_render_complex_hidden_field" >:: test_render_complex_hidden_field;
   "test_add_hidden_field" >:: test_add_hidden_field;
   "test_handle_ui_events" >:: test_handle_ui_events;
   "test_container_event_handling" >:: test_container_event_handling;
   "test_render_validator_valid" >:: test_render_validator_valid;
   "test_render_validator_invalid" >:: test_render_validator_invalid;
   "test_render_validation_summary" >:: test_render_validation_summary;
   "test_render_head_element" >:: test_render_head_element;
   "test_render_body_element" >:: test_render_body_element;
   "test_render_div_element" >:: test_render_div_element;
   "test_render_input_element" >:: test_render_input_element]
;;

