open OUnit

module Concrete_context =
  Test_runtime_context.Runtime_context
;;

module Concrete_session_data =
  Test_types.Concrete_session_data
;;

module Concrete_page_parameters =
  Test_types.Concrete_page_parameters
;;

module Concrete_page_state =
  Srv_page_state.Make
    (Concrete_context)
    (Concrete_session_data)
    (Concrete_page_parameters)
;;

(* fixture setup and teardown *)
let set_up _ =
  let config = Srv_configuration.default_conf in
  let page_name = Netcgi.Argument.simple
                    (Srv_configuration.current_page_field_name config)
                    "test_page.xnunc" in
  let session = Netcgi.Argument.simple
                  (Srv_configuration.session_field_name config)
                  "Session=[a_int=1;\
                            a_float=0.5;\
                            a_char=c;\
                            a_string=&quot;string&quot;;\
                            a_bool=false]" in
  let page_parameters = Netcgi.Argument.simple
                          (Srv_configuration.page_parameters_field_name config)
                          "Parameters=[integer=10;boolean=false]" in
  let cgi = As_test_helper.create_test_activation
              [page_name;
               session;
               page_parameters] in
  let resource_list = Srv_resource_list.create
                        [
                          { Srv_resource_list.name = "test_page.xnunc";
                            Srv_resource_list.resource_type = "xnc:page"
                          }
                        ] in
  let serialized_state = Srv_state_persisting.Serialized.retrieve_state
                           cgi
                           config
                           resource_list in
    Concrete_page_state.create
      "test_page.xnunc"
      (Concrete_page_parameters.Test_page_parameters (0, false))
      { Srv_page_state.cgi = cgi;
        Srv_page_state.state_serialization = serialized_state;
        Srv_page_state.is_self_submitted = false
      }
;;

let tear_down _ =
  ()
;;

(* tests *)
let test_get_page_name state =
  OUnit.assert_equal
    ~printer:Std.identity
    ~msg:"Unexpected page name"
    "test_page.xnunc"
    (Concrete_page_state.get_page_name state);
;;

let test_set_get_page_parameters state =
  let page_parameters =
    Concrete_page_parameters.Test_page_parameters (10, false)
  in
    Concrete_page_state.set_page_parameters state page_parameters;
    OUnit.assert_equal
      ~printer:(fun p ->
                  let fc = Concrete_page_parameters.to_field_collection p in
                    Xnc_field_collection.to_string fc)
      ~msg:"Unexpected page parameters"
      page_parameters
      (Concrete_page_state.get_page_parameters state);
;;

let test_get_cgi_activation state =
  let cgi = Concrete_page_state.get_cgi_activation state in
    OUnit.assert_equal
      ~msg:"Unexpected cgi arguments"
      3
      (List.length (cgi # arguments));
;;

let test_is_self_submitted state =
  OUnit.assert_equal
    ~printer:string_of_bool
    ~msg:"Unexpected self submitted flag"
    false
    (Concrete_page_state.is_self_submitted state);
;;

let test_get_session_data state =
  let session_data = Concrete_session_data.create () in
    OUnit.assert_equal
      ~printer:(fun s ->
                  let fc =
                    Concrete_session_data.to_field_collection s
                  in
                    Xnc_field_collection.to_string fc)
      ~msg:"Unexpected session data"
      session_data
      (Concrete_page_state.get_session_data state);
;;

let test_get_page_structure state =
  let page_structure = Srv_page_structure.create
                         Srv_configuration.default_conf in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected page structure xml version"
      (page_structure # get_xml_version)
      ((Concrete_page_state.get_page_structure state) # get_xml_version);
;;

let test_get_logger state =
  let logger = Srv_page_helper.create_page_logger () in
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected page logger level"
      (Xnc_logger.is_debug logger)
      (Xnc_logger.is_debug (Concrete_page_state.get_logger state));
;;

let test_retrieve_session state =
  Concrete_page_state.retrieve_session state;
  OUnit.assert_equal
    ~printer:(fun s ->
                let fc =
                  Concrete_session_data.to_field_collection s
                in
                  Xnc_field_collection.to_string fc)
    ~msg:"Unexpected retrieved session"
    { Concrete_session_data.a_int = 1;
      Concrete_session_data.a_float = 0.5;
      Concrete_session_data.a_char = 'c';
      Concrete_session_data.a_string = "string";
      Concrete_session_data.a_bool = false
    }
    (Concrete_page_state.get_session_data state);
;;

let test_retrieve_page_parameters state =
  Concrete_page_state.retrieve_page_parameters state;
  OUnit.assert_equal
    ~printer:(fun s ->
                let fc =
                  Concrete_page_parameters.to_field_collection s
                in
                  Xnc_field_collection.to_string fc)
    ~msg:"Unexpected loaded page parameters"
    (Concrete_page_parameters.Test_page_parameters (10, false))
    (Concrete_page_state.get_page_parameters state);
;;

let test_session_to_field_collection state =
  let session_data = Concrete_page_state.get_session_data state in
  let () = session_data.Test_types.Concrete_session_data.a_int <- 0 in
  let fc = Concrete_page_state.session_to_field_collection state in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected session schema int field"
      (Xnc_field_collection.get_int
         (Concrete_session_data.to_field_collection session_data)
         "a_int")
      (Xnc_field_collection.get_int
         fc
         "a_int");
;;

let test_page_parameters_to_field_collection state =
  let parameters = Concrete_page_parameters.Test_page_parameters (1, false) in
  let () = Concrete_page_state.set_page_parameters state parameters in
  let fc = Concrete_page_state.page_parameters_to_field_collection state in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected page parameters int field"
      (Xnc_field_collection.get_int
         (Concrete_page_parameters.to_field_collection
            parameters)
         "integer")
      (Xnc_field_collection.get_int
         fc
         "integer");
;;

let test_invoke_callbacks state =
  let i = ref 0 in
    Concrete_page_state.register_callback
      state
      Srv_types.Pre_completion
      (fun () -> incr i);
    Concrete_page_state.register_callback
      state
      Srv_types.Pre_completion
      (fun () -> i := !i * 2);
    Concrete_page_state.invoke_callbacks state Srv_types.Pre_completion;
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected result after callback invocation"
      2
      !i
;;

let test_is_valid state =
  Concrete_page_state.add_validator
    state
    "test_value"
    (new Srv_input_validator.type_validator
       "42"
       "int"
       "Error during type validation");
  OUnit.assert_bool
    "Page state should be valid"
    (Concrete_page_state.is_valid state);
  Concrete_page_state.add_validator
    state
    "test_invalid"
    (new Srv_input_validator.required_field_validator
       ""
       "Error during validation");
  OUnit.assert_bool
    "Page state should not be valid"
    (not (Concrete_page_state.is_valid state))
;;

(* test suite *)
let suite = "Page state test" >:::
  ["test_get_page_name"
     >:: (bracket set_up test_get_page_name tear_down);
   "test_set_get_page_parameters"
     >:: (bracket set_up test_set_get_page_parameters tear_down);
   "test_get_cgi_activation"
     >:: (bracket set_up test_get_cgi_activation tear_down);
   "test_is_self_submitted"
     >:: (bracket set_up test_is_self_submitted tear_down);
   "test_get_session_data"
     >:: (bracket set_up test_get_session_data tear_down);
   "test_get_page_structure"
     >:: (bracket set_up test_get_page_structure tear_down);
   "test_get_logger"
     >:: (bracket set_up test_get_logger tear_down);
   "test_retrieve_session"
     >:: (bracket set_up test_retrieve_session tear_down);
   "test_retrieve_page_parameters"
     >:: (bracket set_up test_retrieve_page_parameters tear_down);
   "test_session_to_field_collection"
     >:: (bracket set_up test_session_to_field_collection tear_down);
   "test_page_parameters_to_field_collection"
     >:: (bracket set_up test_page_parameters_to_field_collection tear_down);
   "test_invoke_callbacks"
     >:: (bracket set_up test_invoke_callbacks tear_down);
   "test_is_valid"
     >:: (bracket set_up test_is_valid tear_down)]
;;

