open OUnit

(* tests *)
let test_render _ =
  let page = Srv_page_structure.create Srv_configuration.default_conf in
    page # set_lang (Some "en-US");
    page # set_dir (Some "ltr");
    Srv_page_structure.add_element page
      (new Srv_page_element.literal_element
 "<head>
    <title>Ex-nunc test page</title>
  </head>
  <body>");
    Srv_page_structure.add_element page
      (new Srv_page_element.literal_element "<h1>Hello, world!</h1>");
    Srv_page_structure.add_element page
      (new Srv_page_element.literal_element "<p>Hello, world!</p>");
    Srv_page_structure.add_element page
      (new Srv_page_element.literal_element "</body>");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected XHTML string"
      "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\
<!DOCTYPE html \
     PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \
     \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\
<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en-US\" dir=\"ltr\">\
  <head>
    <title>Ex-nunc test page</title>
  </head>
  <body><h1>Hello, world!</h1><p>Hello, world!</p></body></html>"
      (As_test_helper.render_page page)
;;

let test_handle_ui_events _ =
  let page = Srv_page_structure.create Srv_configuration.default_conf in
  let button1 = new Srv_page_element.button_element "Button1" in
  let button2 = new Srv_page_element.button_element "Button2" in
  let i = ref 1 in
    button1 # add_click_handler (fun () -> i := !i * 2);
    button2 # add_click_handler (fun () -> incr i);
    Srv_page_structure.add_element page
      (button1 :> Srv_page_element.page_element);
    Srv_page_structure.add_element page
      (button2 :> Srv_page_element.page_element);
    let arg1 = Netcgi.Argument.simple "Button1" "Button1" in
    let arg2 = Netcgi.Argument.simple "Button2" "Button2" in
    let cgi = As_test_helper.create_test_activation
                [ arg1; arg2 ] in
      Srv_page_structure.handle_ui_events page cgi;
      OUnit.assert_equal
        ~printer:(fun x -> string_of_int x)
        ~msg:"Unexpected result"
        3
        !i
;;

let test_create_form_element _ =
  let cgi = As_test_helper.create_test_activation [] in
  let page = Srv_page_structure.create Srv_configuration.default_conf in
  let _ = Srv_page_structure.create_body_element page in
  let form = Srv_page_structure.create_form_element page cgi in
    Srv_page_structure.add_element form
      (new Srv_page_element.literal_element "<h1>Hello, world!</h1>");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected XHTML string"
      "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\
       <!DOCTYPE html \
            PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \
            \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">\
       <html xmlns=\"http://www.w3.org/1999/xhtml\">\
         <body>\
           <form method=\"post\" action=\"\">\
             <h1>Hello, world!</h1>\
           </form>\
         </body>\
       </html>"
      (As_test_helper.render_page page)
;;

let test_add_hidden_field _ =
  let cgi = As_test_helper.create_test_activation [] in
  let page = Srv_page_structure.create Srv_configuration.default_conf in
  let _ = page # set_dtd "xhtml1-frameset" in
  let fc = Xnc_field_collection.create "Test" in
  let _ = Xnc_field_collection.register_string fc "string_field" "test" in
    ignore(Srv_page_structure.create_body_element page);
    ignore(Srv_page_structure.create_form_element page cgi);
    Srv_page_structure.add_simple_hidden_field page "field1" "value";
    Srv_page_structure.add_complex_hidden_field page "field2" fc;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected XHTML string"
      "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\
       <!DOCTYPE html \
            PUBLIC \"-//W3C//DTD XHTML 1.0 Frameset//EN\" \
            \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd\">\
       <html xmlns=\"http://www.w3.org/1999/xhtml\">\
         <body>\
           <form method=\"post\" action=\"\">\
             <div>\
               <input type=\"hidden\" \
                      name=\"field1\" \
                      value=\"value\" />\
               <input type=\"hidden\" \
                      name=\"field2\" \
                      value=\"Test=[string_field=&quot;test&quot;]\" />\
             </div>\
           </form>\
         </body>\
       </html>"
      (As_test_helper.render_page page)
;;

let test_add_hidden_field_without_form _ =
  let page = Srv_page_structure.create Srv_configuration.default_conf in
    page # set_dtd "xhtml1-strict";
    Srv_page_structure.add_simple_hidden_field page "field" "value";
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected XHTML string"
      "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\
       <!DOCTYPE html \
            PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \
            \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\
       <html xmlns=\"http://www.w3.org/1999/xhtml\">\
       </html>"
      (As_test_helper.render_page page)
;;

let test_create_body_element _ =
  let page = Srv_page_structure.create Srv_configuration.default_conf in
    page # set_dtd "xhtml1-transitional";
    ignore(Srv_page_structure.create_body_element page);
    Srv_page_structure.append_to_body page
      (new Srv_page_element.literal_element "<h1>Hello, world!</h1>");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected XHTML string"
      "<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\
       <!DOCTYPE html \
            PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \
            \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\
       <html xmlns=\"http://www.w3.org/1999/xhtml\">\
         <body>\
           <h1>Hello, world!</h1>\
         </body>\
       </html>"
      (As_test_helper.render_page page)
;;

(* test suite *)
let suite = "Page structure test" >:::
  ["test_render" >:: test_render;
   "test_handle_ui_events" >:: test_handle_ui_events;
   "test_create_form_element" >:: test_create_form_element;
   "test_add_hidden_field" >:: test_add_hidden_field;
   "test_add_hidden_field_without_form" >:: test_add_hidden_field_without_form;
   "test_create_body_element" >:: test_create_body_element]
;;

