open Xnc_stdlib
open Xnc_functional
open OUnit

module U = Unix;;
module T = Srv_sqlite3_state;; (* "T" stands for "tested" module *)

let spf = Printf.sprintf;;
let pf = Printf.printf;;
let norm_path = Neturl.split_path >> Neturl.norm_path >> Neturl.join_path;;
let (//) p1 p2 = Filename.concat p1 p2 |> norm_path;;

let mkdir dirname =
  try U.mkdir dirname 0o755
  with | U.Unix_error (U.EEXIST, "mkdir", dirname) -> ()
;;

let state_printer st =
  match st with
    | None -> "NON PRESENT VALUE"
    | Some s -> (
        let clientip = match s.T.client_ip with
          | None -> "None"
          | Some s -> "Some \"" ^ s ^ "\"" in
        let ua = match s.T.client_user_agent with
          | None -> "None"
          | Some s -> "Some \"" ^ s ^ "\"" in

          Printf.sprintf "{
  key = \"%s\";
  data = \"%s\";
  ctime = %0.17g;
  mtime = %0.17g;
  client_ip = %s;
  client_user_agent = %s;
}" s.T.key s.T.data s.T.ctime s.T.mtime clientip ua
    
    )
;;

Random.self_init ();;

let create_test_dir () =
  let random_string = Random.int 1000000 |> spf "%d" |>
      Digest.string |> Digest.to_hex in
  let test_base = "/tmp/ex_nunc_temp_test" in
    mkdir test_base;
    let random_dir = test_base//random_string in
      mkdir random_dir;
      random_dir
;;  (* END of create_test_dir *)

let test_dir = create_test_dir ();;

let make_rand_string size =
  let str = String.create size in
    for i = 0 to (size - 1) do
        let r = 1 + Random.int 255 in (* every byte value except 0 *)
        str.[i] <- Char.chr r;
    done;
    str
;;

let big_key = make_rand_string (1024*10);;
let big_value = make_rand_string (1024*1024);;

let connect_01 () =
  let db_name = "/impossible_dir"//"connect_01.db" in
    assert_raises
      ~msg:("connect must fail to create this file (" ^ db_name ^ ")")
      (Sqlite3.Error "error opening database: unable to open database file")
      (fun () -> T.connect ~check_schema:false db_name |> ignore)
;;  (* END of connect_01 *)

let set_up () =
  let dir = test_dir//"add_or_replace" in
    mkdir dir;
    let db_name = dir//"test.db" in
      T.connect ~check_schema:true db_name
;;

let tear_down st_mgr =
  T.disconnect st_mgr |> ignore;
;;

let add_or_replace_01 st_mgr =
  T.add_or_replace st_mgr
    ~client_ip:"192.168.10.10"
    ~client_user_agent:"My user agent"
    ~ctime:0.0
    ~mtime:3.15 (* Non representable float number! *)
    "key_01" "value_01";
;;  (* END of add_or_replace_01 *)

let add_or_replace_02 st_mgr =
  T.add_or_replace st_mgr
    ~client_ip:"192.168.10.10"
    ~client_user_agent:"My user agent"
    ~ctime:0.0
    ~mtime:3.15 (* Non representable float number! *)
    "this will be removed" "value_01";
;;  (* END of add_or_replace_02 *)

let add_or_replace_03 st_mgr =
  T.add_or_replace st_mgr
    ~client_ip:"CLIENT IP"
    ~client_user_agent:"CLIENT USER AGENT"
    ~ctime:100.0
    ~mtime:200.0
    "to be replaced" "OLD VALUE";
  T.add_or_replace st_mgr
    ~client_ip:"192.168.10.10"
    ~client_user_agent:"My user agent"
    ~ctime:0.0
    ~mtime:3.15 (* Non representable float number! *)
    "to be replaced" "NEW VALUE";
;;  (* END of add_or_replace_03 *)

let add_or_replace_04 st_mgr =
  T.add_or_replace st_mgr ~ctime:100.0 ~mtime:200.0 big_key big_value;
;;  (* END of add_or_replace_04 *)

let add_or_replace_05 st_mgr =
  assert_raises
    ~msg:("A string containing a 0 value has passed")
    (T.Invalid_string)
    (fun () -> T.add_or_replace st_mgr big_key "INVALID \000";)
;;  (* END of add_or_replace_05 *)

let timestamp_1 = ref 0.0;;
let add_or_replace_06 st_mgr =
  timestamp_1 := Unix.time ();
  T.add_or_replace st_mgr "key_06" "value_06";
;;  (* END of add_or_replace_06 *)

let timestamp_2 = ref 0.0;;
let timestamp_3 = ref 0.0;;
let add_or_replace_07 st_mgr =
  timestamp_2 := Unix.time ();
  T.add_or_replace st_mgr "key_07" "value_07";
  let rand_wait = (1 + Random.int 3) in
    Printf.printf " Please, wait %d seconds %!" rand_wait;
    Unix.sleep rand_wait;
    Printf.printf "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b";
    Printf.printf "                        ";
    Printf.printf "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b%!";
    timestamp_3 := Unix.time ();
    T.add_or_replace st_mgr "key_07" "value_07_replaced";
;;  (* END of add_or_replace_07 *)

let add_or_replace_08 st_mgr =
  let n = 100 in
    Printf.printf " Inserting %d random key/value: " n;
    for i = 1 to n do
      Printf.printf "%04d/%04d\b\b\b\b\b\b\b\b\b\b %!" i n;
      let ctime = Unix.time () |> Random.float in
      let mtime = Unix.time () |> Random.float in
      let key = make_rand_string 64 in
      let value = make_rand_string 1024 in
        T.add_or_replace st_mgr ~ctime:ctime ~mtime:mtime key value;
    done;
      Printf.printf "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b";
      Printf.printf "                                            ";
      Printf.printf "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b";
      Printf.printf "\b\b\b\b\b\b\b\b\b\b\b\b%!";
;;  (* END of add_or_replace_08 *)

let retrieve_01 st_mgr =
  let st = T.retrieve st_mgr "key_01" in
  let expected = {
      T.key = "key_01";
      T.data = "value_01";
      T.ctime = 0.0;
      T.mtime = 3.15;
      T.client_ip = Some "192.168.10.10";
      T.client_user_agent = Some "My user agent";
    } in
    assert_equal ~printer:state_printer
      ~msg:"The expected value cannot be retrieved"
      (Some expected) st;
;;  (* END of retrieve_01 *)

let retrieve_02 st_mgr =
  let st = T.retrieve st_mgr "key not present" in
    assert_equal ~printer:state_printer
      ~msg:"Here we expect a None" None st;
;;  (* END of retrieve_02 *)

let retrieve_03 st_mgr =
  let st = T.retrieve st_mgr "this will be removed" in
    assert_equal ~printer:state_printer
      ~msg:"This value shouldn't be here!" None st;
;;  (* END of retrieve_03 *)

let retrieve_04 st_mgr =
  let st = T.retrieve st_mgr "to be replaced" in
  let expected = {
      T.key = "to be replaced";
      T.data = "NEW VALUE";
      T.ctime = 0.0;
      T.mtime = 3.15;
      T.client_ip = Some "192.168.10.10";
      T.client_user_agent = Some "My user agent";
    } in
    assert_equal ~printer:state_printer
      ~msg:"The expected value cannot be retrieved"
      (Some expected) st;
;;  (* END of retrieve_04 *)

let retrieve_05 st_mgr =
  let st = T.retrieve st_mgr big_key in
  let expected = {
      T.key = big_key;
      T.data = big_value;
      T.ctime = 100.0;
      T.mtime = 200.0;
      T.client_ip = None;
      T.client_user_agent = None;
    } in
    assert_equal ~printer:state_printer
      ~msg:"Error while retrieving a big key/value"
      (Some expected) st;
;;  (* END of retrieve_05 *)

let retrieve_06 st_mgr =
  let st = T.retrieve st_mgr "key_06" in
  let expected = {
      T.key = "key_06";
      T.data = "value_06";
      T.ctime = !timestamp_1;
      T.mtime = !timestamp_1;
      T.client_ip = None;
      T.client_user_agent = None;
    } in
    assert_equal ~printer:state_printer
      ~msg:"Error in the stored timestamp"
      (Some expected) st;
;;  (* END of retrieve_06 *)

let retrieve_07 st_mgr =
  let st = T.retrieve st_mgr "key_07" in
  let expected = {
      T.key = "key_07";
      T.data = "value_07_replaced";
      T.ctime = !timestamp_2;
      T.mtime = !timestamp_3;
      T.client_ip = None;
      T.client_user_agent = None;
    } in
    assert_equal ~printer:state_printer
      ~msg:"Error in the stored mtime timestamp"
      (Some expected) st;
;;  (* END of retrieve_07 *)

let fold_01 st_mgr =
  let f (last_tm, res) st =
    if res = true then
      if last_tm <= st.T.ctime then (st.T.ctime, true) else (0.0, false)
    else (0.0, false) in
      "fold doesn't sort by ctime" @?
        (T.fold ~order:T.By_ctime f ((-1.0), true) st_mgr |> snd)
;; (* END of fold_01 *)

let fold_02 st_mgr =
  let f (last_tm, res) st =
    if res = true then
      if last_tm >= st.T.ctime then (st.T.ctime, true) else (0.0, false)
    else (0.0, false) in
      "fold doesn't sort by ctime (DESC)" @?
        (T.fold ~order:T.By_ctime_desc f (max_float, true) st_mgr |> snd)
;; (* END of fold_02 *)

let fold_03 st_mgr =
  let f (last_tm, res) st =
    if res = true then
      if last_tm <= st.T.mtime then (st.T.mtime, true) else (0.0, false)
    else (0.0, false) in
      "fold doesn't sort by mtime" @?
        (T.fold ~order:T.By_mtime f ((-1.0), true) st_mgr |> snd)
;; (* END of fold_03 *)

let fold_04 st_mgr =
  let f (last_tm, res) st =
    if res = true then
      if last_tm >= st.T.mtime then (st.T.mtime, true) else (0.0, false)
    else (0.0, false) in
      "fold doesn't sort by mtime (DESC)" @?
        (T.fold ~order:T.By_mtime_desc f (max_float, true) st_mgr |> snd)
;; (* END of fold_04 *)

let iter_01 st_mgr =
  let last_tm = ref (-1.0) in
  let res = ref true in
  let f st =
    if !res = true then
      if !last_tm <= st.T.ctime then last_tm := st.T.ctime else res := false
  in
    let () = T.iter ~order:T.By_ctime f st_mgr in
      "iter doesn't sort by ctime" @? !res
;; (* END of iter_01 *)

let remove_01 st_mgr =
  let x = T.remove st_mgr "this will be removed" in
    x
;;  (* END of remove_01 *)

let count_states st_mgr = T.fold (fun count _ -> count + 1) 0 st_mgr;;

let garbage_collect_01 st_mgr =
  let count_before = count_states st_mgr in
  let deleted = T.garbage_collect ~pred:(fun _ -> false) st_mgr in
  let count_after = count_states st_mgr in
    assert_equal ~msg:"garbage_collect returned non zero, which is incorrect!"
      deleted 0;
    assert_equal ~msg:"Some states have been deleted, which is incorrect!"
      count_before count_after
;; (* END of garbage_collect_01 *)

let garbage_collect_02 st_mgr =
  let count_before = count_states st_mgr in
  let deleted = T.garbage_collect ~pred:(fun _ -> true) st_mgr in
  let count_after = count_states st_mgr in
    assert_equal ~msg:"garbage_collect returned unexpected value"
      deleted count_before;
    assert_equal ~msg:"not all states have been deleted"
      count_after 0
;; (* END of garbage_collect_02 *)

(* test suite *)
let suite = "Sqlite3 state storage" >:::
  [
    (
      "connect" >::: [
        "impossible filename" >:: connect_01;
      ]
    );
    (
      "disconnect" >::: [
        (* TODO Don't know how to test this! *)
      ]
    );
    (
      "add_or_replace" >::: [
        "add a value" >:: (bracket set_up add_or_replace_01 tear_down);
        "value to be removed" >:: (bracket set_up add_or_replace_02 tear_down);
        "value to be replaced" >:: (bracket set_up add_or_replace_03 tear_down);
        "add very big key/value" >:: (bracket set_up add_or_replace_04 tear_down);
        "0 value in the value" >:: (bracket set_up add_or_replace_05 tear_down);
        "insert value w/o times" >:: (bracket set_up add_or_replace_06 tear_down);
        "test mtime updating" >:: (bracket set_up add_or_replace_07 tear_down);
        "store 5 values with known times" >:: (bracket set_up add_or_replace_08 tear_down);
      ]
    );
    (
      "remove" >::: [
        "remove a value" >:: (bracket set_up remove_01 tear_down);
      ]
    );
    (
      "retrieve" >::: [
        "retrieve value" >:: (bracket set_up retrieve_01 tear_down);
        "retrieve none" >:: (bracket set_up retrieve_02 tear_down);
        "retrieve removed" >:: (bracket set_up retrieve_03 tear_down);
        "retrieve replaced" >:: (bracket set_up retrieve_04 tear_down);
        "retrieve very big key/value" >:: (bracket set_up retrieve_05 tear_down);
        "retrieve value w/o times" >:: (bracket set_up retrieve_06 tear_down);
        "retrieve updated mtime" >:: (bracket set_up retrieve_07 tear_down);
      ]
    );
    (
      "fold" >::: [
        "test ctime order" >:: (bracket set_up fold_01 tear_down);
        "test ctime order (DESC)" >:: (bracket set_up fold_02 tear_down);
        "test mtime order" >:: (bracket set_up fold_03 tear_down);
        "test mtime order (DESC)" >:: (bracket set_up fold_04 tear_down);
      ]
    );
    (
      "iter" >::: [
        "test ctime order" >:: (bracket set_up iter_01 tear_down);
      ]
    );
    (
      "garbage_collect" >::: [
        "delete none" >:: (bracket set_up garbage_collect_01 tear_down);
        "delete all" >:: (bracket set_up garbage_collect_02 tear_down);
      ]
    );
  ]
;;

