open OUnit

(* Session *)
module Concrete_session_data =
  Test_types.Concrete_session_data
;;

module Concrete_session =
  Srv_state.Make(struct
                   type t = Concrete_session_data.t
                   let to_field_collection =
                     Concrete_session_data.to_field_collection
                   let from_field_collection _ f =
                     Concrete_session_data.from_field_collection f
                 end)
;;

(* tests *)
let test_session_generate_id _ =
  let session = Concrete_session.State.create
                  (Concrete_session_data.create ())
                  "127.0.0.1"
                  "Firefox" in
  let session_id = Concrete_session.State.get_id session in
    OUnit.assert_equal
      ~msg:"Unexpected session id"
      None
      session_id;

  let () = Concrete_session.State.generate_id session in
  let session_id = Option.get (Concrete_session.State.get_id session) in
    OUnit.assert_equal
      ~msg:"Unexpected session id length"
      64
      (String.length session_id)
;;

let test_session_generate_id_once _ =
  let session = Concrete_session.State.create
                  (Concrete_session_data.create ())
                  "127.0.0.1"
                  "Firefox" in
  let () = Concrete_session.State.generate_id_once session in
  let session_id = Concrete_session.State.get_id session in
  let () = Concrete_session.State.generate_id_once session in
  let session_id' = Concrete_session.State.get_id session in
    OUnit.assert_equal
      ~msg:"Different session ids generated by generate_id_once"
      session_id
      session_id'
;;

let test_session_state_to_hidden_field _ =
  let session_manager = Concrete_session.Manager.create
                          Srv_types.Client
                          Srv_configuration.default_conf in
  let session = Concrete_session.State.create
                  (Concrete_session_data.create ())
                  "127.0.0.1"
                  "Firefox" in
  let fc = Concrete_session.Manager.state_to_field_collection
             session_manager
             session in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected default session serialization"
      "Session=[a_int=42;\
                a_float=3.14;\
                a_char=a;\
                a_string=&quot;test&quot;;\
                a_bool=true]"
      (Xnc_field_collection.to_string fc)
;;

let test_session_state_from_hidden_field _ =
  let session_manager = Concrete_session.Manager.create
                          Srv_types.Client
                          Srv_configuration.default_conf in
  let session_data = Concrete_session_data.create () in
  let session = Concrete_session.State.create
                  session_data
                  "127.0.0.1"
                  "Firefox" in
  let fc = Concrete_session_data.to_field_collection session_data in
  let _ = Xnc_field_collection.from_string fc
            "Session=[a_int=42;\
                      a_float=3.14;\
                      a_char=a;\
                      a_string=&quot;test&quot;;\
                      a_bool=true]" in
  let session = Concrete_session.Manager.state_from_field_collection
                  session_manager
                  session
                  fc
                  30 in
  let session_data = Concrete_session.State.get_data session in
    OUnit.assert_equal
      ~msg:"Unexpected value session id"
      None
      (Concrete_session.State.get_id session);
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected value for a_int parameter"
      42
      session_data.Concrete_session_data.a_int;
    OUnit.assert_equal
      ~printer:string_of_float
      ~msg:"Unexpected value for a_float parameter"
      3.14
      session_data.Concrete_session_data.a_float;
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value for a_string parameter"
      "test"
      session_data.Concrete_session_data.a_string
;;

let test_session_state_to_memory _ =
  let session_manager = Concrete_session.Manager.create
                          Srv_types.In_memory
                          Srv_configuration.default_conf in
  let session = Concrete_session.State.create
                  (Concrete_session_data.create ())
                  "127.0.0.1"
                  "Firefox" in
  let fc = Concrete_session.Manager.state_to_field_collection
             session_manager
             session in
    OUnit.assert_bool
      "Session id serialization does not match correct pattern"
      (Str.string_match
         (Str.regexp "^State=\
                         \\[id=&quot;[a-z0-9]+&quot;\\]$")
         (Xnc_field_collection.to_string fc)
         0
      )
;;

let test_session_state_from_memory _ =
  let session_manager = Concrete_session.Manager.create
                          Srv_types.In_memory
                          Srv_configuration.default_conf in
  let session = Concrete_session.State.create
                  (Concrete_session_data.create ())
                  "127.0.0.1"
                  "Firefox" in
  let () = Concrete_session.Manager.persist session_manager session in
  let fc = Concrete_session.Manager.state_to_field_collection
             session_manager
             session in
  let session_id = Xnc_field_collection.get_option fc
                     "id" Xnc_serializable.unwrap_string in
  let session' = Concrete_session.Manager.state_from_field_collection
                   session_manager
                   session
                   fc
                   30 in
  let session_data = Concrete_session.State.get_data session in
  let session_data' = Concrete_session.State.get_data session' in
    OUnit.assert_equal
      ~msg:"Unexpected session id restored from in-memory storage"
      session_id
      (Concrete_session.State.get_id session');
    OUnit.assert_equal
      ~msg:"Unexpected session data restored from in-memory storage"
      session_data
      session_data'
;;

let test_session_state_to_sqlite3 _ =
  let session_manager = Concrete_session.Manager.create
                          Srv_types.Sqlite3
                          Srv_configuration.default_conf in
  let session = Concrete_session.State.create
                  (Concrete_session_data.create ())
                  "127.0.0.1"
                  "Firefox" in
  let fc = Concrete_session.Manager.state_to_field_collection
             session_manager
             session in
    OUnit.assert_bool
      "Session id serialization does not match correct pattern"
      (Str.string_match
         (Str.regexp "^State=\
                         \\[id=&quot;[a-z0-9]+&quot;\\]$")
         (Xnc_field_collection.to_string fc)
         0
      )
;;

let test_session_state_from_sqlite3 _ =
  let session_manager = Concrete_session.Manager.create
                          Srv_types.Sqlite3
                          Srv_configuration.default_conf in
  let session = Concrete_session.State.create
                  (Concrete_session_data.create ())
                  "127.0.0.1"
                  "Firefox" in
  let () = Concrete_session.Manager.persist session_manager session in
  let fc = Concrete_session.Manager.state_to_field_collection
             session_manager
             session in
  let session_id = Xnc_field_collection.get_option fc
                     "id" Xnc_serializable.unwrap_string in
  let session' = Concrete_session.Manager.state_from_field_collection
                   session_manager
                   session
                   fc
                   30 in
  let session_data = Concrete_session.State.get_data session in
  let session_data' = Concrete_session.State.get_data session' in
    OUnit.assert_equal
      ~msg:"Unexpected session id restored from DBM storage"
      session_id
      (Concrete_session.State.get_id session');
    OUnit.assert_equal
      ~msg:"Unexpected session data restored from DBM storage"
      session_data
      session_data'
;;

let test_session_load_from_serialized_state _ =
  let config = Srv_configuration.default_conf in
  let serialized_state =
    { Srv_state_persisting.Serialized.part =
        Srv_state_persisting.Session;
      Srv_state_persisting.Serialized.location =
        Srv_state_persisting.Argument_value;
      Srv_state_persisting.Serialized.content =
        "Session=[a_int=1;\
                  a_float=0.5;\
                  a_char=c;\
                  a_string=&quot;string&quot;;\
                  a_bool=false]"
    } in
  let manager = Concrete_session.Manager.create
                  Srv_types.Client
                  Srv_configuration.default_conf in
  let session = Concrete_session.State.create
                  (Concrete_session_data.create ())
                  "127.0.0.1"
                  "Firefox" in
  let session = Concrete_session.Manager.load_from_serialized_state
                  manager
                  session
                  serialized_state
                  config
                  30 in
    OUnit.assert_equal
      ~msg:"Unexpected session id"
      None
      (Concrete_session.State.get_id session);
    OUnit.assert_equal
      ~msg:"Unexpected session data"
      { Concrete_session_data.a_int = 1;
        Concrete_session_data.a_float = 0.5;
        Concrete_session_data.a_char = 'c';
        Concrete_session_data.a_string = "string";
        Concrete_session_data.a_bool = false
      }
      (Concrete_session.State.get_data session);
;;

(* Page parameters *)
module Concrete_page_parameters =
  Test_types.Concrete_page_parameters
;;

module Concrete_parameters =
  Srv_state.Make(Concrete_page_parameters)
;;

(* tests *)
let test_params_generate_id _ =
  let parameters_state = Concrete_parameters.State.create
                           (Concrete_page_parameters.Test_page_parameters
                              (42, true))
                           "127.0.0.1"
                           "Firefox" in
  let id = Concrete_parameters.State.get_id parameters_state in
    OUnit.assert_equal
      ~msg:"Unexpected parameters id"
      None
      id;

  let () =
    Concrete_parameters.State.generate_id parameters_state in
  let id = Option.get (Concrete_parameters.State.get_id parameters_state) in
    OUnit.assert_bool
      "Parameters id does not match correct pattern"
      (Str.string_match
         (Str.regexp "^[a-z0-9]+$")
         id
         0
      );
;;

let test_params_generate_id_once _ =
  let parameters_state = Concrete_parameters.State.create
                           (Concrete_page_parameters.Test_page_parameters
                              (42, true))
                           "127.0.0.1"
                           "Firefox" in
  let () =
    Concrete_parameters.State.generate_id_once parameters_state in
  let id = Concrete_parameters.State.get_id parameters_state in
  let () =
    Concrete_parameters.State.generate_id_once parameters_state in
  let id' = Concrete_parameters.State.get_id parameters_state in
    OUnit.assert_equal
      ~msg:"Different parameters ids generated by generate_id_once"
      id
      id'
;;

let test_params_state_to_hidden_field _ =
  let manager = Concrete_parameters.Manager.create
                  Srv_types.Client
                  Srv_configuration.default_conf in
  let parameters_state = Concrete_parameters.State.create
                           (Concrete_page_parameters.Test_page_parameters
                              (42, true))
                           "127.0.0.1"
                           "Firefox" in
  let fc = Concrete_parameters.Manager.state_to_field_collection
             manager
             parameters_state in
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected parameters serialization"
      "Parameters=[integer=42;boolean=true]"
      (Xnc_field_collection.to_string fc);
;;

let test_params_state_from_hidden_field _ =
  let manager = Concrete_parameters.Manager.create
                  Srv_types.Client
                  Srv_configuration.default_conf in
  let parameters_state = Concrete_parameters.State.create
                           (Concrete_page_parameters.Test_page_parameters
                              (42, true))
                           "127.0.0.1"
                           "Firefox" in
  let parameters = Concrete_parameters.State.get_data parameters_state in
  let fc = Concrete_page_parameters.to_field_collection parameters in
  let _ = Xnc_field_collection.from_string fc
            "Parameters=[integer=10;\
                         boolean=false]" in
  let parameters_state =
    Concrete_parameters.Manager.state_from_field_collection
      manager
      parameters_state
      fc
      30
  in
    OUnit.assert_equal
      ~msg:"Unexpected parameters id"
      None
      (Concrete_parameters.State.get_id parameters_state);
    OUnit.assert_equal
      ~msg:"Unexpected page parameters"
      (Concrete_page_parameters.Test_page_parameters (10, false))
      (Concrete_parameters.State.get_data parameters_state);
;;

let test_params_state_to_memory _ =
  let manager = Concrete_parameters.Manager.create
                  Srv_types.In_memory
                  Srv_configuration.default_conf in
  let parameters_state = Concrete_parameters.State.create
                           (Concrete_page_parameters.Test_page_parameters
                              (42, true))
                           "127.0.0.1"
                           "Firefox" in
  let fc = Concrete_parameters.Manager.state_to_field_collection
             manager
             parameters_state in
    OUnit.assert_bool
      "Signature id serialization does not match correct pattern"
      (Str.string_match
         (Str.regexp "^State=\
                         \\[id=&quot;[a-z0-9]+&quot;\\]$")
         (Xnc_field_collection.to_string fc)
         0
      );
;;

let test_params_state_from_memory _ =
  let manager = Concrete_parameters.Manager.create
                  Srv_types.In_memory
                  Srv_configuration.default_conf in
  let parameters_state = Concrete_parameters.State.create
                           (Concrete_page_parameters.Test_page_parameters
                              (42, true))
                           "127.0.0.1"
                           "Firefox" in
  let () = Concrete_parameters.Manager.persist manager parameters_state in
  let fc = Concrete_parameters.Manager.state_to_field_collection
             manager
             parameters_state in
  let id = Xnc_field_collection.get_option fc
             "id" Xnc_serializable.unwrap_string in
  let parameters_state' =
    Concrete_parameters.Manager.state_from_field_collection
      manager
      parameters_state
      fc
      30 in
  let parameters = Concrete_parameters.State.get_data parameters_state in
  let parameters' = Concrete_parameters.State.get_data parameters_state' in
    OUnit.assert_equal
      ~msg:"Unexpected parameters id restored from in-memory storage"
      id
      (Concrete_parameters.State.get_id parameters_state');
    OUnit.assert_equal
      ~msg:"Unexpected page parameters restored from in-memory storage"
      parameters
      parameters'
;;

let test_params_state_to_sqlite3 _ =
  let manager = Concrete_parameters.Manager.create
                  Srv_types.Sqlite3
                  Srv_configuration.default_conf in
  let parameters_state = Concrete_parameters.State.create
                           (Concrete_page_parameters.Test_page_parameters
                              (42, true))
                           "127.0.0.1"
                           "Firefox" in
  let fc = Concrete_parameters.Manager.state_to_field_collection
             manager
             parameters_state in
    OUnit.assert_bool
      "Signature id serialization does not match correct pattern"
      (Str.string_match
         (Str.regexp "^State=\
                         \\[id=&quot;[a-z0-9]+&quot;\\]$")
         (Xnc_field_collection.to_string fc)
         0
      );
;;

let test_params_state_from_sqlite3 _ =
  let manager = Concrete_parameters.Manager.create
                  Srv_types.Sqlite3
                  Srv_configuration.default_conf in
  let parameters_state = Concrete_parameters.State.create
                           (Concrete_page_parameters.Test_page_parameters
                              (42, true))
                           "127.0.0.1"
                           "Firefox" in
  let () = Concrete_parameters.Manager.persist manager parameters_state in
  let fc = Concrete_parameters.Manager.state_to_field_collection
             manager
             parameters_state in
  let id = Xnc_field_collection.get_option fc
             "id" Xnc_serializable.unwrap_string in
  let parameters_state' =
    Concrete_parameters.Manager.state_from_field_collection
      manager
      parameters_state
      fc
      30 in
  let parameters = Concrete_parameters.State.get_data parameters_state in
  let parameters' = Concrete_parameters.State.get_data parameters_state' in
    OUnit.assert_equal
      ~msg:"Unexpected parameters id restored from DBM storage"
      id
      (Concrete_parameters.State.get_id parameters_state');
    OUnit.assert_equal
      ~msg:"Unexpected page parameters restored from DBM storage"
      parameters
      parameters'
;;

let test_params_load_from_serialized_state _ =
  let parameters_state = Concrete_parameters.State.create
                           (Concrete_page_parameters.Test_page_parameters
                              (42, true))
                           "127.0.0.1"
                           "Firefox" in
  let config = Srv_configuration.default_conf in
  let serialized_state =
    { Srv_state_persisting.Serialized.part =
        Srv_state_persisting.Page_parameters;
      Srv_state_persisting.Serialized.location =
        Srv_state_persisting.Argument_value;
      Srv_state_persisting.Serialized.content =
        "Parameters=[integer=10;boolean=false]"
    } in
  let manager = Concrete_parameters.Manager.create
                  Srv_types.Client
                  Srv_configuration.default_conf in
  let parameters_state =
    Concrete_parameters.Manager.load_from_serialized_state
      manager
      parameters_state
      serialized_state
      config
      30
  in
    OUnit.assert_equal
      ~msg:"Unexpected parameters id"
      None
      (Concrete_parameters.State.get_id parameters_state);
    OUnit.assert_equal
      ~msg:"Unexpected page parameters"
      (Concrete_page_parameters.Test_page_parameters (10, false))
      (Concrete_parameters.State.get_data parameters_state);
;;

(* test suite *)
let suite = "State test" >:::
  ["test_session_generate_id" >:: test_session_generate_id;
   "test_session_generate_id_once" >:: test_session_generate_id_once;
   "test_session_state_to_hidden_field" >:: test_session_state_to_hidden_field;
   "test_session_state_from_hidden_field"
     >:: test_session_state_from_hidden_field;
   "test_session_state_to_memory" >:: test_session_state_to_memory;
   "test_session_state_from_memory" >:: test_session_state_from_memory;
   "test_session_state_to_sqlite3" >:: test_session_state_to_sqlite3;
   "test_session_state_from_sqlite3" >:: test_session_state_from_sqlite3;
   "test_session_load_from_serialized_state"
     >:: test_session_load_from_serialized_state;
   "test_params_generate_id" >:: test_params_generate_id;
   "test_params_generate_id_once" >:: test_params_generate_id_once;
   "test_params_state_to_hidden_field" >:: test_params_state_to_hidden_field;
   "test_params_state_from_hidden_field"
     >:: test_params_state_from_hidden_field;
   "test_params_state_to_memory" >:: test_params_state_to_memory;
   "test_params_state_from_memory" >:: test_params_state_from_memory;
   "test_params_state_to_sqlite3" >:: test_params_state_to_sqlite3;
   "test_params_state_from_sqlite3" >:: test_params_state_from_sqlite3;
   "test_params_load_from_serialized_state"
     >:: test_params_load_from_serialized_state]
;;

