open OUnit

module Concrete_page_parameters =
struct
  type t =
      No_parameters
    | Start_page_parameters of float
    | Error_page_parameters of string * string
    | Not_found_page_parameters of string
    | Test_page_parameters of int * bool

  let default_start_page_parameters =
    Start_page_parameters 3.14

  let create_error_page_parameters error_description backtrace =
    Error_page_parameters (error_description, backtrace)
  let default_error_page_parameters =
    create_error_page_parameters "" ""
  let get_error_description parameters =
    match parameters with
        Error_page_parameters (error, _) -> error
      | _ -> raise Srv_types.Invalid_page_parameters
  let get_backtrace parameters =
    match parameters with
        Error_page_parameters (_, bt) -> bt
      | _ -> raise Srv_types.Invalid_page_parameters

  let create_not_found_page_parameters missing_resource =
    Not_found_page_parameters missing_resource
  let default_not_found_page_parameters =
    create_not_found_page_parameters ""
  let get_missing_resource parameters =
    match parameters with
        Not_found_page_parameters mr -> mr
      | _ -> raise Srv_types.Invalid_page_parameters

  let to_field_collection parameters =
    let fc = Xnc_field_collection.create "Parameters" in
    let _ = match parameters with
        No_parameters ->
          ()
      | Start_page_parameters f ->
          Xnc_field_collection.register_float fc "float" f
      | Error_page_parameters (error, backtrace) ->
          Xnc_field_collection.register_string fc "error" error;
          Xnc_field_collection.register_string fc "backtrace" backtrace
      | Not_found_page_parameters resource ->
          Xnc_field_collection.register_string fc "resource" resource;
      | Test_page_parameters (integer, boolean) ->
          Xnc_field_collection.register_int fc "integer" integer;
          Xnc_field_collection.register_bool fc "boolean" boolean
    in
      fc

  let from_field_collection parameters fc =
    match parameters with
        No_parameters as p ->
          p
      | Start_page_parameters _ ->
          let f = Xnc_field_collection.get_float fc "float" in
            (Start_page_parameters f)
      | Error_page_parameters _ ->
          let error = Xnc_field_collection.get_string fc "error" in
          let backtrace = Xnc_field_collection.get_string fc "backtrace" in
            (Error_page_parameters (error, backtrace))
      | Not_found_page_parameters _ ->
          let resource = Xnc_field_collection.get_string fc "resource" in
            (Not_found_page_parameters resource)
      | Test_page_parameters _ ->
          let integer = Xnc_field_collection.get_int fc "integer" in
          let boolean = Xnc_field_collection.get_bool fc "boolean" in
            (Test_page_parameters (integer, boolean))

end;;

module Local_page_parameters : Srv_types.PAGE_PARAMETERS =
  Concrete_page_parameters
;;

module Concrete_session_data =
struct
  type t =
      { mutable a_int : int;
        mutable a_float : float;
        mutable a_char : char;
        mutable a_string : string;
        mutable a_bool : bool
      }

  let create () =
    { a_int = 42;
      a_float = 3.14;
      a_char = 'a';
      a_string = "test";
      a_bool = true
    }

  let to_field_collection s =
    let fc = Xnc_field_collection.create "Session" in
      Xnc_field_collection.register_int fc "a_int" s.a_int;
      Xnc_field_collection.register_float fc "a_float" s.a_float;
      Xnc_field_collection.register_char fc "a_char" s.a_char;
      Xnc_field_collection.register_string fc "a_string" s.a_string;
      Xnc_field_collection.register_bool fc "a_bool" s.a_bool;
      fc

  let from_field_collection fc =
    let a_int = Xnc_field_collection.get_int fc "a_int" in
    let a_float = Xnc_field_collection.get_float fc "a_float" in
    let a_char = Xnc_field_collection.get_char fc "a_char" in
    let a_string = Xnc_field_collection.get_string fc "a_string" in
    let a_bool = Xnc_field_collection.get_bool fc "a_bool" in
      { a_int = a_int;
        a_float = a_float;
        a_char = a_char;
        a_string = a_string;
        a_bool = a_bool
      }

end

module Local_session_data : Srv_types.SESSION_DATA =
  Concrete_session_data
;;

(* tests *)
let test_to_field_collection_error_parameters _ =
  let parameters =
    Local_page_parameters.create_error_page_parameters "Test" "" in
  let field_collection = Local_page_parameters.to_field_collection parameters in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      2
      (Xnc_field_collection.length field_collection);
    OUnit.assert_bool
      "Failed registering error parameter"
      (Xnc_field_collection.is_registered field_collection "error");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value for error parameter"
      "Test"
      (Xnc_field_collection.get_string field_collection "error");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value for backtrace parameter"
      ""
      (Xnc_field_collection.get_string field_collection "backtrace");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected serialization"
      "Parameters=[error=&quot;Test&quot;;backtrace=&quot;&quot;]"
      (Xnc_field_collection.to_string field_collection)
;;

let test_to_field_collection_start_page_params _ =
  let parameters = Local_page_parameters.default_start_page_parameters in
  let field_collection = Local_page_parameters.to_field_collection parameters in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      1
      (Xnc_field_collection.length field_collection);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected serialization"
      "Parameters=[float=3.14]"
      (Xnc_field_collection.to_string field_collection)
;;

let test_to_field_collection_not_found_page_params _ =
  let parameters = Local_page_parameters.default_not_found_page_parameters in
  let field_collection = Local_page_parameters.to_field_collection parameters in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      1
      (Xnc_field_collection.length field_collection);
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected serialization"
      "Parameters=[resource=&quot;&quot;]"
      (Xnc_field_collection.to_string field_collection)
;;

let test_to_field_collection_custom_params _ =
  let parameters = Concrete_page_parameters.Test_page_parameters (10, false) in
  let field_collection = Concrete_page_parameters.to_field_collection parameters in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      2
      (Xnc_field_collection.length field_collection);
    OUnit.assert_bool
      "Failed registering integer parameter"
      (Xnc_field_collection.is_registered field_collection "integer");
    OUnit.assert_bool
      "Failed registering boolean parameter"
      (Xnc_field_collection.is_registered field_collection "boolean");
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected value for error parameter"
      10
      (Xnc_field_collection.get_int field_collection "integer");
    OUnit.assert_equal
      ~printer:string_of_bool
      ~msg:"Unexpected value for boolean parameter"
      false
      (Xnc_field_collection.get_bool field_collection "boolean");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected serialization"
      "Parameters=[integer=10;boolean=false]"
      (Xnc_field_collection.to_string field_collection)
;;

let test_from_field_collection_error_parameters _ =
  let parameters_orig = Local_page_parameters.create_error_page_parameters
                          "Test" "" in
  let field_collection = Local_page_parameters.to_field_collection
                           parameters_orig in
  let parameters_new = Local_page_parameters.from_field_collection
                         parameters_orig field_collection in
  OUnit.assert_equal
    ~msg:"Error during error-parameters reconstruction from field_collection"
    parameters_orig
    parameters_new
;;

let test_from_field_collection_start_page_params _ =
  let parameters_orig = Local_page_parameters.default_start_page_parameters in
  let field_collection = Local_page_parameters.to_field_collection
                           parameters_orig in
  let parameters_new = Local_page_parameters.from_field_collection
                         parameters_orig field_collection in
  OUnit.assert_equal
    ~msg:"Error during default start page parameters \
          reconstruction from field_collection"
    parameters_orig
    parameters_new
;;

let test_from_field_collection_not_found_page_params _ =
  let parameters_orig =
    Local_page_parameters.default_not_found_page_parameters in
  let field_collection = Local_page_parameters.to_field_collection
                           parameters_orig in
  let parameters_new = Local_page_parameters.from_field_collection
                         parameters_orig field_collection in
  OUnit.assert_equal
    ~msg:"Error during default not found (404) page parameters \
          reconstruction from field_collection"
    parameters_orig
    parameters_new
;;

let test_from_field_collection_custom_params _ =
  let parameters_orig = Concrete_page_parameters.Test_page_parameters
                          (10, false) in
  let field_collection = Concrete_page_parameters.to_field_collection
                           parameters_orig in
  let parameters_new = Concrete_page_parameters.from_field_collection
                         parameters_orig field_collection in
  OUnit.assert_equal
    ~msg:"Error during custom-parameters reconstruction from field_collection"
    parameters_orig
    parameters_new
;;

let test_to_field_collection_session _ =
  let session = Local_session_data.create () in
  let field_collection = Local_session_data.to_field_collection session in
    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Unexpected number of fields"
      5
      (Xnc_field_collection.length field_collection);
    OUnit.assert_bool
      "Failed registering a_string parameter"
      (Xnc_field_collection.is_registered field_collection "a_string");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected value for a_string parameter"
      "test"
      (Xnc_field_collection.get_string field_collection "a_string");
    OUnit.assert_equal
      ~printer:Std.identity
      ~msg:"Unexpected serialization"
      "Session=[a_int=42;a_float=3.14;a_char=a;a_string=&quot;test&quot;;a_bool=true]"
      (Xnc_field_collection.to_string field_collection)
;;

let test_from_field_collection_session _ =
  let s = Local_session_data.create () in
  let fc = Local_session_data.to_field_collection s in
  let s' = Local_session_data.from_field_collection fc in
    OUnit.assert_equal
      ~msg:"Error during session reconstruction from field_collection"
      s
      s'
;;

(* test suite *)
let suite = "Server module types test" >:::
  ["test_to_field_collection_error_parameters"
     >:: test_to_field_collection_error_parameters;
   "test_to_field_collection_start_page_params"
     >:: test_to_field_collection_start_page_params;
   "test_to_field_collection_not_found_page_params"
     >:: test_to_field_collection_not_found_page_params;
   "test_to_field_collection_custom_params"
     >:: test_to_field_collection_custom_params;
   "test_from_field_collection_error_parameters"
     >:: test_from_field_collection_error_parameters;
   "test_from_field_collection_start_page_params"
     >:: test_from_field_collection_start_page_params;
   "test_from_field_collection_not_found_page_params"
     >:: test_from_field_collection_not_found_page_params;
   "test_from_field_collection_custom_params"
     >:: test_from_field_collection_custom_params;
   "test_to_field_collection_session"
     >:: test_to_field_collection_session;
   "test_from_field_collection_session"
     >:: test_from_field_collection_session]
;;

