open Camlp4.PreCast
open Xnc_functional
open Codegen_common

module FC = Fragment_composer

class type virtual element_type =
object
  val module_name : string
  val class_name : string

  val node : element_type Tree.tree Lazy.t
  val depth : int
  val index : int
  val parent : element_type option

  method virtual tag : element_tag
  method node : element_type Tree.tree
  method depth : int
  method index : int
  method as_element_type : element_type
  method parent : element_type option
  method children : element_type list
  method root : element_type
  method private parent_id : string
  method unique_id : string

  method depth_first_fold : 'a. ('a -> element_type -> 'a) -> 'a -> 'a
  method depth_first_iter: (element_type -> unit) -> unit

  method virtual target_files : target_file_type list

  (* Methods rendering the main page *)
  method render_main_page : page_file_section -> FC.t -> FC.t
  method virtual private render_global_scripts : FC.t -> FC.t
  method virtual private render_form_data : FC.t -> FC.t
  method virtual private render_class_header : FC.t -> FC.t
  method virtual private render_validators : FC.t -> FC.t
  method virtual private render_class_scripts : FC.t -> FC.t
  method virtual private render_page_structure : FC.t -> FC.t
  method virtual private render_class_footer : FC.t -> FC.t

  (* Method rendering the page factory builder *)
  method render_page_factory_builder_file : FC.t -> FC.t

  (* Methods rendering the page parameters file *)
  method render_page_parameters_file : page_param_file_section -> FC.t -> FC.t
  method virtual private render_start_page_param : FC.t -> FC.t
  method virtual private render_param_type_def : FC.t -> FC.t
  method virtual private render_default_generator : FC.t -> FC.t
  method virtual private render_serializer : FC.t -> FC.t
  method virtual private render_deserializer : FC.t -> FC.t
  method virtual private render_params_builder : FC.t -> FC.t

  (* Method rendering the session file *)
  method virtual render_session_file : FC.t -> FC.t
end

class virtual block _depth _idx _parent _node =
object (self)
  val module_name = "Elements"
  val class_name = "block"

  val node = _node
  val depth = _depth
  val index = _idx
  val parent = _parent

  method private banner str comp =
    let l = String.length str in
    let border_w = int_of_float ((80.0 -. (float_of_int l) -. 6.0) /. 2.0) in
    let border = (String.make (80 - 6 - l - border_w) '*') in
    let banner = Printf.sprintf "(%s  %s  %s)\n"
                  (String.make border_w '*') str border in
      FC.append_gen (asterisks ^ banner ^ asterisks) comp

  method private fail _method section =
    Printf.sprintf
      "Module %s, class %s: method %s: can't render the %s section"
      module_name class_name _method section |> failwith

  method virtual tag : element_tag

  method node = Lazy.force node

  method depth = depth

  method index = index

  method as_element_type = (self :> element_type)

  method parent : element_type option = parent

  method children : element_type list =
    self#node |> Tree.get_children |> List.map Tree.get_data

  method root =
    match parent with
      | None -> (self :> element_type)
      | Some p -> p#root

  method private parent_id =
    match parent with
      | None -> ""
      | Some p -> p#unique_id

  method unique_id =
    Printf.sprintf "%s_%d_%s" self#parent_id index class_name

  method depth_first_fold : 'a. ('a -> element_type -> 'a) -> 'a -> 'a =
    fun f acc ->
      List.fold_left
        (fun acc (child : #element_type) -> child#depth_first_fold f acc)
        (f acc (self#as_element_type))
        self#children

  method depth_first_iter f =
    f (self#as_element_type);
    List.iter (fun c -> c#depth_first_iter f) self#children

  method virtual target_files : target_file_type list

  method render_main_page page_section comp =
    match page_section with
      | Global_scripts -> self#render_global_scripts comp
      | Form_data_module -> self#render_form_data comp
      | Class_header -> self#render_class_header comp
      | Build_validators -> self#render_validators comp
      | Class_scripts -> self#render_class_scripts comp
      | Build_page_structure -> self#render_page_structure comp
      | Class_footer -> self#render_class_footer comp
  method private render_global_scripts comp = self#fail "render_main_page" "global scripts"
  method private render_form_data comp = self#fail "render_main_page" "forma data"
  method private render_class_header comp = self#fail "render_main_page" "class header"
  method private render_validators comp = self#fail "render_main_page" "validators"
  method private render_class_scripts comp = self#fail "render_main_page" "class scripts"
  method private render_page_structure comp = self#fail "render_main_page" "build_page_structure method"
  method private render_class_footer comp = self#fail "render_main_page" "class footer"

  method render_page_factory_builder_file comp =
    self#fail "render_page_factory_builder_file" ""

  method render_page_parameters_file section comp =
    match section with
      | Start_page_param -> self#render_start_page_param comp
      | Param_type_def -> self#render_param_type_def comp
      | Default_generators -> self#render_default_generator comp
      | Serialize_param -> self#render_serializer comp
      | Deserialize_param -> self#render_deserializer comp
      | Params_builder -> self#render_params_builder comp

  method private render_start_page_param c = self#fail "render_page_parameters_file" "start page parameters"
  method private render_param_type_def c = self#fail "render_page_parameters_file" "param type def"
  method private render_default_generator c = self#fail "render_page_parameters_file" "default generators"
  method private render_serializer c = self#fail "render_page_parameters_file" "serializer"
  method private render_deserializer c = self#fail "render_page_parameters_file" "deserializer"
  method private render_params_builder c = self#fail "render_page_parameters_file" "param builder"

  method render_session_file comp =
    self#fail "render_session_file" ""
end

class virtual element depth idx _parsed_element parent node =
object (self)
  inherit block depth idx parent node

  val class_name = "element"
  val parsed_element = _parsed_element

  val attrs = new Attributes.attributes _parsed_element

  method attrs = attrs

  method private sanitize_id id =
    Str.global_replace (Str.regexp "[^0-9a-zA-Z_]") "_" id |>
      Str.global_replace (Str.regexp "^[0-9]") "_"

  method unique_id =
    match attrs#value_opt "id" with
      | None -> Printf.sprintf "%s_%d_%s" self#parent_id index class_name
      | Some id -> self#sanitize_id id
end

class ['a] memo =
object (self)
constraint 'a = #element_type
  val table = Hashtbl.create 3

  method virtual add : 'a -> unit
  method add obj = Hashtbl.replace table (obj : < .. > :> < > ) obj

  method virtual mem : 'b. (< .. > as 'b) -> bool
  method mem obj = Hashtbl.mem table (obj : < .. > :> < >)

  method virtual find : 'b. (< .. > as 'b) -> 'a
  method find obj = Hashtbl.find table (obj : < .. > :> < >)
end

