INCLUDE "../camlp4_macros.ml"

open Camlp4.PreCast
open Xnc_functional
open Common_types
open Codegen_common

module FN = Filename
module FC = Fragment_composer

class virtual page_base filename depth idx parsed_element parent_obj node =
object (self)
  inherit Elements.element depth idx parsed_element parent_obj node as super

  val source_fname = filename
  val module_name = "Page"
  val class_name = "page_base"

  val mutable force_debug = false

  method force_debug v = force_debug <- v

  method page_name = FN.basename source_fname
  method dest_module_name = self#page_name |> FN.chop_extension |> String.capitalize
  method dest_class_name = String.lowercase self#dest_module_name

  method as_page_base = (self :> page_base)

  method target_files =
    let children_target =
      List.fold_left
        (fun acc (c : #Elements.element_type) -> acc @ (c#target_files))
        [] self#children in

      [ Parameters_target; Page_target;
        Resource_list_builder_target;
        Page_factory_builder_target ] @ children_target |>
          List.sort Pervasives.compare |> ExtList.List.unique

  method private label_name =
    (self#dest_class_name |> String.capitalize) ^ "_page_parameters"

(******************************************************************************)
(*****                      RENDER_MAIN_PAGE METHODS                      *****)
(******************************************************************************)
  method private render_global_scripts comp =
    comp |> self#banner "GLOBAL SCRIPT SECTION" |>
    fold_left (fun el -> el#render_main_page Global_scripts)
              (Script.find self#node)

  method private render_form_data comp =
    let comp = comp |> FC.append_gen "\n\n" |> self#banner "FORM DATA SECTION" in
    let form_data = Form_data.find self#node in
      match form_data with
        | [] -> comp |> FC.append_gen "module Form_data =
struct
  type t =
      unit
  let create () =
    ()
  let from_cgi_context (cgi : Netcgi.cgi_activation) =
    ()
end\n\n"
        | l -> (List.hd l)#render_main_page Form_data_module comp

  method private render_class_header comp =
    let inh_v, inh_l, inh_p = attrs#value_and_pos "inherit" in
    let inherit_frag comp = if inh_l = -1 && inh_p = -1 then
        FC.append_gen inh_v comp else FC.append_source inh_v inh_l inh_p comp in
      comp |> self#banner "CLASS HEADER SECTION" |>
      FC.append_gen ("class " ^ self#dest_class_name ^
                     " page_context =\nobject(self)\ninherit ") |>
      inherit_frag |>
      FC.append_gen ("\n\"" ^ self#page_name ^ "\" Page_parameters.default_" ^
                     self#dest_class_name ^ "_page_parameters\npage_context\n" ^
                     "val mutable form_data = Form_data.create ()\n\n")

  method private render_validators comp =
    let form = Form.find self#node in
      comp |> self#banner "VALIDATORS COLLECTION SECTION" |>
      FC.append_gen "  method build_validator_collection () =\n" |>
      (fun comp -> match form with
        | Some f -> f#render_main_page Build_validators comp
        | None -> FC.append_gen "    ()\n\n" comp) |>
      self#banner "CONSTANT METHODS SECTION" |>
      FC.append_gen ("  method read_user_input () =\n" ^
                     "    form_data <- Form_data.from_cgi_context " ^
                     "(self#get_cgi_context)\n\n")

  method private render_class_scripts comp =
    comp |> self#banner "CLASS SCRIPT SECTION" |>
    fold_left (fun el -> el#render_main_page Class_scripts)
              (Script.find self#node)

  method private render_page_structure comp =
      comp |> self#banner "BUILD_PAGE_STRUCTURE METHOD SECTION" |>
      FC.append_gen "method build_page_structure parameters =\n" |>
      attrs#render ~before:"self#get_page_structure#set_xml_version "
                   ~after:";\n" ~tr:add_quotes "xml-version" |>
      attrs#render ~before:"self#get_page_structure#set_encoding "
                   ~after:";\n" ~tr:add_quotes "encoding" |>
      attrs#render ~before:"self#get_page_structure#set_dtd "
                   ~after:";\n" ~tr:add_quotes "dtd" |>
      (if force_debug then
        FC.append_gen "self#get_page_structure#set_debug true;\n"
      else
        attrs#render ~before:"self#get_page_structure#set_debug "
                     ~after:";\n" "debug") |>
      fold_left (fun el -> el#render_main_page Build_page_structure)
                self#children

  method private render_class_footer comp =
    comp |> self#banner "CLASS FOOTER SECTION" |> FC.append_gen "end;;\n\n"

(******************************************************************************)
(*****             RENDER_PAGE_FACTORY_BUILDER_FILE METHOD                *****)
(******************************************************************************)
  method render_page_factory_builder_file comp =
    let fragment = Printf.sprintf
"\n  let repository = Page_modules.Page_factory.add_page_constructor
                     repository
                     \"%s\"
                     (new %s.%s) in" self#page_name self#dest_module_name self#dest_class_name in
      FC.append_gen fragment comp

(******************************************************************************)
(*****             RENDER_PAGE_PARAMETERS_FILE METHODS                    *****)
(******************************************************************************)
  method private render_start_page_param c = c

  method private render_param_type_def c =
    match Page_parameters.find self#node with
      | None -> FC.append_gen ("(*| Page \"" ^ self#page_name ^ "\" has no parameters *)") c
      | Some el -> c |>
        FC.append_gen ("  | " ^ self#label_name ^ " of ") |>
        el#render_page_parameters_file Param_type_def

  method private render_default_generator c =
    c |> FC.append_gen ("let default_" ^ self#dest_class_name ^
                        "_page_parameters =\n") |> (
      fun c ->
        match Page_parameters.find self#node with
          | None -> FC.append_gen "  No_parameters\n;;" c
          | Some el -> FC.append_gen ("  " ^ self#label_name ^ " ") c |>
              el#render_page_parameters_file Default_generators |>
              FC.append_gen "\n;;"
    )

  method private render_serializer c =
    match Page_parameters.find self#node with
      | None -> c
      | Some el -> (
          c |> FC.append_gen ("    | " ^ self#label_name) |>
          el#render_page_parameters_file Serialize_param
        )

  method private render_deserializer c =
    match Page_parameters.find self#node with
      | None -> c
      | Some el -> (
          let page_p = Page_parameters.cast el in
            FC.append_gen ("    | " ^ self#label_name ^ " _ ->\n") c |>
            page_p#render_page_parameters_file Deserialize_param |>
            FC.append_gen ("          (" ^ self#label_name ^ " (") |>
            page_p#var_name_concat "," |> FC.append_gen "))\n"
        )

  method private render_params_builder c =
    match Page_parameters.find self#node with
      | None -> (
          FC.append_gen ("let create_" ^ self#dest_class_name ^
                         "_page_parameters = No_parameters;;\n") c)
      | Some el -> (
          let page_p = Page_parameters.cast el in
            c |> FC.append_gen ("let create_" ^ self#dest_class_name ^
                                "_page_parameters ")           |>
            page_p#var_name_concat " "                         |>
            FC.append_gen (" = \n" ^ self#label_name ^ " (")     |>
            page_p#var_name_concat ","                         |>
            FC.append_gen ");;\n"
        )

(******************************************************************************)
(*****              RENDER_RESOURCE_LIST_BUILDER_FILE METHOD              *****)
(******************************************************************************)
  method render_resource_list_builder_file c =
    let fragment = Printf.sprintf
"      { Srv_resource_list.name = \"%s\";
        Srv_resource_list.resource_type = \"xnc:page\"
      };\n" self#page_name in
      FC.append_gen fragment c
end

class standard_page filename depth idx parsed_element parent_obj node =
object (self)
  inherit page_base filename depth idx parsed_element parent_obj node

  val class_name = "standard_page"

  method tag = Codegen_common.Standard_page
end

class start_page filename depth idx parsed_element parent_obj node =
object (self)
  inherit page_base filename depth idx parsed_element parent_obj node

  val class_name = "start_page"

  method tag = Codegen_common.Start_page

  method private render_start_page_param c =
    match Page_parameters.find self#node with
      | None -> FC.append_gen "  No_parameters\n;;" c
      | Some el ->
          c |> FC.append_gen ("  " ^ self#label_name ^ " ") |>
          el#render_page_parameters_file Start_page_param |>
          FC.append_gen "\n;;"
end

class error_page filename depth idx parsed_element parent_obj node =
object (self)
  inherit page_base filename depth idx parsed_element parent_obj node

  val class_name = "error_page"

  method tag = Codegen_common.Error_page

  method private render_class_header comp =
    let inh_v, inh_l, inh_p = attrs#value_and_pos "inherit" in
    let inherit_frag comp = if inh_l = -1 && inh_p = -1 then
        FC.append_gen inh_v comp else FC.append_source inh_v inh_l inh_p comp in
      comp |> self#banner "CLASS HEADER SECTION" |>
      FC.append_gen ("class " ^ self#dest_class_name ^
                     " page_context =\nobject(self)\ninherit ") |>
      inherit_frag |>
      FC.append_gen ("\n\"" ^ self#page_name ^ "\" page_context\n" ^
                     "val mutable form_data = Form_data.create ()\n\n")

  method private render_params_builder c =
    FC.append_gen ("let create_" ^ self#dest_class_name ^ "_page_parameters error_description backtrace =
  Error_page_parameters (error_description, backtrace);;\n\n") c

  method private render_default_generator c = c
end

class not_found_page filename depth idx parsed_element parent_obj node =
object (self)
  inherit page_base filename depth idx parsed_element parent_obj node

  val class_name = "not_found_page"

  method tag = Codegen_common.Not_found_page

  method private render_class_header comp =
    let inh_v, inh_l, inh_p = attrs#value_and_pos "inherit" in
    let inherit_frag comp = if inh_l = -1 && inh_p = -1 then
        FC.append_gen inh_v comp else FC.append_source inh_v inh_l inh_p comp in
      comp |> self#banner "CLASS HEADER SECTION" |>
      FC.append_gen ("class " ^ self#dest_class_name ^
                     " page_context =\nobject(self)\ninherit ") |>
      inherit_frag |>
      FC.append_gen ("\n\"" ^ self#page_name ^ "\" page_context\n" ^
                     "val mutable form_data = Form_data.create ()\n\n")

  method private render_params_builder c =
    FC.append_gen ("let create_" ^ self#dest_class_name ^ "_page_parameters missing_resource =
  Not_found_page_parameters missing_resource;;\n") c

  method private render_default_generator c = c
end

let make_children depth sibl_num block =
  match block with
    | Proc_instr _
    | Print_proc_instr _
    | Common_types.Cdata _ -> failwith "Page.make_children: this exception should never appear"
    | Element e -> begin
        match e.name with
          | "head" -> Head.make depth sibl_num block
          | "body" -> Body.make depth sibl_num block
          | "page-parameters" -> Page_parameters.make depth sibl_num block
          | "script" -> Script.make depth sibl_num block
          | n -> unknown_element n __LOCATION__
      end

let std_memo : standard_page Elements.memo = new Elements.memo
let cast_std (o : Elements.element_type) =
  CAST_OR_ERROR (std_memo, o, __LOCATION__)

let start_memo : start_page Elements.memo = new Elements.memo
let cast_start (o : Elements.element_type) =
  CAST_OR_ERROR (start_memo, o, __LOCATION__)

let error_memo : error_page Elements.memo = new Elements.memo
let cast_error (o : Elements.element_type) =
  CAST_OR_ERROR (error_memo, o, __LOCATION__)

let not_found_memo : not_found_page Elements.memo = new Elements.memo
let cast_not_found (o : Elements.element_type) =
  CAST_OR_ERROR (not_found_memo, o, __LOCATION__)

let make filename (depth : int) (idx : int) (block : Common_types.parsed_block) =
  let element = Common_lib.unwrap_element block in
  let page_type = Common_lib.get_attribute_value element "type" in
    match page_type with
      | "start" -> begin
          let constructor parent_obj node =
            let this_obj = new start_page filename depth idx element parent_obj node in
              start_memo#add this_obj;
              (this_obj :> Elements.element_type) in
            Tree.Comp (constructor, make_children)
        end
      | "standard" -> begin
          let constructor parent_obj node =
            let this_obj = new standard_page filename depth idx element parent_obj node in
              std_memo#add this_obj;
              (this_obj :> Elements.element_type) in
            Tree.Comp (constructor, make_children)
        end
      | "error" -> begin
          let constructor parent_obj node =
            let this_obj = new error_page filename depth idx element parent_obj node in
              error_memo#add this_obj;
              (this_obj :> Elements.element_type) in
            Tree.Comp (constructor, make_children)
        end
      | "not-found" -> begin
          let constructor parent_obj node =
            let this_obj = new not_found_page filename depth idx element parent_obj node in
              not_found_memo#add this_obj;
              (this_obj :> Elements.element_type) in
            Tree.Comp (constructor, make_children)
        end
      | n -> unknown_element n __LOCATION__

let find t =
  let f acc t =
    let obj = Tree.get_data t in
      match acc, obj#tag with
        | Some _, _ -> acc
        | None, Standard_page -> Some ((cast_std obj)#as_page_base)
        | None, Error_page -> Some ((cast_error obj)#as_page_base)
        | None, Start_page -> Some ((cast_start obj)#as_page_base)
        | None, Not_found_page -> Some ((cast_not_found obj)#as_page_base)
        | _, _ -> None in
  !!(Tree.depth_first_fold f None t)

