INCLUDE "../camlp4_macros.ml"

open Camlp4.PreCast
open Xnc_functional
open Common_types
open Codegen_common

module FC = Fragment_composer

class page_parameters depth idx parsed_element parent_obj node =
  object (self)
    inherit Elements.element depth idx parsed_element parent_obj node

    val module_name = "Page_parameters"
    val class_name = "page_parameters"

    method tag = Page_parameters

    method target_files =
      let children_target =
        List.fold_left
          (fun acc (c : #Elements.element_type) -> acc @ (c#target_files))
          [] self#children in
        [ Parameters_target ] @ children_target

  method private render_page_structure comp = comp

  method private render_defaults comp =
    let quote_def var =
      let type_ = var#attrs#value "type" in
        match type_ with
          | "string" -> Printf.sprintf "\"%s\""
          | "char" -> Printf.sprintf "'%s'"
          | _ -> fun x -> x in

    let vars = List.map Variable.cast_param_var self#children in
      if List.length vars = 0 then comp
      else begin
        let first_var, vars = List.hd vars, List.tl vars in
          comp |> FC.append_gen "(" |>
          first_var#attrs#render "default" ~tr:(quote_def first_var) |>
          fold_left (
            fun v ->
              let attrs : Attributes.attributes = v#attrs in
                attrs#render ~before:"," ~tr:(quote_def v) "default"
          ) vars |> FC.append_gen ")"
      end

  method private render_start_page_param c = self#render_defaults c

  method private render_param_type_def c =
    let vars = List.map Variable.cast_param_var self#children in
      if List.length vars = 0 then c
      else begin
        c |> (List.hd vars)#attrs#render "type" |>
        fold_left
          (fun var -> (var#attrs : Attributes.attributes)#render ~before:"*" "type")
          (List.tl vars)
      end

  method private render_default_generator c = self#render_defaults c

  method var_name_concat sep c =
    (* This renders a "sep" separated list of variable names *)
    let vars = List.map Variable.cast_param_var self#children in
      if List.length vars = 0 then c
      else
        let first_var, vars = List.hd vars, List.tl vars in
          c |> first_var#attrs#render "name" |>
          fold_left (
            fun v -> (v#attrs : Attributes.attributes)#render ~before:sep "name"
          ) vars

  method private render_serializer c =
    let vars = List.map Variable.cast_param_var self#children in
      c |> FC.append_gen " (" |> self#var_name_concat "," |> FC.append_gen ") ->\n" |>
      fold_left (
        fun v c ->
          let (attrs : Attributes.attributes) = v#attrs in
          let type_ = attrs#value "type" in
          let name = attrs#value "name" in
            c |>
            FC.append_gen ("Xnc_field_collection.register_" ^ type_ ^
                           " fc \"" ^ name ^  "\" ") |>
                           attrs#render ~after:";\n" "name"
      ) vars

  method private render_deserializer c =
    let vars = List.map Variable.cast_param_var self#children in
      c |>
        fold_left (
          fun v c ->
            let (attrs : Attributes.attributes) = v#attrs in
            let type_ = attrs#value "type" in
            let name = attrs#value "name" in
              c |> FC.append_gen ("        let " ^ name ^
                                  " = Xnc_field_collection.get_" ^ type_ ^
                                  " fc \"" ^ name ^ "\" in\n")
        ) vars
  end

let page_param_memo : page_parameters Elements.memo = new Elements.memo
let cast (o : Elements.element_type) =
  CAST_OR_ERROR (page_param_memo, o, __LOCATION__)

let make_children depth sibl_num block =
  match block with
    | Proc_instr _
    | Print_proc_instr _
    | Common_types.Cdata _ -> failwith "Page_parameters.make_children: this exception should never appear"
    | Element e -> begin
        match e.name with
          | "variable" -> Variable.make Page_parameters_variable depth sibl_num block
          | n -> Codegen_common.unknown_element n __LOCATION__
      end

let make depth idx (block : Common_types.parsed_block) =
  let element = Common_lib.unwrap_element block in
  let constructor parent_obj node =
    let this_obj = new page_parameters depth idx element parent_obj node in
      page_param_memo#add this_obj;
      (this_obj :> Elements.element_type) in
    Tree.Comp (constructor, make_children)

let find t = FIND_UNIQUE(Page_parameters)

