open Xnc_functional

module Pf = Printf

type build_target =
  | Native
  | Profile
  | Byte
  | Debug

type t =
    {
      project_file_name : string;
      target_type : build_target;
      output_filename : string;
      project_file_dir : string;
      build_dir : string;
      page_debug_mode : bool;
    }
;;

let project_file_name = ref None;;
let build_target = ref Native
let output_filename = ref None;;
let project_file_dir = ref None;;
let build_dir = ref "./_xnunc_build";;
let page_debug_mode = ref false;;

let display_application_version () =
  print_endline ("This is " ^ Globals.compiler_name ^ " version " ^
                   Globals.compiler_version);
  exit 0
;;

let program_version_long () =
  print_endline (
    "This compiler, "^ Globals.compiler_name ^ ", is part of the
Ex Nunc Application System version " ^ Globals.compiler_version ^ "

Use --help for a quick help.
Use --copyright for copyright informations. Please, take a look.");
  exit 0
;;

let ex_nunc_copyright_info () =
  print_endline (
    "This compiler, "^ Globals.compiler_name ^ ", is part of the

E X   N U N C   A P P L I C A T I O N   S Y S T E M
                  version " ^ Globals.compiler_version ^ "


Ex Nunc Application System copyright informations:
                  
Copyright (c) 2005-" ^ (string_of_int Globals.current_year) ^ " by
Paolo Donadeo <" ^ Globals.paolo_email ^ ">
  and
Alessandro Strada <" ^ Globals.alessandro_email ^ ">

Permission is  hereby granted,  free of  charge, to
any person  obtaining a  copy of this  software and
associated documentation files (the \"Software\"), to
deal in the Software without restriction, including
without limitation the rights to use, copy, modify,
merge, publish, distribute, sublicense, and/or sell
copies of  the Software,  and to permit  persons to
whom the Software is furnished to do so, subject to
the following conditions:

The  above  copyright  notice and  this  permission
notice  shall   be  included   in  all   copies  or
substantial portions of the Software.

THE   SOFTWARE  IS   PROVIDED   \"AS  IS\",   WITHOUT
WARRANTY   OF  ANY   KIND,   EXPRESS  OR   IMPLIED,
INCLUDING  BUT NOT  LIMITED  TO  THE WARRANTIES  OF
MERCHANTABILITY, FITNESS  FOR A  PARTICULAR PURPOSE
AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS
OR  COPYRIGHT  HOLDERS  BE LIABLE  FOR  ANY  CLAIM,
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT  OR OTHERWISE, ARISING FROM,  OUT OF
OR IN  CONNECTION WITH THE  SOFTWARE OR THE  USE OR
OTHER DEALINGS IN THE SOFTWARE.

Use --help for a quick help.");
  exit 0
;;

let anonymous_arguments_handler param =
  let cwd = Unix.getcwd () in
  let param = Utility.norm_path param in
  let prj_dir = Filename.dirname param in
  let prj_file = Filename.basename param in
    try
      Unix.chdir prj_dir;
      let abs_prj_dir = Unix.getcwd () in
        Unix.chdir cwd;

        project_file_name := Some prj_file;
        project_file_dir := Some abs_prj_dir
    with
      | Unix.Unix_error(Unix.ENOENT as error, calling_f, param)
      | Unix.Unix_error(Unix.EACCES as error, calling_f, param) ->
          Pf.eprintf "%s: %s: %s\n"
            Globals.compiler_name param (Unix.error_message error);
          exit 1
;;

let arguments_spec =
  Arg.align
    [
      "-d",
      Arg.Set_string (build_dir),
      " Sets the build dir (default: \"" ^ !build_dir ^ "\")";

      "--page-debug",
      Arg.Set (page_debug_mode),
      " Sets the debug mode in each page (default: false)";

      "-v",
      Arg.Unit (display_application_version),
      " Print application version";

      "--version",
      Arg.Unit (program_version_long),
      " Print more verbose application server informations";

      "--copyright",
      Arg.Unit (ex_nunc_copyright_info),
      " Print informations on Ex Nunc copyright. Please take a look";

      "-o",
      Arg.String (fun out_fn -> output_filename := Some out_fn),
      " Set output filename (default: name of xrj project file)";

      "-t",
      Arg.String (fun t -> match t with
        | "byte" -> build_target := Byte
        | "native" -> build_target := Native
        | "profile" -> build_target := Profile
        | "debug" -> build_target := Debug
        | _ -> 
          Pf.eprintf "%s: unknown target type: %s\n" Globals.compiler_name t;
          exit 1
      ),
      " Set the exec target type. Valid values are \"byte\", \"native\",
                \"profile\" and \"debug\". Default is native"
    ]
;;

let handle_command_line () =
  Arg.parse
    arguments_spec
    anonymous_arguments_handler
    ("Usage: " ^ Globals.compiler_name ^ " <options> <project file>\nOptions are:");

  let prj_name =
    match !project_file_name with
      | Some name -> name
      | None ->
          let () = Pf.eprintf "%s: no input files\n" Globals.compiler_name in
            exit 1 in
  let output_filename =
    let change_ext ext =
      (prj_name |> Filename.chop_extension |> Filename.basename) ^ ext in
    match !output_filename with
      | Some name -> name
      | None -> begin
        match !build_target with
          | Native -> change_ext ".native"
          | Profile -> change_ext ".p.native"
          | Byte -> change_ext ".byte"
          | Debug -> change_ext ".d.byte"
      end in

    {
      project_file_name = prj_name;
      target_type = !build_target;
      output_filename = output_filename;
      project_file_dir = Option.get !project_file_dir;
      build_dir = !build_dir;
      page_debug_mode = !page_debug_mode
    }
;;

