(** Ex Nunc page compiler common types module
  *
  * TODO Add a decent coment here
  *
  * @author Paolo Donadeo
  * @see <http://ex-nunc.sourceforge.net> http://ex-nunc.sourceforge.net
  * @version 0.2
  *)

type position = int * int
(** The [position] type models a position inside a stream: the first element
  * represents the line from the beginning of the stream, while the second
  * element represents the position from the beginning of the current line.
  *)

type 'a with_position = 'a * position
(** The [with_position] type is a shortcut to attach a position to something.
  *)

type attr_value =
  | Literal of string (** A literal attribute value, e.g. non parametric *)
  | Parametric of string (** A parametric attribute value, contains OCaml code *)
(** The Ex Nunc template language is very similar to XHTML, so elements can have
  * attributes; the main difference between an Ex Nunc file and a XHTML file is
  * that here we permit to have parametric attributes, the syntax is this:
  * {v
  * <xnc:element1>
  *   <xnc:element1 attribute1 = "literal value"
  *                 attribute2 =
  *                 "<?xnc= Printf.sprintf "This is a %s value" parametric ?>"
  *   </xnc:element1>
  * </xnc:element1>
  * v}
  *
  * Parametric arguments must be OCaml expressions with [string] value,
  * otherwise the OCaml compiler will give you an error at compile time.
  *)

type attribute = string with_position * attr_value with_position
(** This type models an attribute of the template language. It's a string with a
  * position, the name of the attribute, and an {!Common_types.attr_value} with
  * the corresponding position.
  *)

type parsed_block =
  | Proc_instr of string with_position (** a processing instruction *)
  | Print_proc_instr of string with_position (** a processing instruction of [string] type *)
  | Cdata of string with_position (** a block of non parsed characters *)
  | Element of parsed_element (** an element of the template language *)
(** The [parsed_block] type models a syntactic element of the Ex Nunc template
  * language. Four type of blocks are possible:
  * {ol
  * {li {b Processing instructions}: similar to XHTML processing instructions,
  * these blocks allow you to embed Objective Caml code in your template. Here
  * is an example:
  * {v
  * <?xnc
  *     (* Any OCaml code here *)
  * ?>
  * v}
  * {i Differently} from PHP or other web oriented scripting languages, Ex Nunc
  * processing instructions have a precise semantic and are not simply expanded
  * in place; the main consequence is that processing instructions are {b not}
  * allowed everywhere in the page. More details in the Ex Nunc Programming
  * manual ({b TODO}: write the manual and put a link here).
  * }
  * {li {b Printing processing instructions}: This is simply syntactic sugar for
  * a processing instruction that evaluates as string. The syntax is:
  * {v
  * <?xnc=
  *     (* Any OCaml code here; the expression must be of type string *)
  * ?>
  * v}
  * If the expression included into {v <?xnc= ... ?> v} isn't of type [string],
  * you will receive an error at compile time.
  * }
  * {li {b Non parsed character blocks}: everything not included in a processing
  * instruction or in an element (see item 4 below) is considered a block of
  * text and is printed out. Like in the processing instruction case, such
  * "wild" blocks of text are {b not} allowed everywhere. A silly example:
  * {v
  * <xnc:page>
  *     <xnc:page-parameters>
  *         <xnc:variable name="foo" type="int">
  *         <xnc:variable name="bar" type="float">
  *     </xnc:page-parameters>
  * This is a simple text block
  * </xnc:page>
  * v}
  * As you see, no CDATA text is present in the "page-parameters" element, this
  * fact is not accidental: page-parameters defines the names and types of a
  * page and in this context free text is a nonsense. The Ex Nunc page compiler
  * will give you a parse error if you try to "force" text -- and even
  * processing instruction -- gere.
  * }
  * {li {b Ex Nunc template language elements}: these elements represent the
  * core parte of the template language. A very simple example is reported
  * above, where we define a "page" with its parameters. Note that the page
  * compiler only recognizes element in the {i pseudo} namespace "xnc":
  * regular VHTML elements like "html" are considered simple,
  * non parsed, text.
  * }
  * }
  *)

and parsed_element =
  {
    name : string; (** the name of the element like [page] in {v <xnc:page> ... </xnc:page> v}*)
    attributes : attribute list; (** the list of all parsed attributes *)
    sub_trees : document; (** the list of children *)
    el_pos : position (** the starting position of the parsed element *)
  }
(** A parsed element (see above) has a name, a list of attributes, a list of
  * children and a position in the file.
  *)

and
    document = parsed_block list
(** A [document] is simply a list of parsed blocks.
  *)

type card = int

type 'a with_range_constraints = No_limits | Range of 'a * 'a

type attribute_type_contraint =
    String
  | Bool
  | Int of int with_range_constraints
  | Float of float with_range_constraints

type attribute_constraint = {
  attr_type : attribute_type_contraint;
  mandatory : bool;
  param_allowed : bool;
  default : string option;
  valid_values : string list option;
}

type element_specific_validation = parsed_element -> unit

type element_constraint = {
  children : (card * card) Xnc_functional.String_map.t;
  attributes_map : attribute_constraint Xnc_functional.String_map.t;
  proc_instr_allowed : bool;
  cdata_allowed : bool;
  is_root_element : bool;
  validation : element_specific_validation;
}

type constraints_table =
    (string * (* The element name *)
     (string * (* child element name *)
      card * card) (* Min and max number *)
                list * (* List of children *)
     (string *
      attribute_type_contraint *
      bool *
      bool *
      string option *
      string list option) list *
     bool *
     bool *
     bool *
     element_specific_validation
    ) list (** The list of constraints *)

exception Unknown_element of string * position
(** Raised when the parser find an unknown element.
  * [string] is the name of the unknown element and [position] its position in
  * the stream.
  *)

exception Proc_instr_not_allowed of string * position
(** Raised when a processing instruction is found where it is forbidden.
  * [string] is the name of the container element and [position] its position in
  * the stream.
  *)

exception Cdata_not_allowed of string * position
(** Raised when a CDATA block is found where it is forbidden. [string] is the
  * name of the container element and [position] its position in the stream.
  *)

exception Illegal_child of string * position * string
(** Raised when an Ex Nunc element contains a forbidden element.
  * The first [string] is the name of the container element, the [position] is
  * its position and the last [string] is the name of the forbidden child
  * element.
  *)

exception Too_few_children of string * position * int * string
(** Raised when an Ex Nunc element don't contain the expected number of
  * children of a certain type.
  * The first [string] is the name of the container element, the [position] is
  * its position; the third [int] argument is the minimum number of children
  * necessary and the last [string] is the name of the missing child element.
  *)

exception Too_many_children of string * position * int * string
(** Raised when an Ex Nunc element don't contain the expected number of
  * children of a certain type.
  * The first [string] is the name of the container element, the [position] is
  * its position; the third [int] argument is the maximum number of children
  * allowed and the last [string] is the name of the child element.
  *)

exception Missing_attribute of string * position * string
(** Raised when a mandatory attribute is missing in Ex Nunc element. The first
  * [string] is the name of the container element, the [position] is its
  * position; the last [string] is the name of the missing attribute.
  *)

exception Unknown_attribute of string * position
(** Raised when the parser find an unknown attribute inside an Ex Nunc element.
  * The first [string] argument is the name of the attribute and the [position]
  * is its position inside the file.
  *)

exception Parametric_param_not_allowed of string * position
(** Raised when the parser find an attribute with a parametric value, e.g.
  * {v
  *   <xnc:page-parameters>
  *     <xnc:variable name="<?xnc= Printf.sprintf "foo" ?>" type="int" />
  *   </xnc:page-parameters>
  * v}
  * but the attribute [name] is supposed to be a constant.
  *)

exception Invalid_attribute_value of string * position
(** Some attributes may take only a list of predefined values, like the [module]
  * attribute of the [script] element, here:
  * {v
  * <xnc:page>
  *   <xnc:page-parameters>
  *     <xnc:variable name="login" type="string" />
  *     <xnc:variable name="password" type="string" />
  *   </xnc:page-parameters>
  *   <xnc:script scope = "module">
  *     exception My_exception
  *     let useless_int_var = 3;;
  *     let useless_int_ref_var = ref 42;;
  *   </xnc:script>
  *   Some text here...
  * </xnc:page>
  * v}
  *
  * Valid values for [scope] are only [class] and [module], so if a different
  * value is found, the parser raises [Invalid_attribute_value].
  *)

exception Attribute_out_of_range of string * position * string * string
(** Integer and float element attributes {i may} have range constraints. In this
  * case, if an out of range attribute value is found, this exception is raised.
  * The first paramter is the name of the attribute involved, the second its
  * value position and the last two strings are the representation of min and
  * max allowed values.
  *)

exception Invalid_attribute_type of string * position * string
(** Ex Nunc template language is a strongly typed language. Element attributes
  * may be strings, boolean, integers of floats. If the value of an attribute is
  * incompatible with the expected type, [Invalid_attribute_type] is raised. The
  * first parameter is the attribute name, the second is the position of the
  * invalid value and the third is the name of the expected type.
  *)

exception Too_many_root_elements of string * position
(** In a Ex Nunc file only one root element is allowed. If two or more are
  * found, the validator raises this exception. Parameters are the name of the
  * duplicated root element and its position.
  *)

exception Non_root_element of string * position
(** If a non root element is found by the parser at the root of the syntactic
  * tree, this exception is raised. The first parameter is the name of the non
  * root element and the second parameter is its position.
  *)

