module L = List
module P = Printf

exception Missing_file_name

type fragment =
  | Source_file of string
  | Dest_file of string
  | Source_fragment of string * int * int (* string, line and pos *)
  | Dest_fragment of string
  | Stop

type t = bool * bool * fragment list

let empty = false, false, []

let new_line_count s =
  let counter = ref 0 in
  let length = String.length s in
    for i = 0 to (length - 1) do if s.[i] = '\n' then incr counter done;
    !counter

let set_source_file name (src_present, dest_present, frags) =
  true, dest_present, (Source_file name)::frags

let set_dest_file name (src_present, dest_present, frags) =
  src_present, true, (Dest_file name)::frags

let append_source text line pos (src_present, dest_present, frags) =
  if not src_present
  then raise Missing_file_name
  else src_present, dest_present, (Source_fragment (text, line, pos))::frags

let append_gen text (src_present, dest_present, frags) =
  if not dest_present
  then raise Missing_file_name
  else src_present, dest_present, (Dest_fragment text)::frags

let render_fragment (buffer, line_count, rendered, source_fn, dest_fn, (gen_string : string option)) frag =
  let actually_render_gen gen_string dest_filename =
    let s =
      if rendered <> 0
      then P.sprintf "\n# %d \"%s\"\n" (line_count + 3) (Option.get dest_fn)
      else "" in
    let s = s ^ gen_string in
      Buffer.add_string buffer s;
      buffer, line_count + (new_line_count s), rendered + 1, source_fn, dest_filename, None in

  match frag with
    | Source_file (s) ->
        buffer, line_count, rendered, Some s, dest_fn, gen_string

    | Dest_file (new_dest) -> (
        match gen_string with
          | None -> buffer, line_count, rendered, source_fn, Some new_dest, gen_string
          | Some gen_string -> actually_render_gen gen_string (Some new_dest)
      )

    | Source_fragment (text, line, pos) -> (
        let buffer, line_count, rendered, source_fn, dest_fn, gen_string =
          match gen_string with
            | None -> buffer, line_count, rendered, source_fn, dest_fn, gen_string
            | Some gen_string -> actually_render_gen gen_string dest_fn in
          let s = P.sprintf "\n# %d \"%s\"\n" line (Option.get source_fn) in
          let s = P.sprintf "%s%s%s" s (String.make (pos - 1) ' ') text in
            Buffer.add_string buffer s;
            buffer, line_count + (new_line_count s), rendered + 1, source_fn, dest_fn, None
      )

    | Dest_fragment (text) -> (
        match gen_string with
          | None -> buffer, line_count, rendered, source_fn, dest_fn, Some text
          | Some s -> buffer, line_count, rendered, source_fn, dest_fn, Some (s ^ text)
      )

    | Stop -> (
        match gen_string with
          | None -> buffer, line_count, rendered, source_fn, dest_fn, gen_string
          | Some gen_string -> actually_render_gen gen_string dest_fn
      )

let simple_render b frag =
  match frag with
    | Source_fragment (text, _, _) -> Buffer.add_string b (" " ^ text); b
    | Dest_fragment text -> Buffer.add_string b text; b
    | _ -> b

let flush ?(annotations=true) (src_present, dest_present, frags) =
  let b =
    if annotations then
      let (b, _, _, _, _, _) =
        L.fold_left render_fragment
          (Buffer.create 1024, 0, 0, None, None, None) (* accumulator *)
          (L.rev (Stop::frags)) in
        b
    else
      L.fold_left simple_render (Buffer.create 1024) (L.rev frags) in

    Buffer.contents b
;;

