(** Ex Nunc text fragment composer
  *
  * This utility module is used to compose different text fragment from
  * different source files into one destination file. Text fragment from source
  * files comes with the line and position informations, while text fragments
  * generated by the compiler do not provide other information.
  *
  * The typical use is this:
  * {v
  *   let dest_fname = "generated.ml" in
  *   let frag = empty in
  *   let frag = set_dest_file frag dest_name in
  *   let frag = set_source_file frag "template_1.xnunc" in
  *   let frag = append_source frag "some string from template_1.xnunc" 10 5 in
  *   let frag = append_source frag "other string from template_1.xnunc" 15 5 in
  *   let frag = set_source_file frag "template_2.xnunc" in
  *   let frag = append_source frag "some string from template_2.xnunc" 42 9 in
  *
  *   let dest_file = open_out dest_fname in
  *   output_string dest_file (flush frag);
  *   close_out dest_file;
  * v}
  *
  * @author Paolo Donadeo
  * @see <http://ex-nunc.sourceforge.net> http://ex-nunc.sourceforge.net
  * @version 0.2
  *)


exception Missing_file_name
(** A test fragment appended before setting up the file name
  *
  * This exception is raised by the {!Fragment_composer.append_source}
  * and {!Fragment_composer.append_gen} functions, if the corresponding file
  * name were not set.
  *)

type t
(** The type that represents a fragment list
  *)

val empty : t
(** Returns an empty list of fragments
  *)

val set_source_file : string -> t -> t
(** Sets up the name of the source file
  *
  * This function sets up the name of the current source file i.e. the name that
  * the rendering function {!Fragment_composer.flush} will put in the line
  * number directives from now on when a source fragment is to be rendered.
  *
  * Can be called several times to change che source file name.
  *)

val set_dest_file : string -> t -> t
(** Sets up the name of the destination file
  *
  * This function sets up the name of the current destination file i.e. the name
  * that the rendering function {!Fragment_composer.flush} will put in the line
  * number directives from now on when a generated fragment is to be rendered.
  *
  * Can be called several times to change che destination file name,
  * {b but this use is discouraged since it makes very little sense}.
  *)

val append_source : string -> int -> int -> t -> t
(** [append_source string line column frag] returns a new list with [string] added
  *
  * Line and column will be used in the line number directive at rendering time
  *
  * Raises {!Fragment_composer.Missing_file_name} if called before
  * {!Fragment_composer.set_source_file}.
  *)

val append_gen : string -> t -> t
(** [append_gen string frag] returns a new list with [string] added
  *
  * Appends a string of text that will be considered a {i generated string},
  * that will appear in the destination file.
  *
  * Raises {!Fragment_composer.Missing_file_name} if called before
  * {!Fragment_composer.set_dest_file}.
  *)

val flush : ?annotations:bool -> t -> string
(** List of text fragments rendering
  *
  * Returns a string that can be printed into a file, tipically with the same
  * filename used in {!Fragment_composer.set_dest_file}.
  *)

