open Xnc_stdlib
open Xnc_functional
open Utility

exception Project_file_parse_error of int * string

module SM = String_map

let start_section = Str.regexp "^\\[\\([^]]*\\)\\]$"

let remove_comments line =
  try
    let index = String.index line '#' in
      String.sub line 0 index
  with
    | Not_found -> line

type line = int * string

type section_return_type =
  | Sources of string list


class virtual section name lines =
object (self)
  val name = name
  val lines = lines

  method virtual parse : section
  method virtual get_content : section_return_type
end

class sources_section name lines =
object (self)
  inherit section name lines

  val files = []

  method parse =
    let files_lines = List.tl lines in
    (* TODO TODO TODO VERIFY IF ALL THE ENUMERATED FILES EXISTS! *)
    let new_files = List.fold_left
                      (fun files (num, line) -> line::files) [] files_lines
                        |> List.rev in
      ({< files = new_files >} :> section)

  method get_content = Sources files
end

let sections_factory name lines =
  let first_line_num = List.hd lines |> fst in
  let actual_section = match name with
    | "sources" -> new sources_section name lines
    | _ -> raise (Project_file_parse_error
                    (first_line_num, "Unknown section name \"" ^ name ^ "\"")) in
  let section = actual_section#parse in
    section

class project_file (ic : in_channel) =
object (self)
  val sections = SM.empty
  val ic = ic
  val lines = []

  method private check_constraints sections_map =
    (* Sources section is mandatory *)
    if SM.mem "sources" sections_map = false
    then raise (Project_file_parse_error
                  (-1, "The \"sources\" section is mandatory and missing"));

  method private split_sections lines =
    let first_num, first_line = List.hd lines in
    if Str.string_match start_section first_line 0 = false
    then raise (Project_file_parse_error
                  (first_num, "The Ex-nunc project file must start with a section"));

    let last_section_name = (Str.matched_group 1 first_line) |> trim_white_space in
    let last_section_lines = [(first_num, first_line)] in
    let last_section_name, last_section_lines, splitted_sections =
      List.fold_left
        (
          fun (last_section_name, last_section_lines, splitted_sections) (num, line) ->
            if Str.string_match start_section line 0
            then (* new section *)
              let splitted_sections =
                (last_section_name, List.rev last_section_lines)::(splitted_sections) in
              let last_section_name = (Str.matched_group 1 line) |> trim_white_space in
              let last_section_lines = [num, line] in
                last_section_name, last_section_lines, splitted_sections
            else (* new line in previous section *)
              let last_section_lines = (num, line)::(last_section_lines) in
                last_section_name, last_section_lines, splitted_sections
        )
        (last_section_name, last_section_lines, [])
        (List.tl lines) in

      (last_section_name, List.rev last_section_lines)::(splitted_sections) |> List.rev

  method private create_sec_objs splitted_sections =
    List.fold_left
      (
        fun sec_map (sec_name, lines) ->
          SM.add sec_name (sections_factory sec_name lines) sec_map
      )
      SM.empty
      splitted_sections

  method read_project_file =
    let process_line (lines, count) line =
      let line = line |> remove_comments |> trim_white_space in
        if String.length line <> 0
        then (count, line)::lines, count + 1
        else (lines, count + 1) in

    let non_empty_lines = in_channel_fold process_line ([], 1) ic |> fst |> List.rev in
    let splitted_sections = self#split_sections non_empty_lines in
    let map = self#create_sec_objs splitted_sections in
      self#check_constraints map;
      {< sections = map; lines = non_empty_lines; >}

  method get_source_files =
    let src_sec = SM.find "sources" sections in
    let content = src_sec#get_content in
      match content with
        | Sources files -> files
end

