open OUnit
open Common_types


let never_closed_element_1 () =
  let test_fragment = "
<xnc:test_element1>
  Some text here...
  <xnc:test_element2>
    Some useless text here...
</xnc:test_element1>" in
  let lexbuf = Lexing.from_string test_fragment in


    OUnit.assert_raises
      ~msg:"The parser doesn't notice that test_element2 is not closed"
      (Xnunc_lexer.Parse_error "Parse error in (6,1): \
element \"test_element2\" in (4,8) never closed")
      (fun () -> Xnunc_lexer.parse_xnunc_lexbuf lexbuf)
;; (* END of "never_closed_element_1" *)


let never_closed_element_2 () =
  let test_fragment = "
<xnc:test_element1>
  Some text here...
  <xnc:test_element2>
    Some useless text here...
  </xnc:test_element2>
  Some other text" in
  let lexbuf = Lexing.from_string test_fragment in
    OUnit.assert_raises
      ~msg:"The parser doesn't notice that test_element1 is not closed"
      (Xnunc_lexer.Parse_error "Parse error in (7,18): \
element \"test_element1\" in (2,6) never closed")
      (fun () -> Xnunc_lexer.parse_xnunc_lexbuf lexbuf)
;; (* END of "never_closed_element_2" *)


let incomplete_attribute () =
  let test_fragment = "
<xnc:test_element1 foo = >
  Some text here...
  <xnc:test_element2>
    Some useless text here...
  </xnc:test_element2>
  Some other text" in
  let lexbuf = Lexing.from_string test_fragment in
    OUnit.assert_raises
      ~msg:"The parser doesn't notice the syntax error of test_element1"
      (Xnunc_lexer.Parse_error "Parse error in (2,20) while scanning \
element \"test_element1\" in (2,6)"
      )
      (fun () -> Xnunc_lexer.parse_xnunc_lexbuf lexbuf)
;; (* END of "incomplete_attribute" *)


let invalid_attribute_name () =
  let test_fragment = "
<xnc:test_element1 5foo = \"bar\">
  Some text here...
  <xnc:test_element2>
    Some useless text here...
  </xnc:test_element2>
  Some other text
</xnc:test_element1>" in
  let lexbuf = Lexing.from_string test_fragment in
    OUnit.assert_raises
      ~msg:"The parser doesn't notice that the \"5foo\" attribute \
      has invalid name"
      (Xnunc_lexer.Parse_error "Parse error in (2,20) while scanning \
element \"test_element1\" in (2,6)"
      )
      (fun () -> Xnunc_lexer.parse_xnunc_lexbuf lexbuf)
;; (* END of "invalid_attribute_name" *)


let no_equal_in_attribute () =
  let test_fragment = "<xnc:test_element1 foo \"bar\">
  Some text here...
  <xnc:test_element2>
    Some useless text here...
  </xnc:test_element2>
  Some other text
</xnc:test_element1>" in
  let lexbuf = Lexing.from_string test_fragment in
    OUnit.assert_raises
      ~msg:"The parser doesn't notice that the \"foo\" attribute \
lacks the equal operator"
      (Xnunc_lexer.Parse_error "Parse error in (1,20) while scanning \
element \"test_element1\" in (1,6)"
      )
      (fun () -> Xnunc_lexer.parse_xnunc_lexbuf lexbuf)
;; (* END of "no_equal_in_attribute" *)


let not_closed_processing_instruction () =
  let test_fragment = "
<xnc:test_element1 foo = \"bar\">
  Some text here...
  <?xnc This is a processing instruction never closed
  <xnc:test_element2>
    Some useless text here...
  </xnc:test_element2>
  Some other text
</xnc:test_element1>" in
  let lexbuf = Lexing.from_string test_fragment in
    OUnit.assert_raises
      ~msg:"The parser doesn't notice that the processing instruction is not closed"
      (Xnunc_lexer.Parse_error "Parse error in (9,21): element \
\"processing instruction\" in (4,8) never closed")
      (fun () -> Xnunc_lexer.parse_xnunc_lexbuf lexbuf)
;; (* END of "not_closed_processing_instruction" *)


let complete_document_validation () =
  let test_fragment = "This is a CDATA block
<xnc:test_element1 foo = \"bar\" param_attr = \"<?xnc= ocaml code here ?>\">
  Some text here...
  <?xnc
    This is a processing instruction
  ?>
  <xnc:test_element2>
    Some useless text here...
    <?xnc
      This is a second processing instruction
    ?>
    <?xnc=
      This is a printing processing instruction
    ?>
  </xnc:test_element2>
  Some other text
</xnc:test_element1>
This is a second CDATA block" in
  let lexbuf = Lexing.from_string test_fragment in

  let parsed_document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in

    OUnit.assert_equal
      ~printer:string_of_int
      ~msg:"Wrong number of parsed blocks"
      3
      (List.length parsed_document);

    OUnit.assert_equal
      ~cmp:Xnunc_lexer.document_compare
      ~printer:Xnunc_lexer.print_document
      ~msg:"The lexer didn't parse the tree correctly"
      [
        (Cdata ("This is a CDATA block\n", (1,1)));
        (
          Element ({
            name = "test_element1";
            attributes =
              [
                ("foo", (2,20)), (Literal("bar"), (2,27));
                ("param_attr", (2,32)), (Parametric(" ocaml code here "), (2,52));
              ];
            sub_trees =
              [
                Cdata("\n  Some text here...\n  ", (2,73));
                Proc_instr("\n    This is a processing instruction\n  ", (4,8));
                Cdata("\n  ", (6,5));
                Element({
                  name = "test_element2";
                  attributes = [];
                  sub_trees =
                    [
                      Cdata("\n    Some useless text here...\n    ", (7,22));
                      Proc_instr("\n      This is a second processing instruction\n    ", (9,10));
                      Cdata("\n    ", (11,7));
                      Print_proc_instr("\n      This is a printing processing instruction\n    ", (12,11));
                      Cdata("\n  ", (14,7));
                    ];
                  el_pos = (7,8);
                });
                Cdata("\n  Some other text\n", (15,23));
              ];
            el_pos = (2,6);
          })
        );
        (Cdata ("\nThis is a second CDATA block", (17,21)));
      ]
      parsed_document;
;; (* END of "complete_document_validation" *)


(* test suite *)
let suite = "Parser test cases" >:::
  [
    "never_closed_element_1" >:: never_closed_element_1;
    "never_closed_element_2" >:: never_closed_element_2;
    "incomplete_attribute" >:: incomplete_attribute;
    "invalid_attribute_name" >:: invalid_attribute_name;
    "no_equal_in_attribute" >:: no_equal_in_attribute;
    "not_closed_processing_instruction" >:: not_closed_processing_instruction;
    "complete_document_validation" >:: complete_document_validation;
  ]
;;

