open OUnit
open Common_types
open Xnc_functional


let test_constraints_table =
  [
    (**************************************************************************)
    (*****                        PAGE ELEMENT                            *****)
    (**************************************************************************)
    "page", (* element name *)
    [
      "page-parameters", 0, 1; (* element child with min and max card *)
      "script", 0, max_int
    ],
    [ (* Attributes *)
      "inherit",
      String, false, false, Some "Page_behavior.page_behavior_base", None
    ],
    true, (* processing instructions allowed? *)
    true, (* cdata block allowed? *)
    true, (* is a root_element? *)
    (fun el ->
       ()
    );
    (**************************************************************************)

    (**************************************************************************)
    (*****                   PAGE-PARAMETER ELEMENT                       *****)
    (**************************************************************************)
    "page-parameters",
    [
      "variable", 1, max_int
    ],
    [ (* Attributes *)
      "int-test-attr", Int(Range(-2, 42)), true, false, None, None;
      "float-test-attr", Float(Range(0.0, 42.0)), true, false, None, None
    ],
    false,
    false,
    false,      
    (fun el -> ());
    (**************************************************************************)

    (**************************************************************************)
    (*****                       SCRIPT ELEMENT                           *****)
    (**************************************************************************)
    "script",
    [
    ],
    [ (* Attributes *)
      "int-test-attr2", Int(No_limits), false, false, None, None;
      "float-test-attr2", Float(No_limits), false, false, None, None;
      "bool-test-attr", Bool, true, false, None, None;
      "scope", String, false, false, Some "module", Some ["module"; "class"]
    ],
    false,
    true,
    false,
    (fun el -> ());
    (**************************************************************************)

    (**************************************************************************)
    (*****                      VARIABLE ELEMENT                          *****)
    (**************************************************************************)
    "variable",
    [
    ],
    [ (* Attributes *)
      "name", String, true, false, None, None;
      "type", String, true, false, None, None;
      "init_value", String, false, false, None, None;
    ],
    false,
    false,
    false,
    (fun el -> ());
    (**************************************************************************)
  ]

let test_validate_document =
  Tree_validation.custom_validate_document test_constraints_table


let int_attribute_out_of_range_test () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters int-test-attr=\"50\" float-test-attr=\"42.0\">
    <xnc:variable name=\"login\" type=\"string\" />
    <xnc:variable name=\"password\" type=\"string\" />
  </xnc:page-parameters>
  <xnc:script scope = \"foo\">
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here...
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = Tree_validation.remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores an out of range integer parameter"
      ( Attribute_out_of_range("int-test-attr", (3, 39), "-2", "42") )
      (fun () -> test_validate_document document)
;; (* END of "int_attribute_out_of_range_test" *)


let float_attribute_out_of_range_test () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters int-test-attr=\"10\" float-test-attr=\"314.272\">
    <xnc:variable name=\"login\" type=\"string\" />
    <xnc:variable name=\"password\" type=\"string\" />
  </xnc:page-parameters>
  <xnc:script scope = \"foo\">
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here...
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = Tree_validation.remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores an out of range float parameter"
      ( Attribute_out_of_range("float-test-attr", (3, 60), "0.", "42.") )
      (fun () -> test_validate_document document)
;; (* END of "float_attribute_out_of_range_test" *)


let bool_invalid_attribute_type_test () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters int-test-attr=\"10\" float-test-attr=\"30.0\">
    <xnc:variable name=\"login\" type=\"string\" />
    <xnc:variable name=\"password\" type=\"string\" />
  </xnc:page-parameters>
  <xnc:script scope = \"class\" bool-test-attr=\"foo\">
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here...
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = Tree_validation.remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores an invalid attribute value"
      ( Invalid_attribute_type("bool-test-attr", (7, 47), "bool") )
      (fun () -> test_validate_document document)
;; (* END of "bool_invalid_attribute_type_test" *)


let int_invalid_attribute_type_test () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters int-test-attr=\"10\" float-test-attr=\"30.0\">
    <xnc:variable name=\"login\" type=\"string\" />
    <xnc:variable name=\"password\" type=\"string\" />
  </xnc:page-parameters>
  <xnc:script scope = \"class\" bool-test-attr=\"false\" int-test-attr2=\"bar\">
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here...
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = Tree_validation.remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores an invalid attribute value"
      ( Invalid_attribute_type("int-test-attr2", (7, 70), "int") )
      (fun () -> test_validate_document document)
;; (* END of "int_invalid_attribute_type_test" *)


let int_invalid_attribute_type_test2 () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters int-test-attr=\"foo\" float-test-attr=\"30.0\">
    <xnc:variable name=\"login\" type=\"string\" />
    <xnc:variable name=\"password\" type=\"string\" />
  </xnc:page-parameters>
  <xnc:script scope = \"class\" bool-test-attr=\"false\" int-test-attr2=\"10\">
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here...
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = Tree_validation.remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores an invalid attribute value"
      ( Invalid_attribute_type("int-test-attr", (3, 39), "int") )
      (fun () -> test_validate_document document)
;; (* END of "int_invalid_attribute_type_test2" *)


let float_invalid_attribute_type_test () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters int-test-attr=\"5\" float-test-attr=\"30.0\">
    <xnc:variable name=\"login\" type=\"string\" />
    <xnc:variable name=\"password\" type=\"string\" />
  </xnc:page-parameters>
  <xnc:script scope = \"class\"
              bool-test-attr=\"false\"
              int-test-attr2=\"10\"
              float-test-attr2=\"foo\">
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here...
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = Tree_validation.remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores an invalid attribute value"
      ( Invalid_attribute_type("float-test-attr2", (10, 33), "float") )
      (fun () -> test_validate_document document)
;; (* END of "float_invalid_attribute_type_test" *)


let float_invalid_attribute_type_test2 () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters int-test-attr=\"5\" float-test-attr=\"foo\">
    <xnc:variable name=\"login\" type=\"string\" />
    <xnc:variable name=\"password\" type=\"string\" />
  </xnc:page-parameters>
  <xnc:script scope = \"class\"
              bool-test-attr=\"false\"
              int-test-attr2=\"10\"
              float-test-attr2=\"12.45\">
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here...
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = Tree_validation.remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores an invalid attribute value"
      ( Invalid_attribute_type("float-test-attr", (3, 59), "float") )
      (fun () -> test_validate_document document)
;; (* END of "float_invalid_attribute_type_test2" *)


let too_many_root_elements_test () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters int-test-attr=\"5\" float-test-attr=\"3.14\">
    <xnc:variable name=\"login\" type=\"string\" />
    <xnc:variable name=\"password\" type=\"string\" />
  </xnc:page-parameters>
  <xnc:script scope = \"class\"
              bool-test-attr=\"false\"
              int-test-attr2=\"10\"
              float-test-attr2=\"12.45\">
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here...
</xnc:page>
Some free text: the following \"page\" should rise an exception!
<xnc:page>
  <xnc:page-parameters int-test-attr=\"5\" float-test-attr=\"3.14\">
    <xnc:variable name=\"login\" type=\"string\" />
    <xnc:variable name=\"password\" type=\"string\" />
  </xnc:page-parameters>
  <xnc:script scope = \"class\"
              bool-test-attr=\"false\"
              int-test-attr2=\"10\"
              float-test-attr2=\"12.45\">
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here...
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = Tree_validation.remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores duplicated root element"
      ( Too_many_root_elements("page", (18, 6)) )
      (fun () -> test_validate_document document)
;; (* END of "too_many_root_elements_test" *)


let non_root_element_test () =
  let test_fragment = "
  <xnc:page-parameters int-test-attr=\"5\" float-test-attr=\"3.14\">
    <xnc:variable name=\"login\" type=\"string\" />
    <xnc:variable name=\"password\" type=\"string\" />
  </xnc:page-parameters>
  <xnc:script scope = \"class\"
              bool-test-attr=\"false\"
              int-test-attr2=\"10\"
              float-test-attr2=\"12.45\">
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here... note that the page root element is missing" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = Tree_validation.remove_phantom_nodes document in

    assert_raises
      ~msg:"Missing root element not"
      ( Non_root_element("page-parameters", (2, 8)) )
      (fun () -> test_validate_document document)
;; (* END of "non_root_element_test" *)


(* test suite *)
let suite = "CUSTOM DTD" >:::
  [
    "int_attribute_out_of_range_test" >:: int_attribute_out_of_range_test;
    "float_attribute_out_of_range_test" >:: float_attribute_out_of_range_test;
    "bool_invalid_attribute_type_test" >:: bool_invalid_attribute_type_test;
    "int_invalid_attribute_type_test" >:: int_invalid_attribute_type_test;
    "int_invalid_attribute_type_test2" >:: int_invalid_attribute_type_test2;
    "float_invalid_attribute_type_test" >:: float_invalid_attribute_type_test;
    "float_invalid_attribute_type_test2" >:: float_invalid_attribute_type_test2;
    "too_many_root_elements_test" >:: too_many_root_elements_test;
    "non_root_element_test" >:: non_root_element_test;
  ]
;;

