open OUnit
open Common_types
open Xnc_functional


let validate_document = Tree_validation.validate_document
let remove_phantom_nodes = Tree_validation.remove_phantom_nodes

let unknown_element_test () =
  let test_fragment = "
<xnc:unknown>
  Some text here...
  <xnc:unknown2>
    Some useless text here...
  </xnc:unknown2>
</xnc:unknown>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in

    assert_raises
      ~msg:"The parser seems to know the \"unknown\" element"
      ( Unknown_element("unknown", (2, 6)) )
      (fun () -> validate_document document)
;; (* END of "unknown_element_test" *)


let proc_instr_not_allowed_test () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters>
    <xnc:variable name=\"foo\" type=\"bar\" />
    <?xnc
      This processing instruction is illegal here!
    ?>
  </xnc:page-parameters>
  <xnc:head>
      <title>A title</title>
  </xnc:head>
  <xnc:body>
      Some text here...
  </xnc:body>
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in

    assert_raises
      ~msg:"The parser do not filter the illegal processing instruction"
      ( Proc_instr_not_allowed("page-parameters", (3, 8)) )
      (fun () -> document |> remove_phantom_nodes |> validate_document)
;; (* END of "proc_instr_not_allowed_test" *)


let cdata_not_allowed_test () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters>
    <xnc:variable name=\"foo\" type=\"bar\" />
    This free text is forbidden in \"page-parameters\" element
  </xnc:page-parameters>
  <xnc:head>
      <title>A title</title>
  </xnc:head>
  <xnc:body>
      Some text here...
  </xnc:body>
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser do not recognize forbidden CDATA clock"
      ( Cdata_not_allowed("page-parameters", (3, 8)) )
      (fun () -> validate_document document)
;; (* END of "cdata_not_allowed_test" *)


let illegal_child_test () =
  let test_fragment = "
<xnc:page>
  <xnc:variable name=\"foo\" type=\"bar\" />
  <xnc:head>
      <title>A title</title>
  </xnc:head>
  <xnc:body>
      Some text here...
  </xnc:body>
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser do not recognize the forbidden \"variable\" element"
      ( Illegal_child("page", (2, 6), "variable") )
      (fun () -> validate_document document)
;; (* END of "illegal_child_test" *)


let too_few_children_test () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters>
  </xnc:page-parameters>
  <xnc:head>
      <title>A title</title>
  </xnc:head>
  <xnc:body>
      Some text here...
  </xnc:body>
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores that \"page-parameters\" is empty"
      ( Too_few_children("page-parameters", (3, 8), 1, "variable") )
      (fun () -> validate_document document)
;; (* END of "too_few_children_test" *)


let too_many_children_test () =
  let test_fragment = "
<xnc:page>
  <xnc:page-parameters>
    <xnc:variable name=\"foo\" type=\"bar\" />
    <xnc:variable name=\"foo2\" type=\"bar2\" />
  </xnc:page-parameters>
  <xnc:page-parameters>
    <xnc:variable name=\"foo3\" type=\"bar3\" />
  </xnc:page-parameters>
  <xnc:head>
      <title>A title</title>
  </xnc:head>
  <xnc:body>
      Some text here...
  </xnc:body>
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores the duplicated illegal <page-parameters> element"
      ( Too_many_children("page", (2, 6), 1, "page-parameters") )
      (fun () -> validate_document document)
;; (* END of "too_many_children_test" *)


let missing_attribute_test () =
  let test_fragment = "
<xnc:page name=\"test\" xnc-version=\"1.0\">
  <xnc:page-parameters>
    <xnc:variable name=\"foo\" type=\"int\" default=\"5\"/>
    <xnc:variable default=\"mad\"/>
  </xnc:page-parameters>
  <xnc:head>
      <title>A title</title>
  </xnc:head>
  <xnc:body>
      Some text here...
  </xnc:body>
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores the missing attribute \"name\" of the \"variable\" element"
      ( Missing_attribute("variable", (5, 10), "name") )
      (fun () -> validate_document document)
;; (* END of "missing_attribute_test" *)


let unknown_attribute_test () =
  let test_fragment = "
<xnc:page name=\"baz\" xnc-version=\"1.0\">
  <xnc:page-parameters>
    <xnc:variable name=\"foo\" type=\"string\" default=\"bar\"/>
    <xnc:variable foo=\"foo_val\" type=\"bool\" default=\"baz\"/>
  </xnc:page-parameters>
  <xnc:head>
      <title>A title</title>
  </xnc:head>
  <xnc:body>
      Some text here...
  </xnc:body>
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores the presence of the unknown attribute \"foo\""
      ( Unknown_attribute("foo", (5, 19)) )
      (fun () -> validate_document document)
;; (* END of "unknown_attribute_test" *)


let parametric_param_not_allowed_test () =
  let test_fragment = "
<xnc:page name=\"foo\" xnc-version=\"1.0\">
  <xnc:page-parameters>
    <xnc:variable name=\"<?xnc= Printf.sprintf \"foo\" ?>\" type=\"bar\" default=\"foo\"/>
  </xnc:page-parameters>
  <xnc:head>
      <title>A title</title>
  </xnc:head>
  <xnc:body>
      Some text here...
  </xnc:body>
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores the illegal parametric value of the \"name\" attribute"
      ( Parametric_param_not_allowed("name", (4, 19)) )
      (fun () -> validate_document document)
;; (* END of "parametric_param_not_allowed_test" *)


let invalid_attribute_value_test () =
  let test_fragment = "
<xnc:page name=\"test\" xnc-version=\"1.0\">
  <xnc:page-parameters>
    <xnc:variable name=\"login\" type=\"string\" default=\"paolo\"/>
    <xnc:variable name=\"password\" type=\"string\" default=\"my_pwd\"/>
  </xnc:page-parameters>
  <xnc:head>
      <title>A title</title>
  </xnc:head>
  <xnc:body>
      <xnc:script scope = \"foo\">
        exception My_exception
        let useless_int_var = 3;;
        let useless_int_ref_var = ref 42;;
      </xnc:script>
      Some text here...
  </xnc:body>
</xnc:page>" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = remove_phantom_nodes document in

    assert_raises
      ~msg:"The parser ignores that the \"scope\" attribute legal values are scope and class"
      ( Invalid_attribute_value("scope", (11, 28)) )
      (fun () -> validate_document document)
;; (* END of "invalid_attribute_value_test" *)


let insert_default_attributes_test () =
  let test_fragment = "
<xnc:page>
  <xnc:script>
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  </xnc:script>
  Some text here...
</xnc:page>
Some free text:" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = remove_phantom_nodes document in
  let document = Tree_validation.insert_document_defaults document in

  OUnit.assert_equal
    ~cmp:Xnunc_lexer.document_compare
    ~printer:Xnunc_lexer.print_document
    ~msg:"The expected default parameters have not been added."
    [
      (
        Element({
          name = "page";
          attributes =
            [
              ("debug", (-1,-1)), (Literal("false"), (-1,-1));     
              ("dtd", (-1,-1)), (Literal("xhtml1-transitional"),(-1,-1));
              ("encoding", (-1,-1)), (Literal("ISO-8859-1"), (-1,-1));
              ("inherit", (-1,-1)), (Literal("Page_modules.Page_behavior.page_behavior_base"), (-1,-1));
              ("type", (-1,-1)), (Literal("standard"), (-1,-1));
              ("xml-version", (-1,-1)), (Literal("1.0"), (-1,-1));
            ];
          sub_trees =
                   [
                     Element({
                       name = "script";
                       attributes =
                         [
                           ("scope", (-1,-1)), (Literal("module"), (-1,-1));
                         ];
                       sub_trees =
                         [
                           Cdata("
    exception My_exception
    let useless_int_var = 3;;
    let useless_int_ref_var = ref 42;;
  ",(3,15)
                           );
                         ];
                       el_pos = (3,8);
                     });
                     Cdata("\n  Some text here...\n", (7,16));
                   ];
          el_pos = (2,6);
        });
      );
      Cdata("\nSome free text:", (9,12));
    ]
    document;
;; (* END of "insert_default_attributes_test" *)


let illegal_params_in_error_page () =
  let test_fragment = "
<xnc:page type=\"error\" name=\"test\" xnc-version=\"1.0\">
  <xnc:page-parameters>
    <xnc:variable name=\"login\" type=\"string\" default=\"paolo\" />
    <xnc:variable name=\"password\" type=\"string\" default=\"foo123\" />
  </xnc:page-parameters>
  <xnc:head>
      <title>A title</title>
  </xnc:head>
  <xnc:body>
      <xnc:script scope = \"class\">
        exception My_exception
        let useless_int_var = 3;;
        let useless_int_ref_var = ref 42;;
      </xnc:script>
      Some text here...
  </xnc:body>
</xnc:page>
Some free text: the following \"page\" should rise an exception!" in
  let lexbuf = Lexing.from_string test_fragment in
  let document = Xnunc_lexer.parse_xnunc_lexbuf lexbuf in
  let document = remove_phantom_nodes document in

    assert_raises
      ~msg:""
      (Illegal_child("page", (2, 6), "page-parameters"))
      (fun () -> validate_document document)
;; (* END of "illegal_params_in_error_page" *)


(* test suite *)
let suite = "Ex-Nunc DTD" >:::
  [
    "unknown_element_test" >:: unknown_element_test;
    "proc_instr_not_allowed_test" >:: proc_instr_not_allowed_test;
    "cdata_not_allowed_test" >:: cdata_not_allowed_test;
    "illegal_child_test" >:: illegal_child_test;
    "too_few_children_test" >:: too_few_children_test;
    "too_many_children_test" >:: too_many_children_test;
    "missing_attribute_test" >:: missing_attribute_test;
    "unknown_attribute_test" >:: unknown_attribute_test;
    "parametric_param_not_allowed_test" >:: parametric_param_not_allowed_test;
    "invalid_attribute_value_test" >:: invalid_attribute_value_test;
    "insert_default_attributes_test" >:: insert_default_attributes_test;
    "illegal_params_in_error_page" >:: illegal_params_in_error_page;
  ]
;;

