open Xnc_stdlib
open Xnc_functional
open Common_types

module Pf = Printf


let count_new_lines s =
  let counter = ref 0 in
  let length = String.length s in
    for i = 0 to (length - 1) do if s.[i] = '\n' then incr counter done;
    !counter
;; (* END OF "count_new_lines" FUNCTION *)

let is_a_white_string s = s |> Utility.trim_white_space |> String.length = 0

let remove_phantom_nodes
      (elements : document)
      : document =

  let rec remove_phantom_nodes_aux
        (elements : parsed_block list)
        (accumulator : parsed_block list)
        : parsed_block list =

      match elements with
        | head :: tail ->
            begin
              match head with
                | Cdata(data,pos) when data |> is_a_white_string ->
                    remove_phantom_nodes_aux tail accumulator
                | Element(xnunc_element) ->
                    let new_fragment = {
                      xnunc_element with
                          sub_trees = remove_phantom_nodes_aux
                                            xnunc_element.sub_trees [];
                    } in
                      remove_phantom_nodes_aux
                        tail
                        ((Element(new_fragment)) :: accumulator)
                | _ ->
                    remove_phantom_nodes_aux tail (head::accumulator)
            end
        | [] -> List.rev accumulator
  in (* END OF "remove_phantom_nodes_aux" FUNCTION *)

    remove_phantom_nodes_aux elements []
;; (* END OF "remove_phantom_nodes" FUNCTION *)


let fill_child_element_map child_list =
  List.fold_left
    (
      fun map (name, card_min, card_max) ->
        String_map.add name (card_min, card_max) map
    )
    String_map.empty
    child_list
;;


let fill_attributes_map attributes_list =
  List.fold_left
    (
      fun map (name, attr_type, mandatory, param_allowed, default, val_list) ->
        String_map.add
          name
          {
            attr_type = attr_type;
            mandatory = mandatory;
            param_allowed = param_allowed;
            default = default;
            valid_values = val_list;
          }
          map
    )
    String_map.empty
    attributes_list
;;


let fill_element_constraints_map contr_list =
  List.fold_left
    (
      fun map (name, children, attributes_list, proc, cdata,
               root_el, specific_validation) ->
        String_map.add
          name
          {
            children = fill_child_element_map children;
            attributes_map = fill_attributes_map attributes_list;
            proc_instr_allowed = proc;
            cdata_allowed = cdata;
            is_root_element = root_el;
            validation = specific_validation;
          }
          map
    )
    String_map.empty
    contr_list
;;


let apply_if_element f node =
  match node with
    | Element(element) -> f element
    | _ -> ()
;; (* END OF "apply_if_element" FUNCTION *)


let rec look_for_unknown_elements element_constraints_map node =
  let is_not_present_in_list el =
    not (String_map.mem el.name element_constraints_map) in

  let raise_exception_if_unknown element =
    if element |> is_not_present_in_list
    then raise (Unknown_element(element.name, element.el_pos));

    List.iter
      (look_for_unknown_elements element_constraints_map)
      element.sub_trees
  in

    apply_if_element raise_exception_if_unknown node;
;; (* END OF "look_for_unknown_elements" FUNCTION *)


let add name map =
  let card = match String_map.find_opt name map with
    | Some el -> el
    | None -> 0 in
    String_map.add name (card + 1) map
;; (* END OF "add" FUNCTION *)


let extract_element_constraint element_constraints_map element_name =
  String_map.find element_name element_constraints_map
;; (* END OF "extract_element_constraint" FUNCTION *)


let check_child_cardinality element child_name count min max =
  if count < min
  then raise (Too_few_children(
    element.name, element.el_pos, min, child_name));

  if count > max
  then raise (Too_many_children(
    element.name, element.el_pos, max, child_name))
;; (* END OF "check_child_cardinality" FUNCTION *)


let verify_if_child_count_is_ok
      element_constraints_map
      (element : parsed_element)
      (child_name : string)
      (count : card)
      : unit =
  let this_element_constraint =
    extract_element_constraint element_constraints_map element.name in
  let min, max =
    try
      String_map.find child_name
        this_element_constraint.children
    with
      | Not_found -> raise (Illegal_child(
          element.name, element.el_pos, child_name)) in

    check_child_cardinality element child_name count min max
;; (* END OF "verify_if_child_count_is_ok" FUNCTION *)


let verify_if_constraint_is_respected
      (cardinality_map : int String_map.t)
      (element : parsed_element)
      (child_name : string)
      (min, max : (card * card))
      : unit =

  let count =
    try String_map.find child_name cardinality_map
    with | Not_found -> 0 in

    check_child_cardinality element child_name count min max
;; (* END OF "verify_if_constraint_is_respected" FUNCTION *)


let rec check_children_constraints_aux element_constraints_map element =

  let this_element_constraint = extract_element_constraint
                                  element_constraints_map
                                  element.name in

  (* cardinality_map is a map from children names to children cardinality *)
  let cardinality_map, has_proc_instr, has_cdata =
    List.fold_left
      (
        fun (map, proc, cdata) child ->
          match child with
            | Proc_instr(_) | Print_proc_instr(_) -> map, true, cdata
            | Cdata(_) -> map, proc, true
            | Element(el) -> add el.name map, proc, cdata
      )
        (String_map.empty, false, false)
      element.sub_trees in

    if this_element_constraint.proc_instr_allowed = false &&
       has_proc_instr = true
    then raise (Proc_instr_not_allowed(element.name, element.el_pos));

    if this_element_constraint.cdata_allowed = false &&
       has_cdata = true
    then raise (Cdata_not_allowed(element.name, element.el_pos));

    String_map.iter
      (verify_if_child_count_is_ok element_constraints_map element)
      cardinality_map;

    String_map.iter
      (verify_if_constraint_is_respected cardinality_map element)
      this_element_constraint.children;

    List.iter
      (check_children_constraints element_constraints_map)
      element.sub_trees
(* END "check_children_constraints_aux" FUNCTION *)

and check_children_constraints element_constraints_map node =
    apply_if_element
      (check_children_constraints_aux element_constraints_map) node;
;; (* END OF "check_children_constraints" FUNCTION *)


let check_mandatory_attributes element_constraints_map element =
  let this_element_constraint =
    extract_element_constraint element_constraints_map element.name in
  let attribute_constraints = this_element_constraint.attributes_map in
    String_map.iter
      (
        fun attr_name attr_constr ->
          if attr_constr.mandatory = true
          then
            try
              ignore
                (
                  List.find
                    (
                      fun ((name, pos), value) ->
                        if name = attr_name then true else false
                    )
                    element.attributes
                )
            with
              | Not_found ->
                  raise (Missing_attribute(
                    element.name, element.el_pos, attr_name))
      )
      attribute_constraints
;; (* END OF "check_mandatory_attributes" FUNCTION *)


let check_parametric_attributes element_constraints_map element =
  let this_element_constraint =
    extract_element_constraint element_constraints_map element.name in
  let attribute_constraints = this_element_constraint.attributes_map in
    List.iter
      (
        fun ((name, pos), (value, value_pos)) ->
          match value with
            | Literal(_) -> ()
            | Parametric(_) ->
                let attribute_constraint =
                  String_map.find name attribute_constraints in
                  if attribute_constraint.param_allowed = false
                  then raise (Parametric_param_not_allowed(name, pos))
      )
      element.attributes
;; (* END OF "check_parametric_attributes" FUNCTION *)


let check_the_type_of_an_attribute attr_constr attr_name value pos =
  match attr_constr.attr_type with
    | String -> ()
    | Bool -> begin
        try ignore (value |> String.lowercase |> bool_of_string)
        with
          | Invalid_argument "bool_of_string" ->
              raise (Invalid_attribute_type(attr_name, pos, "bool"))
      end
    | Int(No_limits) -> begin
        try ignore (value |> int_of_string)
        with
          | Failure "int_of_string" ->
              raise (Invalid_attribute_type(attr_name, pos, "int"))
      end
    | Int(Range (min, max)) ->
        let int_value = try int_of_string value
        with
          | Failure "int_of_string" ->
              raise (Invalid_attribute_type(attr_name, pos, "int")) in
          if int_value < min || int_value > max
          then raise (Attribute_out_of_range(attr_name, pos, (string_of_int min), (string_of_int max)))
    | Float(No_limits) -> begin
        try ignore (value |> float_of_string)
        with
          | Failure "float_of_string" ->
              raise (Invalid_attribute_type(attr_name, pos, "float"))
      end
    | Float(Range (min, max)) ->
        let float_value = try float_of_string value
        with
          | Failure "float_of_string" ->
              raise (Invalid_attribute_type(attr_name, pos, "float")) in
          if float_value < min || float_value > max
          then raise (Attribute_out_of_range(attr_name, pos, (string_of_float min), (string_of_float max)))
;; (* END OF "check_the_type_of_an_attribute" FUNCTION *)


let check_type_attributes element_constraints_map element =
  let this_element_constraint =
    extract_element_constraint element_constraints_map element.name in
  let attribute_constraints = this_element_constraint.attributes_map in
    List.iter
      (
        fun ((name, pos), (value, value_pos)) ->
          match value with
            | Parametric(_) -> ()
            | Literal actual_value ->
                let attribute_constraint =
                  String_map.find name attribute_constraints in

                  check_the_type_of_an_attribute
                    attribute_constraint name actual_value value_pos
      )
      element.attributes
;; (* END OF "check_type_attributes" FUNCTION *)


let check_string_attributes_valid_values element_constraints_map element =
  let this_element_constraint =
    extract_element_constraint element_constraints_map element.name in
  let attribute_constraints = this_element_constraint.attributes_map in
    List.iter
      (
        fun ((name, pos), (value, value_pos)) ->
          match value with
            | Parametric(_) -> ()
            | Literal actual_value ->
                let attribute_constraint =
                  String_map.find name attribute_constraints in
                let valid_values = attribute_constraint.valid_values in
                  match valid_values with
                    | None -> ()
                    | Some valid_list -> try
                        ignore (List.find (fun e -> e = actual_value)
                                  valid_list)
                      with
                        | Not_found ->
                            raise (Invalid_attribute_value(name, value_pos))
      )
      element.attributes
;; (* END OF "check_string_attributes_valid_values" FUNCTION *)


let check_if_attributes_are_known element_constraints_map element =
  let this_element_constraint =
    extract_element_constraint element_constraints_map element.name in
  let attribute_constraints = this_element_constraint.attributes_map in
    List.iter
      (
        fun ((name, pos), value) ->
          try
            ignore (String_map.find name attribute_constraints);
          with
            | Not_found -> raise(Unknown_attribute(name, pos))
      )
      element.attributes
;; (* END OF "check_if_attributes_are_known" FUNCTION *)


let rec check_attributes_constraints element_constraints_map node =
  let check_attributes_constraints_aux element_constraints_map element =
    check_if_attributes_are_known element_constraints_map element;
    check_mandatory_attributes element_constraints_map element;
    check_parametric_attributes element_constraints_map element;
    check_type_attributes element_constraints_map element;
    check_string_attributes_valid_values element_constraints_map element;
    List.iter
      (check_attributes_constraints element_constraints_map)
      element.sub_trees;
  in (* END OF "check_attributes_constraints_aux" FUNCTION *)

    apply_if_element
       (check_attributes_constraints_aux element_constraints_map)
       node
;; (* END OF "check_attributes_constraints" FUNCTION *)


let validate_global_constraints
      element_constraints_map
      (elements : document)
      : unit =
  ignore(List.fold_left
           (
             fun el_found element ->
               match element with
                 | Element(element) ->
                     begin
                       if el_found = true
                       then raise (Too_many_root_elements(
                         element.name, element.el_pos))
                       else
                         let this_element_constraint =
                           extract_element_constraint
                             element_constraints_map
                             element.name in
                           if this_element_constraint.is_root_element = false
                           then raise ( Non_root_element(
                             element.name, element.el_pos))
                           else true
                     end
                 | _ -> el_found
           ) false elements
  )
;; (* END OF "validate_global_constraints" FUNCTION *)


let rec check_specific_validation element_constraints_map node =
  let run_validation element =
    let this_element_constraint = extract_element_constraint
                                    element_constraints_map
                                    element.name in

      this_element_constraint.validation element;

      List.iter
        (check_specific_validation element_constraints_map)
        element.sub_trees in

    apply_if_element run_validation node
;; (* END OF "check_specific_validation" FUNCTION *)


let validate_tree element_constraints_map node =
  (* First verify if the element is known *)
  look_for_unknown_elements element_constraints_map node;
  check_children_constraints element_constraints_map node;
  check_attributes_constraints element_constraints_map node;
  check_specific_validation element_constraints_map node;
;; (* END OF "validate_tree" FUNCTION *)


let custom_validate_document
      (constr_table : Common_types.constraints_table)
      (elements : document)
         : unit =
  let element_constraints_map = fill_element_constraints_map constr_table in

    List.iter ( fun el -> validate_tree element_constraints_map el ) elements;
    validate_global_constraints element_constraints_map elements;
;; (* END OF "custom_validate_document" FUNCTION *)


let validate_document
      (elements : document)
         : unit =
  custom_validate_document Validity_rules.constraints_table elements

;; (* END OF "validate_document" FUNCTION *)


let rec attr_mem_assoc attr_name attr_list =
  match attr_list with
    | [] -> false
    | ((name, _), (_, _)) :: l ->
        String.compare name attr_name = 0 || attr_mem_assoc attr_name l
;; (* END OF "mem_assoc" FUNCTION *)


let attribute_default
      attr_contr_map
      (attributes : attribute list)
      : attribute list =

  let default_collection =
    String_map.fold
      (
        fun attr_name attr_contr def_collection ->
          if attr_contr.mandatory = false && attr_contr.default <> None
          then
            begin
            let new_attribute = (
              (attr_name, (-1, -1)),
              ((Literal(Option.get attr_contr.default )), (-1, -1))) in
              new_attribute :: def_collection
            end
          else
            def_collection
      )
      attr_contr_map
      [] in

    List.fold_left
      (
        fun attributes (((attr_name, _), (_, _)) as def_attr) ->
          if (attr_mem_assoc attr_name attributes) = false
          then
            def_attr :: attributes
          else
            attributes
      )
      attributes
      default_collection
;; (* END OF "attribute_default" FUNCTION *)


let rec insert_element_default
      element_constraints_map
      (old_element : parsed_block)
      : parsed_block =
    match old_element with
      | Proc_instr _ -> old_element
      | Print_proc_instr _ -> old_element
      | Cdata _ -> old_element
      | Element parsed_el ->
          let el_contr =
            String_map.find parsed_el.name element_constraints_map in
            Element (
              { parsed_el with
                    attributes = attribute_default
                                      el_contr.attributes_map
                                      parsed_el.attributes;
                    sub_trees = insert_document_defaults_aux
                                     element_constraints_map
                                     parsed_el.sub_trees
              }
            )
and (* END OF "insert_element_default" FUNCTION *)
insert_document_defaults_aux
          element_constraints_map
          (elements : document)
          : document =
  List.fold_left (
    fun accum old_el ->
      let new_element = insert_element_default element_constraints_map old_el in
        new_element :: accum
  ) [] elements |> List.rev
;; (* END OF "insert_document_defaults_aux" FUNCTION *)


let insert_document_defaults
      (elements : document)
      : document =
  let element_constraints_map =
    fill_element_constraints_map Validity_rules.constraints_table in

    insert_document_defaults_aux
      element_constraints_map
      elements
;; (* END OF "insert_document_defaults" FUNCTION *)


(*
let process_fragment oc (string_count, line_count) f =
  match f with
    | Proc_instr(code, (line, pos)) ->
        begin
          Pf.fprintf oc "\n# %d \"%s\"\n" line in_file_name;
          add_spaces oc (pos - 1);
          Pf.fprintf oc "%s\n" code;
          (string_count, line_count + (new_line_count code) + 3)
        end
    | Print_proc_instr(p_code, (line, pos)) ->
        begin
          Pf.fprintf oc "\n# %d \"%s\"\n" line in_file_name;
          add_spaces oc (pos - 1);
          Pf.fprintf oc "%s;\n" p_code;
          (string_count, line_count + (new_line_count p_code) + 3)
        end
    | Cdata(text) ->
        begin
          Pf.fprintf oc "# %d \"%s\"\n" (line_count + 1) out_file_name;
          Pf.fprintf oc "let string_%03d = " string_count;
          let escaped = Str.global_replace double_quote "\\\"" text in
            Pf.fprintf oc "\"%s\" in\n" escaped;
            (string_count + 1, line_count + (new_line_count escaped) + 2)
        end
    | Element(xnunc_el) ->
        begin
          Pf.printf "Elemento \"%s\" trovato in posizione (%d,%d)\n"
            xnunc_el.name
            (fst xnunc_el.el_pos)
            (snd xnunc_el.el_pos);
          (string_count, line_count)
        end
;;

let (_, last_line) = List.fold_left
                       (process_fragment gen_channel)
                       (1, 1)
                       frags;;

Pf.fprintf gen_channel "# %d \"%s\"\n" (last_line + 1) out_file_name;
output_string gen_channel "()";
close_out gen_channel;;

print_endline ((string_of_int (last_line + 1)) ^ " lines have been generated");
 *)
