open Xnc_stdlib
open Xnc_functional


let norm_path = Neturl.split_path >> Neturl.norm_path >> Neturl.join_path
;;

(* Concatenate paths safaly *)
let (//) p1 p2 = Filename.concat p1 p2 |> norm_path
;;

let string_of_file ch =
  let l = in_channel_length ch in
  let s = String.create l in
    really_input ch s 0 l;
    s
;;

let string_of_file_name name =
  let ic = open_in_bin name in
  let content = string_of_file ic in
    close_in ic;
    content
;;

let string_to_file_name name content =
  let oc = open_out_bin name in
    output_string oc content;
    close_out oc
;;

let timestamp_to_string () =
  let now = Unix.gettimeofday () in
  let struct_now = Unix.localtime now in
    Printf.sprintf "%04d-%02d-%02d %02d:%02d:%02d"
      (struct_now.Unix.tm_year + 1900)
      (struct_now.Unix.tm_mon + 1)
      struct_now.Unix.tm_mday
      struct_now.Unix.tm_hour
      struct_now.Unix.tm_min
      struct_now.Unix.tm_sec
;;

let mkdir ?(perm = 0o755) dir_name =
  try
    Unix.mkdir dir_name perm
  with
    | Unix.Unix_error(Unix.EEXIST, _, _) -> ()
;;

let rec mkdirhier ?(permissions = 0o755) directory =
  let dir_name = Filename.dirname directory in
    try
      mkdir ~perm:permissions directory
    with
      | Unix.Unix_error (Unix.EACCES, raising_fun, parameter) ->
          raise (Unix.Unix_error (Unix.EACCES, "mkdirhier", parameter))
      | Unix.Unix_error (Unix.ENOENT, raising_fun, parameter) ->
          mkdirhier ~permissions:permissions dir_name;
          Unix.mkdir directory permissions
;;

let uniq ?(equal = (=)) l =
  let uniq' acc lst =
    List.fold_left
      (
        fun (last, result) el ->
          if equal last el then last, result else el, el::result
      )
      acc
      lst in

    match List.hd_opt l with
      | None -> l
      | Some first -> uniq' (first,[first]) l |> snd |> List.rev
;;

let is_file name =
  let kind = (Unix.stat name).Unix.st_kind in
    match kind with
      | Unix.S_REG -> true
      | _ -> false
;;

let is_dir name =
  let kind = (Unix.stat name).Unix.st_kind in
    match kind with
      | Unix.S_DIR -> true
      | _ -> false
;;


let rec iter_on_dir_tree f start_dir =
  let handle = Unix.opendir start_dir in
    try
      while true
      do
        let entry = Unix.readdir handle in
          if entry <> "." && entry <> ".."
          then
            let file_name = Filename.concat start_dir entry in
              try
                let kind = (Unix.stat file_name).Unix.st_kind in
                  match kind with
                    | Unix.S_DIR ->
                        let () = f kind file_name in
                          iter_on_dir_tree f file_name;
                    | _ -> f kind file_name
              with
                | Unix.Unix_error (err, raising_fun, parameter) -> ()
      done;
    with
      | End_of_file -> Unix.closedir handle
      | e -> Unix.closedir handle; raise e
;;

let rec fold_on_dir_tree f acc start_dir =
  let rec loop_the_dir handle acc =
    let entry =
      try
        Some (Unix.readdir handle)
      with
        | End_of_file -> None in

      match entry with
        | None -> acc   (* No more entry *)
        | Some entry -> (* another entry to process *)
            begin
              if entry <> "." && entry <> ".." (* skip . and .. *)
              then
                begin
                  let file_name = Filename.concat start_dir entry in
                  let kind =
                    try Some ((Unix.stat file_name).Unix.st_kind)
                    with
                      | Unix.Unix_error _ -> None in

                    match kind with
                      | Some kind ->
                          begin
                            let new_acc = f kind acc file_name in
                              match kind with
                                | Unix.S_DIR ->
                                    loop_the_dir handle
                                      (fold_on_dir_tree f new_acc file_name)
                                | _ -> loop_the_dir handle new_acc
                          end
                      | None ->
                          loop_the_dir handle acc (* cannot stat, continue *)
                end
              else loop_the_dir handle acc (* continue looping *)
            end
  in (* loop_the_dir *)
  
  let handle = Unix.opendir start_dir in
    try
      let result_acc = loop_the_dir handle acc in
        Unix.closedir handle;
        result_acc
    with
      | e -> Unix.closedir handle; raise e
;;

let file_copy old_name new_name =
  string_of_file_name old_name |> string_to_file_name new_name;
  let old_stat = Unix.stat old_name in
    Unix.chmod new_name old_stat.Unix.st_perm
;;

let trim_white_space s =
  let is_white c =
    match c with
      | ' ' | '\t' | '\r' | '\n' -> true
      | _ -> false in

  let rec trim_left s =
    let len = String.length s in
      if len > 0
      then if is_white s.[0] then trim_left (String.sub s 1 (len - 1)) else s
      else s in

  let rec trim_right s =
    let len = String.length s in
      if len > 0
      then
        if is_white s.[len - 1] then trim_right (String.sub s 0 (len - 1))
        else s
      else s in

    s |> trim_left |> trim_right
;;

