{
open Common_types

exception Parse_error of string

let change_line pos new_line = (new_line, (snd pos))
let change_char pos new_char_pos = ((fst pos), new_char_pos)
let get_line pos = fst pos
let get_char pos = snd pos

let buf         = Buffer.create     1024
let add_string  = Buffer.add_string buf
let add_char    = Buffer.add_char   buf
let contents () = Buffer.contents   buf
let clear ()    = Buffer.clear      buf
let flush ()    =
  let s = contents () in
    clear ();
    s

(* the position in the current file *)
let line = ref 1 (* line numbered from 1 *)
let char = ref 1 (* character in the current line numbered from 1 *)

let location () = (!line, !char)

let newline  () =
  add_char '\n';
  incr line;
  char := 1

let newchars n = char := !char + n

let advance_pos s =
  let length = String.length s in
    for i = 0 to (length - 1)
    do
      if  (s.[i] <> '\n') (* TODO: handle DOS format *)
      then
        newchars 1
      else
        begin
          incr line;
          char := 1
        end
    done


let init () =
  clear ();
  line := 1;
  char := 1

let add s =
  add_string s;
  newchars (String.length s)

let addc c =
  add_char c;
  newchars 1

let length lexbuf =
  Lexing.lexeme_end lexbuf - Lexing.lexeme_start lexbuf

let root_element_default_name = "__EX_NUNC_ROOT_ELEMENT__"
let not_closed_error_string =
  format_of_string
    "Parse error in (%d,%d): element \"%s\" in (%d,%d) never closed"

let string_unquote = Xnc_string_transform.unescape_xml
}

let nl = "\r\n" | "\n" | "\r"
let letter = ['a'-'z' 'A'-'Z']
let digit = ['0' - '9']
let name_char = letter | digit | '.' | '-' | '_' | ':'
let name = 	(letter | '_' | ':') (name_char)*
let	space = (' ' | '\t' | '\n' | '\r')
let	spaces = space+
let eq = spaces? '=' spaces?

let char_ref = ("&#" ['0' - '9']+ ';') | ("&#x" ['0'-'9' 'a'-'f' 'A'-'F']+ ';')
let entity_ref = '&' name ';'
let reference = entity_ref | char_ref
let attr_valid = ([^ '<' '&' '"' ] | reference)*

rule scan_element_content starting_loc el = parse
| "<?xnc"
    {
      let cdata_frag = Cdata (flush (), starting_loc) in
        newchars 5;
        let loc = location () in
        let proc_instr_frag = scan_proc_instr loc lexbuf in
        let new_loc = location () in
          cdata_frag ::
          Proc_instr(proc_instr_frag, loc) ::
          (scan_element_content new_loc el lexbuf)
    }

| "<?xnc="
    {
      let cdata_frag = Cdata (flush (), starting_loc) in
        newchars 6;
        let loc = location () in
        let print_proc_instr_frag = scan_proc_instr loc lexbuf in
        let new_loc = location () in
          cdata_frag ::
          Print_proc_instr(print_proc_instr_frag, loc) ::
          (scan_element_content new_loc el lexbuf)
    }

| "<xnc:" (name as name)
    {
      let cdata_frag = Cdata (flush (), starting_loc) in
        newchars 5;
        let loc = location () in
          newchars (String.length name);
          let new_element =
            {
              name = String.lowercase name;
              attributes = [];
              sub_trees = [];
              el_pos = loc
            } in
          let new_element = scan_element new_element lexbuf in
          let new_loc = location () in
            cdata_frag ::
            (Element new_element) ::
            (scan_element_content new_loc el lexbuf)
    }

| "</xnc:" (name as closing_el) ">"
    {
      if ((String.lowercase closing_el) = el.name)
      then
        let element_length = String.length closing_el in
          newchars (element_length + 7);
          [Cdata (flush (), starting_loc)]
      else
        begin
          let (line, pos) = location() in
          let err =
            Printf.sprintf not_closed_error_string
              line pos el.name (fst el.el_pos) (snd el.el_pos)
          in
            raise (Parse_error err)
        end
    }

| nl
    {
      newline ();
      scan_element_content starting_loc el lexbuf
    }

| eof
    {
      if (el.name = root_element_default_name)
      then
        [Cdata (flush (), starting_loc)]
      else
        let (line, pos) = location() in
        let err =
          Printf.sprintf not_closed_error_string
            line pos el.name (fst el.el_pos) (snd el.el_pos)
        in
          raise (Parse_error err)
    }

| _
    {
      addc (Lexing.lexeme_char lexbuf 0);
      scan_element_content starting_loc el lexbuf
    }

and scan_element el = parse
| (name as attr_name) (eq as eq) '"' (attr_valid* as attr_val) '"'
    {
      let name_length = String.length attr_name in
      let name_location = location () in
        newchars name_length;
        advance_pos eq;
        newchars 1; (* The first '"' character *)
        let attr_val_location = location () in
          advance_pos attr_val;
          let attr_val = string_unquote attr_val in
            newchars 1; (* The second '"' character *)

            let new_attribute = (attr_name, name_location),
                                (Literal attr_val, attr_val_location) in
            let new_element =
              { el with attributes =  new_attribute :: el.attributes } in

              scan_element new_element lexbuf
    }

| (name as attr_name) (eq as eq) "\"<?xnc="
    {
      let name_length = String.length attr_name in
      let name_location = location () in
        newchars name_length;
        advance_pos eq;
        newchars 7;
        let code_location = location () in
        let code = scan_parametric_attribute lexbuf in
          advance_pos code;
          newchars 3;

          let new_attribute = (attr_name, name_location),
                              (Parametric code, code_location) in
          let new_element =
            { el with attributes =  new_attribute :: el.attributes } in

            scan_element new_element lexbuf
    }

| ">"
    {
      newchars 1;
      let new_loc = location () in
        { el with
              attributes = List.rev el.attributes;
              sub_trees = scan_element_content new_loc el lexbuf
        }
    }

| "/>"
    {
      newchars 2;
      { el with
            attributes = List.rev el.attributes;
      }
    }

| nl
    {
      newline ();
      scan_element el lexbuf
    }

| (' ' | '\t')+ as white
    {
      newchars (String.length white);
      scan_element el lexbuf
    }

| _
    {
      let (line, pos) = location() in
      let err =
        Printf.sprintf "Parse error in (%d,%d) while scanning element \"%s\" in (%d,%d)"
          line pos el.name (fst el.el_pos) (snd el.el_pos)
      in
        raise (Parse_error err)
    }

and scan_parametric_attribute = shortest
| (_* as code) "?>\"" { code }

and scan_proc_instr starting_loc = parse
| "?>"    { newchars 2; flush () }
| nl      { newline (); scan_proc_instr starting_loc lexbuf }
| eof     {
            let (line, pos) = location() in
            let err =
              Printf.sprintf not_closed_error_string
                line pos "processing instruction"
                (fst starting_loc) (snd starting_loc) in
              raise (Parse_error err)
          }
| _       {
            addc (Lexing.lexeme_char lexbuf 0);
            scan_proc_instr starting_loc lexbuf
          }

{
  let parse_xnunc_lexbuf lexbuf =
    init ();
    let root_element =
      {
        name = root_element_default_name;
        attributes = [];
        sub_trees = [];
        el_pos = (0,0);
      } in
      scan_element_content (1,1) root_element lexbuf
;; (* END OF "parse_xnunc_lexbuf" *)


let rec print_parsed_block block =
  let print = Printf.sprintf in
    match block with
      | Proc_instr(text, (line, pos)) ->
          print "Proc_instr(%s)@(%d, %d)" text line pos
      | Print_proc_instr(text, (line, pos) ) ->
          print "Print_proc_instr(%s)@(%d, %d)" text line pos
      | Cdata(text, (line, pos) ) ->
          print "Cdata(%s)@(%d,%d)" text line pos
      | Element(el) ->
          let attributes_string =
            List.fold_left
              (
                fun acc_string attr ->
                  let name, name_line, name_pos, attr_type,
                      attr_val, val_line, val_pos =
                    match attr with
                      | (name, (name_line, name_pos)),
                        (Literal(attr_val), (val_line, val_pos)) ->
                          name, name_line, name_pos, "Literal",
                          attr_val, val_line, val_pos
                      | (name, (name_line, name_pos)),
                        (Parametric(attr_val), (val_line, val_pos)) ->
                          name, name_line, name_pos, "Parametric",
                          attr_val, val_line, val_pos in
                    
                    acc_string ^
                    (print "%s @ (%d,%d) = %s \"%s\" @ (%d,%d); "
                       name name_line name_pos attr_type attr_val
                       val_line val_pos)
              )
              ""
              el.attributes in
          let sub_trees_string =
            List.fold_left
              (
                fun acc_string block ->
                  acc_string ^
                  (print "%s;\n" (print_parsed_block block))
              )
              ""
              el.sub_trees in
            print "{
  name       = \"%s\";
  attributes = [ %s ];
  sub_trees  = [ %s ];
  el_pos     = (%d,%d);
}"
            el.name
            attributes_string
            sub_trees_string
            (fst el.el_pos)
            (snd el.el_pos)
;; (* END of "print_parsed_block" *)


let print_document doc =
  let print = Printf.sprintf in
  List.fold_left
    (
      fun acc_string block ->
        acc_string ^
        (print "%s;\n" (print_parsed_block block))
    )
    ""
    doc
;; (* END OF "print_document" *)


let rec parsed_block_compare block_1 block_2 =
  match block_1 with
    | Proc_instr(text_1, (line_1, pos_1) ) ->
        begin
          match block_2 with
            | Proc_instr(text_2, (line_2, pos_2) ) ->
                if (text_1 = text_2) && (line_1 = line_2) && (pos_1 = pos_2)
                then true else false
            | _ -> false
        end
    | Print_proc_instr(text_1, (line_1, pos_1) ) ->
        begin
          match block_2 with
            | Print_proc_instr(text_2, (line_2, pos_2) ) ->
                if (text_1 = text_2) && (line_1 = line_2) && (pos_1 = pos_2)
                then true else false
            | _ -> false
        end
    | Cdata(text_1) ->
        begin
          match block_2 with
            | Cdata(text_2) ->
                if (text_1 = text_2) then true else false
            | _ -> false
        end
    | Element(el_1) ->
        begin
          match block_2 with
            | Element(el_2) ->
                if el_1.name = el_2.name
                then
                  begin
                    let attribute_are_equal = try
                      List.fold_left2
                        ( fun acc a b ->
                            if a = b && acc = true then true else false )
                        true
                        el_1.attributes
                        el_2.attributes
                    with (Invalid_argument "List.fold_left2") -> false in
                      if attribute_are_equal = true
                      then
                        begin
                          if el_1.el_pos = el_2.el_pos
                          then
                            document_compare el_1.sub_trees el_2.sub_trees
                          else
                            false
                        end
                      else
                        false
                  end
                else
                  false
            | _ -> false
        end
and (* END of "parsed_block_compare" *)
document_compare doc1 doc2 =
  try
    List.fold_left2
      (
        fun acc a b ->
          if (parsed_block_compare a b) = true && acc = true
          then true else false
      )
      true
      doc1
      doc2
  with
    | (Invalid_argument "List.fold_left2") -> false
;; (* END OF "document_compare" *)
  
}

