open Xnc_stdlib
open Xnc_functional
open Utility
open Common_types

module L = List
module Pf = Printf
module U = Unix
module Fn = Filename
module CL = Command_line

let create_ocamlbuild_files build_dir project_dirs =
  let all_except_dot = List.filter ((<>) ".") project_dirs in

  (* STEP 1: generate the _tags file in the root *)
  let tag_file_body =
    try
      string_of_file_name "_tags"
    with
      | Sys_error _ ->
          List.map (fun s -> "\"" ^ s ^ "\" : include") all_except_dot |>
              String.concat "\n" in
  let old_content =
    try
      string_of_file_name (build_dir//"_tags")
    with
      | Sys_error _ -> "" in
    if tag_file_body <> old_content
    then string_to_file_name (build_dir//"_tags") tag_file_body;

    (* STEP 2: generate myocamlbuild.ml *)
    let myocamlbuild =
      try
        string_of_file_name "myocamlbuild.ml"
      with
        | Sys_error _ -> Globals.default_myocamlbuild_ml in
      string_to_file_name (build_dir//"myocamlbuild.ml") myocamlbuild;

      (* STEP 3: search for user provided _tags file *)
      List.iter
        (
          fun dir ->
            if Sys.file_exists (dir//"_tags")
            then file_copy (dir//"_tags") (build_dir//dir//"_tags")
        ) all_except_dot
;;

let execute_ocamlbuild build_dir target_type =
  let cwd = U.getcwd () in
  let () = U.chdir build_dir in
  let ext =
    match target_type with
      | CL.Native -> ".native"
      | CL.Profile -> ".p.native"
      | CL.Byte -> ".byte"
      | CL.Debug -> ".d.byte" in
  let intermediate_filename = "main" ^ ext in
  let exit_status = U.system ("ocamlbuild " ^ intermediate_filename)  in
    U.chdir cwd;
    match exit_status with | U.WEXITED es | U.WSIGNALED es | U.WSTOPPED es ->
      intermediate_filename, es
;;

class global_validator =
object(self)
  val mutable start_page_exists = 0
  val mutable page_names = []
  val mutable page_calls = String_map.empty

  method collect_info (file : Files_common.source_file) =
    let page_name = file#name |> Fn.basename in
    let name_exists n = if page_name = n then true else false in

      if List.exists name_exists page_names = true
      then failwith("Page \"" ^ page_name ^ "\" is duplicate.");
      (* TODO INTRODUCE AN EXCEPTION INSTEAD OF THIS "failwith" *)

      page_names <- page_name::page_names;

      (* TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
         TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
         TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
         TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO

         QUI C'È DA CONTARE IL NUMERO DI START PAGE E INCREMENTARE start_page_exists
      *)
      start_page_exists <- 1; (* TODO QUESTO È PROVVISORIO!!! OCCORRE CONTARE DAVVERO LE START-PAGE *)

        page_calls <- String_map.add
                        page_name   (* key *)
                        (file#get_target_pages |> L.str_sort |> uniq) (* value *)
                        page_calls;

  method private check_call_integrity =
    String_map.iter (fun caller_page target_pages ->
        L.iter (fun target_page ->
            let is_equal_to_target_page page =
              if page = target_page then true else false in
            if not (L.exists is_equal_to_target_page page_names)
            then failwith ("Page " ^ caller_page ^
                           " contains a jump to non existing page " ^
                           target_page)
              (* TODO INTRODUCE AN EXCEPTION INSTEAD OF THIS "failwith" *)
          ) target_pages
      ) page_calls

  method check_all =
    if start_page_exists <> 1
    then failwith("You must include one and only one start page into the project");
      (* TODO INTRODUCE AN EXCEPTION INSTEAD OF THIS "failwith" *)

    self#check_call_integrity;
end;;

(******************************************************************************)
(*****                     THIS IS THE MAIN FUNCTION                      *****)
(******************************************************************************)
let main () =
  let global_validator = new global_validator in

  let cmd_line_options = CL.handle_command_line () in
  let () = U.chdir cmd_line_options.CL.project_file_dir in
  let build_dir = cmd_line_options.CL.build_dir in
  let () = Utility.mkdir build_dir in

  let project_file =
    try
      open_in (cmd_line_options.CL.project_file_dir//cmd_line_options.CL.project_file_name)
      |> new Project_file.project_file
    with
      | Sys_error error ->
          Pf.eprintf "%s: %s\n" Globals.compiler_name error;
          exit 1 in

  let source_list = project_file#read_project_file#get_source_files in

  let source_file_list =
    try
      L.fold_left (
        fun l file_name ->
          (Files_factory.create_input_file global_validator file_name) :: l
      )
        []
        source_list
    with
      | Sys_error error ->
          Pf.eprintf "%s: %s\n" Globals.compiler_name error;
          exit 1
      | U.Unix_error(U.ENOENT as error, calling_f, param)
      | U.Unix_error(U.EACCES as error, calling_f, param) ->
          Pf.eprintf "%s: %s: %s\n"
            Globals.compiler_name param (U.error_message error);
          exit 1 in

    (* Force the debug mode for all pages if necessary *)
    if cmd_line_options.CL.page_debug_mode then
      List.iter (fun f -> f#force_debug) source_file_list;

    (* Perform all global consistency checks *)
    global_validator#check_all;

    (* Files generated from tamplates *)
    let output_file_map =
      L.fold_left (* for each input file *)
        (
          fun map file ->
            let targets = file#target_files in
              L.fold_left
                (Files_factory.create_output_file build_dir file) map targets
        )
        String_map.empty
        source_file_list in

    (* Constant files *)
    (* The main server file *)
    let main_file_name = build_dir ^ "/main.ml" in
    let main_file = (new Main_file.main_file main_file_name :>
                       Files_common.generated_file) in
    let output_file_map =
      String_map.add main_file_name main_file output_file_map in

    (* The page modules file *)
    let page_modules_file_name = build_dir ^ "/page_modules.ml" in
    let page_modules_file =
      (new Page_modules_file.page_modules_file page_modules_file_name :>
         Files_common.generated_file) in
    let output_file_map =
      String_map.add page_modules_file_name page_modules_file output_file_map in

      (*
       * Create all necessary subdirectories
       *)
      String_map.values output_file_map |>
          List.fold_left (fun acc f -> (Fn.dirname f#name)::acc) [] |>
            L.str_sort |> uniq |> L.iter (Utility.mkdirhier ~permissions:0o755);

      let project_dirs =
        List.fold_left
          (fun acc f -> (Fn.dirname f#name)::acc) [] source_file_list |>
            L.str_sort |> uniq in

      let () = create_ocamlbuild_files build_dir project_dirs in
        String_map.iter (fun _ v -> v#render_file) output_file_map;

        (* Call ocamlbuild *)
        (* TODO this function is only a script for compiling the demo *)
        let tmp_filename, exit_status =
          execute_ocamlbuild build_dir cmd_line_options.CL.target_type in

          if exit_status = 0
          then file_copy (build_dir//tmp_filename) (cmd_line_options.CL.output_filename);

          exit exit_status
;; (* END OF "main" FUNCTION *)

(*
 * The program starts here
 *)
main ()

